#!/usr/bin/env bash

# This script contains wrappers for Linux commands.
#
# You can either source this script to use within another script
#   $ source ./gpudb_bash_utils.sh
# or run it as a program.
#   $ ./gpudb_bash_utils.sh -h

# ---------------------------------------------------------------------------

# All non-sourced scripts should call this with a < 16 char string
# prefaced with 'gpudb', e.g. "gpudb_COMPONENT"
# Set the '/proc/pid/comm' to identify this script, default is 'bash'.
THIS_SCRIPT_COMM=bash # global
function gpudb_set_proc_comm
{
    THIS_SCRIPT_COMM="$1"
    printf "$THIS_SCRIPT_COMM" > /proc/$$/comm
    if [ $? -ne 0 ]; then
        echo_status "Script" "Error setting /proc/$$/comm to '$THIS_SCRIPT_COMM'"
        return 1
    fi
}

# All functions that source this file should call this function to set
# path information used by some functions in this file.
# Usage: gpudb_set_this_script_info "${BASH_SOURCE[0]}" "$@"
function gpudb_set_this_script_info
{
    local BASH_SOURCE_0="$1"
    shift

    # The directory of where this script was run from.
    THIS_SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE_0}" )" && pwd )" # NOTE: you probably want THIS_SCRIPT_LINK_DIR
    THIS_SCRIPT_NAME="${BASH_SOURCE_0##*/}"
    THIS_SCRIPT_LINK=$(readlink -f "${BASH_SOURCE_0}") # filepath that a symlink resolves to
    THIS_SCRIPT_LINK_DIR="$(dirname "${THIS_SCRIPT_LINK}")" # the actual path of the file
    THIS_SCRIPT="$THIS_SCRIPT_DIR/$THIS_SCRIPT_NAME"    # filepath where this was run from
    THIS_SCRIPT_ARGS="$@"
    #echo "THIS: '$THIS_SCRIPT_DIR' '$THIS_SCRIPT_NAME' '$THIS_SCRIPT_ARGS'"
}

# Our info, should be this filename, not caller even if sourced.
UTILS_THIS_SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
UTILS_THIS_SCRIPT_NAME="${BASH_SOURCE[0]##*/}"
UTILS_THIS_SCRIPT_LINK=$(readlink -f "${BASH_SOURCE[0]}") # filepath that a symlink resolves to
UTILS_THIS_SCRIPT_LINK_DIR="$(dirname "$UTILS_THIS_SCRIPT_LINK")" # the actual directory the script resides in
UTILS_THIS_SCRIPT="$UTILS_THIS_SCRIPT_DIR/$UTILS_THIS_SCRIPT_NAME"    # filepath where this was run from
UTILS_THIS_SCRIPT_ARGS="$@"

GPUDB_START_DATETIME=$(date '+%Y-%m-%d_%H.%M.%S')

# ---------------------------------------------------------------------------

#if [[ "$PATH" != *$GPUDB_CORE_LIBS_DIR/bin* ]]; then
#    export PATH="$GPUDB_CORE_LIBS_DIR/bin:$PATH"
#fi

# ---------------------------------------------------------------------------
# Echo, log, stdout control

# Echo a padded string in the format "Component <pad> : State", e.g. "Text : Stopped"
GPUDB_ECHO_STATUS_WIDTH=16
function echo_status
{
    local COMPONENT=$1
    local STATE=$2
    printf "%-${GPUDB_ECHO_STATUS_WIDTH}s : $STATE\n" "$COMPONENT"
}

# ---------------------------------------------------------------------------
# Config file read/write

# Read a conf.ini parameter in the form "KEY SEPARATOR VALUE"
# Usage: get_conf_file_property VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING] [OPTIONAL 'WARN_DEFAULT']

function get_conf_file_property
{
    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="$5"
    local WARN_DEFAULT="$6" # Print warning if string 'WARN_DEFAULT'

    local VALUE_GCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same

    local KEYVAL_LINE=""
    #KEYVAL_LINE=$(grep -E "^[ $(echo -e '\t')]*$KEY[ $(echo -e '\t')]*$SEPARATOR" "$FILENAME")
    KEYVAL_LINE=$($THIS_SCRIPT_DIR/gpudb_conf_parser.py "$FILENAME" "${KEY}")

    if [ $? -ne 0 ]; then
        if [ $# -gt 4 ]; then
            if [[ "a$WARN_DEFAULT" == "aWARN_DEFAULT" ]]; then
                echo "WARN: Missing param in '$FILENAME', using default '$KEY' $SEPARATOR '$DEFAULT_VALUE'"
            fi
            VALUE_GCFP_="$DEFAULT_VALUE"
        else
            >&2 echo "ERROR finding conf param '$KEY' in file '$FILENAME' in '$THIS_SCRIPT', exiting."
            exit 1
        fi
    else
        VALUE_GCFP_=$(echo $KEYVAL_LINE | cut -d "$SEPARATOR" -f2 | sed -e 's/^[ \t]*//' -e 's/[ \t]*$//')
    fi

    eval $OUTPUT_VAR_NAME="'$VALUE_GCFP_'"
}

# Get a key='true/false' conf property, set OUTPUT_VAR_NAME to 1/0
# Usage: get_true_false_conf_file_property VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING] [OPTIONAL 'WARN_DEFAULT']
function get_true_false_conf_file_property
{
    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="$5"
    local WARN_DEFAULT="$6" # Warn if string 'WARN_DEFAULT'

    local VALUE_GTFCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same
    get_conf_file_property VALUE_GTFCFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE" "$WARN_DEFAULT"

    # Using grep since "tr [:upper:] [:lower:]" gives an error when run on some
    # systems in this script, but not as the user on the command line.
    if echo "$VALUE_GTFCFP_" | grep -i '^true$' > /dev/null 2>&1 ; then
        VALUE_GTFCFP_=1
    elif echo "$VALUE_GTFCFP_" | grep -i '^false$' > /dev/null 2>&1 ; then
        VALUE_GTFCFP_=0
    else
        >&2 echo "Error: Unknown value in conf file: '$FILENAME' for param: '$KEY=$VALUE_GTFCFP_', expected either 'true' or 'false' in '$THIS_SCRIPT', exiting."
        exit 1
    fi

    eval $OUTPUT_VAR_NAME="'$VALUE_GTFCFP_'"
}

# Get a conf property that is a directory, handle relative paths, and verify.
# Usage: get_directory_conf_file_property VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_directory_conf_file_property
{
    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="$5"

    local VALUE_GDCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same
    get_conf_file_property VALUE_GDCFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE"

    local PATH_GDCFP_="$VALUE_GDCFP_"
    fix_gpudb_conf_path PATH_GDCFP_ "$VALUE_GDCFP_"

    if [ ! -d "$PATH_GDCFP_" ]; then
        if [[ "a$PATH_GDCFP_" != "a$VALUE_GDCFP_" ]]; then
            echo "Error: Missing or invalid conf param value for '$KEY=$VALUE_GDCFP_' which resolves to '$PATH_GDCFP_' in '$THIS_SCRIPT', exiting."
        else
            echo "Error: Missing or invalid conf param value for '$KEY=$PATH_GDCFP_' in '$THIS_SCRIPT', exiting."
        fi

        exit 1
    fi

    eval $OUTPUT_VAR_NAME="'$PATH_GDCFP_'"
}

# Change a conf.ini parameter in the form "KEY = *" to "KEY = VAL" in FILENAME.
# Usage: change_conf_file_property "KEY_NAME" "=" "NEW_VALUE" conf.ini
function change_conf_file_property
{
    local KEY="$1"
    local SEPARATOR="$2"
    local NEW_VALUE="$3"
    local FILENAME="$4"

    # Also verify that it was written since sed always returns 0.
    if [[ $KEY == *"\n"* ]]; then
        eval "sed -i \"N;s@\(^[ \t]*${KEY}[ \t]*${SEPARATOR}\).*@\1${NEW_VALUE}@\" $FILENAME"
        eval "grep -Ezo \"^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*${SEPARATOR}${NEW_VALUE}\" $FILENAME" > /dev/null
        if [ $? -ne 0 ]; then
            echo "ERROR: Unable to update conf '$FILENAME', '$KEY'='$NEW_VALUE' in '$THIS_SCRIPT', exiting."
            exit 1
        fi
    else
        eval "sed -i \"s@\(^[ \t]*${KEY}[ \t]*${SEPARATOR}\).*@\1${NEW_VALUE}@\" $FILENAME"
        eval "grep -E \"^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*${SEPARATOR}${NEW_VALUE}\" $FILENAME" > /dev/null
        if [ $? -ne 0 ]; then
            echo "ERROR: Unable to update conf '$FILENAME', '$KEY'='$NEW_VALUE' in '$THIS_SCRIPT', exiting."
            exit 1
        fi
    fi
}

# ---------------------------------------------------------------------------
# File directory validation

function gpudb_check_dir
{
    local DIR_NAME="$1"

    if [ ! -d "$1" ]; then
        echo "Error: Missing directory: '$DIR_NAME' in '$THIS_SCRIPT', exiting."
        exit 1
    fi
}

function gpudb_check_file
{
    if [ ! -f "$1" ]; then
        echo "Error: Missing file: '$1' in '$THIS_SCRIPT', exiting."
        exit 1
    fi
}

# ---------------------------------------------------------------------------
# Networking

# Checks if a port is used, logs error and exits 1.
function check_port_usage
{
    local PORT="$1"
    local ERR_MSG="$2"

    if netstat -tuln | grep ":$PORT " | grep LISTEN > /dev/null; then
        echo "$ERR_MSG"
        NETSTAT_TXT=$(netstat -tulpn | grep ":$PORT " 2> /dev/null)
        if [[ "a$NETSTAT_TXT" != "a" ]]; then
            echo "$NETSTAT_TXT"
            ps uwwp $(echo "$NETSTAT_TXT" | awk '{print $7}' | grep -oE '[0-9]+')
        else
            netstat -tuln | grep $WEB_SERVER_PORT
            echo "Unable to determine other app, please run 'netstat -tulpn | grep \":$PORT \"' as root to find PID"
        fi
        exit 1
    fi
}

# ---------------------------------------------------------------------------
# Pids




# ---------------------------------------------------------------------------
#
# ---------------------------------------------------------------------------


if [[ "${BASH_SOURCE[0]}" != "${0}" ]]; then
    #echo "${BASH_SOURCE[0]} is being sourced..."
    return
fi

# ---------------------------------------------------------------------------
# Sometimes nvcc doesn't die when make exits, kill it.

function kill_children()
{
    for i in 1 2 3; do
        local CHILD_PIDS=$(jobs -p)

        echo "CHILD_PIDS: $CHILD_PIDS"

        if [ -z "$CHILD_PIDS" ]; then
            exit
        fi

        echo "killing PIDs: $CHILD_PIDS"
        kill $CHILD_PIDS > /dev/null 2>&1
        sleep 2
    done

    local CHILD_PIDS=$(jobs -p)
    kill -9 $CHILD_PIDS > /dev/null 2>&1

    exit
}

#trap kill_children SIGHUP SIGINT SIGTERM EXIT

# ---------------------------------------------------------------------------


START_TIME=$(date +%s)
END_TIME=$(date +%s)
DIFF_TIME=$(expr $END_TIME - $START_TIME)
DIFF_TIME_STR=$(date '+%H:%M:%S' -u --date="@$DIFF_TIME")
echo "Build took: $(expr $END_TIME - $START_TIME)s, $DIFF_TIME_STR"


