#!/usr/bin/env python

import collections
try:
    from configparser import ConfigParser
except:
    from ConfigParser import SafeConfigParser as ConfigParser
import datetime
import json
import os
import re
import sys
import shutil

# Define types -> comparison engine mapping
VALIDINI = ['.config', '.conf', '.ini']
VALIDPY = ['.py', '.properties']
VALIDJS = ['.js']
VALIDJSON = ['.json']
STATICMARKER = '#static'

# ---------------------------------------------------------------------------------------
# A natural sort function to replace the standard python "sorted" function.  This utility
# will break strings into alphabetic and numeric chunks and then sort the pieces.  So
# for example, the string "a13b" and "a2b" would be broken into ["a", 13, "b"] and
# ["a", 2, "b"] and allows python to sort them with "a2b" coming before "a13b".  We use
# this for example, when we write out the config files so that things like, "rankN.gpu"
# (where N is a number) get sorted out in nice, human readable lists.

natural_sort_re = re.compile('([0-9]+)')
def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [ convert(c) for c in re.split(natural_sort_re, key) ]
    return sorted(l, key = alphanum_key)

def get_base_option_name(option_name):
    """Returns a 'normalized' version of the option name, allowing grouping."""
    return re.sub(r"[0-9]+", "", re.sub(r"\.default\.", ".rank.", option_name))


# ---------------------------------------------------------------------------------------
# DiffLogger & supporting class ConfDiff - Diff tracker and printer

class ConfDiff:
    """A data holder for updated configuration values."""
    def __init__(self, option_name, modified_value, baseline_value, merged_value):
        self.option_name = option_name
        self.modified_value = modified_value
        self.baseline_value = baseline_value
        self.merged_value = merged_value

class DiffLogger:
    """ A fancy pretty printer to write out updated items to the console. """

    def __init__(self, modified_filename, baseline_filename, merged_filename, show_merged = False, parseable = False):
        """This class handles the display of the file differences and potentially the merged output."""
        self.modified_filename = modified_filename
        self.baseline_filename = baseline_filename
        self.merged_filename = merged_filename
        self.show_merged = show_merged
        self.parseable = parseable

        self.option_width = 15
        self.modified_column_width = 15
        self.baseline_column_width = 15
        self.merged_column_width = 15
        self.allowed_header_growth = 20

        self.diffs = []

    def addDiff(self, option_name, modified_value, baseline_value, merged_value):
        """Adds a difference to the list of items to display."""
        self.diffs.append(ConfDiff(option_name, modified_value, baseline_value, merged_value))

    def _getValueDisplayText(self, val):
        """Handles turning 'None' into something useful to show the user."""
        result = val
        if result is None:
            result = "<not present>"
        else:
            result = str(result).replace("\t", "    ")
        return result

    def _getValueTextWidth(self, val):
        maxWidth = 0
        for l in str(val).split("\n"):
            if len(l) > maxWidth:
                maxWidth = len(l)
        return maxWidth

    def _getFilenameDisplayText(self, val, column_width):
        """Handles properly formatting and trimming file names, which are restricted to 
           being only as wide as the largest value in the column plus allowed_header_growth."""
        result = val
        if result is None:
            result = ""
        else:
            if len(result) + 2 > column_width:
                result = "..." + result[-(column_width - 5):]
            result = "(" + result + ")"
        return result

    def _getValueTextHeight(self, val):
        return len(str(val).split("\n"))

    def _calculateColumnWidths(self):
        """This function handles auto-calculating the required width of the columns."""
        for diff in self.diffs:
            self.option_width = max(self.option_width, self._getValueTextWidth(diff.option_name))
            self.modified_column_width = max(self.modified_column_width, self._getValueTextWidth(self._getValueDisplayText(diff.modified_value)))
            self.baseline_column_width = max(self.baseline_column_width, self._getValueTextWidth(self._getValueDisplayText(diff.baseline_value)))
            if self.show_merged:
                self.merged_column_width = max(self.merged_column_width, self._getValueTextWidth(self._getValueDisplayText(diff.merged_value)))

        # Allow the columns to grow slightly for large filenames
        preferred_header_size = 0
        if self.modified_filename is not None and len(self.modified_filename) > self.modified_column_width:
            self.modified_column_width = min(self.modified_column_width + self.allowed_header_growth, len(self.modified_filename))
        if self.baseline_filename is not None and len(self.baseline_filename) > self.baseline_column_width:
            self.baseline_column_width = min(self.baseline_column_width + self.allowed_header_growth, len(self.baseline_filename))
        if self.merged_filename is not None and len(self.merged_filename) > self.merged_column_width:
            self.merged_column_width = min(self.merged_column_width + self.allowed_header_growth, len(self.merged_filename))

    def _padValueColumn(self, val, column_width):
        """Pads a column value based on the provided width."""

        result = self._getValueDisplayText(val)

        if not self.parseable:
            result = result.ljust(column_width)

        return result

    def _padOptionColumn(self, option_name):
        result = option_name

        if not self.parseable:
            result = result.ljust(self.option_width)

        return result

    def _addTableFormattingToRow(self, option_name, modified_value, baseline_value, merged_value):
        """Pads the provided values and adds column delimiters."""
        delimiter = "|"
        if self.parseable:
            delimiter = "\t"

        if self.show_merged:
            merged_value_text = delimiter + self._padValueColumn(merged_value, self.merged_column_width)
        else:
            merged_value_text = ""

        return self._padOptionColumn(option_name) + delimiter + self._padValueColumn(modified_value, self.modified_column_width) + delimiter + self._padValueColumn(baseline_value, self.baseline_column_width) + merged_value_text

    def printTable(self):
        # Precalculate the column widths
        self._calculateColumnWidths()

        # Print the header
        if not self.parseable:
            print("")
            print(self._addTableFormattingToRow("Option", "Modified Value", "Default Value", "Merged Value"))
            if self.modified_filename is not None or self.baseline_filename is not None or self.merged_filename is not None:
                print(self._addTableFormattingToRow("", self._getFilenameDisplayText(self.modified_filename, self.modified_column_width), self._getFilenameDisplayText(self.baseline_filename, self.baseline_column_width), self._getFilenameDisplayText(self.merged_filename, self.merged_column_width)))

            header_break = "-" * self.option_width + \
                           "+" + "-" * self.modified_column_width + \
                           "+" + "-" * self.baseline_column_width
            if self.show_merged:
                header_break += "+" + "-" * self.merged_column_width
            print(header_break)

        # Print the rows
        for diff in self.diffs:
            a = str(self._getValueDisplayText(diff.option_name)).split("\n")
            b = str(self._getValueDisplayText(diff.modified_value)).split("\n")
            c = str(self._getValueDisplayText(diff.baseline_value)).split("\n")
            d = str(self._getValueDisplayText(diff.merged_value)).split("\n")
            for i in range(max([len(a), len(b), len(c), len(d)])):
                linea = "" if i >= len(a) else a[i]
                lineb = "" if i >= len(b) else b[i]
                linec = "" if i >= len(c) else c[i]
                lined = "" if i >= len(d) else d[i]
                print(self._addTableFormattingToRow(linea, lineb, linec, lined))
            #print(self._addTableFormattingToRow(diff.option_name, diff.modified_value, diff.baseline_value, diff.merged_value))

    def printWarnings(self):
        removed_options = []
        added_options = []

        for diff in self.diffs:
            # Added options are values that were not in the "existing" / "modified" config, but were
            # added because there was a parameter in the "baseline" / "default" config.
            if diff.modified_value is None and diff.merged_value is not None:
                added_options.append(diff)
            # Removed options are options that have been removed from the "baseline" / "default"
            # config, and will be removed from the resulting merged output.  We must check
            # that it is missing in both baseline and merged, because there is a possibility
            # that a value in a "group" does not exist, but the group is preserved (for example,
            # when there are many ranks in a modified config, but the baseline config just has
            # the first two ranks as an example).
            if diff.baseline_value is None and diff.merged_value is None:
                removed_options.append(diff)

        if len(removed_options) > 0 or len(added_options) > 0:
            print("")

        if len(removed_options) > 0:
            print("INFO: The following configuration parameters are no longer used and have been removed.")
            for diff in removed_options:
                print("  - " + diff.option_name)
        if len(added_options) > 0:
            print("WARNING: The following configuration parameters have been added with their default values.")
            for diff in added_options:
                print("  - " + diff.option_name + " (" + str(diff.baseline_value) + ")")

# ---------------------------------------------------------------------------------------
# Configuration File Adapters - handle reading config files and generating new ones
# based off a template.
#
# All of them have three functions __init__, getConfig and generateConfig.  Generally
# speaking they should be used like such...
#
# dict_of_values = XXXFileAdapter(some_input_filename)
#
# and
#
# XXXFileAdapter(some_output_filename).generateConfig(some_template_file, dict_of_updates)
#

class JsonFileAdapter:
    """This class knows how to read and write JSON configuration files."""
    def __init__(self, filename):
        self.filename = filename

    def getConfig(self, print_warnings = False):
        """Returns a dictionary of the JSON values in the configuration file."""

        result = Configuration()

        input_file = open(self.filename)
        result.items = json.loads(input_file.read())
        input_file.close()
        return result

    def generateConfig(self, template_file, updates):
        """Accepts a template_file with which to base the new configuration and a dictionary containing the keys and values to
           override from the template."""

        template_config = JsonFileAdapter(template_file).getConfig()

        leaves = updates.getPaths()
        for leaf in leaves:
            updated_value = updates.getValue(leaf)
            template_config.setValue(leaf, updated_value)

        output_file = open(self.filename, 'w')
        output_file.write(json.dumps(template_config.items, indent=4))
        output_file.close()

class JsFileAdapter:
    """This class knows how to read and write Javascript files and interpret any variables therein as configurable
       values."""

    def __init__(self, filename):
        self.filename = filename
        self.assignment_regex = r'^\s*var\s+\w+\s*=[^;]+'

    def getConfig(self, print_warnings = False):
        """Returns a dictionary of the python variables in the python file."""
        template_config = open(self.filename)
        result = Configuration()
        result_dict = result.items
        ignore_keys = []

        pattern = re.compile(self.assignment_regex, flags=re.MULTILINE|re.DOTALL)
        in_match = None

        var_buffer = ""
        var_key_buffer = None
        var_key = None
        quote_char = None
        block_comment = False
        block_comment_last = ""
        line_comment = False
        # Create map of all recognized key => value of input file - this tangled web is a stab at
        # extremely rudimentarly parsing Javascript and assumes that the ONLY JS in the file are
        # "var =" assignments, with the exception of comments.
        for line in template_config:
            for c in line:
                if line_comment:
                    if var_key is not None:
                        var_buffer = var_buffer + c

                elif block_comment:
                    if block_comment_last + c == "*/":
                        block_comment = False
                        block_comment_last = ""
                    else:
                        block_comment_last = c

                    if var_key is not None:
                        var_buffer = var_buffer + c

                elif quote_char is not None:
                    var_buffer = var_buffer + c
                    if c == quote_char and var_buffer[-2] != "\\":
                        quote_char = None

                elif var_key_buffer is not None:
                    if c == "=":
                        # Equals - we're done here.
                        if var_key is None:
                            var_key = var_key_buffer
                        var_key_buffer = None
                    elif c in [" ", "\t"]:
                        # Space - if we have a var_key_buffer, save it,
                        if var_key_buffer is not None and var_key_buffer != "":
                            var_key = var_key_buffer
                    elif re.search("[^A-Za-z0-9\-_]", c) is not None:
                        print("ERROR: bad character '" + c + "' in variable declaration in line '" + line + "' of file " + self.filename)
                        sys.exit(1)
                    else:
                        if var_key is None:
                            # Haven't already assigned a key
                            var_key_buffer = var_key_buffer + c
                        else:
                            # We already have a key - should have gotten space or equals
                            print("ERROR: bad variable declaration in line '" + line + "' of file " + self.filename + " var_key = " + var_key + ", var_key_buffer = " + var_key_buffer)
                            sys.exit(1)

                else:
                    if var_buffer != "" or c not in [" ", "\t", "\n"]:
                        var_buffer = var_buffer + c
                    if c == ";":
                        # process var_buffer
                        if var_key is not None:
                            result_dict[var_key] = var_buffer[:-1];
                            var_key = None
                        else:
                            print("ERROR: unknown identifier '" + c + "' in line '" + line + "' of file " + self.filename)
                            sys.exit(1)

                        var_buffer = ""
                    elif len(var_buffer) >= 2 and var_buffer[-2:] == "//":
                        line_comment = True
                        if var_key is None:
                            var_buffer = var_buffer[0:-2]
                    elif len(var_buffer) >= 2 and var_buffer[-2:] == "/*":
                        block_comment = True
                        if var_key is None:
                            var_buffer = var_buffer[0:-2]
                    elif c in ["'", '"']:
                        quote_char = c
                        quote_char_last = ""
                    elif len(var_buffer) >= 3 and var_buffer[-3:] == "var":
                        if len(var_buffer) > 3 and var_buffer[:-3].strip() != "":
                            print("ERROR: non-var text '" + var_buffer[:-3] + "' in line '" + line + "' of file " + self.filename)
                            sys.exit(1)
                        var_key_buffer = ""
                        var_buffer = ""

            # EOL
            line_comment = False
        # EOF
        if var_buffer.strip() != "":
            print("ERROR: extra characters '" + var_buffer + "' at end of file " + self.filename)
            sys.exit(1)

        return result

    def generateConfig(self, template_file, updates):
        """Accepts a template_file with which to base the new python module and a dictionary containing the variables to
           override from the template.  Note that lines marked with the comment '#static' will be ignored."""

        # We need the template to modify
        template_values = JsFileAdapter(template_file).getConfig()

        template_config = open(template_file)
        template_text = ""
        # Grab all of the file so that we can modify it in-place
        for line in template_config:
            template_text = template_text + line

        # Create output file
        output_config = open(self.filename, 'w')

        for update in updates.getPaths():
            updated_value = updates.getValue(update)
            template_value = template_values.getValue(update)
            if updated_value is not None and template_value is not None:
                update_regex = r"" + update + r"\s*=\s*" + re.escape(template_value)
                template_text = re.sub(update_regex, update + " = " + updated_value, template_text, flags=re.MULTILINE|re.DOTALL)

        output_config.write(template_text)

        template_config.close()
        output_config.close()

class IniFileAdapter:
    """This class knows how to read and write INI configuration files."""
    def __init__(self, filename):
        self.filename = filename

    def getConfig(self, print_warnings = False):
        """Returns the Configuration from the JSON values in the configuration file."""
        config_parser = ConfigParser()
        config_parser.read(self.filename)

        result = Configuration()
        result_dict = result.items
        for section in config_parser.sections():
            if not section in result_dict:
                result_dict[section] = collections.OrderedDict()

            dict_section = result_dict[section]
            for op in config_parser.items(section):
                dict_section[op[0]] = op[1]

        with open(self.filename) as cf:
            current_section = None
            for line in cf:
                if line.strip().startswith("[") and line.strip().endswith("]"):
                    current_section = line.strip().lstrip("[").rstrip("]").strip()

                if current_section is not None and line.startswith("#") and "=" in line:
                    if not current_section in result.optional_items:
                        result.optional_items[current_section] = collections.OrderedDict()

                    option = line.lstrip("#").split("=")[0].strip()
                    if not current_section in result_dict or not option in result_dict[current_section]:
                        result.optional_items[current_section][option] = None

        return result

    def generateConfig(self, template_file, updates):
        """Accepts a template_file with which to base the new configuration and a dictionary containing the keys and values to
           override from the template."""
        template_config_file = open(template_file)
        output_file = open(self.filename, 'w')

        output_options = natural_sort(updates.getPaths())

        template_config = IniFileAdapter(template_file).getConfig()

        current_section = ""
        handled_options = []
        for line in template_config_file:
            stripped = line.strip()

            if stripped.startswith("[") and stripped.endswith("]"):
                # Write out all section headers.
                current_section = stripped.replace("[", "").replace("]", "")
                output_file.write(line)
            elif stripped == "":
                # Copy all newlines straight over.
                output_file.write(line)
            elif stripped.startswith("#") or stripped.startswith(";"):
                # Copy comments over and then copy any overrides
                output_file.write(line)
                if "=" in stripped:
                    optional = stripped.lstrip("#").split("=")[0].strip()
                    base_option = get_base_option_name(optional)
                    base_option_path = current_section + "/" + base_option

                    if base_option_path not in handled_options:
                        output_value = updates.getValue(current_section + "/" + optional)

                        # Only attempt to handle items here which *only* appear as options - if the option
                        # is in the non-optional Paths, then ignore here and handle it later.
                        if not current_section + "/" + optional in template_config.getPaths():
                            for sub_option in output_options:
                                base_sub_option = get_base_option_name(sub_option)
                                if base_sub_option == current_section + "/" + base_option:
                                    sub_option_value = updates.getValue(sub_option)
                                    if sub_option_value is not None and sub_option_value != "":
                                        output_file.write("%s = %s\n" % (sub_option.rpartition("/")[-1], sub_option_value))
                            handled_options.append(base_option_path)
            else:
                # This is an possibility.  See if we need to copy from the modified file or if we pull a default
                # from the 'baseline'.

                keyvalue = stripped.split("=", 1)
                if len(keyvalue) != 2:
                    print("ERROR!  Invalid format in line '%s' - unable to parse into key and value pair." % line)
                    sys.exit(1)

                option = keyvalue[0].strip()
                baseline_value = keyvalue[1].strip()

                base_option = get_base_option_name(option)
                base_option_path = current_section + "/" + base_option

                if base_option_path not in handled_options:
                    output_value = updates.getValue(current_section + "/" + option)

                    for sub_option in output_options:
                        base_sub_option = get_base_option_name(sub_option)
                        if base_sub_option == current_section + "/" + base_option:
                            sub_option_value = updates.getValue(sub_option)
                            if sub_option == current_section + "/" + option and sub_option_value == baseline_value:
                                output_file.write(line)
                            else:
                                output_file.write("%s = %s\n" % (sub_option.rpartition("/")[-1], sub_option_value))

                    handled_options.append(base_option_path)

        output_file.close()

class PyFileAdapter:
    """This class knows how to read and write Python files and interpret any variables therein as configurable 
       values.  Note that any values marked as '#static' will not be included.  For example:
       do_not_touch_this = value #static"""
    def __init__(self, filename):
        self.filename = filename
        self.assignment_regex = r'\w+\s*=.+'

    def getConfig(self, print_warnings = False):
        """Returns a dictionary of the python variables in the python file."""
        template_config = open(self.filename)
        result = Configuration()
        result_dict = result.items
        ignore_keys = []

        # Create map of all recognized key => value of input file
        for line in template_config:
            # Look for any line that is a global variable assignment (no leading whitespace followed by a word,
            # optional whitespace, and then an equals sign and a value).
            if re.match(self.assignment_regex, line) and not STATICMARKER in line.lower():
                keyval = line.split('=', 1)
                key = keyval[0].strip()
                value = keyval[1].strip()
                if not key in result_dict and not key in ignore_keys:
                    result_dict[key] = value
                elif not key in ignore_keys:
                    del result_dict[key]
                    ignore_keys.append(key)
                    if (print_warnings):
                        print("WARNING: Ignoring key '" + key + "' while parsing '" + self.filename + "' because it was found multiple times.  This will cause values to be identified as <not present>.  When merging, this will cause merged values to revert to defaults.")

        return result

    def generateConfig(self, template_file, updates):
        """Accepts a template_file with which to base the new python module and a dictionary containing the variables to
           override from the template.  Note that lines marked with the comment '#static' will be ignored."""
        template_config = open(template_file)

        # Create output file
        output_config = open(self.filename, 'w')
    
        # Determine difference and update baseline line by line
        for line in template_config:
            stripped = line.strip()
            if re.match(self.assignment_regex, line) and not STATICMARKER in line.lower():
                keyval = line.split('=', 1)
                key = keyval[0]
                keystripped = key.strip()

                updated_value = updates.getValue(keystripped)
    
                if updated_value is not None:
                    output_config.write(key.rstrip() + ' = ' + str(updated_value) + '\n')
                else:
                    # TODO: Review this - unlikely to occur - this means that a config value
                    #       in the template needs to be deleted.  More likely encountered
                    #       when doing gpudb.conf, which has 'array' configurations
                    #       (rank[n].host = xxx)
                    output_config.write(line)
            else:
                output_config.write(line)
    
        template_config.close()
        output_config.close()

# ---------------------------------------------------------------------------------------
# This helper class helps flatten/simplify browsing and editing configuration hierarchies.
# The dict behind it is available at .items if needed.

class Configuration:
    """ This helper class helps flatten/simplify browsing and editing configuration hierarchies.
        The dict behind it is available at .items if needed."""

    def __init__(self):
        self.items = {}
        self.optional_items = {}

    def getPaths(self):
        """ Returns a list of paths of all the leaf nodes in the specified dictionary, in the format: 
            this/is/the/path/to/a/leaf
            You can then use this path to call the other helper functions (getDictValue and setDictValue)"""

        return self._getPaths(self.items)

    def getPathsWithOptionals(self):
        return list(set(self._getPaths(self.items)) | set(self._getPaths(self.optional_items)))

    def _getPaths(self, item_dict):
        result = []

        for item in item_dict:
            val = item_dict[item]
            if isinstance(val, dict):
                for name in self._getPaths(val):
                    result.append(item + "/" + name)
            else:
                result.append(item)

        return result

    def getValue(self, path):
        """ Gets the value of a node or None if the node does not exist.  If the node exists with a
            value of None, it will be returned as an empty string ('')."""
        return self._getValue(path, self.items)

    def getOptionalValue(self, path):
        return self._getValue(path, self.optional_items)

    def _getValue(self, path, item_dict):
        result = None

        # Split the first level in the path (path_split[0]) from the rest (path_split[1]).
        path_split = path.split("/", 1)
        if len(path_split) > 0:
            item = path_split[0]
            if item in item_dict:
                if len(path_split) > 1:
                    # There is more hierarchy to traverse.
                    result = self._getValue(path_split[1], item_dict[item])
                else:
                    # This is the requested node - if its value is None, swap that with '' to
                    # allow the consumer to know that there is something here.  This should
                    # help the merge/diff logic.
                    result = item_dict[item]
                    if result is None:
                        result = ""

        return result

    def setValue(self, path, value):
        """ Sets the value of a node (and creates the hierarchy if it does not yet exist.  Path
            should be in the format 'path/to/the/node/to/set'. """
        self._setValue(path, value, self.items)

    def setOptionalValue(self, path, value):
        self._setValue(path, value, self.optional_items)

    def _setValue(self, path, value, item_dict):
        path_split = path.split("/", 1)
        if len(path_split) > 0:
            item = path_split[0]
            if len(path_split) > 1:
                # This is not the node we are looking for - ensure it exists, and recurse
                if item not in item_dict:
                    item_dict[item] = dict()
                self._setValue(path_split[1], value, item_dict[item])
            else:
                # this is it!  Set it.
                item_dict[item] = value

# ---------------------------------------------------------------------------------------
# The main logic of the application

def getUpdates(modified_config, default_config):
    """ The logic to grab the list of configurable items that have been updated and
        what the updated value should be."""

    output_config = Configuration()

    modified_options = modified_config.getPaths()

    handled_options = []
    for option in default_config.getPathsWithOptionals():
        base_option = get_base_option_name(option)

        if base_option not in handled_options:
            # If there is a matching "base" option in the modified config, just take the list
            # from there.  Otherwise, pull all the default values from the 'default/template'
            pull_config = default_config
            for sub_option in modified_options:
                this_base_option = get_base_option_name(sub_option)
                if this_base_option == base_option:
                    pull_config = modified_config
                    break

            for sub_option in pull_config.getPathsWithOptionals():
                this_base_option = get_base_option_name(sub_option)
                if this_base_option == base_option:
                    sub_option_value = pull_config.getValue(sub_option)
                    output_config.setValue(sub_option, sub_option_value)

            handled_options.append(base_option)

    return output_config

def doCompareAndOutput(infile, basefile, output_filename, format_adapter, parseable, print_warnings):
    """ This function takes a XXXFileAdapter reference, the filename with path to the input file,
        the filename with path to the template/default file, and an optional path to the output
        file if the user wishes to output a merged result file (None if the user does not wish
        to generate a merged config)."""
    # Get modified config object using parser
    modified_config = format_adapter(infile).getConfig(print_warnings)

    # Get default config object
    default_config = format_adapter(basefile).getConfig(print_warnings)

    # Get the list of updates
    output_config = getUpdates(modified_config, default_config)

    # Log all the differences
    logger = DiffLogger(infile, basefile, output_filename, output_filename is not None, parseable)
    has_diff = False
    for option in natural_sort(list(set(default_config.getPathsWithOptionals()) | set(modified_config.getPathsWithOptionals()))):
        modified_value = modified_config.getValue(option)
        default_value = default_config.getValue(option)
        output_value = output_config.getValue(option)
        if modified_value is None and default_value is not None or \
                modified_value is not None and default_value is None or \
                (modified_value is not None and default_value is not None and modified_value != default_value):
            logger.addDiff(option, modified_value, default_value, output_value)
            has_diff = True
    if has_diff:
        logger.printTable()

    # If the user is asking to merge the files, then do that now.
    if output_filename is not None:
        if print_warnings:
            logger.printWarnings()
        format_adapter(output_filename).generateConfig(basefile, output_config)

def printHelp():
    print("########################################################")
    print("#######  Welcome to Config File Comparison Tool  #######")
    print("########################################################")
    print()
    print("This tool will compare two configuration files, a 'modified' configuration")
    print("file and a 'baseline' configuration file.  The 'modified' configuration file")
    print("is the file that is considered to be the one that has been updated to match")
    print("the specifics of the user's envrionment.  The 'baseline' configuration file")
    print("is considered to be the file as it would appear in a fresh install.")
    print("The comparison will output items that have changed since a fresh install,")
    print("and which items no longer appear in the modified file.")
    print()
    print("When merging, the same output will be provided as above, but the compare.py")
    print("will also write out a merged file which has only config values that exist")
    print("in the baseline, with values from the modified file unless no matching value")
    print("exists - at which point, the value is pulled in from the baseline.")
    print()
    print("Supported formats: .ini, .conf, .config, .py, and .json")
    print()
    print("Options:")
    print("         --create-backup: Will create a backup of any existing output file at <output-file-name>.yyyy-mm-dd_HH.MM.SS")
    print("         --parseable: Removes the header and column padding - columns are tab delimited.  Use to progammatically ingest output.")
    print("         --print-warnings: When specifying a merged output_file, this will let the user know if there are items that should be reviewed.")
    print()
    print("  Usage: ./gpudb_config_compare.py [options] modified_file baseline_file [output_file]")
    print()
    print("Example: /opt/gpudb/core/bin/gpudb_config_compare.py ./gpudb.conf ./gpudb.conf.template ./gpudb.conf.new")
    print()

# Main program
def run():
    current_arg=1
    create_backup = False
    parseable = False
    print_warnings = False
    while current_arg < len(sys.argv) and sys.argv[current_arg].startswith("-"):
        arg = sys.argv[current_arg]
        if arg == "-h" or arg == "--help":
            printHelp()
            sys.exit(0)
        elif arg == "--parseable":
            parseable = True
        elif arg == "--print-warnings":
            print_warnings = True
        elif arg == "--create-backup":
            create_backup = True
        elif arg == "--":
            curent_arg += 1
            break
        else:
            print("ERROR: Unknown argument '%s'." % arg)
            sys.exit(1)

        current_arg += 1

    if len(sys.argv) < current_arg + 2:
        printHelp()
        print("\n  Error: Please provide paths to both input and baseline files")
        sys.exit(1)

    infile = sys.argv[current_arg]
    basefile = sys.argv[current_arg + 1]
    output_filename = None

    if current_arg + 2 < len(sys.argv):
        output_filename = sys.argv[current_arg + 2]
        if os.path.isfile(output_filename) and create_backup:
            shutil.copyfile(output_filename, output_filename + "." + datetime.datetime.now().strftime("%Y-%m-%d_%H.%M.%S"))
    elif print_warnings:
        printHelp()
        print()
        print("ERROR: Cannot print warnings if we are not merging to an output")
        sys.exit(1)

    if not os.path.isfile(infile) or not os.path.isfile(basefile):
        printHelp()
        print("\n  Error: Could not read input and/or baseline file(s)")
        sys.exit(1)

    filename, fileextension = os.path.splitext(infile)

    # Determine comparison engine based on file extenion
    if fileextension in VALIDINI:
        doCompareAndOutput(infile, basefile, output_filename, IniFileAdapter, parseable, print_warnings)
    elif fileextension in VALIDPY:
        doCompareAndOutput(infile, basefile, output_filename, PyFileAdapter, parseable, print_warnings)
    elif fileextension in VALIDJS:
        doCompareAndOutput(infile, basefile, output_filename, JsFileAdapter, parseable, print_warnings)
    elif fileextension in VALIDJSON:
        doCompareAndOutput(infile, basefile, output_filename, JsonFileAdapter, parseable, print_warnings)
    else:
        print("Error: Unrecognized config file type. We support .config, .conf, .json, .js, .ini, and .py\n")
        sys.exit(1)

# Run program
if __name__ == "__main__":
    run()
