"""Support classess and functions for the Kinetica Machine Learning API."""

import os
import base64
import binascii
import json
import datetime
import uuid
import socket
import sys

#  from starlette.authentication import requires
#  from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.authentication import (
    AuthenticationBackend,
    AuthenticationError,
    SimpleUser,
    UnauthenticatedUser,
    AuthCredentials
)

from kml.api.kml_config import (
    get_profiler_configs,
    running_in_k8s
)

use_profiler = get_profiler_configs()

PROFILE_OUT_DIR = './profile_out'
if not os.path.exists(PROFILE_OUT_DIR):
    os.makedirs(PROFILE_OUT_DIR)

ROT_LOG_MAX_BYTES = 128 * 1024 * 1024
ROT_LOG_COUNT = 5


class BasicAuthBackend(AuthenticationBackend):
    """Implement basic API auth using the Starlette framework."""

    async def authenticate(self, request):
        import gpudb
        from kml.api.kml_config import get_db_conn_str

        #  print('==== aux.py ====')
        if 'Authorization' in request.headers:
            auth = request.headers['Authorization']

            try:
                scheme, credentials = auth.split()
                if scheme.lower() != 'basic':
                    raise AuthenticationError('''Only 'basic' auth supported''')
                decoded = base64.b64decode(credentials).decode('ascii')

            except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
                raise AuthenticationError('Invalid basic auth credentials')

            username, _, password = decoded.partition(':')

            # Attempt authenticaton from Kinetica or external LDAP
            db_conn_str = get_db_conn_str()
            try:
                _ = gpudb.GPUdb(
                    host=db_conn_str,
                    primary_host=db_conn_str,
                    username=username,
                    password=password
                )
                return AuthCredentials(['authenticated']), SimpleUser(username)
            except Exception as e:
                #  print('==== Catching in aux.py ====')
                raise AuthenticationError(e)

        else:
            db_conn_str = get_db_conn_str()
            try:
                _ = gpudb.GPUdb(host=db_conn_str, primary_host=db_conn_str)
                return AuthCredentials(['authenticated']), UnauthenticatedUser()
            except Exception as e:
                #  print('==== Catching in aux.py (else) ====')
                print(e)
                raise AuthenticationError(e)


def attach_log(module='unknown_module', log_name='kml', debug=False):
    """Setup the log file for the calling module."""

    import logging
    from logging.handlers import RotatingFileHandler
    import os
    from kml.api.kml_config import get_log_level

    # Instantiate the root logger
    module = module.replace('kml.api.', '')
    logger = logging.getLogger(module)
    log_level = get_log_level()

    if log_level == 'critical':
        logger.setLevel(logging.CRITICAL)
    elif log_level == 'error':
        logger.setLevel(logging.ERROR)
    elif log_level == 'warning':
        logger.setLevel(logging.WARNING)
    elif log_level == 'info':
        logger.setLevel(logging.INFO)
    elif log_level == 'debug':
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.ERROR)

    # Create the stdout file handler
    console_handler = logging.StreamHandler(sys.stdout)
    if not running_in_k8s():
        host = socket.gethostname()
        path = f"{os.path.dirname(os.path.realpath(__file__))}/logs"
        log_name = f"{log_name}-{host}.log"
        log_name_fq = '/'.join([path, log_name])

        file_handler = RotatingFileHandler(
            log_name_fq,
            mode='a',
            maxBytes=ROT_LOG_MAX_BYTES,
            backupCount=ROT_LOG_COUNT
        )
    else:
        file_handler = None

    # Create a logging format
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)s  (KML) %(name)s - %(message)s'
    )

    # Add the handlers to the logger
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    if file_handler:
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger


def log_comms(inbound_request, inbound_tag=''):
    """Write request and response payloads to disk for debugging."""

    if use_profiler:

        try:
            uniqifier = str(uuid.uuid1())[:8]
            timestamp = datetime.datetime.now().strftime('%Y.%m.%d__%H.%M.%S')

            fname = os.path.join(PROFILE_OUT_DIR,
                                 f'{timestamp}__{inbound_tag}__{uniqifier}.json')

            with open(fname, 'w') as jout:
                if inbound_request:
                    jout.write(json.dumps(inbound_request, indent=4, sort_keys=False))
                else:
                    jout.write('{}')
        except Exception as e:
            print('Exception logging payloads')
            print(e)

    else:
        pass


def json_validator(in_dict):
    """Check JSON compliance of staged response payload."""

    try:
        in_json = json.dumps(in_dict)
        out_dict = json.loads(in_json)
        if in_dict == out_dict:
            return True

    except ValueError as e:
        print(e)


class RespError():
    """Error sub-class."""
    def __init__(self):
        self.message = 'an error message'
        self.code = 'an error code'


class Responder():
    """HTTP standard REST response."""

    def __init__(self):
        """Initialization and instantiate classes."""
        import timeit

        self.success = False
        self.error = RespError()
        self.response = {}
        self.start_time = timeit.default_timer()

    def succeeded(self):
        """Toggle endpoint success flag."""
        self.success = True

    def failed(self, msg=None, code=None):
        """Set error message and failure code."""
        self.error.message = msg
        self.error.code = code
        logger.info(f'{code}: {msg}')

    def emit(self, http_code=911):
        """Fomat JSON response and return it to the host."""
        import timeit
        from starlette.responses import JSONResponse
        #  from starlette_json import SimpleJsonResponse
        import simplejson

        processing_time = round(timeit.default_timer() - self.start_time, 5)

        # Populate the JSON REST response
        ret = {'success': self.success,
               'processing_time': processing_time,
               'response': self.response}

        # Only add errors to response if there are any
        if not self.success:
            ret.update(error={'message': str(self.error.message),
                              'code': self.error.code})

        #  return JSONResponse(ret, status_code=http_code)
        #  json_validator(ret)
        try:
            return JSONResponse(ret, status_code=http_code)
        except Exception as e:
            logger.error(e)
            clean_dict = simplejson.loads(simplejson.dumps(ret, ignore_nan=True))
            #  print(clean_dict)
            return JSONResponse(clean_dict, status_code=http_code)


def get_k8_secrets_hash(bbx_container):
    """Create a secrets hash from the docker location."""
    import hashlib

    sanitized = bbx_container

    if sanitized[0:7] == 'http://':
        sanitized = sanitized[7:]
    elif sanitized[0:8] == 'https://':
        sanitized = sanitized[8:]

    if sanitized[len(sanitized)-1] == '/':
        sanitized = sanitized.strip('/')

    slash_index = -1 if '/' not in sanitized else sanitized.index('/')
    colon_index = -1 if ':' not in sanitized else sanitized.index(':')

    if colon_index < slash_index:
        sanitized = sanitized[0:colon_index] + sanitized[slash_index:]

    return hashlib.md5(sanitized.encode('utf-8')).hexdigest()


# Local Scope:

# Log errors propogated through the response class
logger = attach_log(module='response', log_name='kml', debug=True)
