"""KML app-store endpoint controller."""
import os
import traceback
import json
import datetime
import tempfile
import base64
import copy

from kml.api.kml_config import get_docker_daemon_uri

from kml.api.orm import KCredential

import uuid
import time

import gpudb

from kml.api.orm import (
    Entity,
    ModelDeployment
)
from kml.api.aux import (
    attach_log,
    Responder,
    log_comms
)
from kml.api.kml_config import (
    get_db_conn,
    get_namespace,
    get_db_conn_str,
    get_k8_conn,
    get_pod_resource_defaults,
    get_sys_schemas
)
from kml.api.kmldbops import (
    do_refresh_mv as kml_do_refresh_mv,
    get_tbl_handle as kml_get_tbl_handle,
    get_qualified_table_name
)

# Logging module
logger = attach_log(module=__name__, log_name='kml', debug=True)

from kml.api.kml_config import get_db_conn, get_docker_daemon_uri, get_blackbox_sdk
db = get_db_conn()

DOCKER_DAEMON_URI = get_docker_daemon_uri()
blackbox_sdk_cache, blackbox_sdk_repo, blackbox_sdk_repo_branch = get_blackbox_sdk()

CTNR_FQFN_SPEC = "/opt/gpudb/kml/spec.json"

SCHEMA_SYS, SCHEMA_AUDIT, _, _ = get_sys_schemas()

# Read config
db = get_db_conn()
namespace = get_namespace()
db_conn_str = get_db_conn_str()

k8_apps_v1_api, k8_core_v1_api, _, _, _ = get_k8_conn()

# Defaults
# TODO: this should be configurable
jupyter_store = 'sys_kml_user_data'

CTNR_PULL_SECONDS_BEFORE_ABANDON = 60*60
CTNR_PULL_SECONDS_BETWEEN_POLLS = 15

def delete_k8s_dep(dep_name):
    from kubernetes.client import V1DeleteOptions
    try:
        _ = k8_apps_v1_api.delete_namespaced_deployment(
            name=dep_name,
            namespace=namespace,
            body=V1DeleteOptions(propagation_policy='Foreground',
                                     grace_period_seconds=0)
        )
    except Exception as e:
        logger.warning(f"Error terminating deployment {dep_name}")
        logger.warning(str(e))


def get_events_single_pod(deploy_name):
    """Track container launch and set terminal state."""

    rollout_tm_out = 120
    init_tm_out = 120
    init_c = 0
    rollout_c = 0
    po_found = False

    while not po_found and init_c < init_tm_out:
        time.sleep(1)
        pods = k8_core_v1_api.list_namespaced_pod(namespace)

        for pod in pods.items:
            if pod.metadata.name.startswith(deploy_name):
                po_name = pod.metadata.name
                po_found = True
                logger.debug(f"Pod {po_name} initializing")
        init_c += 1

    if not po_found:
        return False, f"Pod would not initialize in {init_tm_out}s"

    while rollout_c < rollout_tm_out:
        time.sleep(1)
        po_desc = k8_core_v1_api.read_namespaced_pod(
            name=po_name,
            namespace=namespace
        )

        try:
            if po_desc.status.container_statuses[0].state.running:
                logger.info(f"Pod {po_name} running")
                return True, 'Running'
            elif po_desc.status.container_statuses[0].state.waiting.reason == 'ErrImagePull':
                logger.error(f"Pod {po_name} rollout failed: ErrImagePull")
                msg = str(po_desc.status.container_statuses[0].state.waiting.message)
                return False, 'ErrImagePull'
            elif po_desc.status.container_statuses[0].state.waiting.reason == 'ImagePullBackOff':
                logger.error(f"Pod {po_name} rollout failed: ImagePullBackOff")
                msg = str(po_desc.status.container_statuses[0].state.waiting.message)
                return False, 'ImagePullBackOff'
            elif po_desc.status.container_statuses[0].state.waiting.reason == 'CrashLoopBackOff':
                logger.error(f"Pod {po_name} rollout failed: CrashLoopBackOff")
                msg = str(po_desc.status.container_statuses[0].state.waiting.message)
                return False, 'CrashLoopBackOff'

        except Exception as _:
            logger.warning(f"Pod {po_name} state undetectable")
            pass

        rollout_c += 1


def ctnr_image_pull_failed(pod_of_interest):

    for _ in range(int(CTNR_PULL_SECONDS_BEFORE_ABANDON/CTNR_PULL_SECONDS_BETWEEN_POLLS)):
        logger.debug("Pre-pulling container: in progress...")
        time.sleep(CTNR_PULL_SECONDS_BETWEEN_POLLS)
        try:
            pod_list = k8_core_v1_api.list_namespaced_pod(namespace)

            for pod in pod_list.items:
                if pod.metadata.name.startswith(pod_of_interest):
                    if pod.status.phase == "Running":
                        logger.info("Pre-pulling container: successful")
                        return True
                    if pod.status.container_statuses:
                        if len(pod.status.container_statuses) > 0:
                            if pod.status.container_statuses[0]:
                                if pod.status.container_statuses[0].state:
                                    if pod.status.container_statuses[0].state.waiting:
                                        if pod.status.container_statuses[0].state.waiting.reason == "CrashLoopBackOff":
                                            logger.info("Pulling container: failed w/ CrashLoopBackOff")
                                            logger.error(pod.status)
                                            return False
                                        if pod.status.container_statuses[0].state.waiting.reason == "ImagePullBackOff":
                                            logger.info("Pulling container: failed w/ ImagePullBackOff")
                                            logger.error(pod.status)
                                            return False    
                    if pod.status.init_container_statuses:
                        if len(pod.status.init_container_statuses) > 0:
                            if pod.status.init_container_statuses[0]:
                                if pod.status.init_container_statuses[0].state:
                                    if pod.status.init_container_statuses[0].state.waiting:
                                        if pod.status.init_container_statuses[0].state.waiting.reason == "CrashLoopBackOff":
                                            logger.info("Pulling container: failed w/ CrashLoopBackOff")
                                            logger.error(pod.status)
                                            return False
                                        if pod.status.init_container_statuses[0].state.waiting.reason == "ImagePullBackOff":
                                            logger.info("Pulling container: failed w/ ImagePullBackOff")
                                            logger.error(pod.status)
                                            return False 
                                     
        except KeyboardInterrupt:
            break
        except Exception as e:
            logger.error("Pre-pulling container: error getting container image status")
            logger.exception(e)
    logger.info(f"Pre-pulling container: abandoned after {CTNR_PULL_SECONDS_BEFORE_ABANDON} seconds")
    return False


# https://github.com/prafull01/Kubernetes-Utilities/blob/master/kubectl_cp_as_python_client.py
def read_file_from_pod(pod_name, src_path, namespace="default"):
    import tarfile
    import tempfile
    import traceback
    from kubernetes.stream import stream
    from tempfile import TemporaryFile

    _, k8_core_v1_api, _, _, _ = get_k8_conn()
    exec_command = ['tar', 'cf', '-', '--absolute-names', src_path]

    try:
        with TemporaryFile() as tar_buffer:
            resp = stream(k8_core_v1_api.connect_get_namespaced_pod_exec, pod_name, namespace,
                        command=exec_command,
                        stderr=True, stdin=True,
                        stdout=True, tty=False,
                        _preload_content=False)

            while resp.is_open():
                resp.update(timeout=1)
                if resp.peek_stdout():
                    out = resp.read_stdout()
                    # print("STDOUT: %s" % len(out))
                    tar_buffer.write(out.encode('utf-8'))
                if resp.peek_stderr():
                    print("STDERR: %s" % resp.read_stderr())
            resp.close()

            tar_buffer.flush()
            tar_buffer.seek(0)

            with tarfile.open(fileobj=tar_buffer, mode='r:') as tar:
                for member in tar.getmembers():
                    if member.isdir():
                        continue
                    _, tempfilename = tempfile.mkstemp()
                    tar.makefile(member, tempfilename)
                    with open(tempfilename,"rb") as f:
                        data = f.read().decode()
                        return data
    except Exception as e:
        logger.error(e)
        logger.debug(f"Unable to read from path {src_path}")
        #logger.debug(traceback.format_exc())
        return None

async def container_prepull(request):
    from kml.api.utils import string_is_not_blank

    response = Responder()
    try:
        post_inputs = await request.json()
        if 'container_ref' not in post_inputs:
            msg = f"Missing container reference parameter ('container_ref')"
            logger.debug(msg)
            response.failed(msg=msg, code=100201)
            return response.emit(400)

        container_ref = post_inputs['container_ref']  # fully qualified
        logger.info(f"Pre-pulling container: {container_ref}")

        if not string_is_not_blank(container_ref):
            msg = f"Missing container reference"
            logger.debug(msg)
            response.failed(msg=msg, code=100201)
            return response.emit(400)

        from kml.api.yaml_customizer import customize_yaml_ctnr_prepull
        myyaml = customize_yaml_ctnr_prepull(container_ref)
        api_response = k8_apps_v1_api.create_namespaced_daemon_set(body=myyaml, namespace=namespace)

        logger.info(f"Pre-pulling container: {container_ref} -- awaiting download")
        logger.info(f"Metadata Name: {api_response.metadata.name}")
        #  success_flag = ctnr_image_pull_failed(pod_of_interest=api_response.metadata.name)
        success_flag, status_msg = get_events_single_pod(api_response.metadata.name)
        if not success_flag:
            msg = f"Pre-pulling container failed for {container_ref} - {status_msg}"
            logger.debug(msg)
            response.failed(msg=msg, code=100201)
            response.response['name'] = api_response.metadata.name
            return response.emit(400)
        logger.info(f"Pre-pulling container: {container_ref} -- completed")
        response.response['name'] = api_response.metadata.name
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


async def container_introspect(request):

    from kml.api.utils import string_is_not_blank, is_json

    nm_ctnr_dep_temp = None
    response = Responder()
    try:
        post_inputs = await request.json()
        if 'container_ref' not in post_inputs:
            msg = f"Missing container reference parameter ('container_ref')"
            logger.debug(msg)
            response.failed(msg=msg, code=100201)
            return response.emit(400)

        container_ref = post_inputs['container_ref']  # fully qualified
        if not string_is_not_blank(container_ref):
            msg = f"Blank container reference parameter ('container_ref')"
            logger.debug(msg)
            response.failed(msg=msg, code=100201)
            return response.emit(400)

        secrets_handle = None
        if "credential_id_docker" in post_inputs:
            logger.info("Credential ID for Private Registry")
            credential_id_docker = post_inputs["credential_id_docker"]

            if credential_id_docker == -1 or credential_id_docker in ["null", "none"]:
                logger.info(f"Ignoring empty credential_id_docker: "
                            f"{credential_id_docker}")
            else:
                from kml.api.orm import KCredential
                if not KCredential.exists(existing_credential_id=credential_id_docker):
                    msg = f"Referenced Credential ID {credential_id_docker} not found on system"
                    logger.error(msg)
                    response.failed(msg=msg, code=100201)
                    return response.emit(404)

                else:
                    logger.info(f"Utilizing Docker Credential with id "
                                f"{credential_id_docker}")

                if KCredential.load(
                        existing_credential_id=credential_id_docker
                ).credential_type != "DOCKER":
                    msg = f"Referenced Credential ID {credential_id_docker} not a Docker credential"
                    logger.error(msg)
                    response.failed(msg=msg, code=100201)
                    return response.emit(400)
                else:
                    # TODO: show some descriptive stuff here about the private repo, like the URI
                    logger.info(f"Successfully loaded Docker Private Registry Credential "
                                f"via id {credential_id_docker}")
                    secrets_handle = KCredential.load(
                        existing_credential_id=credential_id_docker
                    ).get_handle()

        search_location = CTNR_FQFN_SPEC
        if 'spec_location' in post_inputs:
            spec_location = post_inputs["spec_location"]
            msg = f"Seeking spec in user-provided location {spec_location}"
            logger.debug(msg)
            search_location = spec_location
        else:
            logger.debug(f"Seeking spec in standard location {search_location}")

        logger.info(f"Pulling container for introspection: {container_ref}")

        from kml.api.yaml_customizer import customize_yaml_ctnr_introspect
        myyaml = customize_yaml_ctnr_introspect(container_ref, secrets_handle = secrets_handle)
        api_response = k8_apps_v1_api.create_namespaced_deployment(body=myyaml, namespace=namespace)
        nm_ctnr_dep_temp = api_response.metadata.name

        logger.info(f"Pulling container: {container_ref} -- awaiting download")
        logger.info(f"Metadata Name: {nm_ctnr_dep_temp}")
        #  success_flag = ctnr_image_pull_failed(pod_of_interest=nm_ctnr_dep_temp)
        success_flag, status_msg = get_events_single_pod(nm_ctnr_dep_temp)
        if not success_flag:
            msg = f"Pulling container failed for {container_ref} - {status_msg}"
            logger.debug(msg)
            #  delete_k8s_dep(dep_name=nm_ctnr_dep_temp)
            response.failed(msg=msg, code=100201)
            return response.emit(404)
        logger.info(f"Pulling container: {container_ref} -- completed")
        response.response['name'] = nm_ctnr_dep_temp
        from kml.api.k8_manager import get_all_deployments_familytree
        ftree = get_all_deployments_familytree()
        pname = ftree[nm_ctnr_dep_temp][0]

        response.response['manifest'] = {}
        specs_found = False
        try:
            logger.debug(f"Searching for spec at {search_location}")
            content = read_file_from_pod(pod_name=pname, 
                src_path=search_location, 
                namespace=namespace)
            if content:
                if is_json(content):
                    response.response['manifest'] = json.loads(content)
                else:
                    response.response['manifest'] = content
                specs_found = True
                
        except Exception as e:
            logger.debug(f"Error encountered pulling file for key {skey}")
            logger.error(e)
            logger.debug(traceback.format_exc())

        delete_k8s_dep(dep_name=nm_ctnr_dep_temp)

        if specs_found:
            response.succeeded()
            return response.emit(200)
        else:
            response.failed(
                msg=f"No introspectable specifications found at {search_location}",
                code=1040
            )
            return response.emit(404)
        return response.emit(404)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        if nm_ctnr_dep_temp:
            delete_k8s_dep(dep_name=nm_ctnr_dep_temp)

        response.failed(msg=e, code=1040)        
        return response.emit(509)
