"""KML audit endpoint controller."""

import traceback
import datetime
import base64

from starlette.authentication import AuthenticationError
import gpudb

from kml.api.kml_config import get_db_conn, get_sys_schemas
from kml.api.aux import attach_log, Responder
from kml.api.orm import Audit, ModelInstance, ModelDeployment, FeatureSet, Dataset
from kml.api.kmldbops import get_tbl_handle as kml_get_tbl_handle
from kml.api.kmldbops import has_table as kml_has_table


logger = attach_log(module=__name__, log_name='kml', debug=True)
SCHEMA_SYS, SCHEMA_AUDIT, _, _ = get_sys_schemas()
db = get_db_conn()

MULTI_COL_OUT = ['onehot', 'octetsplit']
INT_TYPES = ['int', 'long']
FP_TYPES = ['float', 'double']
STR_TYPES = ['string']
NUM_OPS = ['>', '<', '=', '<=', '>=']
STR_OPS = ['=', '~=']


def get_audit_fields(request):
    """Return the searchable inference fields."""

    response = Responder()

    try:
        mi_id = request.path_params.get('model_instance_id')
        mi = ModelInstance.load(mi_id)

        terms = []
        all_searchable_fields = (mi.model_inst_config['input_record_type'] +
                                 mi.model_inst_config['output_record_type'])

        for col in all_searchable_fields:
            col_scm = col['col_type'].split('|')
            type_ = col_scm[0]
            if len(col_scm) > 1:
                props = col_scm[1]
            else:
                props = []

            if type_ in (INT_TYPES + FP_TYPES) and 'store_only' not in props:
                terms.append({'field': col['col_name'], 'ops': NUM_OPS})

            if type_ in STR_TYPES and 'store_only' not in props:
                terms.append({'field': col['col_name'], 'ops': STR_OPS})

        response.succeeded()
        response.response.update(auditable_fields=terms,
                                 model_instance_id=mi_id)

        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def create_search_expr(audit_config, mi_id):
    """Generate a compound expression string."""

    filters = []
    dtype = {}

    try:
        mi = ModelInstance.load(mi_id)
        all_searchable_fields = (mi.model_inst_config['input_record_type'] +
                                 mi.model_inst_config['output_record_type'])
        for i in all_searchable_fields:
            dtype[i['col_name']] = i['col_type'].split('|')[0]

        # Build expression filter: search parameters
        if audit_config['fields']:
            for term in audit_config['fields']:

                field = term['field']
                op = term['operator']
                val = term['value']

                # String expression composer
                if dtype.get(field) in STR_TYPES:
                    if op not in STR_OPS:
                        raise ValueError(f"Invalid operator '{op}' for string "
                                         f"field '{field}'")
                    if op == '=':
                        filters.append(f'LCASE({field}) {op} LCASE("{val}")')
                    elif op == '~=':
                        filters.append(f'LIKE(LCASE({field}), LCASE("%{val}%"))')

                # Integer expression composer
                elif dtype.get(field) in INT_TYPES:
                    if op not in NUM_OPS:
                        raise ValueError(f"Invalid operator '{op}' for numeric "
                                         f"field '{field}'")
                    filters.append(f'{field} {op} {val}')

                # Floating point expression composer
                elif dtype.get(field) in FP_TYPES:
                    if op not in NUM_OPS:
                        raise ValueError(f"Invalid operator '{op}' for numeric "
                                         f"field '{field}'")
                    if op == '=':
                        filters.append(f'abs({field} - {val}) < 0.00001')
                    else:
                        filters.append(f'{field} {op} {val}')

                else:
                    raise ValueError(f"Invalid field '{field}' for model ID {mi_id}")

        # Build expression filter: date parameters
        if audit_config['dates']:
            start = audit_config['dates']['start']
            end = audit_config['dates']['end']
            filters.append(f'process_end_dt >= "{start}"')
            filters.append(f'process_end_dt <= "{end}"')

        # Create filter
        expr = {'expression': ' and '.join(filters)}

        logger.info(f'Search expression: {expr}')
        return expr

    except Exception as e:
        raise ValueError(e)


async def get_dep_matches(request):
    """Instantiate an audit object."""

    response = Responder()
    CONFIG_PARAMS = ['fields', 'dates']

    try:
        spec = await request.json()
        mi_id = request.path_params.get('model_instance_id')

        # Validation
        if 'audit_config' not in spec:
            msg = "Bad create request: missing Audit creation payload 'audit_config'"
            logger.error(msg)
            response.failed(msg=msg, code=1040)
            return response.emit(422)

        for key in CONFIG_PARAMS:
            if key not in spec['audit_config']:
                msg = f'Bad create request: missing Audit creation parameter ({key})'
                logger.error(msg)
                response.failed(msg=msg, code=1040)
                return response.emit(422)

        # Generate search expression
        audit_config = spec['audit_config']
        expr = create_search_expr(audit_config=spec['audit_config'],
                                  mi_id=mi_id)
        audit_config.update(search_expression=expr)

        # Create a record of search for internal security auditing
        dt = datetime.datetime.utcnow().isoformat(' ', 'seconds')

        # TODO: put this in a utils file
        if 'Authorization' in request.headers:
            auth = request.headers['Authorization']
            scheme, credentials = auth.split()

            if scheme.lower() != 'basic':
                raise AuthenticationError('Auth header is not "basic"')

            decoded = base64.b64decode(credentials).decode('ascii')
            profile, _, _ = decoded.partition(':')
        else:
            profile = 'anonymous'

        data_audit = Audit.create(
            audit_name='KML audit record',
            audit_desc=f'Created by user: {profile} on {dt}',
            audit_config=audit_config
        )
        data_audit.publish()
        data_audit.get_base_entity().transition_to_created()

        # Perform inital count of matching inferences by deployment
        dep_search = []
        dep_tbl_hdl = kml_get_tbl_handle('sys_kml_model_depl',
                                         schema=SCHEMA_SYS)

        resp = dep_tbl_hdl.get_records_by_column(
            column_names=['entity_id'],
            options={'expression': f'base_model_inst_id={mi_id}'}
        )

        if resp:
            md_ids = resp['entity_id']
        else:
            msg = f'Model instance {mi_id} has no deployments'
            logger.warning(msg)
            response.failed(msg=msg, code=1040)
            return response.emit(404)

        for md_id in md_ids:

            md = ModelDeployment.load(md_id)
            aud_tbl_name = md.model_dep_config['out-tablemonitor']['table_name']

            # Check for audit output table existence. Note that schema is embeded
            # in the table name so audit collection schema is not specified.
            if not kml_has_table(aud_tbl_name, schema=None):
                logger.warning(f"Missing audit output table {aud_tbl_name}")
                continue

            resp = db.get_records(table_name=aud_tbl_name,
                                  options=expr,
                                  offset=0,
                                  limit=0)

            if resp['status_info']['status'] == 'OK':
                n_matches = resp['total_number_of_records']
            else:
                raise ValueError(resp['status_info']['message'])

            total_inf = gpudb.GPUdbTable(name=aud_tbl_name, db=db).count

            dep_search.append({'deployment_id': md_id,
                               'deployment_name': md.base_entity.name,
                               'deployment_desc': md.base_entity.desc,
                               'deployment_type': md.deployment_type,
                               'inference_matches': n_matches,
                               'total_inferences': total_inf})

        response.succeeded()
        response.response.update(deployment_matches=dep_search,
                                 audit_id=data_audit.get_entity_id(),
                                 model_instance_id=mi_id)

        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def get_inferences(request):
    """Return inference data for a specific deployment."""

    response = Responder()

    try:
        offset = request.query_params.get('offset')
        limit = request.query_params.get('limit')
        offset = int(offset) if offset else 0
        limit = int(limit) if limit else 0

        md_id = request.path_params.get('model_deployment_id')
        md = ModelDeployment.load(md_id)

        audit_id = request.path_params.get('audit_id')
        audit = Audit.load(audit_id)
        expr = audit.audit_config['search_expression']

        resp = db.get_records_and_decode(
            table_name=md.model_dep_config['out-tablemonitor']['table_name'],
            options=expr,
            limit=limit,
            offset=offset
        )

        response.succeeded()
        response.response.update(
            matches_returned=len(resp['records']),
            limit=limit,
            offset=offset,
            total_number_of_records=resp['total_number_of_records'],
            audit_id=audit_id,
            model_deployment_id=md_id,
            inferences=resp['records']
        )

        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def generate_audit_graph(request):
    """Return audit graph data for a specific inference."""

    response = Responder()
    nodes = []

    try:
        guid = request.path_params.get('guid')
        md_id = request.path_params.get('model_deployment_id')
        md = ModelDeployment.load(md_id)

        # Load featureset, pre-assume blank FS
        fs = {'config_inline': {}, 'config_relational': {}}
        try:
            fs = FeatureSet.load(md.featureset_id)
        except Exception as e:
            logger.warning(e)

        node_id = 0
        src_node_list = []
        outputs = []
        mi_config = md.base_model_inst.model_inst_config
        _ = [outputs.append(x['col_name']) for x in mi_config['output_record_type']]

        resp = db.get_records_by_column_and_decode(
            table_name=md.model_dep_config['out-tablemonitor']['table_name'],
            column_names=outputs,
            options={'expression': f'guid="{guid}"'}
        )

        if resp['status_info']['status'] == 'OK':
            inf_data = resp['records']
        else:
            raise ValueError(resp['status_info']['message'])

        if not resp['records']:
            raise ValueError(
                f"Unable to find GUID {guid} in table "
                f"{md.model_dep_config['out-tablemonitor']['table_name']}"
            )

        # Inference components
        for output in outputs:
            nodes.append({
                'node_id': node_id,
                'node_type': 'inference',
                'node_children': None,
                'node_details': {
                    'name': output,
                    'value': inf_data[output][0]
                }
            })
            node_id += 1

        # Model deployment
        nodes.append({
            'node_id': node_id,
            'node_type': 'deployment',
            'node_children': list(range(0, node_id)),
            'node_details': {
                'name': md.base_entity.name,
                'description': md.base_entity.desc,
                'mode': md.deployment_type,
                'md_id': md.get_entity_id()
            }
        })
        node_id += 1

        # Model instance
        nodes.append({
            'node_id': node_id,
            'node_type': 'model',
            'node_children': [node_id - 1],
            'node_details': {
                'name': md.base_model_inst.base_entity.name,
                'description': md.base_model_inst.base_entity.desc,
                'function': md.base_model_inst.model_inst_config['blackbox_function'],
                'module': md.base_model_inst.model_inst_config['blackbox_module'],
                'container': md.base_model_inst.model_inst_config['container'],
                'mi_id': md.base_model_inst_id
            }
        })
        model_node_id = node_id
        node_id += 1

        if fs and fs.config_inline:
            # Inline engineered features

            inputs = []
            _ = [inputs.append(x) for x in fs.config_inline['feat_mem']]
            n_features = len(inputs) - 1

            ds = Dataset.load(fs.config_inline['fitting_ds_id'])

            terms = set()
            for input_ in inputs:
                if fs.config_inline['feat_mem'][input_]['function_xform'] not in MULTI_COL_OUT:
                    terms.add(fs.config_inline['feat_mem'][input_]['outschema'][0][0])
                terms.add(input_)
            terms = list(terms)

            resp = db.get_records_by_column_and_decode(
                table_name=md.model_dep_config['out-tablemonitor']['table_name'],
                column_names=terms,
                options={'expression': f'guid="{guid}"'}
            )

            if resp['status_info']['status'] == 'OK':
                feat_data = resp['records']
            else:
                raise ValueError(resp['status_info']['message'])

            if not resp['records']:
                raise ValueError(
                    f"Unable to find GUID {guid} in table "
                    f"{md.model_dep_config['out-tablemonitor']['table_name']}"
                )

            for input_ in inputs:

                # Engineered feature
                xform_type = fs.config_inline['feat_mem'][input_]['function_xform']
                fe_out_name = fs.config_inline['feat_mem'][input_]['outschema'][0][0]
                if xform_type in MULTI_COL_OUT:
                    xform_val = '-'
                    fe_out_name = f"{input_}_is_*"
                else:
                    xform_val = feat_data[fe_out_name]

                nodes.append({
                    'node_id': node_id,
                    'node_type': 'engineered_feature',
                    'node_children': [model_node_id],
                    'node_details': {
                        'fit_to': ds.base_entity.name,
                        'transform': xform_type,
                        'name': fe_out_name,
                        'value': xform_val,
                        'fs_id': md.featureset_id
                    }
                })
                node_id += 1

                # Source features
                src_node_list.append(n_features + node_id)
                nodes.append({
                    'node_id': n_features + node_id,
                    'node_type': 'source_data',
                    'node_children': [node_id - 1],
                    'node_details': {
                        'name': input_,
                        'value': feat_data[input_][0]
                    }
                })

            node_id = n_features + node_id + 1

        else:
            # Features (base inputs with no feature engineering)

            inputs = []
            _ = [inputs.append(x['col_name']) for x in mi_config['input_record_type']]

            resp = db.get_records_by_column_and_decode(
                table_name=md.model_dep_config['out-tablemonitor']['table_name'],
                column_names=inputs,
                options={'expression': f'guid="{guid}"'}
            )

            if resp['status_info']['status'] == 'OK':
                feat_data = resp['records']
            else:
                raise ValueError(resp['status_info']['message'])

            if not resp['records']:
                raise ValueError(
                    f"Unable to find GUID {guid} in table "
                    f"{md.model_dep_config['out-tablemonitor']['table_name']}"
                )

            for input_ in inputs:
                src_node_list.append(node_id)
                nodes.append({
                    'node_id': node_id,
                    'node_type': 'feature',
                    'node_children': [model_node_id],
                    'node_details': {
                        'name': input_,
                        'value': feat_data[input_][0]
                    }
                })
                node_id += 1

        if fs and fs.config_relational:
            # Relational transforms (SQL)
            nodes.append({
                'node_id': node_id,
                'node_type': 'relational_transforms',
                'node_children': src_node_list,
                'node_details': {
                    'name': fs.config_relational['mv_name'],
                    'ddl': fs.config_relational['mv_ddl'],
                    'fs_id': md.featureset_id
                }
            })

        else:
            # Base table with TM
            nodes.append({
                'node_id': node_id,
                'node_type': 'input_table',
                'node_children': src_node_list,
                'node_details': {
                    'name': md.model_dep_config['source_table']
                }
            })

        response.succeeded()
        response.response.update(node_map=nodes,
                                 guid=guid,
                                 model_deployment_id=md_id)

        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def audit_view(request):
    """Retrieve details of a specific audit."""

    response = Responder()

    try:
        audit_id = request.path_params.get('audit_id')
        audit = Audit.load(audit_id)

        response.response['item'] = audit.serialize()
        response.response['found'] = True
        response.succeeded()

        return response.emit(http_code=200)

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)
