"""KML KCredential endpoint controllers."""

import traceback
import base64
import json

from kml.api.orm import KCredential
from kml.api.aux import (
    attach_log,
    Responder
)

logger = attach_log(module=__name__, log_name='kml', debug=True)


async def credential_create(request):

    import kubernetes.client

    from kml.api.kml_config import get_k8_conn, get_namespace
    from kml.api.utils import helper_get_secrets_name_list

    response = Responder()
    try:
        spec = await request.json()
        # TODO: Loop over or provide common function for these
        # validation and string strip blocks

        if "credential_name" not in spec:
            response.failed(
                msg=("Bad create request: missing Credential Name "
                     "parameter (credential_name)"),
                code=1040
            )
            return response.emit(422)
        else:
            spec['credential_name'] = spec['credential_name'].strip()

        if "credential_desc" not in spec:
            logger.info('Credential description not provided. Using credential '
                        'name as the description')
            spec.update(credential_desc=spec.get('credential_name').strip())
        else:
            spec['credential_desc'] = spec['credential_desc'].strip()

        if "credential_type" not in spec:
            response.failed(
                msg=("Bad create request: missing Credential Type "
                     "parameter (credential_type)"),
                code=1040
            )
            return response.emit(422)

        if "credential_config" not in spec:
            response.failed(
                msg=("Bad create request: missing Credential Config "
                     "parameter (credential_config)"),
                code=1040
            )
            return response.emit(422)
        else:
            for key in spec['credential_config'].keys():
                spec['credential_config'][key] = spec['credential_config'][key].strip()

        ptemp = KCredential.create(
            credential_name=spec["credential_name"],
            credential_desc=spec["credential_desc"],
            credential_type=spec["credential_type"],
            credential_config=spec["credential_config"]
        )

        if spec["credential_type"] == 'DOCKER':
            logger.debug("Docker Credential Create: validating required fields")

            # TODO: Use get method on dict key to avoid empty and undef vals

            if "registry_uri" not in spec["credential_config"]:
                # Private Docker Registry FQDN. (https://index.docker.io/v1/ for DockerHub)
                response.failed(
                    msg=("Bad create request for Private Docker Registry Credentials: "
                         "missing FQDN (registry_uri)"),
                    code=1040
                )
                # TODO: also validate FQDN string, not sure how!
                return response.emit(422)

            if "user" not in spec["credential_config"]:
                response.failed(
                    msg=("Bad create request for Private Docker Registry Credentials: "
                         "missing user (user)"),
                    code=1040
                )
                return response.emit(422)

            if "password" not in spec["credential_config"]:
                response.failed(
                    msg=("Bad create request for Private Docker Registry Credentials: "
                         "missing password (password)"),
                    code=1040
                )
                return response.emit(422)

            if "email" not in spec["credential_config"]:
                response.failed(
                    msg=("Bad create request for Private Docker Registry Credentials: "
                         "missing email (email)"),
                    code=1040
                )
                return response.emit(422)

            # Private Docker Registry FQDN. (https://index.docker.io/v1/ for DockerHub)
            docker_priv_reg_fqdn = spec["credential_config"]["registry_uri"]
            docker_priv_reg_user = spec["credential_config"]["user"]
            docker_priv_reg_pass = spec["credential_config"]["password"]
            docker_priv_reg_email = spec["credential_config"]["email"]

            namespace = get_namespace()

            logger.debug(f"Docker Credential Create: Creating k8s secret for "
                         f"{docker_priv_reg_user} on {docker_priv_reg_fqdn}")

            # TODO: Supress not found erros by first checking for existance
            logger.debug(f"Deleting previous versions of secret {ptemp.get_handle()} "
                         f"in case they exist")

            secret_name = ptemp.get_handle()
            _, k8_core_v1_api, _, _, _ = get_k8_conn()

            secrets_names_list = helper_get_secrets_name_list()
            if secret_name in secrets_names_list:
                k8_core_v1_api.delete_namespaced_secret(
                    secret_name,
                    namespace,
                    body=kubernetes.client.V1DeleteOptions()
                )

            # Create the secret
            secret_contents = dict()
            secret_contents["auths"] = dict()
            secret_contents["auths"][docker_priv_reg_fqdn] = dict()
            secret_contents["auths"][docker_priv_reg_fqdn]["username"] = docker_priv_reg_user
            secret_contents["auths"][docker_priv_reg_fqdn]["password"] = docker_priv_reg_pass
            secret_contents["auths"][docker_priv_reg_fqdn]["email"] = docker_priv_reg_email

            secret_contents["auths"][docker_priv_reg_fqdn]["auth"] = base64.encodebytes(
                bytearray(docker_priv_reg_user + ":" + docker_priv_reg_pass, "ascii")
            ).decode("ascii").strip()

            secret_name = ptemp.get_handle()

            encoded_contents = base64.encodebytes(
                bytearray(json.dumps(secret_contents), "ascii")
            ).decode("ascii").strip()

            secret = kubernetes.client.V1Secret()
            secret.metadata = kubernetes.client.V1ObjectMeta(name=secret_name)
            secret.type = "kubernetes.io/dockerconfigjson"
            secret.data = {".dockerconfigjson": encoded_contents}

            logger.debug("Creating k8s secret")
            create_resp = k8_core_v1_api.create_namespaced_secret(
                namespace=namespace,
                body=secret
            )
            #  logger.debug(create_resp)

        ptemp.publish()
        ptemp.get_base_entity().transition_to_created()

        response.response["created"] = ptemp.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)


def credential_view(request):
    response = Responder()
    response.response["found"] = False

    try:
        credential_id = request.path_params.get('credential_id')
        ptemp = KCredential.load(credential_id)
        response.response["item"] = ptemp.serialize()
        response.response["found"] = True
        response.succeeded()
        return response.emit(200)

    except LookupError as le:
        logger.error(le)
        logger.debug(traceback.format_exc())
        response.failed(msg=le, code=1010)
        return response.emit(404)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)


def credential_archive(request):
    response = Responder()
    response.response["found"] = False

    try:
        credential_id = request.path_params.get('credential_id')
        c = KCredential.load(credential_id)
        c.tombstone()
        response.response['item'] = c.serialize()
        response.succeeded()
        return response.emit(200)

    except LookupError as le:
        logger.error(le)
        logger.debug(traceback.format_exc())
        response.failed(msg=le, code=1010)
        return response.emit(404)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)
