"""KML Dataset  endpoint controller."""

import traceback
import json
import tempfile

from kml.api.aux import (
    attach_log,
    Responder
)
from kml.api.orm import Dataset
from kml.api.kml_config import (
    get_db_conn,
    get_sys_schemas
)
from kml.api.kmldbops import (
    get_tbl_handle as kml_get_tbl_handle,
    create_table as kml_create_table
)

SCHEMA_SYS, SCHEMA_AUDIT, SCHEMA_SYS_ARCH, SCHEMA_AUDIT_ARCH = get_sys_schemas()
db = get_db_conn()
logger = attach_log(module=__name__, log_name='kml', debug=True)


async def dataset_testtrainsplit(request):
    import random
    import string

    response = Responder()
    try:
        spec = await request.json()

        if 'src_tbl' not in spec:
            response.failed(
                msg='Bad create request: missing source table (src_tbl)',
                code=1040
            )
            return response.emit(422)
        # TODO: also ensure this table exists and is populated

        if 'pct_train' not in spec:
            response.failed(
                msg='Bad create request: missing train percentage (pct_train)',
                code=1040
            )
            return response.emit(422)
        # TODO: also ensure this is a numeric field

        if 'pct_test' not in spec:
            response.failed(
                msg='Bad create request: missing test percentage (pct_test)',
                code=1040
            )
            return response.emit(422)
        # TODO: also ensure this is a numeric field

        src_tbl = spec['src_tbl']
        pct_train = float(spec['pct_train'])
        pct_test = float(spec['pct_test'])
        if pct_train + pct_test > 1.0:
            response.failed(
                msg='Bad create request: Train+Test percentages over 100%',
                code=1040
            )
            return response.emit(422)

        h_tbl_src = kml_get_tbl_handle(src_tbl)

        logger.info(f"Dataset Source Table Count: {h_tbl_src.count}")
        if h_tbl_src.count == 0:
            response.failed(
                msg=f"Bad create request: Source table {src_tbl} has no records",
                code=1040
            )
            return response.emit(422)
        if h_tbl_src.count < 10:
            logger.warning(f"Suspicious create request: Source table {src_tbl} has "
                           f"very few records ({h_tbl_src.count}), results may not "
                           f"be as expected")

        h_tbl_src_tbl_type = h_tbl_src.get_table_type()

        mycode = ''.join(random.choice(string.ascii_uppercase) for i in range(6))

        tbl_name_train = f"{src_tbl}_{mycode}_train"
        tbl_name_test = f"{src_tbl}_{mycode}_test"

        h_tbl_new_train = kml_create_table(
            tbl_type_list=h_tbl_src_tbl_type,
            tbl_name=tbl_name_train
        )
        h_tbl_new_test = kml_create_table(
            tbl_type_list=h_tbl_src_tbl_type,
            tbl_name=tbl_name_test
        )

        logger.debug(f"Creating train/test datasets from source "
                     f"table {src_tbl} of size {h_tbl_src.count}")
        BATCH_SIZE = 10000
        batches = (h_tbl_src.count // BATCH_SIZE) + 1
        for batch_num in range(batches):
            set_train = []
            set_test = []

            # TODO: IMPORTANT: Batch this up by 20k records a pop
            thisbatch = db.get_records_and_decode(
                table_name=src_tbl,
                offset=BATCH_SIZE * batch_num,
                limit=BATCH_SIZE,
                options={}
            )

            logger.debug(
                f"Handling batch {batch_num+1} of {batches} with count "
                f"{len(thisbatch.records)} from total {thisbatch.total_number_of_records}"
            )

            for t in thisbatch.records:
                if random.random() < pct_train:
                    set_train.append(t)
                elif random.random() <= pct_train + pct_test:
                    set_test.append(t)

            if len(set_train) == 0:
                response.failed(
                    msg=(f"Bad create request: Train set ended up with zero records, "
                         f"the source dataset is probably too small"),
                    code=1040
                )
                return response.emit(422)

            if len(set_test) == 0:
                response.failed(
                    msg=(f"Bad create request: Train set ended up with zero records, "
                         f"the source dataset is probably too small"),
                    code=1040
                )
                return response.emit(422)

            train_insert_status = db.insert_records(
                table_name=tbl_name_train,
                data=set_train
            )
            if 'status_info' in train_insert_status:
                if 'status' in train_insert_status["status_info"]:
                    if train_insert_status["status_info"]["status"] != "OK":
                        logger.error(train_insert_status["status_info"])

            test_insert_status = db.insert_records(
                table_name=tbl_name_test,
                data=set_test
            )
            if 'status_info' in test_insert_status:
                if 'status' in test_insert_status["status_info"]:
                    if test_insert_status["status_info"]["status"] != "OK":
                        logger.error(test_insert_status["status_info"])

        logger.debug("Creating training dataset")
        train_Dataset = Dataset.create(
            dataset_name=f"Training set for table {src_tbl}",
            dataset_desc=f"Training set for table {src_tbl}",
            source_config={
                "tbl": h_tbl_new_train.name,
                "cols": [c.name for c in h_tbl_src_tbl_type.columns],
                "filterex": ""
            }
        )
        logger.debug("Creating testing dataset")
        test_Dataset = Dataset.create(
            dataset_name=f"Testing set for table {src_tbl}",
            dataset_desc=f"Testing set for table {src_tbl}",
            source_config={
                "tbl": h_tbl_new_test.name,
                "cols": [c.name for c in h_tbl_src_tbl_type.columns],
                "filterex": ""
            }
        )

        logger.debug("Publishing training dataset")
        train_Dataset.publish()

        logger.debug("Publishing testing dataset")
        test_Dataset.publish()

        logger.debug(f"Created train datasets (ID: {train_Dataset.get_entity_id()}) "
                     f"from source table {src_tbl}-->{tbl_name_train} of size "
                     f"{h_tbl_new_train.count}")
        logger.debug(f"Created test datasets (ID: {test_Dataset.get_entity_id()}) "
                     f"from source table {src_tbl}-->{tbl_name_test} of size "
                     f"{h_tbl_new_test.count}")

        response.response['created'] = {}
        response.response['created']['train'] = train_Dataset.serialize()
        response.response['created']['test'] = test_Dataset.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


async def dataset_create(request):
    response = Responder()
    try:
        spec = await request.json()

        if 'dataset_name' not in spec:
            response.failed(
                msg='Bad create request: missing Dataset Name (dataset_name)',
                code=1040
            )
            return response.emit(422)

        if 'dataset_desc' not in spec:
            response.failed(
                msg='Bad create request: missing Dataset Description (dataset_desc)',
                code=1040
            )
            return response.emit(422)

        if 'source_config' not in spec:
            response.failed(
                msg='Bad create request: missing Dataset schema (source_config)',
                code=1040
            )
            return response.emit(422)

        if spec['source_config'].get('tbl'):
            spec['source_config']['tbl'] = spec['source_config']['tbl'].strip()
        else:
            response.failed(
                msg='Bad create request: missing Dataset table name (source_config::tbl)',
                code=1040
            )
            return response.emit(422)

        data_Dataset = Dataset.create(
            dataset_name=spec['dataset_name'].strip(),
            dataset_desc=spec['dataset_desc'].strip(),
            source_config=spec['source_config']
        )
        data_Dataset.publish()
        data_Dataset.get_base_entity().transition_to_created()

        response.response['created'] = data_Dataset.serialize()
        response.succeeded()

        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def dataset_view(request):
    response = Responder()
    response.response['found'] = False

    try:
        dataset_id = request.path_params.get('dataset_id')
        retrieved_dataset_obj = Dataset.load(dataset_id)
        response.response['item'] = retrieved_dataset_obj.serialize()
        response.response['found'] = True
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def dataset_clone(request):
    response = Responder()

    try:
        dataset_id = request.path_params.get('dataset_id')
        eve2 = Dataset.clone(existing_dataset_id=dataset_id)
        eve2.get_base_entity().transition_to_created()
        response.response['created'] = eve2.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def dataset_preview(request):

    # TODO: Implement filter allowing not-exactly-7-items preview, already
    # supported by ORM

    # TODO: put a giant try-catch here to separate out the failure of nonexistent
    # Dataset vs bad filter expression

    response = Responder()
    retrieved_dataset_obj = None

    try:
        dataset_id = request.path_params.get('dataset_id')
        retrieved_dataset_obj = Dataset.load(dataset_id)
        preview_set = retrieved_dataset_obj.preview()
        response.response['items'] = preview_set
        response.response['count'] = len(preview_set)
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)

        # A Failure here means the filter expression on the Dataset is probably bad
        # We'll set it to FAILED
        # TODO: Need to determine whether this transition is reasonable
        if retrieved_dataset_obj:
            retrieved_dataset_obj.get_base_entity().transition_to_failed()

        return response.emit(509)


def dataset_grab(request):
    from starlette.responses import Response

    response = Responder()
    try:
        dataset_id = request.path_params.get('dataset_id')
        retrieved_dataset_obj = Dataset.load(dataset_id)
        _, tempfilename = tempfile.mkstemp()
        retrieved_dataset_obj.grab(tempfilename)

        with open(tempfilename, "rb") as f:
            data = f.read()
            FILE_RETURN_HEADER = {}
            FILE_RETURN_HEADER['Access-Control-Allow-Methods'] = 'HEAD, GET'
            FILE_RETURN_HEADER['Access-Control-Allow-Origin'] = '*'
            FILE_RETURN_HEADER['Content-Type'] = 'application/csv'
            FILE_RETURN_HEADER['Content-Disposition'] = 'attachment; filename="'+json.loads(retrieved_dataset_obj.source_config)["tbl"] + ".csv" + '"'
            return Response(data, status_code=200, headers=FILE_RETURN_HEADER)
            # TODO: How do we use "emit" concept here?

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def dataset_archive(request):
    response = Responder()

    try:
        dataset_id = request.path_params.get('dataset_id')
        retrieved_dataset_obj = Dataset.load(dataset_id)
        retrieved_dataset_obj.tombstone()
        response.response['item'] = retrieved_dataset_obj.serialize()
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)
