"""KML model deployment endpoint controllers."""

import traceback
import json
import collections
import datetime
import time

import gpudb
import requests

from kml.api.aux import (
    attach_log,
    log_comms,
    Responder
)
from kml.api.kml_config import (
    get_db_conn,
    get_db_conn_str,
    get_api_base,
    get_k8_conn,
    get_components,
    get_namespace,
    get_db_credentials,
    get_sys_schemas,
    get_pod_resource_defaults,
    get_public_connections,
    running_in_k8s,
    get_ingress_controller_details
)
from kml.api.orm import (
    ModelInstance,
    ModelDeployment,
    FeatureSet,
    Dataset
)
from kml.api.kmldbops import (
    get_tbl_handle as kml_get_tbl_handle,
    has_table as kml_has_table,
    create_table as kml_create_table,
    create_schema_or_collection as kml_create_schema_or_collection,
    check_if_exists_schema_or_collection as kml_check_if_exists_schema_or_collection
)
from kml.api.utils import (
    get_k8_resources,
    get_public_k8s_ip
)

# Configuration
db = get_db_conn()
_, api_host, api_port = get_components(conn='api')
_, db_host, db_port = get_components(conn='db')
namespace = get_namespace()
db_conn_str = get_db_conn_str()
public_api_base, public_db_conn, public_tm_conn = get_public_connections()
_, tm_host, tm_port = get_components('zmq')

DEFAULT_ZMQ_PORT_DEALER = 9009
DEFAULT_ZMQ_FRAMES_PER_BATCH = 10000

SCHEMA_SYS, SCHEMA_AUDIT, _, _ = get_sys_schemas()

# Import credentials from the config file
config_user, config_pass = get_db_credentials()
if config_user and config_pass:
    db_user = config_user
    db_pass = config_pass
else:
    db_user = 'no_cred'
    db_pass = 'no_cred'

if db_user != 'no_cred' and db_pass != 'no_cred':
    from kml.api.controller_utils import check_db_credentials
    check_db_credentials(db_user, db_pass)

# Instantiate logger
logger = attach_log(module=__name__, log_name='kml', debug=True)


def validate_deployment(md):
    """Check required resources in the deployment metadata."""

    response = Responder()

    if (md.deployment_type == 'BATCH' or
            md.deployment_type == 'CONTINUOUS' and
            md.model_dep_config.get('src-tablemonitor')):

        # NOTE: Continuous relational tranform and on-demand configurations are
        # excluded from these checks as they lack input tables

        # Check table existence

        if not md.model_dep_config['source_table']:
            return 'Input table name not specified'

        if not kml_has_table(md.model_dep_config['source_table']):
            # Secondary check, should never get here
            return (
                f"Input table {md.model_dep_config['source_table']} "
                f"not found on database"
            )

        if not md.model_dep_config['sink_table']:
            return 'Output table name not specified'

        if not kml_has_table(md.model_dep_config['sink_table']):
            # Secondary check, should never get here
            return (
                f"Output table {md.model_dep_config['sink_table']} "
                f"not found on database"
            )

    if md.model_dep_config.get('src-tablemonitor'):

        # Check input table for required model features
        # only for non-relational pathway configurations

        req_features = []
        input_cols = []

        if md.base_model_inst.model_inst_config.get('input_record_type'):
            for feature in md.base_model_inst.model_inst_config['input_record_type']:
                req_features.append(feature['col_name'])
        else:
            return 'Model configuration missing input record type'

        if md.model_dep_config['src-tablemonitor']['type_schema'].get('fields'):
            for in_tbl_schema in md.model_dep_config['src-tablemonitor']['type_schema']['fields']:
                input_cols.append(in_tbl_schema['name'])
        else:
            return 'Model configuration missing source fields'

        inp_tbl_hdl = kml_get_tbl_handle(md.model_dep_config['source_table'])
        tbl_cols = list(inp_tbl_hdl.show_table()['properties'][0].keys())
        
        for req_feature in req_features:
            if req_feature not in input_cols:
                return (
                    f"Feature '{req_feature}' not present in input table"
                )
            if req_feature not in tbl_cols:
                # Secondary check, should never get here
                return f"Feature '{req_feature}' not present in input table"

    # If not returned by this point, all parameters are correct
    return None


def set_pod_resources(deploy_details):
    """Create a dict of resource requests and limits for the deployment manifest."""

    (DEFAULT_CPU_REQUEST,
     DEFAULT_CPU_LIMIT,
     DEFAULT_GPU_REQUEST,
     DEFAULT_MEM_REQUEST,
     DEFAULT_MEM_LIMIT) = get_pod_resource_defaults()

    def set_to_default(desc, default_val, units=''):
        logger.info(
            f"Creating Deployment: {desc} unspecified, defaulting "
            f"to {default_val}{units}"
        )
        return default_val

    def param_error(desc, val):
        msg = f"{desc} of {val} is invalid, must be a positive integer value"
        raise ValueError(msg)

    def force_set_limit(desc, val_to_set, units=''):
        logger.info(
            f"Creating Deployment: {desc} limit can not exceed enforced "
            f"limit, setting specified limit to {val_to_set}{units}"
        )
        return val_to_set

    def force_set_request(desc, val_to_set, units=''):
        logger.info(
            f"Creating Deployment: {desc} can not exceed specified limit, "
            f"setting request to the limit of {val_to_set}{units}"
        )
        return val_to_set

    # Set compute target
    compute_target = deploy_details.get('compute_target')
    if compute_target and compute_target.upper() in ['CPU', 'GPU']:
        compute_target = compute_target.upper()
    else:
        logger.info(
            f"Creating Deployment: Unspecified or erroneous compute target "
            f"{deploy_details.get('compute_target')}, defaulting to 'CPU'"
        )
        compute_target = 'CPU'

    # Set compute resource requests and limits (CPU)
    if compute_target == 'CPU':

        gpu_limit = None

        cpu_request = deploy_details.get('cpu_compute_target_allocation') \
            if deploy_details.get('cpu_compute_target_allocation') \
            else set_to_default('CPU request target', DEFAULT_CPU_REQUEST, 'm')

        cpu_limit = deploy_details.get('cpu_compute_limit') \
            if deploy_details.get('cpu_compute_limit') \
            else set_to_default('CPU request limit', DEFAULT_CPU_LIMIT, 'm')

        if not isinstance(cpu_request, int) or cpu_request < 1:
            param_error('CPU request target', cpu_request)

        if not isinstance(cpu_limit, int) or cpu_limit < 1:
            param_error('CPU request limit', cpu_limit)

        if cpu_limit > DEFAULT_CPU_LIMIT:
            cpu_limit = force_set_limit('CPU compute request', DEFAULT_CPU_LIMIT, 'm')

        if cpu_request > cpu_limit:
            cpu_request = force_set_request('CPU compute request', cpu_limit, 'm')

    # Set compute resource requests and limits (NVIDIA GPU)
    if compute_target == 'GPU':

        cpu_request = None
        cpu_limit = None

        # NOTE: Unlike with CPU, when specifing number of GPUs to use, the request is the limit
        if deploy_details.get('gpu_compute_limit'):
            logger.warning(
                "Ingnoring parameter 'gpu_compute_limit', when specifing GPGPU compute, "
                "the requested number of compute devices is the limit"
            )

        gpu_limit = deploy_details.get('gpu_compute_target_allocation') \
            if deploy_details.get('gpu_compute_target_allocation') \
            else set_to_default('GPU request target/limit', DEFAULT_GPU_REQUEST)

        if not isinstance(gpu_limit, int) or gpu_limit < 1:
            param_error('GPU request target', gpu_limit)

        if gpu_limit > DEFAULT_GPU_REQUEST:
            gpu_limit = force_set_limit('GPU compute request/limit', DEFAULT_GPU_REQUEST)

    # Set compute resource requests and limits (Memory)
    mem_request = deploy_details.get('memory_target_allocation') \
        if deploy_details.get('memory_target_allocation') \
        else set_to_default('Memory request', DEFAULT_MEM_REQUEST, 'Mi')

    mem_limit = deploy_details.get('memory_limit') \
        if deploy_details.get('memory_limit') \
        else set_to_default('Memory limit', DEFAULT_MEM_LIMIT, 'Mi')

    if not isinstance(mem_request, int) or mem_request < 1:
        param_error('Memory request', mem_request)

    if not isinstance(mem_limit, int) or mem_limit < 1:
        param_error('Memory limit', mem_limit)

    if mem_limit > DEFAULT_MEM_LIMIT:
        mem_limit = force_set_limit('Memory', DEFAULT_MEM_LIMIT, 'Mi')

    if mem_request > mem_limit:
        mem_request = force_set_request('Memory request', mem_limit, 'Mi')

    return {
        'compute_target': compute_target,
        'cpu_request': cpu_request,
        'mem_request': mem_request,
        'cpu_limit': cpu_limit,
        'gpu_limit': gpu_limit,
        'mem_limit': mem_limit
    }


async def modelinstance_deploy(request):
    """Primary K8s deployment bootstrap function."""

    from collections import OrderedDict

    import copy
    import requests

    from kml.api.naming_svc import get_audit_table_names
    from kml.api.systable_spec import TBL_SCHEMA_inf_input, TBL_SCHEMA_inf_output

    response = Responder()
    response.response["deployed"] = False

    try:

        model_instance_id = request.path_params.get('model_instance_id')
        logger.info(f"Creating Deployment: Using underlying Model Instance {model_instance_id}")
        mi = ModelInstance.load(model_instance_id)

        # TODO: consider the case of NULL payloads where json parsing will fail
        spec = await request.json()
        log_comms(spec, "inbound_mi_deploy")
        deploy_details = {}
        if spec:
            deploy_details = spec
        else:
            raise ValueError("Received deployment create call with empty payload")

        deployment_name = mi.base_entity.name
        if 'deployment_name' in deploy_details:
            deployment_name = deploy_details["deployment_name"].strip()
        else:
            logger.warning(f"Creating Deployment: Received deployment request w/o "
                           f"deployment_name, we'll carry it over from the model "
                           f"instance: {deployment_name}")

        deployment_desc = mi.base_entity.desc
        if 'deployment_desc' in deploy_details:
            deployment_desc = deploy_details["deployment_desc"].strip()
        else:
            logger.warning(f"Creating Deployment: Received deployment request w/o "
                           f"deployment_desc, we'll carry it over from the model "
                           f"instance: {deployment_desc}")

        fs_details = None
        if "featureset_id" in spec:
            logger.info(f"Creating Deployment: Creating BlackBox Model with user-provided "
                        f"FeatureSet")
            featureset_id = spec["featureset_id"]
            fs_details = FeatureSet.load(featureset_id)
            logger.info(f"Creating Deployment: Utilizing existing FeatureSet with id "
                        f"{featureset_id}")
        else:
            logger.info(f"Creating Deployment: Creating BlackBox Model with sys-generated "
                        f"stub-identity FeatureSet")
            fs_details = FeatureSet.create(
                featureset_name=f"Identity featureset stub for dep {deployment_name}",
                featureset_desc=(f"Identity featureset stub for dep {deployment_name}: "
                                 f"{deployment_desc}"),
                config_inline={},
                config_relational={}
            )
            logger.info(f"Creating Deployment: Utilizing stub identity featureset with "
                        f"id {fs_details.get_entity_id()}")

        replicas = 1
        if 'replicas' in deploy_details:
            if not isinstance(deploy_details['replicas'], int):
                msg = (f"Creating Deployment: Desired replica count specified but received "
                       f"non-integer parameter")
                raise ValueError(msg)

            if deploy_details['replicas'] == 0:
                msg = f"Creating Deployment: Desired replica count of zero is invalid"
                raise ValueError(msg)

            replicas = int(deploy_details['replicas'])
            logger.info(f"Creating Deployment: Setting to {replicas} desired worker replicas")
        else:
            logger.info(f"Creating Deployment: Intent to use single worker replica "
                        f"(no replica argument received, using default of 1)")

        user_provided_env_vars = []
        if 'env_vars' in deploy_details:
            user_provided_env_vars = deploy_details["env_vars"]
            if len(user_provided_env_vars) == 0:
                logger.info(f"Creating Deployment: No user-provided environment variables "
                            f"provided")
            else:
                for anev in user_provided_env_vars:
                    if "var_name" in anev and "var_value" in anev:
                        logger.info(f"Creating Deployment: Using user proved env var: "
                                    f"{anev['var_name']}: {anev['var_value']}")
                    else:
                        raise ValueError(f"Creating Deployment: User provided environment "
                                         f"variable provided w/o required keys var_name "
                                         f"and var_value")

        # Set the pod resource requests/limits to be injected into the manifest on deployment
        pod_resources = set_pod_resources(deploy_details)

        # Initialize custom path parameter for ondemand deployment.
        # This is overridden if one is provided and the deployment mode is ONDEMAND
        custom_path_slug = None

        blackbox_module = "bb_module_default"
        blackbox_function = "blackbox_function_default"
        if "blackbox_module" in mi.model_inst_config:
            # TODO: ensure this isnt empty string and not null
            blackbox_module = mi.model_inst_config["blackbox_module"]
        if "blackbox_function" in mi.model_inst_config:
            blackbox_function = mi.model_inst_config["blackbox_function"]
            # TODO: ensure this isnt empty string and not null

        deployment_mode = "ON_DEMAND"
        if 'deployment_mode' not in deploy_details:
            logger.warning(f"Creating Deployment: No deployment mode argument received, "
                           f"defaulting to ON-DEMAND")
            logger.info(f"Creating Deployment: Attempting deployment in deployment "
                        f"{deployment_mode} mode")

        elif 'deployment_mode' in deploy_details:
            deployment_mode = deploy_details['deployment_mode'].strip()
            if deployment_mode in ['ON_DEMAND', 'CONTINUOUS', 'BATCH']:
                logger.info(f"Creating Deployment: Attempting deployment in deployment "
                            f"{deployment_mode} mode")
            else:
                msg = (f"Creating Deployment: Encountered unknown deployment mode "
                       f"{deployment_mode}, deployment attempt aborted")
                logger.error(msg)
                response.failed(msg, code=1422)
                return response.emit(422)

        sink_table = None
        source_table = None
        ZMQ_FRAMES_PER_BATCH = -1

        if deployment_mode == "ON_DEMAND":

            if fs_details.config_relational:
                msg = f"Creating Deployment: You provided a FeatureSet with Relational Features, those are only supported in CONTINUOUS mode"
                logger.error(msg)
                response.failed(msg=msg, code=1422)
                return response.emit(422)

            if 'source_table' in deploy_details:
                msg = f"Creating Deployment: Received on-demand mode deployment request w/ parameter source_table, which does not make sense"
                logger.error(msg)
                response.failed(msg=msg, code=1422)
                return response.emit(422)

            if 'sink_table' in deploy_details:
                msg = f"Creating Deployment: Received on-demand mode deployment request w/ parameter sink_table, which does not make sense"
                logger.error(msg)
                response.failed(msg=msg, code=1422)
                return response.emit(422)

            custom_path_slug = (
                deploy_details.get('custom_path_slug').strip()
                if deploy_details.get('custom_path_slug')
                else None
            )

        if deployment_mode == "CONTINUOUS":
            if 'source_table' not in deploy_details:
                if not fs_details.config_relational:
                    msg = (f"Creating Deployment: Received continuous mode deployment request w/o required "
                           "parameter source_table or relational transform config in featureset")
                    logger.error(msg)
                    response.failed(msg=msg, code=1422)
                    return response.emit(422)
            else:
                deploy_details["source_table"] = deploy_details["source_table"].strip()
                source_table = deploy_details["source_table"]
                if not kml_has_table(tbl_name=source_table):
                    msg = f"Creating Deployment: input source table {source_table} specified but does not exist"
                    logger.error(msg)
                    response.failed(msg=msg, code=1422)
                    return response.emit(422)

            if 'sink_table' not in deploy_details:
                msg = f"Creating Deployment: Received continuous mode deployment request w/o parameter sink_table, will create one automatically"
                logger.warning(msg)
                sink_table = f"{source_table}__model_out"
                logger.info(f"Creating Deployment: Setting sink table to {sink_table} in mode {deployment_mode}")
            else:
                deploy_details["sink_table"] = deploy_details["sink_table"].strip()
                sink_table = deploy_details["sink_table"]
                # TODO: If source table DOES exist, ensure schema matches, if not -- FAIL

        if deployment_mode == "BATCH":
            if fs_details.config_relational:
                msg = "You provided a FeatureSet with Relational Features, those are only supported in CONTINUOUS mode"
                logger.error(msg)
                response.failed(msg=msg, code=1422)
                return response.emit(422)
            if 'source_table' not in deploy_details:
                logger.error("Received batch mode deployment request w/o required parameter "
                             "source_table")
                response.failed(msg="Received batch mode deployment request w/o required parameter "
                                    "source_table",
                                code=1422)
                return response.emit(422)
            else:
                deploy_details["source_table"] = deploy_details["source_table"].strip()
                source_table = deploy_details["source_table"]

            if not kml_has_table(tbl_name=source_table):
                logger.error(f"Input source table {source_table} specified but does not exist")
                response.failed(msg=f"Input source table {source_table} specified but does "
                                    f"not exist",
                                code=1422)

            if 'sink_table' not in deploy_details:
                logger.warning("Received batch mode deployment request w/o parameter sink_table, "
                               "will create one automatically")
                sink_table = f"{source_table}__model_out"
                logger.info(f'Setting sink table to {sink_table} in mode {deployment_mode}')
            else:
                deploy_details["sink_table"] = deploy_details["sink_table"].strip()
                sink_table = deploy_details["sink_table"]
                # TODO: If sink_table table DOES exist, ensure schema matches, if not -- FAIL

            if 'zmq_frames_per_batch' in deploy_details:
                ZMQ_FRAMES_PER_BATCH = int(deploy_details["zmq_frames_per_batch"])
            else:
                ZMQ_FRAMES_PER_BATCH = DEFAULT_ZMQ_FRAMES_PER_BATCH

            logger.info(f"ZMQ frame size: {ZMQ_FRAMES_PER_BATCH}")

        PERSIST_AUDIT = "TRUE"
        if 'persist_audit' in deploy_details:
            if deploy_details['persist_audit'].upper() in ["FALSE", "NO"]:
                if deployment_mode == "ON_DEMAND":
                    logger.error("Cannot suppress audit outputs for ON_DEMAND mode")
                    response.failed(msg="Cannot suppress audit outputs for ON_DEMAND mode",
                                    code=1422)
                    return response.emit(422)
                PERSIST_AUDIT = "FALSE"
        logger.info(f"Creating Deployment: Audit Persistance {PERSIST_AUDIT}")

        USE_MULTIHEAD_IO = "FALSE"
        if 'multihead_io' in deploy_details:
            if deploy_details['multihead_io'].upper() in ["TRUE", "YES"]:
                if deployment_mode == "ON_DEMAND":
                    logger.error("Cannot use multihead IO with ON_DEMAND mode")
                    response.failed(msg="Cannot use multihead IO with ON_DEMAND mode",
                                    code=1422)
                    return response.emit(422)
                USE_MULTIHEAD_IO = "TRUE"
        logger.info(f"Creating Deployment: Multihead IO {USE_MULTIHEAD_IO}")

        model_dep_config__enriched = {
            "sink_table": sink_table,
            "source_table": source_table,
            "zmq_frames_per_batch": ZMQ_FRAMES_PER_BATCH,
            "user_provided_env_vars": user_provided_env_vars,
            "persist_audit": PERSIST_AUDIT,
            "multihead_io": USE_MULTIHEAD_IO,
            "custom_path_slug": custom_path_slug
        }

        pod_resources.update(replicas=replicas)

        md = ModelDeployment.create(
            deployment_name=deployment_name,
            deployment_desc=deployment_desc,
            base_model_inst_id=mi.get_entity_id(),
            featureset_id=fs_details.get_entity_id(),
            deployment_type=deployment_mode,
            model_dep_config=model_dep_config__enriched,
            resource_config=pod_resources
        )

        md.publish()
        md.get_base_entity().transition_to_created()
        md_id = md.get_entity_id()
        response.response["draft_entity_id"] = md_id
        logger.info(f"Creating Deployment: Draft deployment id {md_id}")

        # #####################################################################
        # STEP 1: Establish Sources and Sinks Tables Names
        # #####################################################################

        tbl_name_audit_input, tbl_name_audit_output = get_audit_table_names(
            in_deployment_id=md.get_entity_id()
        )

        # References/Pointers for table connections
        #tbl_handle_audit_input = None
        #tbl_handle_audit_output = None

        # #####################################################################
        # STEP 2: Create Input+Output Column Definitions (schema generation)
        # #####################################################################

        # DESIRED END RESULT
        columns_audit_input = OrderedDict()
        columns_audit_output = OrderedDict()

        # ------ FOUNDATIONAL FIELDS
        # Add Metadata columns
        for f in TBL_SCHEMA_inf_input:
            columns_audit_input[f[0]] = f
        for f in TBL_SCHEMA_inf_output:
            columns_audit_output[f[0]] = f

        for f in md.base_model_inst.model_inst_config["input_record_type"]:
            columns_audit_input[f['col_name']] = [f['col_name']] + f['col_type'].split('|')
            columns_audit_output[f['col_name']] = [f['col_name']] + f['col_type'].split('|')

        if fs_details.config_inline:
            logger.info(f"Creating Deployment {md_id}: Adding native transform schema:")
            for f in fs_details.config_inline['target_schema']:
                #logger.info(f'        {f}')
                columns_audit_input[f[0]] = f
                columns_audit_output[f[0]] = f
        else:
            logger.info(f"Creating Deployment {md_id}: No native transform identified. Proceeding without featureset")

        # ------ OUTPUT FIELDS
        # For BLACKBOX, output record schemas are user-defined
        for f in md.base_model_inst.model_inst_config["output_record_type"]:
            columns_audit_output[f['col_name']] = [f['col_name']] + f['col_type'].split('|')

        # Make all columns nullable/text_search
        for col_spec in [columns_audit_input, columns_audit_output]:
            for k in col_spec:
                if "nullable" not in col_spec[k]:
                    col_spec[k].append("nullable")
                if "str" in col_spec[k] and "text_search" not in col_spec[k]:
                    col_spec[k].append("text_search")

        schema_audit_in = [v for k,v in columns_audit_input.items()]
        schema_audit_out = [v for k,v in columns_audit_output.items()]
        schema_sink_table = schema_audit_out

        # #####################################################################
        # #####################################################################


        # #####################################################################
        # STEP 3: Create Audit Input+Output Tables and Reference for SRC Table
        # #####################################################################

        tbl_ref_source_data = None
        if source_table:
            tbl_ref_source_data = kml_get_tbl_handle(tbl_name=source_table)
        logger.info(f"Creating Deployment {md_id}: Creating Audit Input Table using Schema")
        tbl_ref_audit_input = kml_create_table(tbl_type_list=schema_audit_in,
                                               tbl_name=tbl_name_audit_input,
                                               schema=SCHEMA_AUDIT)

        logger.info(f"Creating Deployment {md_id}: Creating Audit Output (and Sink) Table using Schema")
        tbl_ref_audit_output = kml_create_table(tbl_type_list=schema_audit_out,
                                               tbl_name=tbl_name_audit_output,
                                               schema=SCHEMA_AUDIT)

        #schema_inbound = tbl_ref_audit_input.get_table_type().schema_string
        #schema_outbound = tbl_ref_audit_output.get_table_type().schema_string

        # #####################################################################
        # STEP 4: Create Output Tables
        # #####################################################################

        # Consider that on-demand deployment dont even have sink tables
        #   thus only do these checks if there is a sink table at play
        if sink_table:
            logger.info(f"Creating Deployment {md_id}: Checking to see if table {sink_table} exists")
            sink_table_existence =  kml_has_table(sink_table)
            if sink_table_existence:
                logger.info(f"Creating Deployment {md_id}: Sink table {sink_table} already exists")

                # TODO:Upgrage sink table instead of just throwing error

                sink_tbl_hdl = kml_get_tbl_handle(sink_table)
                existing_cols = [x.name for x in sink_tbl_hdl.get_table_type().columns]
                req_cols = []
                for col_sch in schema_sink_table:
                    req_cols.append(col_sch[0])
                for name in req_cols:
                    if name not in existing_cols:
                        msg = (
                            f"Output table already exists but is missing column '{name}'. "
                            f"Drop table and allow system to create it"
                        )
                        response.failed(msg, code=1422)
                        return response.emit(400)
            else:
                sink_table_split = sink_table.split('.')
                if len(sink_table_split) > 1:
                    if not kml_check_if_exists_schema_or_collection(sink_table_split[0]):
                        logger.info(f"Creating Deployment {md_id}: Sink table schema "
                                    f"{sink_table_split[0]} does not exits, creating")
                        _ = kml_create_schema_or_collection(sink_table_split[0])
                    else:
                        logger.info(f"Creating Deployment {md_id}: Sink table schema "
                                    f"{sink_table_split[0]} exits, proceeding")

                logger.info(f"Creating Deployment {md_id}: Sink table {sink_table} does not exist, creating")
                _ = kml_create_table(tbl_type_list=schema_sink_table,
                                     tbl_name=sink_table,
                                     schema=None)

        # #####################################################################
        # STEP 5: Create Table-Monitors
        # #####################################################################

        src_tm = None
        if source_table:
            src_tm = tbl_ref_source_data.create_table_monitor()
        inp_tm = tbl_ref_audit_input.create_table_monitor()
        out_tm = tbl_ref_audit_output.create_table_monitor()
        # force dict-ify and json-ify table monitor type schemas since the GPUDB library response
        # has strange newlines and formatting
        if source_table:
            src_tm["type_schema"] = json.loads(src_tm["type_schema"])
        inp_tm["type_schema"] = json.loads(inp_tm["type_schema"])
        out_tm["type_schema"] = json.loads(out_tm["type_schema"])

        # TODO: Is this required any longer?
        if source_table:
            src_tm.pop('info', None)  # Is this required any longer?
            src_tm.pop('status_info', None)  # Is this required any longer?
        inp_tm.pop('info', None)  # Is this required any longer?
        inp_tm.pop('status_info', None)  # Is this required any longer?
        out_tm.pop('info', None)  # Is this required any longer?
        out_tm.pop('status_info', None)  # Is this required any longer?

        #if source_table:
        #    schema_data_src = json.dumps(src_tm["type_schema"]).replace (" ","")
        #schema_inbound = json.dumps(inp_tm["type_schema"]).replace (" ","")
        #schema_outbound = json.dumps(out_tm["type_schema"]).replace (" ","")
        #if source_table:
        #    logger.info(f"Creating Table Monitor on source table {source_table.strip()}")
        logger.info(f"Creating Deployment {md_id}: Creating Table Monitor on audit in table {tbl_name_audit_input.strip()}")
        logger.info(f"Creating Deployment {md_id}: Creating Table Monitor on audit out table {tbl_name_audit_output.strip()}")

        #######################################################################

        # #####################################################################
        # STEP 6: Persist infrastructure to system tables
        # #####################################################################

        if source_table:
            model_dep_config__enriched["src-tablemonitor"] = src_tm
        else:
            model_dep_config__enriched["src-tablemonitor"] = None

        if inp_tm:
            model_dep_config__enriched["inp-tablemonitor"] = inp_tm
        if out_tm:
            model_dep_config__enriched["out-tablemonitor"] = out_tm

        md.set_config(model_dep_config__enriched)

        # #####################################################################
        # STEP 6.5: Create materialized view for input and check input columns
        # #####################################################################

        # TODO: Also add to FeatureSet Create
        # TODO: Check to see if the MV needs to be created at this point
        if fs_details.config_relational:
            #if not db.has_table(fs_details.config_relational['mv_name'])['table_exists']:
            if not kml_has_table(fs_details.config_relational['mv_name']):
                resp = db.execute_sql(fs_details.config_relational['mv_ddl'])

                if resp['status_info']['status'] != 'OK':
                    msg = f"Creating Deployment {md_id}: resp['status_info']['message']"
                    logger.error(msg)
                    raise ValueError(msg)

            mv_src_tbl_name = fs_details.config_relational['mv_name']
            #mv_src_tbl_hdl = gpudb.GPUdbTable(name=mv_src_tbl_name, db=db)
            mv_src_tbl_hdl = kml_get_tbl_handle(tbl_name=mv_src_tbl_name)
            mv_src_cols = {}
            for col in json.loads(mv_src_tbl_hdl.get_table_type().schema_string)['fields']:
                mv_src_cols[col['name']] = col['type']

            mv_timestamp_col = fs_details.config_relational['mv_timestamp_col']
            if mv_timestamp_col not in mv_src_cols.keys():
                msg = f"Creating Deployment {md_id}: Required timestamp column {mv_timestamp_col} is not present in the input materialized view table"
                logger.error(msg)
                raise ValueError(msg)

        # Validate deployment object
        msg = validate_deployment(md)
        if msg:
            response.failed(msg, code=1422)
            return response.emit(400)

        # #####################################################################
        # STEP 10: Respond to creator with created details
        # #####################################################################

        response.response["created"] = md.serialize()
        response.response["deployed"] = False
        _ = response.response.pop("draft_entity_id")
        response.succeeded()
        log_comms(response.response, "outbound_mi_deploy")
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def model_dep_stop(request):
    """Stop deployed model."""

    # Create and initialize a responder object
    response = Responder()

    try:
        dep_id = request.path_params.get('model_deployment_id')
        logger.info(f"Stopping Deployment {dep_id}: Retrieving details")
        dep = ModelDeployment.load(dep_id)

        if dep.base_entity.entity_state not in ['FAILED', 'STOPPED', 'COMPLETED']:
            logger.info(
                f"Stopping Deployment {dep_id}: Transitioning to STOPPED state"
            )
            dep.get_base_entity().transition_to_stopped()
        else:
            logger.info(
                f"Deployment {dep_id} is in {dep.base_entity.entity_state} state. "
                f"Attempting to terminate pods but not transitioning STOPPED."
            )

        logger.info(f"Stopping Deployment {dep_id}: Initiating teardown")
        model_deploy_teardown(dep_id)


        response.response['stopped'] = True
        response.response['deployment_id'] = dep_id
        logger.info(f"Stopping Deployment {dep_id}: Success")
        response.succeeded()
        return response.emit(http_code=200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)


def deploy_helper_pods(dep_id, dep_mode, env_vars, custom_path_slug=None, md=None):

    from kml.api.yaml_customizer import (
        customize_yaml_input_mgr,
        customize_yaml_depmgr
    )
    from kml.api.naming_svc import get_ondemand_default_slug

    NUM_POLL_TRIES_AWAITING_STARTUP = 100
    WAIT_TIME_BETWEEN_POLLS = 3

    def check_is_pod_running(pod_name):
        pod_list = k8_core_v1_api.list_namespaced_pod(namespace)
        for pod in pod_list.items:
            if pod.metadata.name.startswith(pod_name):
                if pod.status.phase == "Running":
                    logger.info(f"Deploying Model {dep_id}: Pod {pod_name}, still "
                                f"awaiting startup")
                    return (True, pod.status.pod_ip)
                else:
                    logger.info(f"Deploying Model {dep_id}: Pod {pod_name}, still "
                                f"awaiting startup")
                    return (False, None)
        return (False, None)

    k8_apps_v1_api, k8_core_v1_api, _, _, k8_ext_v1_api = get_k8_conn()

    bbyaml_dealer = customize_yaml_depmgr(
        depl_id=dep_id,
        envvars=env_vars,
        pod_resources=md.resource_config,
    )

    pod_name_dealer = bbyaml_dealer['metadata']['name']
    logger.info(f"Deploying Model {dep_id}: Dealer {pod_name_dealer}, attempting deployment")
    api_response = k8_apps_v1_api.create_namespaced_deployment(
        body=bbyaml_dealer,
        namespace=namespace
    )

    # Configure ingress path
    if dep_mode == 'ON_DEMAND':
        use_ingress = True
        slug = (
            get_ondemand_default_slug(dep_id)
            if not custom_path_slug
            else custom_path_slug
        )
    else:
        use_ingress = False
        slug = None

    # Customize the K8s manifests
    inp_yaml, svc_yaml, igr_yaml = customize_yaml_input_mgr(
        depl_id=dep_id,
        envvars=env_vars,
        pod_resources=md.resource_config,
        use_ingress=use_ingress,
        slug=slug
    )

    pod_name_ipmgr = inp_yaml['metadata']['name']
    logger.info(f"Deploying Model {dep_id}: Data input mgr {pod_name_ipmgr}, "
                f"attempting deployment")

    # Launch the data input manager container
    api_response = k8_apps_v1_api.create_namespaced_deployment(
        body=inp_yaml,
        namespace=namespace
    )
    # TODO: Validate response on api_response

    # Launch the acompaing service pod if on-demand mode
    if use_ingress:

        # Service
        logger.info(f"Deploying Model {dep_id}: Data input mgr {pod_name_ipmgr} "
                    f"on-demand clusterIP service, attempting creation")
        api_response = k8_core_v1_api.create_namespaced_service(
            body=svc_yaml,
            namespace=namespace
        )

        # Ingress resource
        logger.info(f"Deploying Model {dep_id}: Data input mgr {pod_name_ipmgr} "
                    f"on-demand ingress resource, attempting creation")
        api_response = k8_ext_v1_api.create_namespaced_ingress(
            body=igr_yaml,
            namespace=namespace
        )

        # TODO: Validate response on api_responses

    dealer_ip = None
    for _ in range(NUM_POLL_TRIES_AWAITING_STARTUP):
        time.sleep(WAIT_TIME_BETWEEN_POLLS)
        dealer_deployed, dealer_ip = check_is_pod_running(pod_name=pod_name_dealer)
        ipputmgr_deployed, _ = check_is_pod_running(pod_name=pod_name_ipmgr)
        if dealer_deployed and ipputmgr_deployed:
            return dealer_ip

    raise Exception(f"Deploying Model {dep_id}: Timed out awaiting launch of Deployment "
                    f"Manager {pod_name_dealer} and Data Input Manager {pod_name_ipmgr}")


def ctnr_image_pull_failed(in_status_list):
    import sys
    logger.info("Checking for failed container pulls")
    try:
        if not in_status_list:
            return False
        for i in in_status_list:
            if i.state:
                if i.state and not i.state.running:
                    if i.state.waiting and i.state.waiting.reason:
                        if i.state.waiting.reason == "ErrImagePull":
                            logger.error(f"Deploying Model: Found container with pull errors")
                            logger.error(i)
                            return True
    except:
        logger.error(f"Deploying Model: Unexpected error:", sys.exc_info()[0])
    return False


def successfully_launched(pod_of_interest):

    _, k8_core_v1_api, _, _, _ = get_k8_conn()

    logger.info(f"Checking for replica set {pod_of_interest} startup")
    NUM_POLL_TRIES_AWAITING_STARTUP = 300
    WAIT_TIME_BETWEEN_POLLS = 1

    for _ in range(NUM_POLL_TRIES_AWAITING_STARTUP):
        # TODO: Find a better way to do this using
        #       k8_apps_v1_api.read_namespaced_deployment(doi, namespace)
        pod_list = k8_core_v1_api.list_namespaced_pod(namespace)

        for pod in pod_list.items:
            if pod.metadata.name.startswith(pod_of_interest):
                #  logger.debug(f"POD STATUS: {pod.status.phase}")
                if pod.status.phase == "Running":
                    return True
                else:
                    if pod.status.phase == 'Pending':
                        if ctnr_image_pull_failed(in_status_list=pod.status.container_statuses):
                            logger.error(f"Deploying Model: k8s Failed to pull image")
                            raise Exception("Failed to pull image")

                        if pod.status.conditions and len(pod.status.conditions) > 0:
                            for c in pod.status.conditions:
                                if c.reason and c.message:

                                    #  logger.debug(f"Kubernetes: {c.reason}")
                                    #  logger.debug(f"Kubernetes: {c.message}")

                                    if c.reason == 'Unschedulable' \
                                            and 'cpu' in c.message \
                                            and 'Insufficient' in c.message:
                                        logger.error(f"Kubernetes: {c.reason}")
                                        logger.error(f"Kubernetes: {c.message}")
                                        raise Exception(f"Deploying Model: Could not launch "
                                                        f"worker, insufficient CPU capacity")

                                    if c.reason == 'Unschedulable' \
                                            and 'nvidia.com/gpu' in c.message \
                                            and 'Insufficient' in c.message:
                                        logger.error(f"Kubernetes: {c.reason}")
                                        logger.error(f"Kubernetes: {c.message}")
                                        raise Exception(f"Deploying Model: Could not launch "
                                                        f"worker, insufficient GPUs available")

                                    if c.reason == 'Unschedulable' \
                                            and 'memory' in c.message \
                                            and 'Insufficient' in c.message:
                                        logger.error(f"Kubernetes: {c.reason}")
                                        logger.error(f"Kubernetes: {c.message}")
                                        raise Exception(f"Deploying Model: Could not launch "
                                                        f"worker, insufficient memory available")

                    logger.info(f"Deploying Model: Dealer {pod_of_interest}, still awaiting startup")
        time.sleep(WAIT_TIME_BETWEEN_POLLS)

    # Other possible states?
    return False


def model_dep_start(request):
    """Start model which has been created (and possibly previously started/stopped)."""

    from kml.api.yaml_customizer import customize_yaml_depl_bbx_worker

    k8_apps_v1_api, _, _, _, _ = get_k8_conn()

    # Create and initialize a responder object
    response = Responder()

    try:
        dep_id = request.path_params.get('model_deployment_id')
        logger.info(f"Deploying Model {dep_id}: Starting")
        md = ModelDeployment.load(dep_id)

        # Check for available GPGPU resources and fail fast
        # Only for on-prem. Cloud assumes an available GPU scaleset
        if running_in_k8s:
            if md.resource_config.get('compute_target') == 'GPU':
                avail_gpus = get_k8_resources()['total_available']['GPU']
                req_replicas = md.resource_config['replicas']
                req_gpus_per_replica = md.resource_config['gpu_limit']
                req_gpus = req_replicas * req_gpus_per_replica
                if req_gpus > avail_gpus:
                    msg = (
                        f"Insufficient GPGPU resources available. Requested {req_gpus} "
                        f"({req_gpus_per_replica} GPUs X {req_replicas} replicas) but only "
                        f"{avail_gpus} are available"
                    )
                    logger.error(msg)
                    response.failed(msg=msg, code=1422)
                    return response.emit(422)

        # TODO: Fail Fast if Model Deployment is alredy in "STARTING" or "RUNNING" states
        deployment_mode = md.deployment_type
        mi = ModelInstance.load(md.base_model_inst_id)

        # #####################################################################
        # STEP 0: Re-Create Table Monitors
        # #####################################################################
        # TODO: Revisit this

        # #####################################################################
        # STEP 1a: Launch Input Manager Container
        # STEP 1b: Launch Deployment Manager (Dealer) Container
        # #####################################################################

        env_var_to_inject = {}
        env_var_to_inject['KML_DEP_MODE'] = md.deployment_type
        env_var_to_inject['KML_DEPL_ID'] = str(md.get_entity_id()) # Worker reference to dealer
        env_var_to_inject['KML_API_BASE'] = public_api_base
        env_var_to_inject['DB_CONN_STR'] = public_db_conn
        env_var_to_inject['DB_USER'] = db_user
        env_var_to_inject['DB_PASS'] = db_pass

        # DB <--> Input Manager/Dealer
        env_var_to_inject['ZMQ_HOST'] = tm_host
        env_var_to_inject['ZMQ_PORT'] = tm_port
        # Dealer <--> Worker Pod(s)
        env_var_to_inject['ZMQ_DEALER_PORT'] = str(DEFAULT_ZMQ_PORT_DEALER)

        # MAY (or may not) be used in the future to split out rowsets for better parallelization
        # perhaps the user can tune it using this:
        env_var_to_inject['ZMQ_FRAMES_PER_BATCH'] = md.model_dep_config["zmq_frames_per_batch"]
        # ...or perhaps the system can always split by 1/replicas
        env_var_to_inject['DOWNSTREAM_REPLICAS'] = md.resource_config["replicas"]

        env_var_to_inject['SCHEMA_SYS'] = SCHEMA_SYS
        env_var_to_inject['SCHEMA_AUDIT'] = SCHEMA_AUDIT

        env_var_to_inject['USE_MULTIHEAD_IO'] = md.model_dep_config["multihead_io"]

        #input_mgr_ip = help_deploy_input_mgr(dep_id=md.get_entity_id(), dep_mode=deployment_mode)
        #dealer_ip = help_deploy_dealer(dep_id=md.get_entity_id())
        dealer_ip = deploy_helper_pods(
            dep_id=md.get_entity_id(),
            dep_mode=md.deployment_type,
            env_vars=env_var_to_inject,
            custom_path_slug=md.model_dep_config.get('custom_path_slug'),
            md=md
        )

        # For restarts of existing deployments, first check the instantaneous
        # entity state to ensure it has not already been moved to running
        # by the input manager
        if ModelDeployment.load(dep_id).base_entity.entity_state != 'RUNNING':
            md.get_base_entity().transition_to_starting()
        else:
            logger.info(f"Deploying Model {dep_id}: Model already running, "
                        f"skipping set to STARTING")

        # #####################################################################
        # STEP 3: Prepare & Launch Worker Container(s)
        # #####################################################################

        # Additional environment variables for worker pods
        # Dealer <--> Worker Pod(s)
        env_var_to_inject['ZMQ_CONN_STR'] = public_tm_conn
        env_var_to_inject['ZMQ_DEALER_HOST'] = dealer_ip
        env_var_to_inject['ZMQ_DEALER_PORT'] = str(DEFAULT_ZMQ_PORT_DEALER)

        env_var_to_inject['BLACKBOX_MODULE'] = mi.model_inst_config["blackbox_module"]
        env_var_to_inject['BLACKBOX_FUNCTION'] = mi.model_inst_config["blackbox_function"]

        env_var_to_inject['SCHEMA_AUDIT'] = SCHEMA_AUDIT
        env_var_to_inject['PERSIST_AUDIT'] = md.model_dep_config["persist_audit"]

        if "user_provided_env_vars" in md.model_dep_config:
            for anev in md.model_dep_config["user_provided_env_vars"]:
                k=anev['var_name']
                v=anev['var_value']
                env_var_to_inject[k]=v

        ctnr = mi.model_inst_config['container']
        secrets_handle = None
        if "k8s_secrets_handle" in mi.model_inst_config:
            # TODO: What is better here, to pull the handle here using credential_id, or directly?
            secrets_handle = mi.model_inst_config["k8s_secrets_handle"]
            logger.info(f"Deploying Model {dep_id}: Launching {ctnr} with "
                        f"secrets handle {secrets_handle}")
        else:
            logger.info(f"Deploying Model {dep_id}: Launching {ctnr} without "
                        f"any registry permissions")

        bbyaml = customize_yaml_depl_bbx_worker(
            bbx_container=ctnr,
            depl_id=md.get_entity_id(),
            envvars=env_var_to_inject,
            pod_resources=md.resource_config,
            secrets_handle = secrets_handle
        )

        woi = bbyaml['metadata']['name']
        logger.info(f"Deploying Model {dep_id}: Worker(s) {woi}, attempting deployment")
        api_response = k8_apps_v1_api.create_namespaced_deployment(
            body=bbyaml,
            namespace=namespace
        )
        logger.info(f"Deploying Model {dep_id}: Worker(s) {woi}, awaiting startup")

        # if md.deployment_type == "ON_DEMAND":
        #     logger.info(f"***** Deploying Model {dep_id}: Awaiting On-Demand k8s service to start")
        #     # TODO: check for readiness instead of a blind pause
        #     time.sleep(10)

        try:
            if not successfully_launched(woi):
                # XXX: This is suspicious - we should NOT be polling K8s for readiness!!!
                # Input Mgr tells us when infrastructure is ready

                #model_deploy_terminate(md.get_entity_id())
                md.get_base_entity().transition_to_terminated()
                model_deploy_teardown(md.get_entity_id())

                msg = f"Deploying Model {dep_id}: Could not launch worker(s)"
                logger.error(msg)
                response.failed(msg=msg, code=1422)
                return response.emit(422)
        except Exception as e:
            md.get_base_entity().transition_to_terminated()
            model_deploy_teardown(md.get_entity_id())
            msg = f"Deploying Model {dep_id}: Could not launch worker(s): {str(e)}"
            logger.error(msg)
            response.failed(msg=msg, code=1422)
            return response.emit(422)

        response.response['started'] = md.base_entity.serialize()
        response.succeeded()
        return response.emit(http_code=200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)


async def doinfer(request):
    import zmq
    import threading

    import requests
    from zmq.eventloop.future import Context

    from kml.api.naming_svc import (
        get_cluster_name_input_mgr,
        get_cluster_name_ingress
    )

    (k8_apps_v1_api,
     k8_core_v1_api,
     k8_custom_objects_api,
     k8_batch_v1_api,
     k8_ext_v1_api) = get_k8_conn()

    response = Responder()

    try:
        model_deployment_id = request.path_params.get('model_deployment_id')
        md = ModelDeployment.load(model_deployment_id)

        depl_id = request.path_params.get('model_deployment_id')
        igr_name = get_cluster_name_ingress(depl_id)

        depl_state = md.get_base_entity().get_state()
        if depl_state == "STOPPED":
            msg = (f"Model Deployment {depl_id} is in STOPPED state, thus inferences "
                   f"are not allowed. Please start the model to infer.")
            logger.error(msg)
            response.failed(msg=msg, code=1422)
            return response.emit(422)
        elif depl_state != "RUNNING":
            msg = (f"Model Deployment {depl_id} is not in RUNNING state, thus inferences "
                   f"are not allowed. Please start the model to infer.")
            logger.error(msg)
            response.failed(msg=msg, code=1422)
            return response.emit(422)

        resp = k8_ext_v1_api.read_namespaced_ingress(
            name=igr_name,
            namespace=namespace
        )
        path = resp.spec.rules[0].http.paths[0].path.replace('(/|$)(.*)', '')

        return_schema = json.dumps(md.model_dep_config["out-tablemonitor"]["type_schema"])
        record_type = gpudb.RecordType.from_type_schema("", return_schema, {})

        logger.info(f'Received Inference Request for deployment {model_deployment_id}')

        receive_dt = datetime.datetime.utcnow().isoformat(' ')[:-3]

        inference_inputs = await request.json()
        if not isinstance(inference_inputs, list):
            logger.warning("Inference payloads should always be lists, received "
                           "lone object. Converting to list.")
            inference_inputs = [inference_inputs]
        logger.info(f"Received payload for inference with {len(inference_inputs)} items")

        zmq_context = zmq.eventloop.future.Context()
        mysocket = zmq_context.socket(zmq.SUB)
        mysocket.connect(public_tm_conn)
        topicstring = md.model_dep_config["out-tablemonitor"]["topic_id"]
        topicfilter = bytes(topicstring, 'utf-8').decode()
        mysocket.setsockopt_string(zmq.SUBSCRIBE, topicfilter)

        dep_name = get_cluster_name_input_mgr(model_deployment_id)

        ingress_ip = get_public_k8s_ip()
        _, _, ing_http_port, ing_https_port = get_ingress_controller_details()
        if ing_http_port:
            ingress_host = f"http://{ingress_ip}:{ing_http_port}"
        elif ing_https_port:
            ingress_host = f"https://{ingress_ip}:{ing_https_port}"
        else:
            ingress_host = f"http://{ingress_ip}"

        def delay_inserter(wait_before_insert, drop_uri, insertable):

            ON_DEMAND_READY_TIMEOUT = 30  # TODO: move all timeouts to system.ini
            on_demand_ready = False

            for _ in range(ON_DEMAND_READY_TIMEOUT):
                try:
                    ping_uri = f"{ingress_host}{path}"
                    logger.debug(f"Ingress URL for On Demand model: {ping_uri}")
                    r = requests.get(ping_uri)
                    # logger.debug(f"Handler ping URI: {ping_uri}")
                    # logger.debug(f"Ping on on-demand handler: {r.status_code}")
                    # logger.debug(f"On-demand handler test response: {r.content}")
                    if r.status_code == 200:
                        on_demand_ready = True
                        break
                    else:
                        logger.info('Awaiting input manager service readiness')
                        time.sleep(1)
                except Exception as e:
                    pass

            if not on_demand_ready:
                raise ValueError(f"On-demand K8s microservice {dep_name} failed to start")

            # NOTE: Is this delay actually needed?
            #  time.sleep(wait_before_insert)
            logger.info(f"Initiated flow for request set of size {len(insertable)}")

            r = requests.post(
                f"{ingress_host}{path}/ondemand/{model_deployment_id}/infer",
                json=insertable
            )

            if r.status_code != 200:
                logger.error(r.content)
                response.failed(
                    msg="Unable to communicate with on-demand service",
                    code=1040
                )
                return response.emit(509)


        # TODO: Chunk large requests into batchs of 10k or 20k MAX,
        # confg pool count accordingly
        threads = []
        drop_uri = ""
        t = threading.Thread(
            group=None,
            target=delay_inserter,
            args=(2, drop_uri, inference_inputs)
        )
        threads.append(t)
        t.start()

        logger.info("Waiting for Inference Worker response on Topic %s" % topicstring)
        inference_outputs = {}
        while len(inference_outputs) < len(inference_inputs):

            all_responses = await mysocket.recv_multipart(flags=0, copy=True, track=False)
            logger.info(f"Received Potential Response on topic {topicstring} "
                         f"with {len(all_responses)-1} items")

            for resx in all_responses[1:]:
                return_payload = [dict(x) for x in gpudb.GPUdbRecord.decode_binary_data(record_type, resx)][0]
                if return_payload["guid"] not in inference_outputs:
                    inference_outputs[return_payload["guid"]] = []
                inference_outputs[return_payload["guid"]].append(return_payload)

            #if len(inference_outputs) > 0:
                # BIG Assumption here -- that if we got one in the batch, we got all
                # TODO: REVISIT THIS MAJOR ASSUMPTION

                #  logger.info(f"Returning inference batch of {len(inference_outputs)}, "
                #              f"inbound request had {len(inference_inputs)}")
                #  logger.debug(inference_outputs)
            logger.info(f"Inputs {len(inference_inputs)}, outputs so far: {len(inference_outputs)}")
            if len(inference_outputs) >= len(inference_inputs):
                if len(inference_outputs) > len(inference_inputs):
                    logger.warning("Received more outputs than inputs...potential problem")

                response.response["inference"] = inference_outputs
                response.response["inference_count"] = len(inference_outputs)
                response.succeeded()
                mysocket.close()
                zmq_context.term()
                return response.emit(200)
                #else:
                #    logger.debug("Decoded message off wire for task %s,z not my task, i dont care, lets continue to wait" %
                #        return_payload[0]["guid"])
        # We should never reach here
        # TODO: this would be a timeout, how do we handle this?
        mysocket.close()
        zmq_context.term()
        response.failed(msg="Timeout waiting for response", code=1040)
        return response.emit(509)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


async def model_deploy_setstatus(request):
    from kml.api.orm import Entity

    response = Responder()
    response.response['state-modified'] = False
    try:
        model_deployment_id = request.path_params.get('model_deployment_id')
        spec = await request.json()
        md = ModelDeployment.load(model_deployment_id)
        destination_state = spec['destination-state']
        if destination_state not in Entity.ENTITY_STATES:
            msg = f"Unknown destination state {destination_state}"
            raise ValueError(msg)

        ts = datetime.datetime.utcnow().isoformat(' ')
        logger.info(
            f"Setting model deployment {model_deployment_id} status to {destination_state} "
            f"just after {ts}"
        )
        md.change_state(destination_state)
        response.response['state-modified'] = True
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def model_deploy_complete(request):
    """Set deployment entity to completed and tear down infrastructure."""

    # Create and initialize a responder object
    response = Responder()

    try:
        dep_id = request.path_params.get('model_deployment_id')

        deployment_record = ModelDeployment.load(dep_id)
        deployment_record.get_base_entity().transition_to_completed()

        model_deploy_teardown(dep_id)

        # Create response
        response.response['completed'] = True
        response.response['deployment_id'] = dep_id
        response.succeeded()

        return response.emit(http_code=200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)


def model_deploy_terminate(request):
    """Set deployment entity to terminated and tear down infrastructure."""

    # Create and initialize a responder object
    response = Responder()

    try:
        dep_id = request.path_params.get('model_deployment_id')

        deployment_record = ModelDeployment.load(dep_id)
        deployment_record.get_base_entity().transition_to_terminated()

        model_deploy_teardown(dep_id)

        # Create response
        response.response['terminated'] = True
        response.response['deployment_id'] = dep_id
        response.succeeded()

        return response.emit(http_code=200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1010)
        return response.emit(509)


def model_deploy_teardown(dep_id):
    """Remove all model deployment infrastructure from K8s (internal use)."""

    import zmq
    from zmq.eventloop.future import Context
    from kubernetes.client import V1DeleteOptions

    from kml.api.naming_svc import (
            get_cluster_name_deployment_blackbox,
            get_cluster_name_dep_mgr,
            get_cluster_name_input_mgr,
            get_cluster_name_service,
            get_cluster_name_ingress
    )

    (k8_apps_v1_api,
     k8_core_v1_api,
     k8_custom_objects_api,
     k8_batch_v1_api,
     k8_ext_v1_api) = get_k8_conn()

    # Enable fast termination
    delete_options = V1DeleteOptions(propagation_policy='Foreground',
                                     grace_period_seconds=0)

    # Remove ZMQ connection the pool
    def helper_remove_zmq_item(atopic):

        try:
            logger.info(f'Clearing ZMQ topic {atopic}')
            zmq_context = zmq.eventloop.future.Context()
            zmq_socket = zmq_context.socket(zmq.SUB)
            zmq_socket.connect(public_tm_conn)
            topic_filter = bytes(atopic, 'utf-8').decode()
            zmq_socket.setsockopt_string(zmq.UNSUBSCRIBE, topic_filter)
        except Exception as e:
            logger.warning(f'Problem unsubscribing from zmq topic {atopic}, proceeding')
            #  response.failed(msg=e, code=1040)
            #  logger.error(e)
            #  logger.debug(traceback.format_exc())
            #  return response.emit(http_code=509)

    try:

        # Load deployment metadata
        deployment_record = ModelDeployment.load(dep_id)

        # Compose the K8s primary entity names
        dep_name = get_cluster_name_deployment_blackbox(dep_id)
        mgr_name = get_cluster_name_dep_mgr(dep_id)
        inp_name = get_cluster_name_input_mgr(dep_id)

        # Clear table monitors and ZeroMQ topic
        # FIXME: Currently disabled due to loss of function. Do we need this?
        #
        #  if "src-tablemonitor" in deployment_record.model_dep_config:
        #      if "topic_id" in deployment_record.model_dep_config["src-tablemonitor"]:
        #          src_topic = deployment_record.model_dep_config['src-tablemonitor']['topic_id']
        #          src_tm_resp = db.clear_table_monitor(topic_id=src_topic)
        #          helper_remove_zmq_item(src_topic)
        #  if "inp-tablemonitor" in deployment_record.model_dep_config:
        #      if "topic_id" in deployment_record.model_dep_config["inp-tablemonitor"]:
        #          inp_topic = deployment_record.model_dep_config['inp-tablemonitor']['topic_id']
        #          inp_tm_resp = db.clear_table_monitor(topic_id=inp_topic)
        #          helper_remove_zmq_item(inp_topic)
        #  if "out-tablemonitor" in deployment_record.model_dep_config:
        #      if "topic_id" in deployment_record.model_dep_config["out-tablemonitor"]:
        #          out_topic = deployment_record.model_dep_config['out-tablemonitor']['topic_id']
        #          out_tm_resp = db.clear_table_monitor(topic_id=out_topic)
        #          helper_remove_zmq_item(out_topic)

        # Remove the on-demand clusterIP service pod (in necessary)
        if deployment_record.deployment_type == 'ON_DEMAND':
            svc_name = get_cluster_name_service(dep_id)
            try:
                api_response = k8_core_v1_api.delete_namespaced_service(
                    name=svc_name,
                    namespace=namespace,
                    body=delete_options
                )
            except Exception as e:
                logger.warning(f"Error terminating service pod {svc_name}, proceeding")
                logger.warning(str(e))

        # Remove the ingress definition (in necessary)
        if deployment_record.deployment_type == 'ON_DEMAND':
            igr_name = get_cluster_name_ingress(dep_id)
            try:
                api_response = k8_ext_v1_api.delete_namespaced_ingress(
                    name=igr_name,
                    namespace=namespace,
                    body=delete_options
                )
            except Exception as e:
                logger.warning(f"Error removing ingress definition {igr_name}, proceeding")
                logger.warning(str(e))

        # Remove data input manager
        try:
            api_response = k8_apps_v1_api.delete_namespaced_deployment(
                name=inp_name,
                namespace=namespace,
                body=delete_options
            )
        except Exception as e:
            logger.warning(f"Error terminating input manager {inp_name}, proceeding")
            logger.warning(str(e))

        # Remove deployment manager (dealer)
        try:
            api_response = k8_apps_v1_api.delete_namespaced_deployment(
                name=mgr_name,
                namespace=namespace,
                body=delete_options
            )
        except Exception as e:
            logger.warning(f"Error terminating deployment manager {mgr_name}, proceeding")
            logger.warning(str(e))

        # Remove worker deployments (and associated pods)
        try:
            api_response = k8_apps_v1_api.delete_namespaced_deployment(
                name=dep_name,
                namespace=namespace,
                body=delete_options
            )
        except Exception as e:
            logger.warning(f"Error terminating model deployment {dep_name}, proceeding")
            logger.warning(str(e))

    except Exception as e:
        raise Exception(e)
        #  logger.error(e)
        #  logger.debug(traceback.format_exc())
        #  response.failed(msg=e, code=1010)
        #  return response.emit(509)


def model_deploy_view(request):
    """Retrieve info on a specific model deployment."""
    response = Responder()
    try:
        model_deployment_id = request.path_params.get('model_deployment_id')
        deployment_record = ModelDeployment.load(model_deployment_id)
        response.response['item'] = deployment_record.serialize()
        response.response['found'] = True
        response.succeeded()
        logger.info(f"Model deployment {model_deployment_id} view")
        return response.emit(http_code=200)

    except LookupError as e:
        response.failed(msg=e, code=404)
        #logger.error(e)
        return response.emit(http_code=404)

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


def model_deploy_archive(request):
    """Retrieve info on a specific model deployment."""
    response = Responder()
    try:
        model_deployment_id = request.path_params.get('model_deployment_id')
        deployment_record = ModelDeployment.load(model_deployment_id)
        deployment_record.tombstone()

        # TODO: Reject request if the deployment is not in COMPLETED/TERMINATED/FAILED state

        response.response['item'] = deployment_record.serialize()
        response.succeeded()
        return response.emit(http_code=200)
    except LookupError as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        return response.emit(http_code=404)
    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)
