"""KML FeatureSet  endpoint controller."""

import traceback
import tempfile
import json

from kml.api.orm import FeatureSet
from kml.api.aux import (
    attach_log,
    log_comms,
    Responder
)
from kml.api.kml_config import (
    get_components,
    get_namespace
)
from kml.api import controller_transform
from kml.api.kmldbops import get_tbl_handle as kml_get_tbl_handle

# Configuration
_, api_host, api_port = get_components(conn='api')
namespace = get_namespace()

logger = attach_log(module=__name__, log_name='kml', debug=True)

VALID_XFORMS = [
    'onehot',
    'bucket',
    'ln',
    'vocabulary',
    'normalize',
    'zscale',
    'inverse',
    'identity',
    'log',
    'extracthour',
    'octetsplit'
]


async def featureset_create(request):
    """Build relational olap and native in-line feature sets."""

    response = Responder()
    spec = await request.json()
    log_comms(spec, 'inbound_feats_create')

    req_params = [
        'featureset_name',
        'featureset_desc',
        'config_relational',
        'config_inline'
    ]
    req_params_inline = [
        'dataset_id_fitting',
        'xforms'
    ]
    req_params_relational = [
        'mv_name',
        'mv_timestamp_col'
    ]

    config_inline = {}
    config_relational = {}

    try:
        # Parameter check: General FeatureSet
        for param in req_params:
            if param not in spec:
                msg = f"Bad create request: missing FeatureSet creation parameter ({param})"
                logger.error(msg)
                response.failed(msg=msg, code=1040)
                return response.emit(422)
            if param in ['featureset_name', 'featureset_desc']:
                spec[param] = spec[param].strip()

        # --- Validate Configurations for Relational Transforms
        if spec.get('config_relational'):
            for param in req_params_relational:
                if param not in spec['config_relational']:
                    msg = (f"Bad create request: missing FeatureSet relational "
                           f"config parameter ({param})")
                    logger.error(msg)
                    response.failed(msg=msg, code=1040)
                    return response.emit(422)

            # Prune strings
            for key in req_params_relational:
                spec['config_relational'][key] = spec['config_relational'][key].strip()

            logger.info(f"Getting view {spec['config_relational']['mv_name']}")
            view_info = kml_get_tbl_handle(
                spec['config_relational']['mv_name']
            ).show_table()

            if view_info['status_info']['status'] == 'ERROR':
                msg = f"Problem with relational view: {view_info['status_info']['message']}"
                logger.error(msg)
                response.failed(msg=msg, code=1040)
                return response.emit(422)

            if view_info['additional_info'][0]['refresh_method'] != 'periodic':
                msg = f"Problem with relational view: Refresh method is not periodic"
                logger.error(msg)
                response.failed(msg=msg, code=1040)
                return response.emit(422)

            # TODO: Validate that referenced view exists spec['config_relational']['mv_name']

            # This will be persisted to the system tables
            config_relational = spec['config_relational']

            # From the referenced view, extract metadata
            config_relational["mv_refresh"] = int(
                view_info['additional_info'][0]['refresh_period']
            )
            request_avro = json.loads(
                view_info['additional_info'][0]['request_avro_json']
            )
            config_relational["mv_ddl"] = request_avro['options']['view_definition']
            #  TODO: save base tables

        # --- Validate Configurations for Inline Transforms
        if spec.get('config_inline'):
            for param in req_params_inline:
                if param not in spec['config_inline']:
                    msg = (f'Bad create request: missing FeatureSet inline config '
                           f'parameter ({param})')
                    logger.error(msg)
                    response.failed(msg=msg, code=1040)
                    return response.emit(422)

            # --- Validate Each Inidividual Transformation Configurations for Inline Transforms
            for feat in spec['config_inline']['xforms']:
                req_xform = feat['function_xform'][0]
                if req_xform not in VALID_XFORMS:
                    msg = (f'Bad create request: unexpected xform function '
                           f'{req_xform} specified for DB native in-line transform')
                    logger.error(msg)
                    response.failed(msg=msg, code=1040)
                    return response.emit(422)

            # This will be persisted to the system tables
            # Fit and build native expression pipeline
            config_inline = controller_transform.transform_create(
                spec['config_inline']['dataset_id_fitting'],
                spec['config_inline']['xforms']
            )

        just_created_FeatureSet = FeatureSet.create(
            featureset_name=spec['featureset_name'],
            featureset_desc=spec['featureset_desc'],
            config_inline=config_inline,
            config_relational=config_relational
        )

        just_created_FeatureSet.publish()
        just_created_FeatureSet.get_base_entity().transition_to_created()
        response.response['created'] = just_created_FeatureSet.serialize()
        response.succeeded()
        log_comms(response.response, 'outbound_mi_deploy')

        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


async def featureset_refit(request):

    # TODO: Later: Re-fit featureset to new dataset
    response = Responder()
    spec = await request.json()
    response.succeeded()
    response.response['message'] = 'method not implemented'
    log_comms(response.response, 'method not implemented')
    return response.emit(201)


async def featureset_import(request):
    response = Responder()
    spec = await request.json()
    log_comms(spec, 'inbound_feats_import')

    # Parameter check: General FeatureSet
    if 'featureset_name' not in spec:
        response.failed(
            msg=('Bad create request: missing FeatureSet Name '
                 'parameter (featureset_name)'),
            code=1040
        )
        return response.emit(422)

    if 'featureset_desc' not in spec:
        response.failed(
            msg=('Bad create request: missing FeatureSet Description '
                 'parameter (featureset_desc)'),
            code=1040
        )
        return response.emit(422)

    if 'feat_mem' not in spec:
        response.failed(
            msg=('Bad create request: missing FeatureSet memory '
                 'payload (feat_mem)'),
            code=1040
        )
        return response.emit(422)

    config_inline = {}
    config_inline['xforms'] = []

    # NOTE: Imports do not need to be fit, hence this should be left missing (perhaps null?)
    # config_inline['dataset_id_fitting'] = -1

    config_inline['feat_mem'] = spec['feat_mem']
    config_inline['target_schema'] = [
        ['guid', 'string', 'uuid', 'init_with_uuid'],
        ['receive_dt', 'string', 'datetime']
    ]
    config_inline['xf_map'] = {
        'guid': 'new_uuid()',
        'receive_dt': 'current_datetime()'
    }

    # Create a single FeatureSet instance
    try:
        if 'kml::metadata' in spec['feat_mem']:
            del spec['feat_mem']['kml::metadata']
        if 'metadata' in spec['feat_mem']:
            del spec['feat_mem']['metadata']

        for (k, v) in spec['feat_mem'].items():

            # Prep output schema
            #  logger.info(f"Prepping out schema for input col {k}")
            config_inline['target_schema'].append(v["inschema"])
            for osch in v['outschema']:
                if osch[0] != v["inschema"][0]:
                    config_inline['target_schema'].append(osch)

            # Prep xforms
            if v['function_xform'] == "onehot":
                config_inline['xforms'].append(
                    {
                        "in_col_name": k,
                        "out_col_name": None,
                        "function_xform": [v['function_xform']]
                    }
                )
            elif v['function_xform'] == "octetsplit":
                config_inline['xforms'].append(
                    {
                        "in_col_name": k,
                        "out_col_name": None,
                        "function_xform": [v['function_xform']]
                    }
                )
            else:
                config_inline['xforms'].append(
                    {
                        "in_col_name": k,
                        "out_col_name": v['outschema'][0][0],
                        "function_xform": [v['function_xform']]
                    }
                )

            # Prep xf_map
            # throw in all the output expressions
            #  logger.debug(f"Expression {v['kexp']}")
            config_inline['xf_map'].update(v['kexp'])
            # ...also throw in input cols (pass-thru)
            config_inline['xf_map'][k] = k

        # TODO: IS THIS EVEN RELEVANT FOR IMPORTS?
        config_relational = {}
        just_created_FeatureSet = FeatureSet.create(
            featureset_name=spec['featureset_name'],
            featureset_desc=spec['featureset_desc'],
            config_inline=config_inline,
            config_relational=config_relational
        )

        just_created_FeatureSet.publish()
        just_created_FeatureSet.get_base_entity().transition_to_created()
        response.response['created'] = just_created_FeatureSet.serialize()
        response.succeeded()
        log_comms(response.response, "feature_memory_import")
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def featureset_view(request):
    response = Responder()
    response.response['found'] = False

    try:
        featureset_id = request.path_params.get('featureset_id')
        retrieved_featureset_obj = FeatureSet.load(existing_featureset_id=featureset_id)
        response.response['found'] = True
        response.response['item'] = retrieved_featureset_obj.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def featureset_export(request):

    from starlette.responses import Response

    response = Responder()

    try:
        featureset_id = request.path_params.get('featureset_id')
        retrieved_featureset_obj = FeatureSet.load(existing_featureset_id=featureset_id)
        _, tempfilename = tempfile.mkstemp()
        #  logger.debug(f'Scratch file: {tempfilename}')
        #  logger.debug(retrieved_featureset_obj)
        retrieved_featureset_obj.grab(tempfilename)

        with open(tempfilename, "r") as f:
            data = f.read()
            FILE_RETURN_HEADER = {}
            FILE_RETURN_HEADER['Access-Control-Allow-Methods'] = 'HEAD, GET'
            FILE_RETURN_HEADER['Access-Control-Allow-Origin'] = '*'
            FILE_RETURN_HEADER['Content-Type'] = 'application/csv'
            FILE_RETURN_HEADER['Content-Disposition'] = 'attachment; filename="feat_mem.json"'
            return Response(data, status_code=200, headers=FILE_RETURN_HEADER)
            # TODO: How do we use "emit" concept here?

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def featureset_clone(request):
    response = Responder()

    try:
        featureset_id = request.path_params.get('featureset_id')
        eve2 = FeatureSet.clone(existing_featureset_id=featureset_id)
        eve2.publish()
        eve2.get_base_entity().transition_to_created()
        response.response['created'] = eve2.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def featureset_preview(request):
    # TODO
    pass


def featureset_archive(request):
    response = Responder()

    try:
        featureset_id = request.path_params.get('featureset_id')
        fs = FeatureSet.load(featureset_id)
        fs.tombstone()
        response.response['item'] = fs.serialize()
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def featureset_publish(request):
    response = Responder()

    try:
        featureset_id = request.path_params.get('featureset_id')
        fs = FeatureSet.load(featureset_id)
        fs.base_entity.publish()
        # TODO: Find specific errors from publish response and handle in less generic ways
        response.response['published'] = True
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)