"""KML ingest endpoint controller."""

import os
import traceback
import json

from kml.api.aux import (
    attach_log,
    Responder,
    log_comms
)
from kml.api.kml_config import (
    get_api_base,
    get_k8_conn,
    get_sys_container,
    get_db_conn_str,
    get_db_credentials,
    get_namespace,
    get_sys_schemas,
    get_public_connections
)
from kml.api.orm import KCredential
from kml.api.kmldbops import get_db_track as kml_get_db_track

# Configuration
db_user, db_pass = get_db_credentials()
namespace = get_namespace()

KEYSTORE_PROPERTIES_WITH_PASS="""
bootstrap.servers=__INSERT_HOST_PORT__
security.protocol=SSL
ssl.truststore.location=/tmp/kafka.client.truststore.jks
ssl.truststore.password=__INSERT_PASSWORD__
key.converter=org.apache.kafka.connect.storage.StringConverter
value.converter=org.apache.kafka.connect.json.JsonConverter
key.converter.schemas.enable=false
value.converter.schemas.enable=false
offset.storage.file.filename=/tmp/connect.offsets
offset.flush.interval.ms=10000"""

KEYSTORE_PROPERTIES_ZERO_PASS="""
bootstrap.servers=__INSERT_HOST_PORT__
key.converter=org.apache.kafka.connect.storage.StringConverter
value.converter=org.apache.kafka.connect.json.JsonConverter
key.converter.schemas.enable=false
value.converter.schemas.enable=false
offset.storage.file.filename=/tmp/connect.offsets
offset.flush.interval.ms=10000"""


SASL_PROPERTIES="""
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# These are defaults. This file just demonstrates how to override some settings.
# bootstrap.servers=localhost:9092
rest.port=8090
bootstrap.servers=pkc-4nym6.us-east-1.aws.confluent.cloud:9092
ssl.endpoint.identification.algorithm=https

security.protocol=SASL_SSL
sasl.mechanism=PLAIN
sasl.jaas.config=org.apache.kafka.common.security.plain.PlainLoginModule required username=\'__INSERT_USERNAME__\' password=\'__INSERT_PASSWORD__\';

producer.security.protocol=SASL_SSL
producer.sasl.mechanism=PLAIN
producer.sasl.jaas.config=org.apache.kafka.common.security.plain.PlainLoginModule required username=\'__INSERT_USERNAME__\' password=\'__INSERT_PASSWORD__\';

consumer.security.protocol=SASL_SSL
consumer.sasl.mechanism=PLAIN
consumer.sasl.jaas.config=org.apache.kafka.common.security.plain.PlainLoginModule required username=\'__INSERT_USERNAME__\' password=\'__INSERT_PASSWORD__\';

# The converters specify the format of data in Kafka and how to translate it into Connect data. Every Connect user will
# need to configure these based on the format they want their data in when loaded from or stored into Kafka
key.converter=org.apache.kafka.connect.storage.StringConverter
value.converter=org.apache.kafka.connect.json.JsonConverter
# Converter-specific settings can be passed in by prefixing the Converter's setting with the converter we want to apply
# it to
key.converter.schemas.enable=false
value.converter.schemas.enable=false

offset.storage.file.filename=/tmp/connect.offsets
# Flush much faster than normal, which is useful for testing/debugging
offset.flush.interval.ms=10000

# Set to a list of filesystem paths separated by commas (,) to enable class loading isolation for plugins
# (connectors, converters, transformations). The list should consist of top level directories that include
# any combination of:
# a) directories immediately containing jars with plugins and their dependencies
# b) uber-jars with plugins and their dependencies
# c) directories immediately containing the package directory structure of classes of plugins and their dependencies
# Note: symlinks will be followed to discover dependencies or plugins.
# Examples:
# plugin.path=/usr/local/share/java,/usr/local/share/kafka/plugins,/opt/connectors,
# plugin.path=/usr/share/java
plugin.path=/opt/kafka/libs/kafka-2.6.0-connector-kinetica-7.1.1.0-jar-with-dependencies.jar

"""

# External connection strings
public_api_base, public_db_conn, _ = get_public_connections()
KML_API_BASE = public_api_base
db_conn_str = public_db_conn

# Remove http or https protocol for KIO convention
db_host_port_str = public_db_conn.lstrip('https://')
use_ssl = True if 'https' in public_db_conn else False

# Configure log file
logger = attach_log(module=__name__, log_name='kml', debug=True)


async def ingest_create(request):
    from kml.api.orm import Ingest

    response = Responder()

    try:
        spec = await request.json()

        if 'ingest_name' not in spec:
            response.failed(
                msg=('Bad create request: missing Ingest Name '
                     'parameter (ingest_name)'),
                code=1040
            )
            return response.emit(422)

        if not spec.get('ingest_desc'):
            logger.info('Ingest description not provided. Using ingest name '
                        'as the description')
            spec.update(ingest_desc=spec.get('ingest_name'))

        for key in spec.keys():
            if key == 'ingest_config':
                for conf in spec['ingest_config']:
                    if isinstance(spec['ingest_config'][conf], str):
                        spec['ingest_config'][conf] = spec['ingest_config'][conf].strip()
            else:
                spec[key] = spec[key].strip()

        data_ingest = Ingest.create(
            ingest_name=spec['ingest_name'],
            ingest_desc=spec['ingest_desc'],
            ingest_mode=spec['ingest_mode'],
            ingest_type=spec['ingest_type'],
            ingest_config=spec['ingest_config'],
            file_id=spec.get('file_id')
        )
        data_ingest.publish()
        data_ingest.get_base_entity().transition_to_created()

        response.response['created'] = data_ingest.serialize()
        response.succeeded()

        return response.emit(200)

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(509)


def ingest_view(request):
    """Retrieve info on a specific ingest."""
    from kml.api.orm import Ingest

    # Create and initialize a responder object
    response = Responder()

    try:
        ingest_id = request.path_params.get('ingest_id')
        ingest_record = Ingest.load(ingest_id)
        response.response['item'] = ingest_record.serialize()
        response.response['found'] = True
        response.succeeded()
        return response.emit(http_code=200)

    except LookupError as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        return response.emit(http_code=404)

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


def ingest_terminate(request):
    """Delete an ingest job (and it's pods) running on the Kubernetes cluster."""
    from kubernetes.client import V1DeleteOptions
    from kml.api.naming_svc import get_cluster_name_ingest
    from kml.api.orm import Ingest


    k8_apps_v1_api, k8_core_v1_api, k8_custom_objects_api, k8_batch_v1_api, _ = get_k8_conn()

    # Create and initialize a responder object
    response = Responder()

    try:
        ingest_id = request.path_params.get('ingest_id')
        ingest_record = Ingest.load(ingest_id)
        mode = ingest_record.serialize()['ingest_mode']

        delete_options = V1DeleteOptions(propagation_policy='Foreground')

        try:
            if mode == 'BATCH':
                job_name = get_cluster_name_ingest(in_ingest_id=ingest_id)
                response.response['terminated'] = job_name
                api_response = k8_batch_v1_api.delete_namespaced_job(
                    name=job_name,
                    namespace=namespace,
                    body=delete_options
                )

            elif mode == 'CONTINUOUS':
                dep_name = get_cluster_name_ingest(in_ingest_id=ingest_id)
                response.response['terminated'] = dep_name
                api_response = k8_apps_v1_api.delete_namespaced_deployment(
                    name=dep_name,
                    namespace=namespace,
                    body=delete_options
                )
        except Exception as e:
            logger.error("Error trying to purge containerized ingest from k8s (perhaps it is already gone?")
            logger.exception(e)

        ingest_record.get_base_entity().transition_to_terminated()
        response.succeeded()
        return response.emit(http_code=200)

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


def ingest_status():
    """Return the status of all deployed Kubernetes jobs."""
    # TODO: Is this endpoint still used? (needed?)


    k8_apps_v1_api, k8_core_v1_api, k8_custom_objects_api, k8_batch_v1_api, _ = get_k8_conn()

    # Create and initialize a responder object
    response = Responder()

    try:
        api_response = k8_batch_v1_api.list_namespaced_job(namespace=namespace,
                                                           watch=False)

        status = list()
        for job in api_response.items:
            #  active is True if job.status.active else False
            active = 1 if job.status.active else 0
            succeeded = 1 if job.status.succeeded == 1 else 0
            # TODO: add start and completed times
            if 'kml-ing' in job.metadata.name:
                status.append({'name': job.metadata.name,
                               'succeeded': succeeded,
                               'active': active})

        response.response['count'] = len(status)
        response.response['items'] = status
        response.succeeded()
        return response.emit(http_code=200)

    except Exception as e:
        response.failed(msg=e, code=999)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


def ingest_clone(request):
    """Clone an existing Ingest."""
    from kml.api.orm import Ingest

    # Create and initialize a responder object
    response = Responder()

    try:
        ingest_id = request.path_params.get('ingest_id')
        clone = Ingest.clone(existing_ingest_id=ingest_id)
        clone.publish()
        clone.get_base_entity().transition_to_created()

        response.response['created'] = clone.serialize()
        response.response['found'] = True
        response.succeeded()
        return response.emit(http_code=200)

    except LookupError as e:
        response.failed(msg=e, code=400)
        logger.error(e)
        return response.emit(http_code=404)

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


def ingest_archive(request):
    """Archive an existing Ingest."""
    from kml.api.orm import Ingest

    # Create and initialize a responder object
    response = Responder()

    try:
        ingest_id = request.path_params.get('ingest_id')
        ingest_record = Ingest.load(ingest_id)

        if ingest_record.get_base_entity().is_running():
            response.failed(
                msg=(f"Attempted archive on ingest (id {ingest_id}), failed because "
                     f"ingest is currently running"),
                code=1040
            )
            return response.emit(509)

        ingest_record.tombstone()
        response.response['item'] = ingest_record.serialize()
        response.succeeded()
        return response.emit(http_code=200)
    except LookupError as e:
        response.failed(msg=e, code=400)
        logger.error(e)
        return response.emit(http_code=404)
    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


def ingest_start(request):
    """Launch a Kubernetes Job to perform a data ingest."""
    import yaml
    from kml.api.orm import Ingest
    from kml.api.naming_svc import get_cluster_name_ingest
    from kml.api.yaml_customizer import customize_yaml_byoc_ing
    from kml.api.kml_config import get_db_credentials

    k8_apps_v1_api, k8_core_v1_api, k8_custom_objects_api, k8_batch_v1_api, _ = get_k8_conn()

    # Create and initialize a responder object
    response = Responder()

    # Load the ingest parameters from the system table
    data_ingest = None
    try:
        ingest_id = request.path_params.get('ingest_id')
        data_ingest = Ingest.load(ingest_id)
        job_name = get_cluster_name_ingest(in_ingest_id=data_ingest.get_entity_id())
    except LookupError as e:
        response.failed(msg=e.args[0], code=400)
        logger.error(e)
        return response.emit(http_code=404)

    except Exception as e:
        response.failed(msg=e.args[0], code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


    #### Special Section for BYOC ####
    ################################## BYOC Section Start
    # TODO: This needs to be more elegantly incorporated into this endpoint

    if data_ingest.ingest_type == "BYOC":
        logger.info("Processing Bring-Your-Own-Container Ingest")

        try:
            configs = data_ingest.ingest_config

            config_user, config_pass = get_db_credentials()
            env_var_to_inject = configs["env_vars"]
            env_var_to_inject['DB_USER'] = config_user
            env_var_to_inject['DB_PASS'] = config_pass
            env_var_to_inject['KML_ING_ID'] = ingest_id

            ingest_yaml = customize_yaml_byoc_ing(
                ing_id=ingest_id,
                byoc_container=configs["src_uri"],
                dest_tabl=configs["table_name"],
                envvars=env_var_to_inject
            )

            api_response = k8_apps_v1_api.create_namespaced_deployment(
                body=ingest_yaml,
                namespace=namespace
            )

            depname = ingest_yaml['metadata']['name']
            data_ingest.get_base_entity().transition_to_starting()

            response.response['deployment'] = depname
            response.succeeded()
            data_ingest.change_state('STARTING')
            log_comms(response.response, "inspect_ingest_start_outbound")
            return response.emit(http_code=200)

        except Exception as e:
            response.failed(msg=e, code=1040)
            logger.error(e)
            logger.debug(traceback.format_exc())
            return response.emit(http_code=509)

    ################################## BYOC Section Completed

    try:
        # Load KIO parameterization from saved ingest
        ingest_config = data_ingest.ingest_config

        # Select the K8 template
        if ingest_config['source'] == 'local':
            job_template = 'yaml/template-ingest-job-fileload.yaml'
        elif ingest_config['source'] == 'kafka':
            job_template = 'yaml/template-ingest-kafka.yaml'
        else:
            job_template = 'yaml/template-ingest-job.yaml'

    except Exception as e:
        response.failed(msg=e.args[0], code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)

    # Load K8 config template
    try:
        path = os.path.dirname(os.path.realpath(__file__))
        ingest_yaml = yaml.load(
            open('/'.join([path, job_template])), Loader=yaml.FullLoader
        )
    except Exception as e:
        response.failed(msg=e.args[0], code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)

    try:
        # Inject new pod name
        ingest_yaml['metadata']['name'] = job_name
        ingest_yaml['metadata']['namespace'] = namespace

        ###############################################
        # Native Kinetica load (ie. table to new table)
        ###############################################

        if ingest_config['source'] == 'kinetica':

            ref_container = get_sys_container("ctnr_kio")
            ingest_yaml['spec']['template']['spec']['containers'][0]['image'] = ref_container

            kc = KCredential.load(ingest_config['cred_id'])
            ingest_cred = KCredential.decrypt_secrets('KINETICA_DB',
                                                      kc.credential_config)

            kio_args = list(['--source'])

            if ingest_cred['user'] and ingest_cred['password']:
                kio_args.append(''.join(['kinetica', '://',
                                         ingest_cred['user'], ':',
                                         ingest_cred['password'], '@',
                                         ingest_cred['connection_str'], '::',
                                         ingest_config['table_name']]))
            else:
                kio_args.append(''.join(['kinetica', '://',
                                         ingest_cred['connection_str'], '::',
                                         ingest_config['table_name']]))

            kio_args.append('--destination')

            if db_user and db_pass:
                kio_args.append(''.join(['kinetica', '://',
                                         db_user, ':',
                                         db_pass, '@',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))
            else:
                kio_args.append(''.join(['kinetica', '://',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))

            if use_ssl:
                kio_args.append('--d-https')
                kio_args.append('--s-https')
                
            if ingest_config.get('collection_name'):
                kio_args.append('--d-collection-name')
                kio_args.append(ingest_config['collection_name'])

            # Spark not available for export so overide with Python engine
            kio_args.append('--engine')
            kio_args.append('odo')

            # Inject new arguments for pod based KIO
            ingest_yaml['spec']['template']['spec']['containers'][0]['args'] = kio_args

        ########################
        # PostgreSQL direct load
        ########################

        if ingest_config['source'] == 'postgresql':

            ref_container = get_sys_container("ctnr_kio")
            ingest_yaml['spec']['template']['spec']['containers'][0]['image'] = ref_container

            kc = KCredential.load(ingest_config['cred_id'])
            ingest_cred = KCredential.decrypt_secrets('POSTGRES',
                                                      kc.credential_config)

            kio_args = list(['--source'])

            if ingest_cred['user'] and ingest_cred['password']:
                kio_args.append(''.join(['postgresql', '://',
                                         ingest_cred['user'], ':',
                                         ingest_cred['password'], '@',
                                         ingest_cred['connection_str'], '/',
                                         str(ingest_cred['pg_database']), '::',
                                         ingest_config['table_name']]))
            else:
                kio_args.append(''.join(['postgresql', '://',
                                         ingest_cred['connection_str'], '/',
                                         str(ingest_cred['pg_database']), '::',
                                         ingest_config['table_name']]))

            kio_args.append('--destination')

            if db_user and db_pass:
                kio_args.append(''.join(['kinetica', '://',
                                         db_user, ':',
                                         db_pass, '@',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))
            else:
                kio_args.append(''.join(['kinetica', '://',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))

            if use_ssl:
                kio_args.append('--d-https')

            if ingest_config.get('collection_name'):
                kio_args.append('--d-collection-name')
                kio_args.append(ingest_config['collection_name'])

            # Inject new arguments for pod based KIO
            ingest_yaml['spec']['template']['spec']['containers'][0]['args'] = kio_args

        ################################################
        # Local csv, parquet, parquetdataset or orc load
        ################################################

        elif ingest_config['source'] == 'local':
            from kml.api.orm import KMLFile

            try:
                path_and_file = KMLFile.load(data_ingest.file_id).file_name

            except LookupError as e:
                response.failed(msg=e.args[0], code=1040)
                logger.error(e)
                logger.debug(traceback.format_exc())
                return response.emit(http_code=404)

            except Exception as e:
                response.failed(msg=e.args[0], code=1040)
                logger.error(e)
                logger.debug(traceback.format_exc())
                return response.emit(http_code=509)

            bb_cmd = list(['wget', '-O', path_and_file])

            bb_cmd.append(''.join([KML_API_BASE,
                                   '/file/',
                                   str(data_ingest.file_id)]))

            kio_args = list(['--source', path_and_file])

            kio_args.append('--destination')

            if db_user and db_pass:
                kio_args.append(''.join(['kinetica', '://',
                                         db_user, ':',
                                         db_pass, '@',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))
            else:
                kio_args.append(''.join(['kinetica', '://',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))

            if ingest_config.get('collection_name'):
                kio_args.append('--d-collection-name')
                kio_args.append(ingest_config['collection_name'])

            # Inject new arguments for pod based KIO
            ingest_yaml['spec']['template']['spec']['containers'][0]['args'] = kio_args
            ingest_yaml['spec']['template']['spec']['initContainers'][0]['command'] = bb_cmd

        ############################################
        # AWS S3 csv, parquet or parquetdataset load
        ############################################

        elif ingest_config['source'] == 's3':

            ref_container = get_sys_container("ctnr_kio")
            ingest_yaml['spec']['template']['spec']['containers'][0]['image'] = ref_container

            kc = KCredential.load(ingest_config['cred_id'])
            ingest_cred = KCredential.decrypt_secrets('AWS_S3', kc.credential_config)

            kio_args = list(['--source'])

            if ingest_cred.get('s3_path'):
                kio_args.append(''.join(['s3', '://',
                                         ingest_cred['s3_bucket'], '/',
                                         ingest_cred['s3_path'], '/',
                                         ingest_config['file_name']]))
            else:
                kio_args.append(''.join(['s3', '://',
                                         ingest_cred['s3_bucket'], '/',
                                         ingest_config['file_name']]))

            kio_args.append('--destination')

            if db_user and db_pass:
                kio_args.append(''.join(['kinetica', '://',
                                         db_user, ':',
                                         db_pass, '@',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))
            else:
                kio_args.append(''.join(['kinetica', '://',
                                         db_host_port_str, '::',
                                         ingest_config['target_table']]))

            if use_ssl:
                kio_args.append('--d-https')
                
            if ingest_cred.get('aws_access_key_id'):
                kio_args.append('--aws-access-key-id')
                kio_args.append(ingest_cred['aws_access_key_id'])
                kio_args.append('--aws-secret-access-key')
                kio_args.append(ingest_cred['aws_secret_access_key'])

            s3_file_name = ingest_config['file_name']
            s3_file_type = s3_file_name[s3_file_name.rindex('.')+1:]

            if s3_file_type == 'csv':
                if ingest_config.get('delimiter'):
                    delim = ingest_config['delimiter']
                    kio_args.append('--s-delimiter')
                    kio_args.append(f'{delim}')
                if ingest_config.get('headerless'):
                    kio_args.append('--s-headerless')
                if ingest_config.get('escape_character'):
                    esc_char = ingest_config['escape_character']
                    kio_args.append('--s-escape-character')
                    kio_args.append(f'{esc_char}')

            if s3_file_type in ['parquet', 'parquetdataset']:
                if ingest_config.get('recursive'):
                    kio_args.append('--s-recursive')

            if ingest_config.get('collection_name'):
                kio_args.append('--d-collection-name')
                kio_args.append(ingest_config['collection_name'])

            # Inject new arguments for pod based KIO
            ingest_yaml['spec']['template']['spec']['containers'][0]['args'] = kio_args

        #################################
        # Kafka producer stream connector
        #################################

        elif ingest_config['source'] == 'kafka':

            # Set container image
            ref_container = get_sys_container("ctnr_kafkaconnector")
            ingest_yaml['spec']['template']['spec']['containers'][0]['image'] = ref_container

            # NOTE: If the connector name is unique, it will always retrieve from
            # begining on a particular topic. If the standard connector name is used
            # the load will begin only with what was added to the broker queue since
            # the last connection.
            #
            import uuid
            connector_name = f"Connector_{str(uuid.uuid4())}"
            #  connector_name = 'KafkaConnector_001'

            kc = KCredential.load(ingest_config['cred_id'])
            ingest_cred = KCredential.decrypt_secrets('KAFKA', kc.credential_config)

            # Create the sink properties file
            sink = []

            sink.append(f'name={connector_name}')
            sink.append(f"topics={ingest_cred['topics']}")
            sink.append('connector.class=com.kinetica.kafka.KineticaSinkConnector')
            sink.append('tasks.max=1')
            sink.append(f"kinetica.url={db_conn_str}")

            if db_user and db_pass:
                sink.append(f"kinetica.username={db_user}")
                sink.append(f"kinetica.password={db_pass}")

            sink.append('kinetica.timeout=120000')
            sink.append('kinetica.batch_size=100')
            sink.append('kinetica.create_table=true')
            sink.append('kinetica.retry_count=3')
            sink.append('kinetica.single_table_per_topic=true')

            target_table = ingest_config['target_table']

            if float(kml_get_db_track()) >= 7.1:
                if len(target_table.split('.')) == 1:
                    logger.info(
                        'Adding schema ki_home for Kafka connector compatibility'
                    )
                    target_table = f"ki_home.{target_table}"
            sink.append(f"kinetica.dest_table_override={target_table}")

            # Disabled due to 7.0 style collection specification incompatability
            # with current Kafka connector
            #
            #  if kml_get_db_track() == '7.0':
            #      logger.debug('** is 7.0')
            #      if ingest_config.get('collection_name'):
            #          logger.debug('** has collection_name')
            #          sink.append(
            #              f"kinetica.collection_name={ingest_config['collection_name']}"
            #          )

            sink.append('kinetica.add_new_fields_as_columns=false')
            sink.append('kinetica.make_missing_field_nullable=false')

            sink_inject = '\n'.join(sink)

            # AUTHENTICATION METHOD: KEYSTORE w/ Password
            if 'auth_method' in ingest_cred and ingest_cred['auth_method']=='KEYSTORE':

                cont_path = '/opt/kafka'
                inject_kinetica_conn_config = f'echo -e \"{sink_inject}\" > {cont_path}/config/quickstart-kinetica-sink.properties && cat {cont_path}/config/quickstart-kinetica-sink.properties'
                start_connector = f'{cont_path}/bin/connect-standalone.sh {cont_path}/config/connect-standalone.properties {cont_path}/config/quickstart-kinetica-sink.properties'

                if ingest_cred['key_password'] and ingest_cred['keystore']:
                    logger.debug("Authenticating with Keystore w/ password")
                    inject_key = f'echo -e \"{ingest_cred["keystore"]}\" > /tmp/keystore.b64'
                    decode_key = 'base64 -d /tmp/keystore.b64 > /tmp/kafka.client.truststore.jks'

                    conn_inject = KEYSTORE_PROPERTIES_WITH_PASS
                    conn_inject = conn_inject.replace("__INSERT_HOST_PORT__", ingest_cred['connection_str'])
                    conn_inject = conn_inject.replace("__INSERT_PASSWORD__", ingest_cred['key_password'])
                    inject_kafka_sa_config = f'echo -e \"{conn_inject}\" > {cont_path}/config/connect-standalone.properties && cat {cont_path}/config/connect-standalone.properties'

                    init_args = ' && '.join([inject_key,
                                            decode_key,
                                            inject_kinetica_conn_config,
                                            inject_kafka_sa_config,
                                            start_connector])
                else:
                    logger.debug("Authenticating with Keystore w/o password")
                    conn_inject = KEYSTORE_PROPERTIES_ZERO_PASS
                    conn_inject = conn_inject.replace("__INSERT_HOST_PORT__", ingest_cred['connection_str'])
                    inject_kafka_sa_config = f'echo -e \"{conn_inject}\" > {cont_path}/config/connect-standalone.properties && cat {cont_path}/config/connect-standalone.properties'
                    init_args = ' && '.join([inject_kinetica_conn_config,
                                            inject_kafka_sa_config,
                                            start_connector])

            # AUTHENTICATION METHOD: KEYSTORE w/o Password
            elif 'auth_method' not in ingest_cred or ingest_cred['auth_method']=='NONE':

                logger.debug("Authenticating with Keystore w/o password")

                cont_path = '/opt/kafka'
                inject_kinetica_conn_config = f'echo -e \"{sink_inject}\" > {cont_path}/config/quickstart-kinetica-sink.properties && cat {cont_path}/config/quickstart-kinetica-sink.properties'
                start_connector = f'{cont_path}/bin/connect-standalone.sh {cont_path}/config/connect-standalone.properties {cont_path}/config/quickstart-kinetica-sink.properties'

                logger.debug("Using empty password")
                conn_inject = KEYSTORE_PROPERTIES_ZERO_PASS
                conn_inject = conn_inject.replace("__INSERT_HOST_PORT__", ingest_cred['connection_str'])
                inject_kafka_sa_config = f'echo -e \"{conn_inject}\" > {cont_path}/config/connect-standalone.properties && cat {cont_path}/config/connect-standalone.properties'
                init_args = ' && '.join([inject_kinetica_conn_config,
                                        inject_kafka_sa_config,
                                        start_connector])

            # AUTHENTICATION METHOD: SSL w/ Plaintext SASL
            elif 'auth_method' in ingest_cred and ingest_cred['auth_method']=='SASL_SSL__PLAIN':
                logger.debug("Authenticating with SSL using SASL w/ Plaintext password")
                conn_inject = SASL_PROPERTIES
                conn_inject = conn_inject.replace("__INSERT_HOST_PORT__", ingest_cred['connection_str'])
                conn_inject = conn_inject.replace("__INSERT_USERNAME__", ingest_cred['username'])
                conn_inject = conn_inject.replace("__INSERT_PASSWORD__", ingest_cred['key_password'])

                cont_path = '/opt/kafka'
                inject_kinetica_conn_config = f'echo -e \"{sink_inject}\" > {cont_path}/config/quickstart-kinetica-sink.properties && cat {cont_path}/config/quickstart-kinetica-sink.properties'
                inject_kafka_sa_config = f'echo -e \"{conn_inject}\" > {cont_path}/config/connect-standalone.properties && cat {cont_path}/config/connect-standalone.properties'
                start_connector = f'{cont_path}/bin/connect-standalone.sh {cont_path}/config/connect-standalone.properties {cont_path}/config/quickstart-kinetica-sink.properties'

                init_args = ' && '.join([inject_kinetica_conn_config,
                                         inject_kafka_sa_config,
                                         start_connector])

            ingest_yaml['spec']['template']['spec']['containers'][0]['args'] = ['-c', init_args]

        # Deploy the KIO tool with the tranfer specifications
        if ingest_config['source'] != 'kafka':
            try:
                api_response = k8_batch_v1_api.create_namespaced_job(
                    body=ingest_yaml,
                    namespace=namespace
                )

                data_ingest.get_base_entity().transition_to_starting()

                response.response['job'] = job_name
                response.response['entity_id'] = ingest_id
                response.response['target_table'] = ingest_config['target_table']
                response.succeeded()
                data_ingest.change_state('STARTING')
                return response.emit(http_code=200)

            except Exception as e:
                response.failed(msg=e.args[0], code=1040)
                logger.error(e)
                logger.debug(traceback.format_exc())
                return response.emit(http_code=509)

        # Deploy in continuous mode as a K8s deployment
        else:
            try:
                api_response = k8_apps_v1_api.create_namespaced_deployment(
                    body=ingest_yaml,
                    namespace=namespace
                )

                data_ingest.get_base_entity().transition_to_starting()
                response.response['deployment'] = job_name
                response.response['entity_id'] = ingest_id
                data_ingest.get_base_entity().transition_to_starting()
                response.response['target_table'] = ingest_config['target_table']
                response.succeeded()
                data_ingest.change_state('STARTING')
                return response.emit(http_code=200)

            except Exception as e:
                response.failed(msg=e.args[0], code=1040)
                logger.error(e)
                logger.debug(traceback.format_exc())
                return response.emit(http_code=509)

    except Exception as e:
        response.failed(msg=e.args[0], code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(http_code=509)


async def ingest_setstatus(request):
    from kml.api.orm import Ingest, Entity

    response = Responder()
    response.response['state-modified'] = False
    try:
        ingest_id = request.path_params.get('ingest_id')
        spec = await request.json()

        ig = Ingest.load(ingest_id)
        destination_state = spec['destination-state']
        if destination_state not in Entity.ENTITY_STATES:
            raise ValueError("Unknown destination state %s" % destination_state)
        ig.change_state(destination_state)
        response.response['state-modified'] = True
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)
