"""KML ingest endpoint controller."""

import traceback

from kml.api.aux import (
    attach_log,
    Responder,
    log_comms
)
from kml.api.kml_config import (
    get_api_base,
    get_components,
    get_db_conn_str,
    get_db_credentials,
    get_namespace
)
from kml.api.utils import (
    string_is_not_blank
)
from kml.api.orm import KCredential


# Configuration
KML_API_BASE = get_api_base()
_, db_host, db_port = get_components(conn='db')
db_conn_str = get_db_conn_str()
db_user, db_pass = get_db_credentials()
namespace = get_namespace()

# Configure log file
logger = attach_log(module=__name__, log_name='kml', debug=True)


async def ingest_import(request):

    from kml.api.orm import Ingest

    # Create and initialize a responder object
    response = Responder()

    # Parse http body and write to system tables
    try:
        spec = await request.json()
        log_comms(spec, "inspect_ingest_import_inbound")

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(400)

    # Create a single ingest instance
    try:

        if "src_uri" not in spec["ingest_config"]:
            msg = ("BYOC Ingest Import request missing source Docker Repository "
                   "via payload parameter src_uri")
            response.failed(msg=msg, code=1040)
            logger.error(msg)
            return response.emit(422)

        if not string_is_not_blank(spec["ingest_config"]["src_uri"]):
            msg = ("BYOC Ingest Import request has empty source Docker Repository "
                   "via payload parameter src_uri")
            response.failed(msg=msg, code=1040)
            logger.error(msg)
            return response.emit(422)

        # TODO: Test for input parameter name
        # TODO: Test for input parameter desc
        # TODO: Test for input parameter ingest_mode

        data_ingest = Ingest.create(
            ingest_name=spec['name'].strip(),
            ingest_desc=spec['desc'].strip(),
            ingest_mode=spec['ingest_mode'],
            ingest_type="BYOC",
            ingest_config=spec["ingest_config"]
        )
        data_ingest.publish()
        data_ingest.get_base_entity().transition_to_created()

    except Exception as e:
        response.failed(msg=e, code=1040)
        logger.error(e)
        logger.debug(traceback.format_exc())
        return response.emit(509)

    # Store the ingest instance in the KML system table
    response.response['created'] = data_ingest.serialize()
    response.succeeded()
    log_comms(response.response, "inspect_ingest_import_outbound")
    logger.info(f"BYOC ingest create (id = {data_ingest.get_entity_id()})")
    return response.emit(200)
