"""KML model endpoint controller."""

import traceback

from kml.api.aux import (
    attach_log,
    log_comms,
    Responder
)
from kml.api.orm import ModelInstance
from kml.api.kml_config import get_namespace

# Configuration
namespace = get_namespace()

logger = attach_log(module=__name__, log_name='kml', debug=True)


async def model_inst_create(request):
    response = Responder()

    try:
        spec = await request.json()

        log_comms(spec, "inbound_mi_create")

        if "model_inst_name" not in spec:
            response.failed(msg="Bad create request: missing Model Instance "
                                "Name parameter (model_inst_name)",
                            code=1060)
            return response.emit(422)

        if "model_inst_desc" not in spec:
            response.failed(msg="Bad create request: missing Model Instance Description "
                                "parameter (model_inst_desc)",
                            code=1060)
            return response.emit(422)

        if "model_inst_config" not in spec:
            response.failed(msg="Bad create request: missing Model Instance Config "
                                "parameter (model_inst_config)",
                            code=1060)
            return response.emit(422)

        if "model_type" not in spec["model_inst_config"]:
            response.failed(msg="Bad create request: missing model_type parameter",
                            code=1060)
            return response.emit(422)

        if spec["model_inst_config"]["model_type"] not in ModelInstance.MODEL_TYPES:
            response.failed(msg="Bad create request: model_type parameter not valid",
                            code=1060)
            return response.emit(422)

        # Trim leading and trailing whitespace from input strings
        for key in ['container', 'blackbox_module', 'blackbox_function']:
            spec['model_inst_config'][key] = spec['model_inst_config'][key].strip()

        # TODO: Replace this with in-line if-else
        my_resource_config = None
        if "resource_config" in spec:
            my_resource_config = spec["resource_config"]

        featureset_id = None

        if spec["model_inst_config"]["model_type"] != "BLACKBOX":
            raise ValueError("Not currently supported on standard build")

        if spec["model_inst_config"]["model_type"] == "BLACKBOX":
            logger.info("Creating Model instance for BLACKBOX")

            if "credential_id_docker" in spec:
                logger.info("Creating BlackBox Model with user-provided Docker "
                            "Credential ID for Private Registry")
                credential_id_docker = spec["credential_id_docker"]

                if credential_id_docker == -1 or credential_id_docker in ["null", "none"]:
                    logger.info(f"Ignoring empty credential_id_docker: "
                                f"{credential_id_docker}")
                else:
                    from kml.api.orm import KCredential
                    if not KCredential.exists(existing_credential_id=credential_id_docker):
                        logger.error(f"Referenced Credential ID({credential_id_docker}) "
                                     f"not found on system")
                        # TODO: should we abort this operation?
                    else:
                        logger.info(f"Utilizing Docker Credential with id "
                                    f"{credential_id_docker}")

                    if KCredential.load(
                            existing_credential_id=credential_id_docker
                    ).credential_type != "DOCKER":
                        logger.error(f"Referenced Credential ID({credential_id_docker}) "
                                     f"is not a Docker credential")
                        # TODO: should we abort this operation?
                    else:
                        # TODO: show some descriptive stuff here about the private repo, like the URI
                        logger.info(f"Successfully loaded Docker Private Registry Credential "
                                    f"via id {credential_id_docker}")
                        spec["model_inst_config"]["credential_id"] = credential_id_docker
                        spec["model_inst_config"]["k8s_secrets_handle"] = KCredential.load(
                            existing_credential_id=credential_id_docker
                        ).get_handle()

            # if "featureset_id" in spec:
            #     logger.info("Creating BlackBox Model with user-provided FeatureSet")
            #     featureset_id = spec["featureset_id"]
            #     logger.info(f"Utilizing FeatureSet with id {featureset_id}")
            # else:
            #     logger.info("Creating BlackBox Model with sys-generated stub-identity FeatureSet")
            #     model_inst_name = spec["model_inst_name"]

            #     fs_identity_stub = FeatureSet.create(
            #         featureset_name=f"stub for {model_inst_name}",
            #         featureset_desc=f"stub for {model_inst_name}",
            #         config_inline={},
            #         config_relational={}
            #     )

            #     featureset_id = fs_identity_stub.get_entity_id()
            # logger.info(f"Utilizing FeatureSet with id {featureset_id}")

        mitemp = ModelInstance.create(
            model_inst_name=spec["model_inst_name"].strip(),
            model_inst_desc=spec["model_inst_desc"].strip(),
            # base_model_id=spec["base_model_id"],
            model_inst_config=spec["model_inst_config"],
            resource_config=my_resource_config,
            featureset_id=featureset_id
        )
        mitemp.publish()
        mitemp.get_base_entity().transition_to_created()

        logger.info(f"Created Model Instance {mitemp.get_entity_id()}")
        response.response["created"] = mitemp.serialize()
        response.succeeded()
        log_comms(response.response, "outbound_mi_deploy")
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1060)
        return response.emit(509)


async def model_inst_push_metrics(request):
    response = Responder()

    try:
        model_instance_id = request.path_params.get('model_instance_id')
        mitemp = ModelInstance.load(model_instance_id)
        inmetrics = await request.json()
        mitemp.set_metrics(inmetrics)
        logger.debug(f"Updated Model Instance {model_instance_id} with training metrics")
        response.response["updated"] = mitemp.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1060)
        return response.emit(509)


def model_inst_view(request):
    response = Responder()

    try:
        model_instance_id = request.path_params.get('model_instance_id')
        mitemp = ModelInstance.load(model_instance_id)
        response.response["item"] = mitemp.serialize()
        response.response["found"] = True
        #  logger.info(f"Returning Model Instance with ID {mitemp.get_entity_id()}")
        response.succeeded()
        return response.emit(201)

    except LookupError as e:
        response.failed(msg=e, code=404)
        # logger.error(e)
        return response.emit(http_code=404)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1060)
        return response.emit(509)


def model_inst_clone(request):
    response = Responder()

    try:
        model_instance_id = request.path_params.get('model_instance_id')
        miclone = ModelInstance.clone(model_instance_id)
        miclone.publish()
        miclone.get_base_entity().transition_to_created()
        response.response["created"] = miclone.serialize()
        response.succeeded()
        return response.emit(201)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1060)
        return response.emit(509)


def model_inst_archive(request):
    response = Responder()

    try:
        model_instance_id = request.path_params.get('model_instance_id')
        mitemp = ModelInstance.load(model_instance_id)
        mitemp.tombstone()
        response.response["item"] = mitemp.serialize()
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1060)
        return response.emit(509)
