"""RapidsLab (Manageaged and embedded Jupyterlab) controller."""

import traceback

from starlette.authentication import AuthenticationError

from kml.api.aux import (
    attach_log,
    Responder
)
from kml.api.kml_config import (
    get_namespace,
    get_k8_conn,
    get_sys_schemas,
    get_db_credentials,
    get_public_connections,
    get_ingress_controller_details
)
from kml.api.kmldbops import get_tbl_handle as kml_get_tbl_handle
from kml.api.controller_utils import check_db_credentials
from kml.api.utils import get_public_k8s_ip
from kml.api.naming_svc import (
        get_cluster_name_jupyter,
        get_jupyter_default_slug,
        get_cluster_name_service,
        get_cluster_name_ingress
)

# Import credentials from the config file
config_user, config_pass = get_db_credentials()
if config_user and config_pass:
    db_user = config_user
    db_pass = config_pass
    check_db_credentials(db_user, db_pass)
else:
    db_user = ''
    db_pass = ''

# Read config
namespace = get_namespace()
SCHEMA_SYS, SCHEMA_AUDIT, _, _ = get_sys_schemas()
_, public_db_conn, _ = get_public_connections()

# Logging module
logger = attach_log(module=__name__, log_name='kml', debug=True)

# Defaults
jupyter_store = 'sys_kml_user_data'


async def jupyter_start(request):
    """Launch the Jupyter Lab container with RAPIDS connector."""

    import base64
    from kml.api.yaml_customizer import customize_yaml_rapids_jupyter
    from kml.api.utils import compute_resources_available

    (k8_apps_v1_api,
     k8_core_v1_api,
     k8_custom_objects_api,
     k8_batch_v1_api,
     k8_ext_v1_api) = get_k8_conn()

    response = Responder()

    try:
        spec = await request.json()
        load_override = 1 if spec.get('override') else 0
        ingress_ip = get_public_k8s_ip()

        # Check if a GPU is available for deployment
        if not compute_resources_available(replicas=1, allocation_per_replica=1):
            msg = 'Insufficient GPU resources'
            response.response.update(
                jupyter_url=None,
                state='stopped',
                load_override=load_override
            )
            response.failed(msg=msg, code=503)
            return response.emit(503)

        # Determine username (aka profile) from the req header
        if 'Authorization' in request.headers:
            auth = request.headers['Authorization']
            scheme, credentials = auth.split()

            if scheme.lower() != 'basic':
                raise AuthenticationError('Auth header is not "basic"')

            decoded = base64.b64decode(credentials).decode('ascii')
            profile, _, _ = decoded.partition(':')

        else:
            msg = 'User not logged in or credentials missing from request header'
            response.failed(msg=msg, code=1040)
            return response.emit(509)

        # Check status of Jupyter environment
        tbl_hdl = kml_get_tbl_handle(jupyter_store, schema=SCHEMA_SYS)
        mode = tbl_hdl.get_records_by_column(
            column_names='is_active',
            options={f'expression': f'profile="{profile}"'}
        )

        # Check for first time use of this user profile
        if not mode:
            tbl_hdl.insert_records(
                [{
                    'profile': profile,
                    'is_active': 0,
                    'modified_dt': None,
                    'notebooks_data': None,
                    'jupyter_url': None,
                    'jupyter_deployment': None
                }]
            )
            mode = {'is_active': [0]}

        # Check for running Jupyter Lab for this user
        jupyter_dep = get_cluster_name_jupyter(profile)
        pod_state = get_jupyter_running_state(jupyter_dep)

        if pod_state != 'stopped':

            status = tbl_hdl.get_records_by_column(
                column_names=['jupyter_url', 'jupyter_deployment'],
                options={f'expression': f'profile="{profile}"'}
            )

            jupyter_url = status['jupyter_url'][0] if status.get('jupyter_url') else None

            msg = (f"Jupyter Lab environment for user profile '{profile}' is running or in "
                   "the process of shutting down")

            response.response.update(
                jupyter_url=jupyter_url,
                state=pod_state,
                load_override=load_override
            )
            response.failed(msg=msg, code=403)
            return response.emit(403)

        # Return if Jupyter environment is potentially in use
        if mode['is_active'][0] == 1 and load_override == 0:

            msg = (f"Jupyter Lab environment for user profile '{profile}' did not exit "
                   "cleanly. Resend with override parameter set to 'true' to force load")

            status = tbl_hdl.get_records_by_column(
                column_names=['jupyter_url', 'jupyter_deployment'],
                options={f'expression': f'profile="{profile}"'})

            if status.get('jupyter_url') and status.get('jupyter_deployment'):
                response.response.update(
                    jupyter_url=status['jupyter_url'][0],
                    state=pod_state,
                    load_override=load_override
                )
                response.failed(msg=msg, code=403)
                return response.emit(403)

        # Check that a zombie clusterIP service does not exist
        jupyter_svc = get_cluster_name_service(profile)
        svcs = k8_core_v1_api.list_namespaced_service(namespace=namespace)
        for svc in svcs.items:
            if svc.metadata.name == jupyter_svc:
                try:
                    logger.warning('Found unattached service. Removing from K8s')
                    resp_svc = k8_core_v1_api.delete_namespaced_service(
                        name=jupyter_svc,
                        namespace=namespace
                    )
                except Exception as e:
                    logger.warning(e)
                    pass

        # Generate a path for the RapidsLab ingress
        if spec.get('custom_path_slug'):
            slug = spec.get('custom_path_slug').strip()
        else:
            slug = get_jupyter_default_slug(profile)

        # Proceeding with Jupyter Lab launch
        envvars = {
            'KML_PROFILE': profile,
            'KML_DB_CONN_STR': public_db_conn,
            'KML_JUPYTER_STORE': jupyter_store,
            'KML_LOAD_OVERRIDE': str(load_override),
            'KML_USER': db_user,
            'KML_PASSWD': db_pass,
            'KML_ING_PATH': slug
        }

        # Inject parameters into k8s deployment specs
        jupyter_yaml, svc_yaml, ingress_yaml = customize_yaml_rapids_jupyter(
            profile=profile,
            envvars=envvars,
            slug=slug
        )

        # Launch Jupyter Lab
        resp_jupyter = k8_apps_v1_api.create_namespaced_deployment(
            body=jupyter_yaml,
            namespace=namespace
        )

        # Launch service routing pod
        resp_svc = k8_core_v1_api.create_namespaced_service(
            body=svc_yaml,
            namespace=namespace
        )

        # Create ingress definition
        resp_ingress = k8_ext_v1_api.create_namespaced_ingress(
            body=ingress_yaml,
            namespace=namespace
        )

        # Prepare response
        _, _, ing_http_port, ing_https_port = get_ingress_controller_details()
        if ing_http_port:
            jupyter_url = f"http://{ingress_ip}:{ing_http_port}/{slug}"
        elif ing_https_port:
            jupyter_url = f"https://{ingress_ip}:{ing_https_port}/{slug}"
        else:
            jupyter_url = f"http://{ingress_ip}/{slug}"
        logger.debug(f"RapidsLab URI: {jupyter_url}")
        dep_name = resp_jupyter.metadata.name

        # Persist current URL and deployment name to the user data table
        tbl_hdl.update_records(
            expressions=[f'profile="{profile}"'],
            new_values_maps={
                'jupyter_url': jupyter_url,
                'jupyter_deployment': dep_name
            }
        )

        response.response.update(
            jupyter_url=jupyter_url,
            state='starting',
            load_override=load_override,
            jupyter_dep=dep_name
        )
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def get_jupyter_running_state(dep_name):
    """Return actual operational state of a pod (located by parent run label)."""

    (k8_apps_v1_api,
     k8_core_v1_api,
     k8_custom_objects_api,
     k8_batch_v1_api,
     k8_ext_v1_api) = get_k8_conn()

    dep_active = False
    pod_active = False

    try:
        pods = k8_core_v1_api.list_namespaced_pod(namespace=namespace)
        deps = k8_apps_v1_api.list_namespaced_deployment(namespace=namespace)

        for dep in deps.items:
            if dep.metadata.name:
                if dep.metadata.name == dep_name:
                    dep_active = True

        for pod in pods.items:
            if pod.metadata.labels:
                if 'run' in pod.metadata.labels:
                    if pod.metadata.labels['run'] == dep_name:
                        if pod.status.phase == 'Running':
                            pod_active = True

        if dep_active and pod_active:
            return 'started'
        elif dep_active and not pod_active:
            return 'starting'
        elif not dep_active and pod_active:
            return 'stopping'
        else:
            return 'stopped'

    except Exception as e:
        raise ValueError(e)


def jupyter_status(request):
    """Return metadata about a users Jupyter Lab session."""

    import base64
    from kubernetes.client import V1DeleteOptions

    delete_options = V1DeleteOptions()
    (k8_apps_v1_api,
     k8_core_v1_api,
     k8_custom_objects_api,
     k8_batch_v1_api,
     k8_ext_v1_api) = get_k8_conn()

    response = Responder()

    try:
        # Determine username (aka profile) from the req header
        if 'Authorization' in request.headers:
            auth = request.headers['Authorization']
            scheme, credentials = auth.split()

            if scheme.lower() != 'basic':
                raise AuthenticationError('Auth header is not "basic"')

            decoded = base64.b64decode(credentials).decode('ascii')
            profile, _, _ = decoded.partition(':')

        else:
            msg = 'User not logged in or credentials missing from request header'
            response.failed(msg=msg, code=1040)
            return response.emit(509)

        # Check to see if Jupyter Lab (Rapids) is running for this user
        tbl_hdl = kml_get_tbl_handle(jupyter_store, schema=SCHEMA_SYS)

        # Continue to determine URL and dep name
        status = tbl_hdl.get_records_by_column(
            column_names=['modified_dt',
                          'jupyter_url',
                          'jupyter_deployment'],
            options={f'expression': f'profile="{profile}"'}
        )

        if status:

            if status['modified_dt'][0]:
                dt = status['modified_dt'][0].split(':')
                last_used = ':'.join([dt[0], dt[1]])
            else:
                last_used = None

            if status['jupyter_deployment'][0]:
                jupyter_dep = status['jupyter_deployment'][0]
            else:
                jupyter_dep = get_cluster_name_jupyter(profile)

            jupyter_url = status['jupyter_url'][0] if status.get('jupyter_url')[0] else None

        else:
            last_used = None
            jupyter_url = None
            jupyter_dep = None

        # Poll K8s cluster for accurate running state
        pod_state = get_jupyter_running_state(jupyter_dep)

        # Check for errors in URL and running state
        if pod_state == 'started' and not jupyter_url:
            jpt_svc_name = get_cluster_name_service(profile)
            jpt_igr_name = get_cluster_name_ingress(profile)

            # On error, terminate the deployment, supporting service and ingress
            resp_jupyter = k8_apps_v1_api.delete_namespaced_deployment(
                name=jupyter_dep,
                namespace=namespace,
                body=delete_options
            )
            resp_svc = k8_core_v1_api.delete_namespaced_service(
                name=jpt_svc_name,
                namespace=namespace,
                body=delete_options
            )
            resp_ingress= k8_ext_v1_api.delete_namespaced_ingress(
                name=jpt_igr_name,
                namespace=namespace,
                body=delete_options
            )

            pod_state = 'stopping'

        response.response.update(
            jupyter_url=jupyter_url,
            jupyter_dep=jupyter_dep,
            last_used=last_used,
            profile=profile,
            state=pod_state
        )
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def jupyter_stop(request):
    """Terminate a Jupyter Lab deployment and it's associated service pod."""

    import base64
    from kubernetes.client import V1DeleteOptions

    delete_options = V1DeleteOptions()
    (k8_apps_v1_api,
     k8_core_v1_api,
     k8_custom_objects_api,
     k8_batch_v1_api,
     k8_ext_v1_api) = get_k8_conn()

    response = Responder()

    try:
        # Determine username (aka profile) from the req header
        if 'Authorization' in request.headers:
            auth = request.headers['Authorization']
            scheme, credentials = auth.split()

            if scheme.lower() != 'basic':
                raise AuthenticationError('Auth header is not "basic"')

            decoded = base64.b64decode(credentials).decode('ascii')
            profile, _, _ = decoded.partition(':')

        else:
            msg = 'User not logged in or credentials missing from request header'
            response.failed(msg=msg, code=1040)
            return response.emit(509)

        # Check to see if Jupyter Lab (Rapids) has a DB entry
        # otherwise create deployment name as a fallback
        tbl_hdl = kml_get_tbl_handle(jupyter_store, schema=SCHEMA_SYS)
        try:
            status = tbl_hdl.get_records_by_column(
                column_names=['jupyter_deployment'],
                options={f'expression': f'profile="{profile}"'}
            )
            dep_name = status['jupyter_deployment'][0]
            if not dep_name:
                dep_name = get_cluster_name_jupyter(profile)

        except Exception as e:
            # Fall back to standard deployment name
            logger.warning(e)
            dep_name = get_cluster_name_jupyter(profile)

        # Determine the run state of a deployment
        pod_state = get_jupyter_running_state(dep_name)

        if pod_state == 'stopped':

            msg = f"Jupyter Lab for user '{profile}' is not running."
            response.response.update(terminated=False,
                                     jupyter_dep=dep_name)
            response.failed(msg=msg, code=403)
            return response.emit(403)

        else:

            # terminate jupyter deployment
            try:
                _ = k8_apps_v1_api.delete_namespaced_deployment(
                    name=dep_name,
                    namespace=namespace,
                    body=delete_options
                )
            except Exception as e:
                logger.warning(f"Error terminating Jupyter deployment {dep_name}, proceeding")
                logger.warning(str(e))

            # terminate service pod
            jpt_svc_name = get_cluster_name_service(profile)

            try:
                _ = k8_core_v1_api.delete_namespaced_service(
                    name=jpt_svc_name,
                    namespace=namespace,
                    body=delete_options
                )
            except Exception as e:
                logger.warning(f"Error terminating service pod {jpt_svc_name}, proceeding")
                logger.warning(str(e))

            # remove ingress definition
            jpt_igr_name = get_cluster_name_ingress(profile)
            try:
                _ = k8_ext_v1_api.delete_namespaced_ingress(
                    name=jpt_igr_name,
                    namespace=namespace,
                    body=delete_options
                )
            except Exception as e:
                logger.warning(f"Error removing ingress definition {jpt_igr_name}, proceeding")
                logger.warning(str(e))

            response.response.update(terminated=True, jupyter_dep=dep_name)
            response.succeeded()
            return response.emit(200)


    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)
