"""KML direct deployment controller."""

import os
import yaml
import time
import threading
from hashlib import sha256
from datetime import datetime

import gpudb
from kubernetes.client import V1DeleteOptions

from kml.api.aux import (
    attach_log,
    Responder
)
from kml.api.kml_config import (
    get_db_conn,
    get_db_conn_str,
    get_api_base,
    get_k8_conn,
    get_namespace,
    get_db_credentials,
    get_sys_schemas,
    get_public_connections
)

# Configuration
db = get_db_conn()
namespace = get_namespace()
db_conn_str = get_db_conn_str()
SCHEMA_SYS, _, _, _ = get_sys_schemas()
WATCHER_TIMEOUT = 10

# Import credentials from the config file
db_user, db_pass = get_db_credentials()

# Instantiate logger
logger = attach_log(module=__name__, log_name='kml', debug=True)

# Database table for tracking direct launch containers
tracking_tbl_hdl = gpudb.GPUdbTable(
    name=f'{SCHEMA_SYS}.sys_kml_slipway',
    db=db
)

# Get public connection strings for containers
public_api_base, public_db_conn, public_tm_conn = get_public_connections()


def delete_deployment(name):
    """Teardown kubernetes deployment and delete record."""

    k8_apps_v1_api, _, _, _, _ = get_k8_conn()

    delete_options = V1DeleteOptions(
        propagation_policy='Background',
        grace_period_seconds=1
    )

    try:
        resp = k8_apps_v1_api.delete_namespaced_deployment(
            name=name,
            namespace=namespace,
            body=delete_options,
        )
        tracking_tbl_hdl.delete_records(
            expressions=[f'deployment_name="{name}"']
        )
        return True

    except Exception as e:
        logger.warning(f'Unable to terminate deployment {name}, {e.reason}')
        logger.info(f'Removing record of deployment {name} from the system')
        tracking_tbl_hdl.delete_records(
            expressions=[f'deployment_name="{name}"']
        )
        return False


def pod_launch_watcher(name, replicas):
    """Track container launch and set terminal state."""

    k8_apps_v1_api, _, _, _, _ = get_k8_conn()
    running = False
    polls = 0

    while polls < WATCHER_TIMEOUT and not running:

        polls += 1

        try:
            resp = k8_apps_v1_api.read_namespaced_deployment(
                name=name,
                namespace=namespace
            )
            if resp.status.available_replicas == replicas:
                logger.info(f'Replica set for deployment {name} is available')
                tracking_tbl_hdl.update_records(
                    expressions=f'deployment_name = "{name}"',
                    new_values_maps={'status': 'running'}
                )                                                              

                running = True

        except Exception as e:
            pass

        time.sleep(1)

    if not running:
        logger.warning(
            'Running state for all pods not achived, watcher thread timing out'
        )
        tracking_tbl_hdl.update_records(
            expressions=f'deployment_name = "{name}"',
            new_values_maps={'status': 'failed'}
        )                                                              


def parse_ctnr_name(container):
    """Parse fully qualified container name."""

    repo = None
    name = None
    tag = None

    if '/' in container:
        if ':' in container:
            repo, ctnr = container.split('/')
            name, tag = ctnr.split(':')
        else:
            repo, name = container.split('/')
    elif ':' in container:
        name, tag = container.split(':')
    else:
        name = container

    return repo, name, tag


def gen_unique_dep_name():
    """Generate a unique deployment name."""

    suffix = sha256(str(time.time()).encode('utf-8')).hexdigest()[:8]

    return f'kml-slp-{suffix}'


def gen_dep_manifest(container, name, replicas, env_vars, run_cmd):
    """Create a customized deployment manifest."""

    template = 'tmplt-kml-generic.yaml'
    path = os.path.dirname(os.path.realpath(__file__))
    manifest = yaml.load(
        open(f'{path}/yaml/{template}'),
        Loader=yaml.FullLoader
    )

    manifest['spec']['template']['spec']['containers'][0]['image'] = container
    manifest['metadata']['name'] = name
    manifest['metadata']['namespace'] = namespace
    manifest['spec']['replicas'] = replicas
    manifest['spec']['template']['spec']['containers'][0]['env'] = []

    std_env_vars = [
        ('KML_CONN_STR', public_api_base),
        ('DB_CONN_STR', public_db_conn),
        ('DB_USER', db_user),
        ('DB_PASS', db_pass)
    ]

    for k, v in std_env_vars:
        manifest['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': str(v)}
        )

    for k, v in env_vars.items():
        manifest['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': str(v)}
        )

    logger.info(yaml.dump(manifest))

    return manifest


async def direct_runner_start(request):
    """Launch a container into kubernetes."""

    response = Responder()
    k8_apps_v1_api, _, _, _, _ = get_k8_conn()

    try:
        spec = await request.json()

        if spec.get('container'):
            container = spec.get('container')
        else:
            raise ValueError('Missing container parameter')

        replicas = spec.get('replicas') if spec.get('replicas') else 1
        env_vars = spec.get('env_vars') if spec.get('env_vars') else {}
        run_cmd = spec.get('run_cmd') if spec.get('run_cmd') else None
        desc = spec.get('description') if spec.get('description') else ''

        name = gen_unique_dep_name()

        manifest = gen_dep_manifest(
            container=container,
            name=name,
            replicas=replicas,
            env_vars=env_vars,
            run_cmd=run_cmd
        )

        resp = k8_apps_v1_api.create_namespaced_deployment(
            body=manifest,
            namespace=namespace
        )

        payload = {
            'container_name': container,
            'deployment_name': name,
            'description': desc,
            'launch_time': datetime.utcnow().isoformat(' ')[:-7],
            'status': 'starting'
        }
        tracking_tbl_hdl.insert_records([payload])

        watcher_thread = threading.Thread(
            group=None,
            target=pod_launch_watcher,
            args=(name, replicas)
        )
        watcher_thread.start()

        response.response['name'] = name
        response.response['action'] = 'started'
        response.succeeded()

        return response.emit(http_code=201)

    except Exception as e:
        logger.exception(e)
        response.failed(msg=e, code=1001)
        return response.emit(500)


async def direct_runner_stop(request):
    """Remove a container from kubernetes."""

    response = Responder()

    try:
        spec = await request.json()

        if spec.get('name'):
            name = spec.get('name')
        else:
            raise ValueError('Missing name parameter')

        success = delete_deployment(name)

        if success:
            response.response['name'] = name
            response.response['action'] = 'stopped'
            response.succeeded()
            return response.emit(http_code=200)

        else:
            response.response['name'] = name
            response.response['action'] = 'none'
            response.succeeded()
            return response.emit(http_code=404)

    except Exception as e:
        logger.error(e)
        logger.exception(e)
        response.failed(msg=e, code=1001)
        return response.emit(500)


def direct_runner_list(request):
    """List running slipway kubernetes containers."""

    response = Responder()

    try:
        resp = tracking_tbl_hdl.get_records()

        response.response['items'] = resp
        response.succeeded()
        return response.emit(http_code=200)

    except Exception as e:
        logger.error(e)
        logger.exception(e)
        response.failed(msg=e, code=1001)
        return response.emit(500)


def direct_runner_purge(request):
    """Remove all slipway containers and records."""

    response = Responder()

    try:
        resp = tracking_tbl_hdl.get_records_by_column(
            column_names=['deployment_name']
        )

        if resp:
            for name in resp['deployment_name']:
                logger.info(f'Stopping and deleting deployment {name}')
                delete_deployment(name)

            response.response['items'] = resp['deployment_name']
            response.succeeded()
            return response.emit(http_code=200)

        else:
            response.response['items'] = []
            response.succeeded()
            return response.emit(http_code=200)

    except Exception as e:
        logger.error(e)
        logger.exception(e)
        response.failed(msg=e, code=1001)
        return response.emit(500)
