"""KML ML Transform endpoint controller."""

import traceback
import json

from kml.api.orm import (
    Dataset,
    FeatureSet
)
from kml.api.aux import (
    attach_log,
    Responder
)
from kml.api.kmldbops import (
    get_tbl_handle as kml_get_tbl_handle,
    create_table as kml_create_table,
    has_table as kml_has_table,
    clear_table as kml_clear_table
)

logger = attach_log(module=__name__, log_name='kml', debug=True)


def transform_validate(col_type, xf_type):
    """Check that the input column type is of the right type to support the transform."""

    # NOTE: Use NVL function for data filling?
    #  logger.debug(f"Column {col_type} has type {xf_type}")
    if xf_type in ['normalize', 'log', 'ln', 'zscale',
                   'inverse', 'bucket', 'extracthour']:
        return True if col_type in ['int', 'long', 'float', 'double'] else False

    if xf_type in ['vocabulary', 'onehot', 'octetsplit']:
        return True if col_type in ['string'] else False

    if xf_type == 'identity':
        return True if col_type else False


def get_formatted_schema(tbl_name):
    """Return the list of lists formatted GPUdb schema appropriate for table creation."""

    tbl_hdl = kml_get_tbl_handle(tbl_name)
    col_type = {x.name: x.column_type for x in tbl_hdl.get_table_type().columns}
    col_prop = tbl_hdl.get_table_type().column_properties

    table_schema = []
    for name in col_type:
        col_schema = []
        col_schema.append(name)
        col_schema.append(col_type[name])
        for prop in col_prop[name]:
            col_schema.append(prop)
        table_schema.append(col_schema)

    return table_schema


def source_data_validate(dataset_cols, data_table_cols, fs_in_cols,
                         src_tbl_name, ds_id):
    """Check that all columns specified in FS manifest are in the data table."""

    for col in fs_in_cols:

        if col not in data_table_cols:
            msg = (f'Column {col} is required for feature engineering but is '
                   f'not found in table {src_tbl_name}')
            raise Exception(msg)

        if col not in dataset_cols:
            msg = (f'Column {col} is required for feature engineering but is '
                   f'not found in dataset {ds_id}')
            raise Exception(msg)


def transform_create(ds_id, xform_config):
    """Setup expression template for data transformations."""

    sink_tbl_schema = []
    feat_mem = {}
    xf_map = {}

    try:
        ds_obj = Dataset.load(existing_dataset_id=ds_id).serialize()
        src_tbl_name = ds_obj['source_config'].get('tbl')
        if not src_tbl_name:
            raise ValueError(f"No table name associated with dataset {ds_id}")
        src_tbl_hdl = kml_get_tbl_handle(src_tbl_name)
        if src_tbl_hdl.count == 0:
            raise ValueError(f"Table {src_tbl_name} (dataset {ds_id}) was selected for "
                             f"featureset fitting but is empty. Unable to proceed")

        src_tbl_schema = json.loads(
            src_tbl_hdl.get_table_type().schema_string
        )['fields']

        src_tbl_col_type = {col['name']: col['type'] for col in src_tbl_schema}

        # Check that all of the desired columns for the feature engineering are
        # present in the dataset entity and in the actual table
        source_data_validate(
            dataset_cols=ds_obj['source_config']['cols'],
            data_table_cols=src_tbl_col_type.keys(),
            fs_in_cols=[x['in_col_name'] for x in xform_config],
            src_tbl_name=src_tbl_name,
            ds_id=ds_obj['entity']['entity_id']
        )

        src_tbl_schema_dict = {
            c[0]: c for c in get_formatted_schema(src_tbl_name)
        }

        # Add UUID and timestamp
        sink_tbl_schema.append(['guid', 'string', 'uuid', 'init_with_uuid'])
        xf_map.update(guid='new_uuid()')
        sink_tbl_schema.append(['receive_dt', 'string', 'datetime'])
        xf_map.update(receive_dt='current_datetime()')

        # Create indentity field map (and schema) for all input columns
        for col in src_tbl_schema:
            xf_map[col['name']] = col['name']
        for col_schema in get_formatted_schema(src_tbl_name):
            sink_tbl_schema.append(col_schema)

        # Create field map and append schema according to transform type
        for xf in xform_config:

            # Validate input columns against requested transform method
            # TODO: Why is this sometimes a list and sometimes not?
            # TODO: DONT do this isinstance hack, rather figure out why
            #       this input isnt consistently a scalar

            if isinstance(src_tbl_col_type[xf['in_col_name']], list):

                if not transform_validate(col_type=src_tbl_col_type[xf['in_col_name']][0],
                                          xf_type=xf['function_xform'][0]):

                    raise ValueError(f"Schema for column {xf['in_col_name']} type mismatch "
                                     f"for transform method {xf['function_xform'][0]}")

            else:
                if not transform_validate(col_type=src_tbl_col_type[xf['in_col_name']],
                                          xf_type=xf['function_xform'][0]):

                    raise ValueError(f"Schema for column {xf['in_col_name']} type mismatch "
                                     f"for transform method {xf['function_xform'][0]}")

            feat_mem[xf['in_col_name']] = {}
            feat_mem[xf['in_col_name']]['function_xform'] = xf['function_xform'][0]

            # Z-Scale (aka standardization) transform
            if xf['function_xform'][0] == 'zscale':

                stats = src_tbl_hdl.aggregate_statistics(
                    column_name=xf['in_col_name'],
                    stats='mean,stdv'
                ).stats

                mean = stats['mean']
                stdv = stats['stdv']

                feat_mem[xf['in_col_name']]['stats'] = {}
                feat_mem[xf['in_col_name']]['stats']['mean'] = stats['mean']
                feat_mem[xf['in_col_name']]['stats']['stdv'] = stats['stdv']

                featexp = f'''divz(({xf['in_col_name']}-{mean}),{stdv},0)'''
                outschema = [[xf['out_col_name'], 'double']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                xf_map[xf['out_col_name']] = featexp
                sink_tbl_schema.extend(outschema)

            # Normalize data transform
            elif xf['function_xform'][0] == 'normalize':

                stats = src_tbl_hdl.aggregate_statistics(
                    column_name=xf['in_col_name'],
                    stats='min,max'
                ).stats

                col_min = stats['min']
                spread = stats['max'] - col_min

                feat_mem[xf['in_col_name']]['stats'] = {}
                feat_mem[xf['in_col_name']]['stats']['min'] = stats['min']
                feat_mem[xf['in_col_name']]['stats']['max'] = stats['max']
                feat_mem[xf['in_col_name']]['stats']['spread'] = spread

                featexp = f'''divz(({xf['in_col_name']}-{col_min}),{spread},0)'''
                outschema = [[xf['out_col_name'], 'double']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # Bucket (aka categorical) transform
            elif xf['function_xform'][0] == 'bucket':

                # TODO: Make this configurable (or auto calculated)
                n_bins = 5
                stats = src_tbl_hdl.aggregate_statistics(
                    column_name=xf['in_col_name'],
                    stats='min,max'
                ).stats

                col_min = stats['min']
                col_max = stats['max']
                local_min = 0
                local_max = col_max - col_min

                div = local_max / n_bins
                col = xf['in_col_name']

                feat_mem[xf['in_col_name']]['stats'] = {}
                feat_mem[xf['in_col_name']]['stats']['min'] = stats['min']
                feat_mem[xf['in_col_name']]['stats']['max'] = stats['max']
                feat_mem[xf['in_col_name']]['stats']['bins'] = n_bins
                feat_mem[xf['in_col_name']]['stats']['div'] = div

                featexp = (
                    f'''if(({col}-{col_min})<{local_min},0,'''
                    f'''if(({col}-{col_min})>{local_max},{n_bins+2},'''
                    f'''floor(({col}-{col_min})/{div})+1))'''
                )

                outschema = [[xf['out_col_name'], 'int', 'int16']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # Vocabulary (aka re-code) transform
            elif xf['function_xform'][0] == 'vocabulary':

                unique_terms = src_tbl_hdl.aggregate_group_by(
                    column_names=xf['in_col_name'],
                    offset=0,
                    limit=src_tbl_hdl.count
                ).data.get(xf['in_col_name'])

                matches = ','.join([f'"{x}"' for x in unique_terms])
                cat_vals = ','.join([str(x) for x in range(1, len(unique_terms) + 1)])

                featexp = ''.join(
                    ['case(',
                     str(xf['in_col_name']), ',{',
                     matches, '},{',
                     cat_vals, '},0)']
                )

                outschema = [[xf['out_col_name'], 'int', 'int16']]

                my_unique_terms = unique_terms
                my_unique_terms.append('OTHER')
                my_cat_vals = cat_vals.split(',')
                my_cat_vals.append('0')

                feat_mem[xf['in_col_name']]['stats'] = {}
                feat_mem[xf['in_col_name']]['stats']['val_map'] = dict(zip(my_unique_terms,
                                                                           my_cat_vals))

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # Identity (aka passthrough) transform
            elif xf['function_xform'][0] == 'identity':

                if xf['out_col_name'] != xf['in_col_name']:
                    raise Exception('Identity transforms must have the same in_col_name '
                                    'and out_col_name values in the featuresset spec')

                feat_mem[xf['in_col_name']]['stats'] = {}
                outschema = [[xf['out_col_name'], src_tbl_col_type[xf['in_col_name']]]]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: xf['in_col_name']}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = xf['in_col_name']

                # In schema is not required here, as in schema is out schema
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                #  sink_tbl_schema.extend(outschema)

            # Log base 10 transform
            elif xf['function_xform'][0] == 'log':

                feat_mem[xf['in_col_name']]['stats'] = {}
                featexp = f'''log10({xf['in_col_name']})'''
                outschema = [[xf['out_col_name'], 'double']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # Extract the hour from a datetime string
            elif xf['function_xform'][0] == 'extracthour':

                feat_mem[xf['in_col_name']]['stats'] = {}
                featexp = f'''hour({xf['in_col_name']})'''
                outschema = [[f"{xf['out_col_name']}_hour", 'int', 'int8']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # Natural log base e
            elif xf['function_xform'][0] == 'ln':

                feat_mem[xf['in_col_name']]['stats'] = {}

                featexp = f'''ln({xf['in_col_name']})'''
                outschema = [[xf['out_col_name'], 'double']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # Inverse transform, div/0 set to 0
            elif xf['function_xform'][0] == 'inverse':

                feat_mem[xf['in_col_name']]['stats'] = {}
                #  featexp = f'''if({xf['in_col_name']}!=0,1/{xf['in_col_name']},0)'''
                featexp = f'''1/{xf['in_col_name']}'''
                outschema = [[xf['out_col_name'], 'double']]

                feat_mem[xf['in_col_name']]['kexp'] = {xf['out_col_name']: featexp}
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                xf_map[xf['out_col_name']] = featexp
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]
                sink_tbl_schema.extend(outschema)

            # One-hot encoding tranform to separatly labled columns
            elif xf['function_xform'][0] == 'onehot':

                # One hot encoded values are value driven and lack explicit output col names
                if xf['out_col_name'] is not None:
                    raise Exception('One hot encoded values do not have explicit output columns '
                                    'names. out_col_spec must be `null` in the featureset spec')

                unique_terms = src_tbl_hdl.aggregate_group_by(
                    column_names=xf['in_col_name'],
                    offset=0,
                    limit=src_tbl_hdl.count
                ).data.get(xf['in_col_name'])

                outschema = []
                featexp = {}

                for term in unique_terms:
                    fmt_term = term.replace(" ", "_")
                    outschema.append([f"{xf['in_col_name']}_is_{fmt_term}", 'int', 'int8'])
                    featexp[f"{xf['in_col_name']}_is_{fmt_term}"] = \
                        f'''if({xf['in_col_name']}="{term}",1,0)'''

                # Handle fallouts
                outschema.append([f"{xf['in_col_name']}_is_other", 'int', 'int8'])
                featexp[f"{xf['in_col_name']}_is_other"] = \
                    f'''if({xf['in_col_name']} not in {tuple(unique_terms)},1,0)'''

                feat_mem[xf['in_col_name']]['stats'] = {}
                feat_mem[xf['in_col_name']]['stats']['unique_terms'] = unique_terms

                feat_mem[xf['in_col_name']]['kexp'] = featexp
                feat_mem[xf['in_col_name']]['outschema'] = outschema

                sink_tbl_schema.extend(outschema)

                xf_map.update(featexp)
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]

            # Octet split tranform to separatly labled columns
            elif xf['function_xform'][0] == 'octetsplit':

                if xf['out_col_name'] is not None:
                    raise Exception('Octetspit transforms do not have explicit output columns '
                                    'names. out_col_spec must be `null` in the featureset spec')

                feat_mem[xf['in_col_name']]['stats'] = {}
                outschema = []
                featexp = {}

                for part in range(1, 5):
                    outschema.append([f"{xf['in_col_name']}_{part}", 'int', 'int16'])
                    featexp[f"{xf['in_col_name']}_{part}"] = \
                        f'''cast(split({xf['in_col_name']},".",{part}),int16)'''

                xf_map.update(featexp)

                feat_mem[xf['in_col_name']]['kexp'] = featexp
                feat_mem[xf['in_col_name']]['outschema'] = outschema
                feat_mem[xf['in_col_name']]['inschema'] = src_tbl_schema_dict[xf['in_col_name']]

                sink_tbl_schema.extend(outschema)

            # Unknown transform requested
            else:
                raise ValueError(f"Unknown transform feature, {xf['function_xform'][0]}")
                sink_tbl_schema.append(['error_unkown'])

        # Return transform map and target table schema for storage in FeatureSet entity
        # TODO: sorting the schema may not be required here
        return {'target_schema': sorted(sink_tbl_schema),
                'feat_mem': feat_mem,
                'xf_map': xf_map,
                'fitting_ds_id': ds_id,
                'xforms': xform_config}

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        raise ValueError(e)


async def featureset_apply(request):
    """Run an in-line transform featureset directly without deployment."""

    response = Responder()

    try:
        featureset_id = request.path_params.get('featureset_id')
        spec = await request.json()

        src_tbl_name = spec.get('source_table')
        sink_tbl_name = spec.get('sink_table')

        fs_obj = FeatureSet.load(existing_featureset_id=featureset_id)
        xf_map = fs_obj.serialize()['config_inline']['xf_map']
        sink_tbl_scm = fs_obj.serialize()['config_inline']['target_schema']

        new_sink_tbl = False
        #  if db.has_table(sink_tbl_name)['table_exists']:
        if kml_has_table(sink_tbl_name, schema=None):
            logger.warning(f'Sink table {sink_tbl_name} exists, '
                           'clearing and remaking')
            #  db.clear_table(sink_tbl_name)
            kml_clear_table(sink_tbl_name, schema=None)
            new_sink_tbl = True
        #  _ = gpudb.GPUdbTable(_type=sink_tbl_scm, name=sink_tbl_name, db=db)
        _ = kml_create_table(tbl_type_list=sink_tbl_scm,
                             tbl_name=sink_tbl_name,
                             schema=None)

        sink_tbl_handle = kml_get_tbl_handle(sink_tbl_name, schema=None)
        status = sink_tbl_handle.append_records(
            source_table_name=src_tbl_name,
            field_map=xf_map
        )

        if status['status_info']['status'] == 'OK':
            response.succeeded()
            response.response['featureset_applied'] = featureset_id
            response.response['source_table'] = src_tbl_name
            response.response['sink_table'] = sink_tbl_name
            response.response['recreate_sink_table'] = new_sink_tbl
            return response.emit(201)
        else:
            raise Exception(status['status_info']['message'])

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)
