"""KML utilities endpoint controller."""
import traceback
import json
import uuid
import time

import gpudb

from kml.api.orm import (
    Entity,
    ModelDeployment
)
from kml.api.aux import (
    attach_log,
    Responder,
    log_comms
)
from kml.api.kml_config import (
    get_db_conn,
    get_namespace,
    get_db_conn_str,
    get_k8_conn,
    get_pod_resource_defaults,
    get_sys_schemas
)
from kml.api.kmldbops import (
    do_refresh_mv as kml_do_refresh_mv,
    get_tbl_handle as kml_get_tbl_handle,
    get_qualified_table_name
)

SCHEMA_SYS, SCHEMA_AUDIT, _, _ = get_sys_schemas()

# Read config
db = get_db_conn()
namespace = get_namespace()
db_conn_str = get_db_conn_str()

# Logging module
logger = attach_log(module=__name__, log_name='kml', debug=True)

# Defaults
# TODO: this should be configurable
jupyter_store = 'sys_kml_user_data'

CTNR_PULL_SECONDS_BEFORE_ABANDON = 60*60
CTNR_PULL_SECONDS_BETWEEN_POLLS = 15


def id_to_metadata_mapper(entity_type):
    toret = {}
    tbl = kml_get_tbl_handle(Entity.table_name, schema=SCHEMA_SYS)
    loadedObject = tbl.get_records(
        limit=-9999,
        options={"expression": f"entity_type='{entity_type}'"}
    )
    if len(loadedObject) > 0:
        for lo in loadedObject:
            toret[str(lo['entity_id'])] = {}
            toret[str(lo['entity_id'])]["name"] = lo['entity_name']
            toret[str(lo['entity_id'])]["desc"] = lo['entity_desc']
    return toret


def fix_JSON(json_message=None):
    result = None
    try:
        result = json.loads(json_message)
    except Exception as e:
        # Find the offending character index:
        idx_to_replace = int(e.message.split(' ')[-1].replace(')', ''))
        # Remove the offending character:
        json_message = list(json_message)
        json_message[idx_to_replace] = ' '
        new_message = ''.join(json_message)
        return fix_JSON(json_message=new_message)
    return result


def parse_ql_query_params(query):
    """Configure URL inline params for quicklist functions."""

    offset = query.query_params.get('offset')
    limit = query.query_params.get('limit')
    state_query = query.query_params.get('state')
    archived = query.query_params.get('archived')

    # Defaults if not specified
    offset = int(offset) if offset else 0
    limit = int(limit) if limit else -9999

    # Optionally build filter expression for entity state
    if state_query:
        filters = ' or '.join(
            [f'entity_state = "{x.upper()}"' for x in state_query.split(',')]
        )
    else:
        filters = ''

    # Optionally build filter expression for archive state
    if archived and archived.isalpha() and archived.upper() == 'FALSE':
        if filters:
            filters = ' and '.join(['visibility != "ARCHIVED"', filters])
        else:
            filters = 'visibility != "ARCHIVED"'

    expr = {'expression': filters} if filters else {}
    #  logger.info(f"Quicklist filter: {expr}")

    return limit, offset, expr


def get_ql_data(ql_name, schema, expr, limit, offset):

    kml_do_refresh_mv(
        mv_name=ql_name,
        schema=SCHEMA_SYS
    )

    table_hdl = kml_get_tbl_handle(
        tbl_name=ql_name,
        schema=SCHEMA_SYS
    )

    table_name = get_qualified_table_name(
        tbl_name=ql_name,
        schema=SCHEMA_SYS
    )

    resp = db.get_records(
        table_name=table_name,
        options=expr,
        offset=0,
        limit=0
    )
    if resp['status_info']['status'] == 'OK':
        total_recs = resp['total_number_of_records']
    else:
        raise ValueError(resp['status_info']['message'])

    recs = table_hdl.get_records(
        options=expr,
        offset=offset,
        limit=limit
    )

    return total_recs, recs


def quick_list_model_instance(request):
    """Model instance inventory."""

    response = Responder()

    try:
        limit, offset, expr = parse_ql_query_params(query=request)

        total_recs, recs = get_ql_data(
            ql_name='sys_kml_inventory_quick_model_instance',
            schema=SCHEMA_SYS,
            expr=expr,
            limit=limit,
            offset=offset
        )

        results = []

        for v in recs:
            to_insert = v
            to_insert["model_inst_config"] = json.loads(v["model_inst_config"])

            results.append(to_insert)

        #  response.response['include_archived'] = False if expr else True
        response.response['inventory'] = results
        response.response['total_records_available'] = total_recs
        response.response['n_records_returned'] = len(recs)
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def quick_list_model_deployment(request):
    """Model deployment inventory."""

    response = Responder()

    try:
        base_model_cache = id_to_metadata_mapper("MODEL_INSTANCE")

        limit, offset, expr = parse_ql_query_params(query=request)

        total_recs, recs = get_ql_data(
            ql_name='sys_kml_inventory_quick_model_deployment',
            schema=SCHEMA_SYS,
            expr=expr,
            limit=limit,
            offset=offset
        )

        results = []

        for v in recs:
            to_insert = v
            to_insert["model_dep_config"] = json.loads(v["model_dep_config"])

            if ("base_model_inst_id" in v and
                    str(v["base_model_inst_id"]) in base_model_cache):

                to_insert["base_model_inst_name"] = (
                    base_model_cache[str(v["base_model_inst_id"])]["name"]
                )
                to_insert["base_model_inst_desc"] = (
                    base_model_cache[str(v["base_model_inst_id"])]["desc"]
                )

            results.append(to_insert)

        #  response.response['include_archived'] = False if expr else True
        response.response['inventory'] = results
        response.response['total_records_available'] = total_recs
        response.response['n_records_returned'] = len(recs)
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def quick_list_featureset(request):
    """Featureset inventory."""

    response = Responder()
    try:
        limit, offset, expr = parse_ql_query_params(query=request)

        # Hide featureset stubs
        if expr.get('expression') and 'visibility' in expr['expression']:
            expr.update(
                expression='visibility != "ARCHIVED" and entity_state = "CREATED"'
            )
        else:
            expr.update(
                expression='entity_state = "CREATED"'
            )

        total_recs, recs = get_ql_data(
            ql_name='sys_kml_inventory_quick_featureset',
            schema=SCHEMA_SYS,
            expr=expr,
            limit=limit,
            offset=offset
        )

        results = []

        for v in recs:
            to_insert = v
            to_insert["config_relational"] = json.loads(v["config_relational"])
            to_insert["config_inline"] = json.loads(v["config_inline"])
            results.append(to_insert)

        #  response.response['include_archived'] = False if expr else True
        response.response['inventory'] = results
        response.response['total_records_available'] = total_recs
        response.response['n_records_returned'] = len(recs)
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def quick_list_dataset(request):
    """Dataset inventory."""

    response = Responder()

    try:
        limit, offset, expr = parse_ql_query_params(query=request)

        total_recs, recs = get_ql_data(
            ql_name='sys_kml_inventory_quick_dataset',
            schema=SCHEMA_SYS,
            expr=expr,
            limit=limit,
            offset=offset
        )

        results = []

        for v in recs:
            to_insert = v
            to_insert["source_config"] = json.loads(fix_JSON(v["source_config"]))
            to_insert["tbl"] = to_insert["source_config"]["tbl"]
            results.append(to_insert)

        #  response.response['include_archived'] = False if expr else True
        response.response['inventory'] = results
        response.response['total_records_available'] = total_recs
        response.response['n_records_returned'] = len(recs)
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def quick_list_ingest(request):
    """Ingest inventory."""

    response = Responder()

    try:
        limit, offset, expr = parse_ql_query_params(query=request)

        total_recs, recs = get_ql_data(
            ql_name='sys_kml_inventory_quick_ingest',
            schema=SCHEMA_SYS,
            expr=expr,
            limit=limit,
            offset=offset
        )

        results = []

        for v in recs:
            to_insert = v
            to_insert["ingest_config"] = json.loads(v["ingest_config"])
            to_insert["target_table"] = {}
            if "target_table" in to_insert["ingest_config"]:
                to_insert["target_table"] = to_insert["ingest_config"]["target_table"]
            results.append(to_insert)

        #  response.response['include_archived'] = False if expr else True
        response.response['inventory'] = results
        response.response['total_records_available'] = total_recs
        response.response['n_records_returned'] = len(recs)
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def quick_list_credential(request):
    """Credential inventory."""

    response = Responder()

    try:
        limit, offset, expr = parse_ql_query_params(query=request)

        total_recs, recs = get_ql_data(
            ql_name='sys_kml_inventory_quick_credential',
            schema=SCHEMA_SYS,
            expr=expr,
            limit=limit,
            offset=offset
        )

        results = []

        if len(recs) > 0:
            for v in recs:
                to_insert = v
                to_insert["credential_type"] = json.loads(v["credential_type"])
                to_insert["credential_config"] = json.loads(v["credential_config"])
                results.append(to_insert)

        #  response.response['include_archived'] = False if expr else True
        response.response['inventory'] = results
        response.response['total_records_available'] = total_recs
        response.response['n_records_returned'] = len(recs)
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def index(request):
    response = Responder()
    try:
        response.response['message'] = "Welcome to Kinetica Active Analytics Workbench"
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def ping(request):
    """Returns API and DB version string."""
    import datetime
    from pathlib import Path

    from kml.api.kmldbops import get_db_track

    response = Responder()
    try:
        versions = {}
        all_ver_data = {}

        db_ver = get_db_track()

        if Path('VERSION').is_file():
            with open('VERSION', 'r') as f:
                for line in f:
                    comp_name, ver_name = line.strip().split('=')
                    all_ver_data[comp_name] = ver_name

            versions.update(workbench_api=all_ver_data.get('kml_VERSION'))
            versions.update(workbench_ui=all_ver_data.get('kml-ui_VERSION'))
            versions.update(workbench_libs=all_ver_data.get('kml-libs_VERSION'))
            versions.update(database=db_ver)

            response.response['version'] = versions
            response.response['timestamp'] = datetime.datetime.utcnow().isoformat(' ')
            response.succeeded()
            return response.emit(200)

        else:

            with open('../../VERSION') as f:
                content = f.readlines()
            content = [x.strip() for x in content]

            vers = {}
            for v in content:
                level = [x.strip() for x in v.split('=')]
                vers[level[0]] = level[1]

            dev_api_ver = '.'.join([vers['MAJOR'],
                                    vers['MINOR'],
                                    vers['REVISION'],
                                    vers['ABI_VERSION']])

            versions.update(workbench_api=dev_api_ver)
            versions.update(workbench_ui='n/a')
            versions.update(workbench_libs='virtual environment')
            versions.update(database=db_ver)

            response.response['version'] = versions
            response.response['timestamp'] = datetime.datetime.utcnow().isoformat(' ')
            response.succeeded()
            return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


async def db_latency(request):
    """Test latency between KML API and Database.

    Metrics are caculated based on a standardized score of 1. Larger
    values represent a larger latency between the DB and the API for
    the specific database operation.

    The reference configuration is:
        OS:        Ubuntu 18.04 LTS
        CPU:       Intel Core i7-8750H
        GPU:       Nvidia GeForce GTX 1070
        RAM:       32 GiB
        CUDA:      10.2
        CPython:   3.6.8
        Starlette: 0.13.3
        Uvicorn:   0.11.3
        GPUdb:     v7.1.0.0.20200618134209
        gpudb-api: 7.1.0b2 (python)

        number of records inserted = 1000
        relative timings are approximate centered at 1 due to system load

        total db operations time is reported in seconds

    """

    response = Responder()
    try:
        n_recs = int(request.path_params.get('n_records'))

        TBL_NAME = f"zkml_latency_test_{str(uuid.uuid4())}"
        TBL_SCM = [
            ['x', 'string', 'char1'],
            ['y', 'double'],
            ['z', 'int', 'int16']
        ]
        baseline = {
            'create_table': 0.0307,
            'get_recs_changed': 0.0221,
            'get_recs_column': 0.0226,
            'get_stats': 0.0163,
            'ins_recs': 0.0144,
            'update_recs': 0.8572,
            'clear_table': 0.0166
        }

        t0 = time.time()
        metrics = {}

        # Create table
        latency_tbl_hdl = gpudb.GPUdbTable(
            _type=TBL_SCM,
            name=TBL_NAME,
            db=db
        )
        t1 = time.time()
        split = t1 - t0
        metrics.update(create_table=(split / baseline['create_table']))

        # Insert records
        #  logger.info(f"Inserting {n_recs} random records into {latency_tbl_hdl.name}")
        latency_tbl_hdl.insert_records_random(n_recs)
        t2 = time.time()
        split = t2 - t1
        metrics.update(ins_recs=(split / baseline['ins_recs']))

        # Statistics
        #  logger.info("Generating statistics")
        stats = latency_tbl_hdl.aggregate_statistics(
            column_name='x',
            stats='cardinality, count'
        )['stats']
        card_x = stats['cardinality']
        #  logger.info(f"Cardinality of alpha column: {card_x}")
        t3 = time.time()
        split = t3 - t2
        metrics.update(get_stats=(split / baseline['get_stats']))

        # Get records with filter
        #  logger.info("Getting records")
        match_data = latency_tbl_hdl.get_records_by_column(
            column_names='x',
            options={'expression': 'y>0 and z<0'}
        )
        #  logger.info(f"Found {len(match_data['x'])} matches")
        match_unique = set(match_data['x'])
        #  logger.info(f"Of these, {len(match_unique)} are unique matches")
        t4 = time.time()
        split = t4 - t3
        metrics.update(get_recs_column=(split / baseline['get_recs_column']))

        # Update records with filter
        #  logger.info("Updating records")
        for pos, match_x in enumerate(match_unique):
            latency_tbl_hdl.update_records(
                expressions=[f'x="{match_x}"'],
                new_values_maps={'y': pos, 'z': int(card_x)}
            )
        t5 = time.time()
        split = t5 - t4
        metrics.update(update_recs=(split / baseline['update_recs']))

        # Count changed rows
        latency_tbl_hdl.get_records(
            limit=-9999,
            options={'expression': f'z={card_x}'}
        )
        #  logger.info(f"Updated {len(data)} records")
        t6 = time.time()
        split = t6 - t5
        metrics.update(get_recs_changed=(split / baseline['get_recs_changed']))

        # Cleanup
        db.clear_table(
            table_name=TBL_NAME,
            options={'no_error_if_not_exists': 'true'}
        )
        t7 = time.time()
        split = t7 - t6
        metrics.update(clear_table=(split / baseline['clear_table']))

        t_final = time.time()
        logger.info(f"Database latency metric: {round(t_final - t0, 5)}")

        metrics = {key: round(metrics[key], 1) for key in metrics.keys()}

        response.response['n_recs'] = n_recs
        response.response['total_db_ops_seconds'] = round(t_final - t0, 5)
        response.response['scaled_metrics'] = metrics
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


async def echo(request):
    response = Responder()
    try:
        my_data = await request.json()
        my_id = request.path_params.get('my_id')
        my_name = request.path_params.get('my_name')

        response.response['id'] = my_id
        response.response['name'] = my_name
        response.response['data'] = my_data
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def infra(request):
    response = Responder()
    # Return the API host conn, DB conn and K8 conn utilized by this instance
    try:
        response.response['message'] = "HTTP connections"
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)

def pod_resource_defaults(request):
    """Return default request and limit."""

    response = Responder()

    try:
        (cpu_request,
         cpu_limit,
         gpu_request,
         mem_request,
         mem_limit) = get_pod_resource_defaults()

        response.response['pod_resource_defaults'] = (
            {
                'cpu_min': cpu_request,
                'cpu_max': cpu_limit,
                'gpu_min': 1,
                'gpu_max': gpu_request,
                'mem_min': mem_request,
                'mem_max': mem_limit,
                'cpu_units': 'millicore (m)',
                'gpu_units': 'GPGPU',
                'mem_units': 'mebibyte (Mi)'
            }
        )

        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def mycluster(request):
    from kml.api.utils import get_k8_resources

    response = Responder()

    try:
        cluster = get_k8_resources()

        response.response['cluster'] = cluster
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def mycluster_deployments(request):
    response = Responder()
    try:
        from kml.api.k8_manager import get_all_deployments
        response.response['deployments'] = []

        toret = []
        for ad in get_all_deployments():
            toret.append(ad)

        response.response['deployments'] = toret
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def mycluster_deployments_and_replicas(request):
    response = Responder()
    try:
        from kml.api.k8_manager import get_all_deployments_familytree

        toret = []

        ftree = get_all_deployments_familytree()
        for adep in ftree.keys():
            onedep = {}
            toret.append(onedep)

            onedep["deployment"] = adep
            onedep["replicas"] = []
            for apod in ftree[adep]:
                onedep["replicas"].append(apod)

        response.response['deployments'] = toret
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def mycluster_deployment_logs(request):
    response = Responder()
    try:
        model_deployment_id = request.path_params.get('model_deployment_id')
        md = ModelDeployment.load(model_deployment_id)
        # we do this as a validation, this will barf if the deployment does not exist

        from kml.api.k8_manager import get_pod_logs
        from kml.api.k8_manager import get_all_deployments_familytree
        from kml.api.naming_svc import get_cluster_name_deployment_blackbox

        toret = {}
        depname = None
        container_name = None

        if md.base_model_inst.model_inst_config["model_type"] == "BLACKBOX" and md.deployment_type == "CONTINUOUS":
            container_name = "blackbox-worker"
            toret["deployment-type"] = md.deployment_type
            depname = get_cluster_name_deployment_blackbox(model_deployment_id)
        elif md.base_model_inst.model_inst_config["model_type"] == "BLACKBOX":
            toret["deployment-type"] = md.deployment_type
            depname = get_cluster_name_deployment_blackbox(model_deployment_id)

        ftree = get_all_deployments_familytree()
        allreplicas = ftree[depname]

        toret["deployment"] = depname

        toret["replicas"] = []

        for apod in allreplicas:
            ar = {}
            ar["replica"] = apod
            ar["replica-log"] = []
            thislog = get_pod_logs(apod, container=container_name)
            if thislog:
                for aline in thislog.splitlines():
                    ar["replica-log"].append(aline)
            else:
                ar["replica-log"].append("NO LOGS FOUND")
            toret["replicas"].append(ar)

        response.response['deployments'] = toret
        response.succeeded()

        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def mycluster_deployment_info(request):
    response = Responder()
    try:
        model_deployment_id = request.path_params.get('model_deployment_id')

        status = ModelDeployment.load(model_deployment_id)
        # we do this as a validation, this will barf if the deployment does not exist

        # from kml.api.k8_manager import get_pod_logs
        # response.response['items'] = get_pod_logs(deployment_id)
        logger.info(status)
        response.succeeded()
        return response.emit(200)
    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def ensure_no_entity_id_conflict():
    from kml.api.manage_systables import tables_audit_inventory

    def get_max_audit_id(in_audit_tables):
        maxid = 0
        for a in in_audit_tables:
            if a.startswith('dep-'):
                thisid = int(a[4:9])
                if thisid > maxid:
                    maxid = thisid
        return maxid

    #  logger.debug(f"Checking for potential ID conflicts")
    next_id = Entity.get_next_entity_id()
    all_audit_tables = tables_audit_inventory()
    max_audit_id = get_max_audit_id(all_audit_tables)
    #  logger.debug(f"Next available entity ID is {next_id}")
    #  logger.debug(f"Max recorded audit ID is {max_audit_id}")

    if next_id < max_audit_id:
        return False
    return True

def check_db_credentials(db_user, db_pass):
    """Confirm validity of database credentials."""

    #  print('==== controller_utils.py ====')
    try:
        _ = gpudb.GPUdb(
            host=db_conn_str,
            primary_host=db_conn_str,
            username=db_user,
            password=db_pass
        )
    except gpudb.GPUdbException as e:
        #  print('==== Catching in controller_utils.py ====')
        raise ValueError(e)


async def register_events(request):
    """Record events from running containers."""

    import datetime

    response = Responder()
    try:
        # get the event payload
        event_json = await request.json()
        log_comms(event_json, "event_capture")

        # SAVE RAW EVENT DATA
        tbl_hdl = kml_get_tbl_handle("sys_kml_events", schema=SCHEMA_SYS)

        # TODO: Handle LISTS of events also
        reporter_type = "UNKNOWN"
        if "reporter_type" in event_json:
            reporter_type = event_json["reporter_type"]

        insertable = {
                "guid": str(uuid.uuid4()),
                "receive_dt": datetime.datetime.now().isoformat(' ')[:-3],
                "event_type": "UNKNOWN",
                "reporter_type": reporter_type,
                "event_detail": json.dumps(event_json)
            }

        # TODO: Handle LISTS of events also
        if "event_type" in event_json and event_json["event_type"]:
            insertable["event_type"] = event_json["event_type"]

        _ = tbl_hdl.insert_records(insertable)

        if insertable["event_type"] == "LIFECYCLE":
            # SAVE PARSED ***LIFECYCLE*** EVENT DATA

            # Creating table:
            #
            #  TBL_SCHEMA_events_lifecycle = [
            #      ["guid", "string", "char64", "nullable", "text_search"],
            #      ["receive_dt", "string", "datetime", "nullable", "text_search"],
            #      ["deployment_id", "long"],
            #      ["deployment_name", "string", "char32", "nullable", "text_search"],
            #      ["k8s_replicaset", "string", "char32", "nullable", "text_search"],
            #      ["k8s_pod_name", "string", "char32", "nullable", "text_search"],
            #      ["event_sub_type", "string", "char16", "nullable", "text_search", "dict"]
            #  ]

            # TODO: How expensive is making a table handle?
            # Should this be done only once at the top of the file?
            tbl_hdl_lifecycle = kml_get_tbl_handle("sys_kml_events_lifecycle",
                                                   schema=SCHEMA_SYS)

            # TODO: Handle LISTS of events also
            insertable_lifecycle = {
                    "guid": insertable["guid"],
                    "receive_dt": insertable["receive_dt"],
                    "event_type": "LIFECYCLE",
                    "reporter_type": reporter_type,
                    "deployment_id": event_json.get("deployment_id"),
                    "deployment_name": event_json.get("deployment_name"),
                    "k8s_replicaset": event_json.get("k8s_replicaset"),
                    "k8s_pod_name": event_json.get("k8s_pod_name"),
                    "event_sub_type": event_json.get("event_sub_type"),
                }
            _ = tbl_hdl_lifecycle.insert_records(insertable_lifecycle)

        elif insertable["event_type"] == "METRICS":
            # SAVE PARSED ***METRICS*** EVENT DATA

            # Creating table:
            #
            #  TBL_SCHEMA_events_metrics = [
            #      ["guid", "string", "char64", "nullable", "text_search", "dict"],
            #      ["receive_dt", "string", "datetime", "nullable", "text_search"],
            #      ["deployment_id", "long"],
            #      ["deployment_name", "string", "char32", "nullable", "text_search", "dict"],
            #      ["k8s_replicaset", "string", "char32", "nullable", "text_search", "dict"],
            #      ["k8s_pod_name", "string", "char32", "nullable", "text_search", "dict"],
            #      ["recs_received", "long", "nullable"],
            #      ["recs_inf_success", "long", "nullable"],
            #      ["recs_inf_failure", "long", "nullable"],
            #      ["recs_inf_persisted", "long", "nullable"],
            #      ["throughput_inf", "long", "nullable"],
            #      ["throughput_e2e", "long", "nullable"]
            #  ]

            # TODO: How expensive is making a table handle?
            # Should this be done only once at the top of the file?
            tbl_hdl_metrics = kml_get_tbl_handle("sys_kml_events_metrics",
                                                 schema=SCHEMA_SYS)

            # TODO: Handle LISTS of events also
            insertable_metric = {
                    "guid": insertable["guid"],
                    "receive_dt": insertable["receive_dt"],
                    "event_type": "METRICS",
                    "reporter_type": reporter_type,
                    "deployment_id": event_json.get("deployment_id"),
                    "deployment_name": event_json.get("deployment_name"),
                    "k8s_replicaset": event_json.get("k8s_replicaset"),
                    "k8s_pod_name": event_json.get("k8s_pod_name"),
                    "seq_id": event_json.get("seq_id"),
                    "recs_received": event_json.get("recs_received"),
                    "recs_relayed": event_json.get("recs_relayed"),
                    "recs_inf_success": event_json.get("recs_inf_success"),
                    "recs_inf_failure": event_json.get("recs_inf_failure"),
                    "recs_inf_persisted": event_json.get("recs_inf_persisted"),
                    "throughput_inf": event_json.get("throughput_inf"),
                    "throughput_e2e": event_json.get("throughput_e2e"),
                }
            _ = tbl_hdl_metrics.insert_records(insertable_metric)

        #######################################################################
        else:
            logger.warn(f"Unrecognized event type {insertable['event_type']}")
            logger.warn(event_json)

        response.response['captured'] = True
        response.response['persistable'] = insertable
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)


def retrieve_events(request):
    """Retrieve stored container events from database."""

    import datetime

    response = Responder()
    try:
        returnable = []
        #  tbl = gpudb.GPUdbTable( name="sys_kml_events", db=db)
        tbl = kml_get_tbl_handle("sys_kml_events", schema=SCHEMA_SYS)
        # TODO: Chunk this 20k by 20k
        loadedObjects = []
        offset = 0
        limit = 5000
        while offset < tbl.count:
            loadedObjects.extend(tbl.get_records(limit=limit, offset=offset))
            offset = offset + limit
        for lobs in loadedObjects:
            try:
                parsed_details = json.loads(lobs["event_detail"])
                lobs["event_detail"] = parsed_details
            except Exception as e:
                logger.warn("Invalid JSON for event detail")
                logger.warn(lobs["event_detail"])
            returnable.append(lobs)

        response.response['asof'] = datetime.datetime.utcnow().isoformat(' ')
        response.response['count'] = len(returnable)
        response.response['events'] = returnable
        response.succeeded()
        return response.emit(200)

    except Exception as e:
        logger.error(e)
        logger.debug(traceback.format_exc())
        response.failed(msg=e, code=1040)
        return response.emit(509)
