import traceback

import kubernetes.client
from kubernetes.client.apis.logs_api import LogsApi

from kml.api.kml_config import get_k8_conn, get_namespace
k8_apps_v1_api, k8_core_v1_api, k8_custom_objects_api, k8_batch_v1_api, _ = get_k8_conn()
namespace = get_namespace()

from kml.api.aux import attach_log

logger = attach_log(module=__name__, log_name='kml', debug=True)

def get_all_deployments_familytree():

    # See K8s bindings docs:
    #   https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/CoreV1Api.md#list_pod_for_all_namespaces
    #   https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1PodList.md

    field_selector = 'status.phase=Running' # str | A selector to restrict the list of returned objects by their fields. Defaults to everything. (optional)
    include_uninitialized = True # bool | If true, partially initialized resources are included in the response. (optional)
    limit = 2048 # int | limit is a maximum number of responses to return for a list call. If more items exist, the server will set the `continue` field on the list metadata to a value that can be used with the same initial query to retrieve the next set of results. Setting a limit may return fewer than the requested amount of items (up to zero items) in the event all requested objects are filtered out and kubernetes.clients should only use the presence of the continue field to determine whether more results are available. Servers may choose not to support the limit argument and will return all of the available results. If limit is specified and the continue field is empty, kubernetes.clients may assume that no more results are available. This field is not supported if watch is true.  The server guarantees that the objects returned when using continue will be identical to issuing a single list call without a limit - that is, no objects created, modified, or deleted after the first request is issued will be included in any subsequent continued requests. This is sometimes referred to as a consistent snapshot, and ensures that a kubernetes.client that is using limit to receive smaller chunks of a very large result can ensure they see all possible objects. If objects are updated during a chunked list the version of the object that was present at the time the first list result was calculated is returned. (optional)

    def find_parent_deployment(podname):
        n = 2
        needle = "-"
        parts = podname.split(needle, n+1)
        if len(parts)<=n+1:
            return -1
        detatchment_point = len(podname)-len(parts[-1])-len(needle)
        return podname[:detatchment_point]

    try:
        family_tree = {}
        api_response = k8_core_v1_api.list_pod_for_all_namespaces(
          field_selector = field_selector,
          #include_uninitialized = include_uninitialized,
          limit = limit,
          pretty = True)
        for a in api_response.items:
            pod_name = a.metadata.name
            if pod_name.startswith("kml"):
                dep_name = find_parent_deployment(pod_name)
                if dep_name not in family_tree:
                    family_tree[dep_name] = []
                family_tree[dep_name].append(pod_name)
        return family_tree
    except Exception as e:
        logger.error("Exception when calling CoreV1Api->list_pod_for_all_namespaces: %s\n" % e)
        logger.debug(traceback.format_exc())

def get_all_deployments():

    family_tree = get_all_deployments_familytree()
    return family_tree.keys()

def get_pod_logs(podname, container = None, namespace = namespace, trailing_lines = 240, byte_limit = 1024 * 1024):

    # See K8s bindings docs:
    #   https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/CoreV1Api.md#read_namespaced_pod_log

    pretty = 'pretty_example' # str | If 'true', then the output is pretty printed. (optional)
    follow = False # bool | Follow the log stream of the pod. Defaults to false. (optional)
    pretty = 'pretty_example' # str | If 'true', then the output is pretty printed. (optional)
    previous = False # bool | Return previous terminated container logs. Defaults to false. (optional)
    timestamps = True # bool | If true, add an RFC3339 or RFC3339Nano timestamp at the beginning of every line of log output. Defaults to false. (optional)

    try:
        if container:
            # CONTINUOUS containers DO have container names
            api_response = k8_core_v1_api.read_namespaced_pod_log(
                podname,
                namespace,
                follow=follow,
                limit_bytes = byte_limit,
                pretty=pretty,
                previous=previous,
                #since_seconds=trailing_seconds,
                container=container,
                tail_lines=trailing_lines,
                timestamps=timestamps)
            return api_response
        else:
            # ONDEMAND containers dont have container names
            api_response = k8_core_v1_api.read_namespaced_pod_log(
                podname,
                namespace,
                follow=follow,
                limit_bytes = byte_limit,
                pretty=pretty,
                previous=previous,
                #since_seconds=trailing_seconds,
                tail_lines=trailing_lines,
                timestamps=timestamps)
            return api_response

    except Exception as e:
        logger.error("Exception when calling CoreV1Api->list_pod_for_all_namespaces: %s\n" % e)
        logger.debug(traceback.format_exc())
