import os
import sys
import time
import configparser

import gpudb

# Number of retries to attach to DB (1 sec wait between tries)
MAX_DB_CONNECT_ATTEMPTS = 50000
ATTEMPT_WAIT_TIME = 10  # seconds


def running_in_k8s():
    """Detect if api is running in the cloud environment."""

    return (
        True
        if os.environ.get('KINETICA_LOGS_TO_STDOUT')
        and os.environ['KINETICA_LOGS_TO_STDOUT'] != 0
        else False
    )


def open_config_file():
    """Load the contents of the configuration file."""

    try:
        path = os.path.dirname(os.path.realpath(__file__))
        kml_ini = '/'.join([path, 'kml.ini'])

        config = configparser.ConfigParser()
        config.read(kml_ini)
    except Exception as e:
        print(e)

    return config


def open_system_params_file():
    """Load the contents of the system parameters file."""

    try:
        path = os.path.dirname(os.path.realpath(__file__))
        sys_ini = '/'.join([path, 'system.ini'])

        config = configparser.ConfigParser()
        config.read(sys_ini)
    except Exception as e:
        print(e)

    return config


def get_log_level():
    """Return the requested logging level."""

    config = open_config_file()
    log_level = config['response']['log_level'].strip()

    return log_level if log_level else 'WARNING'


def get_api_base():
    """Return a constructed URL for the API base path."""

    config = open_config_file()

    return f"{config['api']['api_connection'].strip()}/kml"


def get_db_conn_str():
    """Return the full connection string for the database."""

    config = open_config_file()

    return config['database']['db_connection'].strip()


def get_tm_conn_str():
    """Return the full connection string the ZMQ socket."""

    config = open_config_file()

    return config['database']['table_monitor_connection'].strip()


def get_public_connections():
    """Return the all public IP based connection strings."""

    config = open_config_file()

    if running_in_k8s():

        api_pub = config['api']['api_connection']
        db_pub = config['database']['db_connection']
        tm_pub = config['database']['table_monitor_connection']

        return f"{api_pub.strip()}/kml", db_pub.strip(), tm_pub.strip()

    else:

        api_pub = (
            config['api']['api_public_connection']
            if config['api']['api_public_connection'].strip()
            else config['api']['api_connection']
        )
        db_pub = (
            config['database']['db_public_connection']
            if config['database']['db_public_connection'].strip()
            else config['database']['db_connection']
        )
        tm_pub = (
            config['database']['table_monitor_public_connection']
            if config['database']['table_monitor_public_connection'].strip()
            else config['database']['table_monitor_connection']
        )

        return f"{api_pub.strip()}/kml", db_pub.strip(), tm_pub.strip()


def get_components(conn):
    """Return the host and port for the database, api and zmq socket."""

    config = open_config_file()

    if conn.lower() == 'db':
        parts = config['database']['db_connection'].strip().split(':')
    elif conn.lower() == 'zmq':
        if running_in_k8s():
            parts = config['database']['table_monitor_connection'].strip().split(':')
        else:
            parts = (
                config['database']['table_monitor_public_connection'].strip().split(':')
                if config['database']['table_monitor_public_connection']
                else config['database']['table_monitor_connection'].strip().split(':')
            )
    else:
        parts = config['api']['api_connection'].strip().split(':')

    return parts[0], parts[1].lstrip('/'), parts[2]


def get_sys_schemas():
    """Read system schema names from the config file."""

    config = open_config_file()

    schema_sys = config['database']['schema_sys'].strip()
    schema_audit = config['database']['schema_audit'].strip()
    schema_sys_archived = config['database']['schema_sys_archived'].strip()
    schema_audit_archived = config['database']['schema_audit_archived'].strip()

    return schema_sys, schema_audit, schema_sys_archived, schema_audit_archived


def get_db_credentials():
    """Read credentials from the config file."""
    from kml.api.utils import kdecrypt

    config = open_config_file()

    db_user = ''
    db_pass = ''

    # Check for db user credentials
    if 'user' in config['database']:
        db_user = config['database']['user']
    if 'pass' in config['database']:
        db_pass_encrypted = config['database']['pass']
        db_pass = (
            kdecrypt(db_pass_encrypted.strip('"'))
            if not running_in_k8s()
            else db_pass_encrypted
        )

    return db_user, db_pass


def get_db_conn():
    from kml.api.utils import kdecrypt
    """Return a database handle."""

    config = open_config_file()

    # Database parameters
    db_conn_str = config['database']['db_connection'].strip()
    db_user = ''
    db_pass = ''

    # Check for db user credentials
    if 'user' in config['database']:
        db_user = config['database']['user']
    if 'pass' in config['database']:
        db_pass_encrypted = config['database']['pass']
        db_pass = (
            kdecrypt(db_pass_encrypted.strip('"'))
            if not running_in_k8s()
            else db_pass_encrypted
        )

    # Create database handle
    connection_established = False
    for attempt in range(1, MAX_DB_CONNECT_ATTEMPTS):
        #  print(f"** Connection attempt: {attempt}")
        try:
            db_handle = gpudb.GPUdb(
                encoding='BINARY',
                host=db_conn_str,
                primary_host=db_conn_str,
                username=db_user,
                password=db_pass,
                disable_failover=True,
                disable_auto_discovery=True
            )
            connection_established = True
            return db_handle

        except:
            print(f"Sleeping on DB Connect: {ATTEMPT_WAIT_TIME} seconds")
            time.sleep(ATTEMPT_WAIT_TIME)

    if not connection_established:
        print(
            f"Unable to connect to the database: Maximum attempts exceeded",
            flush=True
        )
        sys.exit(1)


    #      try:
    #          db_handle = gpudb.GPUdb(
    #              encoding='BINARY',
    #              host=db_conn_str,
    #              primary_host=db_conn_str,
    #              username=db_user,
    #              password=db_pass,
    #              disable_failover=True,
    #              disable_auto_discovery=True
    #          )
    #          connection_established = True
    #          return db_handle

    #      except Exception as e:
    #          if 'Unauthorized' in e.message or 'Insufficient credentials' in e.message:
    #              print('Incorrect database credentials, unable to proceed')
    #              print('Use the /opt/gpudb/kml/utils/kpass utility to update the AAW')
    #              print('configuration file and restart with "/opt/gpudb/kml/bin/kml start".')
    #              sys.exit(1)
    #          print(f"Unable to connect to the database: {e.message}", flush=True)
    #          print(f"Waiting for {ATTEMPT_WAIT_TIME} seconds", flush=True)
    #          time.sleep(ATTEMPT_WAIT_TIME)
    #          print(f"Reconnecting, attempt {attempt}", flush=True)

    #  if not connection_established:
    #      print(f"Unable to connect to the database: Maximum attempts exceeded",
    #            flush=True)
    #      sys.exit(1)


def get_k8_conn():
    """Return instantiated class bindings to K8s APIs."""

    import kubernetes as k8
    config = open_config_file()

    try:
        if running_in_k8s():
            k8.config.load_incluster_config()
            #  print('Using in-cluster k8s service account config')
        else:
            if config['cluster']['kube_config']:
                k8.config.load_kube_config(
                    config_file=config['cluster']['kube_config']
                )
                #  print(f"Using k8s cluster specified with "
                #        f"{config['cluster']['kube_config']}")
            else:
                k8.config.load_kube_config()
                #  print('Using default K8s configuration')

        # Instantiate the Kubernetes APIs
        k8_apps_v1_api = k8.client.AppsV1Api()
        k8_core_v1_api = k8.client.CoreV1Api()
        k8_custom_objects_api = k8.client.CustomObjectsApi()
        k8_batch_v1_api = k8.client.BatchV1Api()
        k8_ext_v1_api = k8.client.ExtensionsV1beta1Api()

        return (
            k8_apps_v1_api,
            k8_core_v1_api,
            k8_custom_objects_api,
            k8_batch_v1_api,
            k8_ext_v1_api
        )

    except Exception as e:
        print(e)
        msg = """
              Error connecting to Kubernetes cluster. Kubernetes is required
              for AAW operation. Check that there is a kubeconfig file at
              /opt/gpudb/.kube/config or /etc/kubernetes/admin.conf or that one
              has been specified in kml.ini (the AAW configuration file). Run:
                     export PATH=/opt/gpudb/kml/python3/bin:$PATH
                     kubectl get nodes
              to test the connection to the Kubernetes node (or cluster).
              """
        print(msg)
        raise Exception('ERROR: Unable to connect to the Kubernetes cluster')


def get_k8_rbac_conn():
    """Return rule-based access control API bindings."""

    import kubernetes as k8
    config = open_config_file()

    # Database parameters
    use_cluster = config['cluster']['use_cluster']

    if use_cluster.strip().upper() in ['FALSE', 'NO']:
        raise ValueError('Attempting K8 connection despite use_cluster set to NO')

    try:
        if running_in_k8s():
            k8.config.load_incluster_config()
            # print('Using in-cluster k8s service account config')
        else:
            if config['cluster']['kube_config']:
                k8.config.load_kube_config(
                    config_file=config['cluster']['kube_config']
                )
                #  print(f"Using k8s cluster specified with "
                #        f"{config['cluster']['kube_config']}")
            else:
                k8.config.load_kube_config()
                # print('Using default K8s configuration')

        # Instantiate the Kubernetes APIs
        return k8.client.RbacAuthorizationV1Api()

    except Exception as e:
        print(e)
        print('Error connecting to K8')
        raise Exception('Error connecting to K8')


def get_namespace():
    """Return the configured Kubernetes namespace."""

    config = open_config_file()
    namespace = config['cluster']['k8s_namespace']

    if namespace:
        # User specified
        if namespace[0:4].lower() != 'kml-':
            namespace = f'kml-{namespace}'
        return namespace
    else:
        # Default
        return 'kml-default'


def get_image_pull_policy():
    """Return the image pull policy for K8s containers."""

    config = open_system_params_file()
    always_pull_image = (
        True
        if config['kubernetes']['always_pull_images'].upper() == 'TRUE'
        else False
    )
    global_policy = 'Always' if always_pull_image else 'IfNotPresent'
    bbx_policy = 'Always' if always_pull_image else 'IfNotPresent'

    if bbx_policy == 'IfNotPresent':
        config = open_config_file()
        always_pull_image = (
            True
            if config['cluster']['always_pull_user_images'].upper() == 'TRUE'
            else False
        )
        if always_pull_image:
            bbx_policy = 'Always'

    return global_policy, bbx_policy


def get_ingress_controller_details():
    """Return the service name, namespace & ports of the K8s ingress controller."""

    config = open_system_params_file()

    ing_svc_name = config['kubernetes']['ingress_service_name']
    ing_svc_ns = (
        config['kubernetes']['ingress_service_namespace'].strip()
        if not running_in_k8s()
        else 'nginx'
    )
    
    ing_http_port = (
        str(config['kubernetes']['ingress_http_port'])
        if config['kubernetes']['ingress_http_port']
        else None
    )
    ing_https_port = (
        str(config['kubernetes']['ingress_https_port'])
        if config['kubernetes']['ingress_https_port']
        else None
    )

    return ing_svc_name, ing_svc_ns, ing_http_port, ing_https_port


def get_docker_daemon_uri():
    """Return and test the Docker URI."""

    config = open_system_params_file()
    docker_daemon_uri = config['blackbox_wizard']['docker_daemon_uri']

    # TODO: Connectivity Test
    try:
        pass
    except Exception as e:
        print(e)
        print('Error connecting to Docker Daemon')
        raise Exception('Error connecting to Docker Daemon')

    return docker_daemon_uri


def get_blackbox_sdk():
    """Return the names of the black box SDK components."""

    config = open_system_params_file()
    blackbox_sdk_cache = config['blackbox_wizard']['blackbox_sdk_cache']
    blackbox_sdk_repo = config['blackbox_wizard']['blackbox_sdk_repo']
    blackbox_sdk_repo_branch = config['blackbox_wizard']['blackbox_sdk_repo_branch']

    return blackbox_sdk_cache, blackbox_sdk_repo, blackbox_sdk_repo_branch


def get_sys_container(container_ref):
    """Return the names for the system base containers."""

    if container_ref not in ['ctnr_kio',
                             'ctnr_kafkaconnector',
                             'ctnr_bb_basesdk',
                             'ctnr_kml_depmgr',
                             'ctnr_kml_rapids_jupyter',
                             'ctnr_kml_input_mgr']:
        raise Exception(f'Unrecognized container reference {container_ref}')

    config = open_system_params_file()

    return config['system_container'][container_ref]


def get_profiler_configs():
    """Return flag for log_comms enablement."""

    config = open_config_file()
    profiler_mode = config['response']['enable_profiling']
    profiler_mode = True if profiler_mode.upper() == 'TRUE' else False

    return profiler_mode


def get_systables_ver():
    """Return the required system table version."""

    config = open_system_params_file()
    return config['general']['req_sys_table_ver']


def get_required_db_ver():
    """Return the required database version for KML operation."""

    config = open_system_params_file()

    # Return format is list: [major, minor, revision, api]
    return config['general']['req_database_ver'].split('.')


def get_state_manager_configs():
    """Return configurations for KML State Manager."""

    config = open_config_file()
    return config['state_manager']['wait_between_checks']


def get_pod_resource_defaults():
    """Return defaults for pod resource allocation."""

    config = open_config_file()

    return (
        int(config['cluster']['cpu_request']),
        int(config['cluster']['cpu_limit']),
        int(config['cluster']['gpu_request']),
        int(config['cluster']['mem_request']),
        int(config['cluster']['mem_limit'])
    )
