"""Support classess and functions for the Kinetica Machine Learning API."""

import json

import gpudb

from kml.api.kml_config import (
    get_db_conn,
    get_db_credentials,
    get_sys_schemas
)
from kml.api.aux import attach_log

SCHEMA_SYS, SCHEMA_AUDIT, SCHEMA_SYS_ARCH, SCHEMA_AUDIT_ARCH = get_sys_schemas()

# Configure log file
logger = attach_log(module=__name__, log_name='kml', debug=True)

# Config
db = get_db_conn()
db_user, _ = get_db_credentials()


def get_db_track():
    # NOTE: Support for 7.0 series DB has been removed.
    #  propz = db.show_system_properties()['property_map']['version.gpudb_core_version']
    return "7.1"


DB_TRACK = get_db_track()


def has_table(tbl_name, schema=None):
    tbl_name = tbl_name.strip()

    if DB_TRACK == "7.0":
        return db.has_table(table_name=tbl_name)['table_exists']
    elif DB_TRACK == "7.1":
        tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"
        return db.has_table(table_name=tbl_ref)['table_exists']
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def get_tbl_row_count(tbl_name, schema=None):
    tbl_name = tbl_name.strip()

    if DB_TRACK == "7.0":
        return gpudb.GPUdbTable(name=tbl_name, db=db).size()
    elif DB_TRACK == "7.1":
        tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"
        #  logger.debug(f"Getting row count for {tbl_ref}")
        st = db.show_table(table_name=tbl_ref, options={"get_sizes": "true"})
        #  logger.debug(st)
        return st['total_full_size']
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def do_refresh_mv(mv_name, schema=None):

    if DB_TRACK == "7.0":
        db.alter_table(table_name=mv_name,
                       action='refresh',
                       value='refresh',
                       options={})
    elif DB_TRACK == "7.1":
        tbl_ref = mv_name if not schema else f"{schema}.{mv_name}"
        db.alter_table(table_name=tbl_ref,
                       action='refresh',
                       value='refresh',
                       options={})
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def get_aggregate_min_max(tbl_name, col_name, schema=None):

    if DB_TRACK == "7.0":
        return db.aggregate_min_max(table_name=tbl_name.strip(),
                                    column_name=col_name)
    elif DB_TRACK == "7.1":
        tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"
        return db.aggregate_min_max(table_name=tbl_ref.strip(),
                                    column_name=col_name)
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def get_tbl_handle(tbl_name, schema=None):

    if DB_TRACK == "7.0":
        try:
            return gpudb.GPUdbTable(name=tbl_name, db=db)
        except Exception as e:
            if not has_table(tbl_name):
                errstr = (f"Table '{tbl_name}' in schema '{schema}' does not exist")
                logger.error(errstr)
                raise Exception(errstr)
            raise e

    elif DB_TRACK == "7.1":
        try:
            tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"
            return gpudb.GPUdbTable(name=tbl_ref, db=db)
        except Exception as e:
            if not has_table(tbl_name):
                schema = 'ki_home' if not schema else schema
                errstr = (f"Table '{tbl_name}' in schema '{schema}' does not exist")
                logger.error(errstr)
                raise Exception(errstr)
            raise e
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def check_if_exists_schema_or_collection(scname):

    if DB_TRACK == "7.0":
        if not db.has_table(scname)["table_exists"]:
            #  logger.debug(f"Schema/collection {scname} does not exist")
            return False
        th = gpudb.GPUdbTable(name=scname, db=db)
        return th.is_collection
    elif DB_TRACK == "7.1":
        return db.has_schema(scname)['schema_exists']
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def create_schema_or_collection(scname):

    if DB_TRACK == "7.0":
        th = gpudb.GPUdbTable(name=scname, db=db, options={'is_collection': "true"})
        return th
    elif DB_TRACK == "7.1":
        return db.create_schema(scname, options={'no_error_if_exists': "true"})
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def create_table(tbl_type_list, tbl_name, schema=None):

    if DB_TRACK == "7.0":
        myoptions = {}
        if schema:
            myoptions = {'collection_name': schema}
        return gpudb.GPUdbTable(_type=tbl_type_list,
                                name=tbl_name,
                                db=db,
                                options=myoptions)

    elif DB_TRACK == "7.1":
        tbl_name = tbl_name.strip()
        tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"

        if schema and not db.has_schema(schema)['schema_exists']:
            _ = create_schema_or_collection(scname=schema)

        if db.has_table(table_name=tbl_ref)['table_exists']:
            return gpudb.GPUdbTable(name=tbl_ref,
                                    db=db)
        else:
            return gpudb.GPUdbTable(_type=tbl_type_list,
                                    name=tbl_ref,
                                    db=db)
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def clear_table(tbl_name, schema=None):

    if DB_TRACK == "7.0":
        return db.clear_table(tbl_name, options={'no_error_if_not_exists': 'true'})
    elif DB_TRACK == "7.1":
        tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"
        return db.clear_table(tbl_ref, options={'no_error_if_not_exists': 'true'})
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def show_members(scname):
    return db.show_table(table_name=scname,
                         options={'show_children': 'true'})['table_names']


def create_mv_jointbl(view_name, table_names, column_names,
                      expressions, schema=None):
    """Create or refresh materialized views and joins."""

    if DB_TRACK == "7.0":

        db.clear_table(view_name, options={'no_error_if_not_exists': 'true'})

        mv_create_resp = db.create_materialized_view(
            table_name=view_name,
            options={'refresh_method': 'manual',
                     'persist': True,
                     'collection_name': schema}
        )

        if mv_create_resp['status_info']['status'] != 'OK':
            logger.error(json.dumps(mv_create_resp, indent=4, sort_keys=True))
            logger.error(mv_create_resp)

        view_id = mv_create_resp['view_id']

        jt_create_resp = db.create_join_table(
            join_table_name=view_name,
            table_names=[f"{tn}" for tn in table_names],
            column_names=column_names,
            expressions=expressions,
            options={'view_id': view_id,
                     'collection_name': schema}
        )

        if jt_create_resp['status_info']['status'] != 'OK':
            logger.error(json.dumps(jt_create_resp, indent=4, sort_keys=True))
            logger.error(jt_create_resp)

    elif DB_TRACK == "7.1":
        tbl_ref = view_name if not schema else f"{schema}.{view_name}"

        db.clear_table(tbl_ref, options={'no_error_if_not_exists': 'true'})

        mv_create_resp = db.create_materialized_view(
            table_name=tbl_ref,
            options={'refresh_method': 'manual',
                     'persist': True}
        )

        if mv_create_resp['status_info']['status'] != 'OK':
            logger.error(json.dumps(mv_create_resp, indent=4, sort_keys=True))
            logger.error(mv_create_resp)

        view_id = mv_create_resp['view_id']

        jt_create_resp = db.create_join_table(
            join_table_name=tbl_ref,
            table_names=[f"{schema}.{tn}" for tn in table_names],
            column_names=column_names,
            expressions=expressions,
            options={'view_id': view_id,
                     'collection_name': 'kml'}
        )

        if jt_create_resp['status_info']['status'] != 'OK':
            logger.error(json.dumps(jt_create_resp, indent=4, sort_keys=True))
            logger.error(jt_create_resp)

    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def move_table_to_diff_schema(tbl_name, schema_target, schema_current=None):

    if DB_TRACK == "7.0":
        status = db.alter_table(table_name=tbl_name,
                                action='move_to_collection',
                                value=schema_target)
        return status
    elif DB_TRACK == "7.1":
        tbl_ref = tbl_name if not schema_current else f"{schema_current}.{tbl_name}"
        status = db.alter_table(table_name=tbl_ref,
                                action='move_to_schema',
                                value=schema_target)
        return status
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def rename_table(tbl_name, tbl_name_desired, schema=None):

    if DB_TRACK == "7.0":
        status = db.alter_table(table_name=tbl_name,
                                action='rename_table',
                                value=tbl_name_desired)
        return status
    elif DB_TRACK == "7.1":
        tbl_ref = tbl_name if not schema else f"{schema}.{tbl_name}"
        status = db.alter_table(table_name=tbl_ref,
                                action='rename_table',
                                value=tbl_name_desired)
        return status
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
    return False


def get_qualified_table_name(tbl_name, schema=None):
    """Return a correctly formated table name for current DB."""

    tbl_name = tbl_name.strip()

    if DB_TRACK == "7.0":
        return tbl_name
    elif DB_TRACK == "7.1":
        return tbl_name if not schema else f"{schema.strip()}.{tbl_name}"
    else:
        raise Exception(f"Unknown DB version {DB_TRACK}")
