import sys
import argparse
import time
import traceback

import urllib3
import requests
from requests.exceptions import ConnectionError
import kubernetes as k8
import gpudb

from kml.api.kml_config import (
    get_api_base,
    get_namespace,
    get_db_conn,
    get_db_credentials,
    get_state_manager_configs,
    get_sys_schemas
)
from kml.api.kmldbops import (
    get_tbl_handle as kml_get_tbl_handle,
    do_refresh_mv as kml_do_refresh_mv,
    show_members as kml_show_members,
    get_db_track as kml_get_db_track
)
from kml.api.aux import attach_log
from kml.api.naming_svc import get_container_type


SCHEMA_SYS, _, _, _ = get_sys_schemas()

inflight_tbl_hdl = kml_get_tbl_handle(
    tbl_name="sys_kml_inventory_inflight",
    schema=SCHEMA_SYS
)

# Configure log file
logger = attach_log(module='kmlsweep', log_name='kml', debug=True)

check_delay_seconds = get_state_manager_configs()
if not check_delay_seconds.isnumeric():
    logger.error(f"Invalid setting for state_manager::wait_between_checks: "
                 f"{check_delay_seconds}, will default to 60 seconds instead")
    check_delay_seconds = 60
if int(check_delay_seconds) > 120:
    logger.warning(f"Setting for state_manager::wait_between_checks: "
                   f"{check_delay_seconds} is too high and not recommended")

logger.info(f"Will wait {check_delay_seconds} between sync checks")

# Read configuration file
namespace = get_namespace()
uri_base = get_api_base()
db_track = kml_get_db_track()
db = get_db_conn()
if float(db_track) >= 7.1:
    logger.info('Setting HA sync mode to REPL_NONE for DB >= 7.1')
    db.ha_sync_mode = gpudb.GPUdb.HASynchronicityMode.NONE

# Configure DB credentials
db_user, db_pass = get_db_credentials()
if db_user and db_pass:
    requests_cred = (db_user, db_pass)
else:
    requests_cred = None

try:
    k8.config.load_kube_config()
    k8_apps_v1_api = k8.client.AppsV1Api()
    k8_core_v1_api = k8.client.CoreV1Api()
    k8_batch_v1_api = k8.client.BatchV1Api()
    logger.info('Successfully connected to K8s')
except Exception as e:
    logger.error(e)
    logger.error('Cound not connect to K8, cannot proceed')
    sys.exit(1)


def get_k8_job_status(k8s_job_name, k8_job_dump):

    if k8_job_dump and k8_job_dump.items:

        for i in k8_job_dump.items:
            if i.metadata.name == k8s_job_name:
                if i.status.failed:
                    return "FAILED"
                if not i.status.succeeded:
                    if i.status.active > 0:
                        # TODO: Why are Pendings seen as RUNNING,
                        # can this be further delineated?
                        return "RUNNING"
                    else:
                        return "STARTING"
                if i.status.conditions and len(i.status.conditions) > 0:
                    if (i.status.conditions[0].type == 'Complete' and
                            i.status.conditions[0].status == 'True'):
                        return "COMPLETED"
    return "UNKNOWN"


def get_k8_depl_status(k8s_dep_name, k8_dep_dump):

    if k8_dep_dump:
        for i in k8_dep_dump.items:
            if i.metadata.name == k8s_dep_name:

                if i.status.ready_replicas and i.status.ready_replicas > 0:
                    return 'RUNNING'

                if len(i.status.conditions) > 1:
                    c0 = i.status.conditions[0]
                    c1 = i.status.conditions[1]

                    if (c0.reason == "MinimumReplicasUnavailable" and
                            'progressing' in c1.message and
                            c1.status == "True"):
                        return "STARTING"

                    if (c1.reason == "MinimumReplicasUnavailable" and
                            c1.status == "False"):
                        return "FAILED"

    return "UNKNOWN"


def get_k8_state():
    from kml.api.naming_svc import get_entity_id_from_k8s_name

    refreshed_k8_job_dump = k8_batch_v1_api.list_namespaced_job(
        namespace,
        timeout_seconds=10,
        watch=False
    )
    refreshed_k8_dep_dump = k8_apps_v1_api.list_namespaced_deployment(
        namespace,
        timeout_seconds=10,
        watch=False
    )

    enitity_to_k8s_map = {}

    for i in refreshed_k8_job_dump.items:
        entity_id = get_entity_id_from_k8s_name(i.metadata.name)
        enitity_to_k8s_map[entity_id] = {}
        enitity_to_k8s_map[entity_id]["k8s_name"] = i.metadata.name
        enitity_to_k8s_map[entity_id]["immortal"] = False
        enitity_to_k8s_map[entity_id]["k8s_status"] = get_k8_job_status(
            i.metadata.name,
            refreshed_k8_job_dump
        )
        enitity_to_k8s_map[entity_id]["entity_sub_type"] = get_container_type(
            i.metadata.name
        )

    for i in refreshed_k8_dep_dump.items:
        if i.metadata.name[:7] in ['kml-jpt', 'kml-slp']:
            continue
        entity_id = get_entity_id_from_k8s_name(i.metadata.name)
        if get_container_type(i.metadata.name) not in ["DEPLOYMENT_MANAGER",
                                                       "DATA_INPUT_MANAGER"]:
            enitity_to_k8s_map[entity_id] = {}
            enitity_to_k8s_map[entity_id]["k8s_name"] = i.metadata.name
            enitity_to_k8s_map[entity_id]["immortal"] = True
            enitity_to_k8s_map[entity_id]["k8s_status"] = get_k8_depl_status(
                i.metadata.name,
                refreshed_k8_dep_dump
            )
            enitity_to_k8s_map[entity_id]["entity_sub_type"] = get_container_type(
                i.metadata.name
            )

    return enitity_to_k8s_map


def get_kml_joblist():
    sys_schema_members = kml_show_members(SCHEMA_SYS)
    if "sys_kml_inventory_inflight" not in sys_schema_members:
        logger.error("Critical view missing: sys_kml_inventory_inflight")

    # Get state of Jobs and Deployments from Kubernetes Perspective
    enitity_to_k8s_map = get_k8_state()

    # Gather all the items to return keyed by entity id
    toret = []
    kml_do_refresh_mv(mv_name="sys_kml_inventory_inflight", schema=SCHEMA_SYS)
    in_flight_jobs_latest = inflight_tbl_hdl.get_records()

    # Merge these two perspectives
    for lo in in_flight_jobs_latest:
        eid = lo["entity_id"]
        newitem = {}
        newitem["entity_id"] = lo["entity_id"]
        newitem["entity_type"] = lo["entity_type"]
        newitem["entity_state"] = lo["entity_state"]

        if eid in enitity_to_k8s_map:
            newitem["entity_sub_type"] = enitity_to_k8s_map[eid]["entity_sub_type"]
            newitem["immortal"] = enitity_to_k8s_map[eid]["immortal"]
            newitem["k8sparent"] = enitity_to_k8s_map[eid]["k8s_name"]
            newitem["k8s_status"] = enitity_to_k8s_map[eid]["k8s_status"]
        else:
            # FIXME: Supress the excessive logging for MODEL_INSTANCE w/o a K8s entity
            # This should be fixed properly.
            #  if lo['entity_type'] in ['INGEST', 'DEPLOYMENT']:
            #      logger.warning(f"KML Entity {eid} of type {lo['entity_type']} "
            #                     f"not found on k8s")
            newitem["entity_sub_type"] = None
            newitem["immortal"] = None
            newitem["k8sparent"] = None
            newitem["k8s_status"] = None

        toret.append(newitem)

    return toret


def update_status_kml(entity_id, entity_type, entity_sub_type, destination_state):

    if destination_state not in ['COMPLETED', 'FAILED', 'RUNNING',
                                 'TERMINATED', 'STARTING', 'STOPPED']:
        # enforce restrictions given the context
        raise ValueError('Status can only be updated to RUNNING, COMPLETED, '
                         'STOPPED, or FAILED')

    if entity_type not in ['INGEST', 'DEPLOYMENT']:
        # enforce restrictions given the context
        raise ValueError(f"Encountered unexpected entity_type {entity_type} -- "
                         f"status can only be updated on Job type entities: "
                         f"INGEST, DEPLOYMENT")

    update_uri = None

    if entity_type == 'INGEST':
        update_uri = f'{uri_base}/ingest/{entity_id}/setstatus'
    elif entity_sub_type == 'BLACK_BOX_WORKER':
        update_uri = f'{uri_base}/model/deployment/{entity_id}/setstatus'
    else:
        raise ValueError(f"Unexpected combination of entity_type ({entity_type}) "
                         f"and entity_sub_type ({entity_sub_type})")

    try:
        _ = requests.post(
            update_uri,
            json={'destination-state': destination_state},
            auth=requests_cred,
            timeout=15
        )
    except Exception as e:
        time.sleep(30)
        logger.error(e)
        raise ValueError('KML API not responding')


def show_cluster_state(detailed_list):

    print(f'{"--":>5}  {"----------":<15} {"---------":<11} {"---------":<11} {"--------":<11}  {"----":<17} {"--------":<18}')
    print(f'{"ID":>5}  {"K8s Parent":<15} {"KML State":<11} {"k8s State":<11} {"Lifetime":<11}  {"Type":<17} {"Sub-Type":<18}')
    print(f'{"--":>5}  {"----------":<15} {"---------":<11} {"---------":<11} {"--------":<11}  {"----":<17} {"--------":<18}')
    for kp in detailed_list:
        entity_id = kp["entity_id"]
        entity_type = kp["entity_type"]
        entity_state = kp["entity_state"]
        entity_sub_type = kp["entity_sub_type"] if kp.get("entity_sub_type") else ""
        k8sparent = kp["k8sparent"] if kp.get("k8sparent") else ""
        k8s_status = kp["k8s_status"] if kp.get("k8s_status") else ""

        immortal = "Unknown"
        if "immortal" in kp:
            if kp["immortal"]:
                immortal = "Deployment"
            else:
                immortal = "Finite Job"

        print(f'{entity_id:>5}  {k8sparent:<15} {entity_state:<11} {k8s_status:<11} {immortal:<10}   {entity_type:<17} {entity_sub_type:<22}')


def syncClusterState():

    while True:
        try:
            time.sleep(int(check_delay_seconds))
            #  logger.info(f"Will wait {check_delay_seconds} between sync checks")

            detailed_list = sorted(get_kml_joblist(), key=lambda i: i['entity_id'])
            #  show_cluster_state(detailed_list)

            #  logger.info(f'detailed list: {detailed_list}')

            for kp in detailed_list:
                entity_id = kp["entity_id"]
                entity_type = kp["entity_type"]
                entity_sub_type = kp["entity_sub_type"]
                immortal = kp["immortal"]
                entity_state_kml = kp["entity_state"]
                entity_state_k8s = kp["k8s_status"]

                #  logger.debug(kp)

                # Propagate user-specified termination to k8s if it has
                # not already been done so
                if (entity_state_kml == "TERMINATED" and
                        entity_state_k8s in ["UNKNOWN", "STARTING", "RUNNING"]):

                    logger.debug(f"Failing entity {kp['entity_id']}: API blame")
                    update_status_kml(entity_id=entity_id,
                                      entity_type=entity_type,
                                      entity_sub_type=entity_sub_type,
                                      destination_state="FAILED")

                # Propagate on cluster failure back to kml system and db
                if (entity_state_kml not in ["FAILED", "COMPLETED"] and
                        entity_state_k8s == "FAILED"):

                    logger.debug(f"Failing entity {kp['entity_id']}: K8s blame")
                    update_status_kml(entity_id=entity_id,
                                      entity_type=entity_type,
                                      entity_sub_type=entity_sub_type,
                                      destination_state="FAILED")

                # Propagate on cluster deployment success back to kml system and db
                # Ingest containers only, models intentionally trigger their own
                # running state.
                if (entity_state_kml == "STARTING" and
                        entity_state_k8s == "RUNNING" and
                        entity_type == 'INGEST'):

                    update_status_kml(entity_id=entity_id,
                                      entity_type=entity_type,
                                      entity_sub_type=entity_sub_type,
                                      destination_state="RUNNING")

                # Lets handle specific cases for immortal deployments/services
                if immortal:
                    pass

                # Lets handle specific cases for finite life jobs
                else:
                    # Propagate on cluster completion back to kml system and db
                    if (entity_state_kml != "COMPLETED" and
                            entity_state_k8s == "COMPLETED"):
                        update_status_kml(entity_id=entity_id,
                                          entity_type=entity_type,
                                          entity_sub_type=entity_sub_type,
                                          destination_state="COMPLETED")

        except urllib3.exceptions.NewConnectionError as nce:
            logger.error('Problems connecting to KML, will re-try after wait')
            logger.error(nce)

        except ConnectionError as e:    # This is the correct syntax
            logger.error('Problems connecting to KML, will re-try after wait')
            logger.error(e)

        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())


def dump_cluster_state():

    try:
        detailed_list = sorted(get_kml_joblist(), key=lambda i: i['entity_id'])
        show_cluster_state(detailed_list)

    except Exception as e:
        logger.error('Error encountered while dumping cluster state')
        logger.error(e)
        logger.error(traceback.format_exc())


def main():
    logger.info('Starting KML State Manager')
    parser = argparse.ArgumentParser()

    global check_delay_seconds
    parser.set_defaults(check_delay=check_delay_seconds)
    parser.set_defaults(keep_completed=False)
    parser.set_defaults(dump_cluster_state=False)

    parser.add_argument('--check-delay',
                        type=int,
                        default=check_delay_seconds,
                        help='Seconds delay between checks (defaults to 2 seconds)')
    parser.add_argument('--see-k8s',
                        action='store_true',
                        dest='dump_cluster_state',
                        help='Dump state of K8s Cluster')

    args = parser.parse_args()

    check_delay_seconds = args.check_delay
    logger.info(f'Using delay of {check_delay_seconds} seconds between checks')

    # We either want to dump the state of the cluster
    if args.dump_cluster_state:
        dump_cluster_state()
        sys.exit(0)

    # ...or we want to continuously check and sync db-k8s status
    logger.info(f"Syncing KML-K8s Work Status")
    logger.info(f"Using Kubernetes namespace {namespace}")
    _ = syncClusterState()
    # the above should run forever, so no exit required


if __name__ == '__main__':
    main()
