import traceback
import argparse
import datetime

import gpudb

#  from kml.api.systable_spec import *
from kml.api.aux import attach_log
from kml.api.kml_config import get_db_conn
#from kml.api.setup_tables_system import coll_audit_archive, coll_sys_archive
from kml.api.systable_spec import TBL_UNIVERSE_KML_SYSTABLES
from kml.api.kml_config import get_sys_schemas
SCHEMA_SYS, SCHEMA_AUDIT, SCHEMA_SYS_ARCH, SCHEMA_AUDIT_ARCH = get_sys_schemas()

# Configure log file
logger = attach_log(module="manage_systables", log_name='kml', debug=True)

# Get DB handle
db = get_db_conn()


def tables_audit_inventory():
    """Generate list of all system base tables."""

    from kml.api.kml_config import get_sys_schemas
    from kml.api.kmldbops import show_members as kml_show_members
    _, SCHEMA_AUDIT, _, _ = get_sys_schemas()
    return kml_show_members(scname=SCHEMA_AUDIT)

def tables_audit_inspect():
    from kml.api.kmldbops import check_if_exists_schema_or_collection

    if check_if_exists_schema_or_collection(scname=SCHEMA_AUDIT):
        at = tables_audit_inventory()
        logger.info(f"Active audit tables: ({len(at)}) found")
        for a in at:
            logger.info(f"\tFound audit table: {a}")
    else:
        logger.info(f"Skipping audit inspection -- schema {SCHEMA_AUDIT} does not exist")

def tables_audit_clear():
    from kml.api.kmldbops import clear_table as kml_clear_table
    from kml.api.kmldbops import has_table as kml_has_table
    from kml.api.kmldbops import check_if_exists_schema_or_collection

    if check_if_exists_schema_or_collection(scname=SCHEMA_AUDIT):
        for gat in tables_audit_inventory():
            status = kml_clear_table(tbl_name=gat, schema=SCHEMA_AUDIT)
            if 'status_info' not in status or status['status_info']['status'] != 'OK':
                logger.error(f"Error deleting audit table {gat}")
            else:
                logger.info(f"   Successfully purged audit table {gat}")
    else:
        logger.info(f"Skipping audit clear -- schema {SCHEMA_AUDIT} does not exist")

def tables_audit_archive():
    """Backup all audit tables."""
    from kml.api.kmldbops import move_table_to_diff_schema as kml_move_table_to_diff_schema
    from kml.api.kmldbops import rename_table as kml_rename_table

    dt = datetime.datetime.utcnow().isoformat(' ', 'seconds')
    clean_dt = dt.replace(' ', '_').replace('-', '_').replace(':', '_')

    for tbl_name in tables_audit_inventory():
        new_tbl_name = f'archive_{clean_dt}_{tbl_name}'

        logger.info(f"Renaming table {tbl_name} to {new_tbl_name}")
        status = kml_rename_table(tbl_name=tbl_name,
                                  tbl_name_desired=new_tbl_name,
                                  schema=SCHEMA_AUDIT)
        #status = db.alter_table(table_name=tbl_name,
        #                        action='rename_table',
        #                        value=new_tbl_name)

        if status['status_info']['status'] != 'OK':
            logger.error(f'Error re-naming {tbl_name} to {new_tbl_name}')
        logger.info(f"Moving table {new_tbl_name} to schema {SCHEMA_AUDIT_ARCH}")
        status = kml_move_table_to_diff_schema(tbl_name=new_tbl_name,
                                               schema_target=SCHEMA_AUDIT_ARCH,
                                               schema_current=SCHEMA_AUDIT)
        #status = db.alter_table(table_name=new_tbl_name,
        #                        action='move_to_schema',
        #                        value=SCHEMA_AUDIT_ARCH)
        #print(status)

        if status['status_info']['status'] != 'OK':
            logger.error(f'Error moving {new_tbl_name} to collection {SCHEMA_AUDIT_ARCH}')


def tables_system_inventory():
    """Generate list of all system base tables."""

    resp = db.show_table(table_name='kml',
                         options={'show_children': 'true'})
    all_names = resp['table_names']
    all_desc = resp['table_descriptions']

    sys_tbl_names = []
    for i, table in enumerate(all_names):
        if 'VIEW' not in all_desc[i]:
            sys_tbl_names.append(table)

    return sys_tbl_names


def tables_system_archive():
    """Backup all system tables."""

    from kml.api.kmldbops import move_table_to_diff_schema as kml_move_table_to_diff_schema
    from kml.api.kmldbops import rename_table as kml_rename_table

    dt = datetime.datetime.utcnow().isoformat(' ', 'seconds')
    clean_dt = dt.replace(' ', '_').replace('-', '_').replace(':', '_')

    for tbl_name in tables_system_inventory():
        new_tbl_name = f'archive_{clean_dt}_{tbl_name}'

        logger.info(f"Renaming table {tbl_name} to {new_tbl_name}")
        status = kml_rename_table(tbl_name=tbl_name, tbl_name_desired=new_tbl_name, schema=SCHEMA_SYS)
        #status = db.alter_table(table_name=tbl_name,
        #                        action='rename_table',
        #                        value=new_tbl_name)

        if status['status_info']['status'] != 'OK':
            logger.error(f'Error re-naming {tbl_name} to {new_tbl_name}')

        logger.info(f"Moving table {new_tbl_name} to schema {SCHEMA_SYS_ARCH}")
        status = kml_move_table_to_diff_schema(tbl_name=new_tbl_name, schema_target=SCHEMA_SYS_ARCH, schema_current=SCHEMA_SYS)
        #status = db.alter_table(table_name=new_tbl_name,
        #                        action='move_to_collection',
        #                        value=SCHEMA_SYS_ARCH)

        if status['status_info']['status'] != 'OK':
            logger.error(f'Error moving {new_tbl_name} to collection {SCHEMA_SYS_ARCH}')

def tables_system_clear():
    from kml.api.kmldbops import clear_table as kml_clear_table
    from kml.api.kmldbops import has_table as kml_has_table
    for tbl_name in TBL_UNIVERSE_KML_SYSTABLES:
        if kml_has_table(tbl_name, schema=SCHEMA_SYS):
            try:
                if kml_has_table(tbl_name, schema=SCHEMA_SYS):
                    logger.info(f"Deleting table {tbl_name}")
                    status = kml_clear_table(tbl_name=tbl_name,
                                             schema=SCHEMA_SYS)
                    if 'status_info' in status and 'status' in status['status_info']:
                        logger.info(f"\tdeletion status for {tbl_name} is: {status['status_info']['status']}")
                        if status['status_info']['status'] != 'OK':
                            logger.error(f"\tdeletion status for {tbl_name} is: {status}")
                    else:
                        logger.error(f"\tdeletion status for {tbl_name} is: {status}")
                    if kml_has_table(tbl_name, schema=SCHEMA_SYS):
                        logger.info(f"Deletion of table {tbl_name} failed")
                else:
                    logger.info(f"\tDeletion skipped for {tbl_name}...table non-existent")
            except Exception as e:
                logger.error(f"Trouble deleting {tbl_name}")
                #logger.error(status)
                logger.error(e)
                logger.error(traceback.format_exc())
        else:
            logger.info(f"Table {tbl_name} already non-existent, nothing to clear")

def main():
    try:
        logger.info("DB Connection - Attempting")
        db = get_db_conn()
        logger.info("DB Connection - Acquired")

        parser = argparse.ArgumentParser()

        parser.set_defaults(audit_inspect=False)
        parser.set_defaults(audit_clear=False)
        parser.set_defaults(audit_archive=False)

        parser.set_defaults(systable_clear=False)
        parser.set_defaults(systable_archive=False)

        parser.add_argument("--audit-inspect",
                            action="store_true",
                            dest="audit_inspect",
                            help="Inspect all active KML Audit Tables")
        parser.add_argument("--audit-clear",
                            action="store_true",
                            dest="audit_clear",
                            help="Truncate and drop all KML Audit Tables")
        parser.add_argument("--audit-archive",
                            action="store_true",
                            dest="audit_archive",
                            help="Archive all KML Audit Tables")

        parser.add_argument("--systable-clear",
                            action="store_true",
                            dest="systable_clear",
                            help="Truncate and drop all KML System Tables")
        parser.add_argument("--systable-archive",
                            action="store_true",
                            dest="systable_archive",
                            help="Archive all KML System Tables")
        args = parser.parse_args()

        if args.audit_inspect:
            try:
                logger.info("Inspecting audit tables")
                tables_audit_inspect()
                logger.info("Inspecting audit tables...completed successfully")
            except Exception as e:
                logger.error(e)
                logger.error(traceback.format_exc())

        if args.audit_clear:
            try:
                logger.info("Clearing all audit tables")
                logger.debug("Inventory of audit tables before clearing")
                tables_audit_inspect()
                tables_audit_clear()
                logger.info("Clearing all audit tables...completed successfully")
                logger.debug("Inventory of audit tables after clearing")
                tables_audit_inspect()

            except Exception as e:
                logger.error(e)
                logger.error(traceback.format_exc())

        if args.audit_archive:
            try:
                logger.info("Archiving all audit tables...starting")
                logger.debug("Inventory of audit tables before clearing")
                tables_audit_inspect()
                tables_audit_archive()
                logger.info("Clearing all audit tables...completed successfully")
                logger.debug("Inventory of audit tables after clearing")
                tables_audit_inspect()
                logger.info("Archiving all audit tables...completed successfully")

            except Exception as e:
                logger.error(e)
                logger.error(traceback.format_exc())

        if args.systable_clear:
            try:
                logger.info("Purging all system tables")
                tables_system_clear()
                logger.info("Purging all system tables...completed successfully")
            except Exception as e:
                logger.error(e)
                logger.error(traceback.format_exc())

        if args.systable_archive:
            try:
                logger.info("Archiving all system tables")
                tables_system_archive()
            except Exception as e:
                logger.error(e)
                logger.error(traceback.format_exc())

    except Exception as e:
        logger.error(e)
        logger.error(traceback.format_exc())


if __name__ == '__main__':
    logger.info("KML System & Audit Table Manager...starting")
    main()
    logger.info("KML System & Audit Table Manager...completed")
