import os
import sys
import json
import time
import random
import collections
import datetime

import fasteners
import gpudb

#from kml.api.systable_spec import *
from kml.api.kml_config import get_db_conn

from kml.api.kml_config import get_sys_schemas
SCHEMA_SYS, SCHEMA_AUDIT, SCHEMA_SYS_ARCH, SCHEMA_AUDIT_ARCH = get_sys_schemas()

from kml.api.kmldbops import get_tbl_handle as kml_get_tbl_handle
from kml.api.kmldbops import has_table as kml_has_table

from kml.api.aux import attach_log

logger = attach_log(module=__name__, log_name='kml', debug=True)

# Lock path used for synchronized Entity Creation block
LOCK_FILE = "./kml_entity_create_lock"

# Get DB handle
db = get_db_conn()

class Entity:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_entity
    table_name='sys_kml_entity'
    table_schema=TBL_SCHEMA_sys_kml_entity
    ENTITY_TYPES=['INGEST', 'DATASET', 'FEATURESET', 'PROJECT', 'MODEL_INSTANCE', 'DEPLOYMENT', 'PRIMITIVE', 'AUDIT', 'CREDENTIAL']
    VISIBILITY_STATES = ['VISIBLE', 'ARCHIVED']
    DOCUMENT_STATES = ['DRAFT', 'PUBLISHED']
    ENTITY_STATES = ['PENDING', 'CREATED', 'STARTING', 'RUNNING', 'COMPLETED', 'FAILED', 'TERMINATED', 'STOPPED']

    def __init__(self, entity_id, entity_name, entity_desc, entity_type, entity_state, doc_state, visibility, create_dt, update_dt, free_map_of_maps):
        self.entity_id=entity_id
        if entity_name:
            self.name = entity_name[:64]
        else:
            entity_name = ""

        if entity_desc:
            self.desc = entity_desc[:256]
        else:
            self.desc = ""
        self.entity_type = entity_type
        self.entity_state=entity_state
        self.doc_state=doc_state
        self.visibility=visibility
        self.create_dt = create_dt
        self.update_dt = update_dt
        self.free_map_of_maps = free_map_of_maps

    def persist_insert(self):
        entity_datum = collections.OrderedDict()
        entity_datum["entity_id"] = self.entity_id
        entity_datum["entity_name"] = self.name
        entity_datum["create_dt"] = self.create_dt
        entity_datum["update_dt"] = self.update_dt
        entity_datum["entity_type"] = self.entity_type
        entity_datum["entity_state"] = self.entity_state
        entity_datum["doc_state"] = self.doc_state
        entity_datum["visibility"] = self.visibility
        entity_datum["entity_desc"] = self.desc
        entity_datum["free_map_of_maps"] = json.dumps(self.free_map_of_maps)
        #sink_table = gpudb.GPUdbTable( name = self.table_name, db = db)
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)
        _=sink_table.insert_records(entity_datum)

    @staticmethod
    def roster():
        kmltypes = ["PROJECT", "MODEL_INSTANCE", "DEPLOYMENT", "FEATURESET", "DATASET", "INGEST", "AUDIT", "CREDENTIAL", "PRIMITIVE", "FILE", "AUDIT"]
        toret = {t:[] for t in kmltypes}

        tbl = kml_get_tbl_handle(tbl_name=Entity.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(options={}, limit=-9999)
        for lo in loadedObject:
            eid = int(lo['entity_id'])
            etype = lo['entity_type']
            toret[etype].append(eid)

        tbl = kml_get_tbl_handle(tbl_name = KMLFile.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999)
        for lo in loadedObject:
            toret["FILE"].append(int(lo['file_id']))

        return toret

    @staticmethod
    @fasteners.interprocess_locked(LOCK_FILE)
    # The above lock ensures distributed uvicorn workers dont grab the same next primary key
    def create(entity_name, entity_desc, entity_type):
        if entity_type not in Entity.ENTITY_TYPES:
            raise ValueError('Unknown entity type ' + str(entity_type))
        if not entity_name:
            raise ValueError('Entity must have non-blank name, received: ' + str(entity_name))

        my_timestamp = datetime.datetime.utcnow().isoformat(' ')[:-3]
        base_entity = Entity(entity_id=Entity.get_next_entity_id(),
            entity_name=entity_name,
            entity_desc=entity_desc,
            entity_type=entity_type,
            entity_state="PENDING",
            doc_state="PUBLISHED",
            visibility="VISIBLE",
            create_dt=my_timestamp,
            update_dt=my_timestamp,
            free_map_of_maps={}
            )
        base_entity.persist_insert()
        return base_entity

    def renew_last_updated_timestamp(self):
        self.update_dt=datetime.datetime.utcnow().isoformat(' ')[:-3]
        entity_datum = collections.OrderedDict()
        entity_datum["update_dt"] = self.update_dt
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)

        sink_table.update_records(
            expressions=["entity_id=" +str(self.entity_id)],
            new_values_maps={
                "update_dt": self.update_dt
                })

    def get_state(self):
        return self.entity_state
    def is_running(self):
        return self.entity_state=='RUNNING'
    def is_completed(self):
        return self.entity_state=='COMPLETED'
    def transition_to_created(self):
        self.change_state('CREATED')
    def transition_to_starting(self):
        self.change_state('STARTING')
    def transition_to_running(self):
        self.change_state('RUNNING')
    def transition_to_failed(self):
        self.change_state('FAILED')
    def transition_to_stopped(self):
        self.change_state('STOPPED')
    def transition_to_terminated(self):
        self.change_state('TERMINATED')
    def transition_to_completed(self):
        self.change_state('COMPLETED')

    def change_state(self, to_state):
        if to_state not in Entity.ENTITY_STATES:
            raise ValueError("Unknown destination state %s" % to_state)

        if self.entity_state in ['COMPLETED','FAILED','TERMINATED']:
            # we're in a terminal state, cannot change state....
            if to_state==self.entity_state:
                logger.warning(f"WARNING: Destination state and current state are identical, {to_state}, no change made")
            else:
                logger.warning("Object is already in a terminal state %s, cannot change state to %s" % (self.entity_state, to_state))

        if self.entity_state in ['RUNNING'] and to_state=="CREATED":
            raise ValueError("Cannot go from Running state backwards to Created state")

        self.entity_state=to_state
        self.update_dt=datetime.datetime.utcnow().isoformat(' ')[:-3]

        sink_table = kml_get_tbl_handle(tbl_name = self.table_name, schema=SCHEMA_SYS)
        sink_table.update_records(
            expressions=["entity_id=" +str(self.entity_id)],
            new_values_maps={
                "entity_state": self.entity_state,
                "update_dt": self.update_dt
                })

    def publish(self):
        self.doc_state="PUBLISHED"
        self.update_dt=datetime.datetime.utcnow().isoformat(' ')[:-3]

        sink_table = kml_get_tbl_handle(tbl_name = self.table_name, schema=SCHEMA_SYS)

        sink_table.update_records(
            expressions=["entity_id=" +str(self.entity_id)],
            new_values_maps={
                "doc_state": self.doc_state,
                "update_dt": self.update_dt
                })

    def tombstone(self):
        self.visibility="ARCHIVED"
        self.update_dt=datetime.datetime.utcnow().isoformat(' ')[:-3]

        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)

        sink_table.update_records(
            expressions=["entity_id=" +str(self.entity_id)],
            new_values_maps={
                "visibility": self.visibility,
                "update_dt": self.update_dt
                })

    @staticmethod
    def get_next_entity_id():
        from kml.api.kmldbops import get_aggregate_min_max as kml_get_aggregate_min_max
        # TODO: Need to wrap this into a mutex and have a global-next-available-id
        tbl_h = kml_get_tbl_handle(tbl_name=Entity.table_name, schema=SCHEMA_SYS)
        if tbl_h.count==0:
            return 1

        minmax=kml_get_aggregate_min_max(tbl_name=Entity.table_name, col_name="entity_id", schema=SCHEMA_SYS)
        if "status_info" not in minmax:
            raise Exception("Encountered error querying DB")
        if "status" not in minmax["status_info"]:
            raise Exception(f"Encountered error querying DB {minmax['status_info']}")
        if minmax["status_info"]["status"]=='ERROR':
            raise Exception(f"Encountered error querying DB {minmax['status_info']}")
        return int(minmax["max"])+1

    @staticmethod
    def load(existing_entity_id):
        #tbl = gpudb.GPUdbTable( name = Entity.table_name, db = db)
        tbl = kml_get_tbl_handle(tbl_name=Entity.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": "entity_id="+str(existing_entity_id) })
        if len(loadedObject)==0:
            raise ValueError('No Entity found with id ' + str(existing_entity_id))
        if len(loadedObject)>1:
            raise ValueError('Major issue: more than one Entity found with id ' + str(existing_entity_id))
        if len(loadedObject)==1:
            for lo in loadedObject:
                #TODO: we should not iterate here, just choose the one-and-only object
                return Entity(
                    entity_id=lo['entity_id'],
                    entity_name=lo['entity_name'],
                    entity_desc=lo['entity_desc'],
                    entity_type=lo['entity_type'],
                    entity_state=lo['entity_state'],
                    doc_state=lo['doc_state'],
                    visibility=lo['visibility'],
                    create_dt=lo['create_dt'],
                    update_dt=lo['update_dt'],
                    free_map_of_maps=json.loads(lo['free_map_of_maps'])
                    )

    def get_entity_id(self):
        return self.entity_id

    def serialize(self):
        serialization = {}
        serialization['entity_id'] = self.entity_id
        serialization['entity_name'] = self.name
        serialization['entity_desc'] = self.desc
        serialization['entity_type'] = self.entity_type
        serialization['entity_state'] = self.entity_state
        serialization['doc_state'] = self.doc_state
        serialization['visibility'] = self.visibility
        serialization['create_dt'] = self.create_dt
        serialization['update_dt'] = self.update_dt
        serialization['free_map_of_maps'] = self.free_map_of_maps
        return serialization

    def __str__(self):
        return "Entity %s %s" % (str(self.entity_id), json.dumps(self.serialize(), indent=4, sort_keys=True))


class Project:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_project
    table_name='sys_kml_project'
    table_schema=TBL_SCHEMA_sys_kml_project

    def __init__(self, base_entity, project_config):
        self.base_entity = base_entity
        self.project_config = project_config

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity
    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def create(project_name, project_desc, project_config):

        if not Project.validate_config(project_config):
            raise ValueError('Project Configuration given is invalid ' + json.dumps(project_config))

        base_entity = Entity.create(entity_name=project_name,
            entity_desc=project_desc,
            entity_type="PROJECT")
        proj_inst = Project(base_entity=base_entity,
            project_config=project_config)
        proj_inst.get_base_entity().transition_to_created()
        proj_inst.persist_insert()
        return proj_inst

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    def persist_insert(self):
        model_datum = collections.OrderedDict()
        model_datum["entity_id"] = self.base_entity.get_entity_id()
        model_datum["project_config"] = json.dumps(self.project_config)
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(model_datum)

    @staticmethod
    def load(existing_project_id):
        tbl = kml_get_tbl_handle(tbl_name=Project.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"entity_id={existing_project_id}" })
        if len(loadedObject)>0:
            for lo in loadedObject:
                return Project(base_entity=Entity.load(lo['entity_id']),
                    project_config=json.loads(lo['project_config']))
        else:
            raise LookupError('No Project found with id ' + str(existing_project_id))

    @staticmethod
    def clone(existing_project_id):
        to_copy_from=Project.load(existing_project_id)
        return Project.create(
            project_name=to_copy_from.base_entity.name,
            project_desc=to_copy_from.base_entity.desc,
            project_config=to_copy_from.project_config)

    @staticmethod
    def exists(existing_project_id):
        tbl = kml_get_tbl_handle(tbl_name=Project.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"entity_id={existing_project_id}" })
        if len(loadedObject)>0:
            return True
        return False

    @staticmethod
    def validate_config(project_config):
        #TODO: Create project configuration validation steps
        #TODO: Minimally, determine if it is a valid JSON
        # for now, we assume all project configs are valid
        return True

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['project_config'] = self.project_config
        return serialization

    def __str__(self):
        return "Project %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True))


class Ingest:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_ingest
    table_name = "sys_kml_ingest"
    table_schema = TBL_SCHEMA_sys_kml_ingest

    INGEST_MODES=["BATCH", "CONTINUOUS"]
    INGEST_TYPES=["BYOC", "KIO", "KAFKA"]

    def __init__(self, base_entity, ingest_mode, ingest_type,
                 ingest_config, file_id=None):

        self.base_entity = base_entity
        self.ingest_mode = ingest_mode
        self.ingest_type = ingest_type
        self.ingest_config = ingest_config
        self.file_id = file_id

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.get_base_entity().publish()
    def tombstone(self):
        self.get_base_entity().tombstone()
    def get_base_entity(self):
        return self.base_entity

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def create(ingest_name, ingest_desc, ingest_mode, ingest_type,
               ingest_config, file_id=None):

        if not Ingest.validate_config(ingest_config):
            raise ValueError(f"Ingest Configuration given is invalid "
                             f"{json.dumps(ingest_config)}")

        if ingest_mode not in Ingest.INGEST_MODES:
            raise ValueError(f"Unknown ingest mode {ingest_mode}")

        if ingest_type not in Ingest.INGEST_TYPES:
            raise ValueError(f"Unknown ingest type {ingest_type}")

        base_entity = Entity.create(entity_name=ingest_name,
                                    entity_desc=ingest_desc,
                                    entity_type="INGEST")

        ingest_inst = Ingest(base_entity=base_entity,
                             ingest_mode=ingest_mode,
                             ingest_type=ingest_type,
                             ingest_config=ingest_config,
                             file_id=file_id)

        ingest_inst.persist_insert()

        return ingest_inst

    def persist_insert(self):
        injectable = collections.OrderedDict()
        injectable['entity_id'] = self.base_entity.get_entity_id()
        injectable['ingest_mode'] = self.ingest_mode
        injectable['ingest_type'] = self.ingest_type
        injectable['ingest_config'] = json.dumps(self.ingest_config)
        injectable['file_id'] = self.file_id

        sink_table = kml_get_tbl_handle(tbl_name=self.table_name,
                                        schema=SCHEMA_SYS)
        insert_status = sink_table.insert_records(injectable)

    def set_config(self, in_config):
        if not Ingest.validate_config(in_config):
            raise ValueError(f"Ingest Configuration given is invalid "
                             f"{json.dumps(in_config)}")

        sink_table = kml_get_tbl_handle(tbl_name=self.table_name,
                                        schema=SCHEMA_SYS)
        sink_table.update_records(
            expressions=[f"entity_id={self.base_entity.get_entity_id()}"],
            new_values_maps={'ingest_config': json.dumps(in_config)}
        )

    @staticmethod
    def load(existing_ingest_id):
        tbl = kml_get_tbl_handle(tbl_name=Ingest.table_name,
                                 schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999,
            options={"expression": f"entity_id={existing_ingest_id}"}
        )
        if len(loadedObject) > 0:
            for lo in loadedObject:
                return Ingest(base_entity=Entity.load(lo['entity_id']),
                              ingest_mode=lo['ingest_mode'],
                              ingest_type=lo['ingest_type'],
                              ingest_config=json.loads(lo['ingest_config']),
                              file_id=lo['file_id'])
        else:
            raise LookupError(f'No Ingest found with id {existing_ingest_id}')

    def is_state_change_acceptable(self, to_state):
        # TODO: Validate if this change is acceptable
        return True

    @staticmethod
    def clone(existing_ingest_id):
        to_copy_from = Ingest.load(existing_ingest_id)

        newfile=None
        if to_copy_from.file_id:
            newfile = KMLFile.clone(to_copy_from.file_id).file_id

        return Ingest.create(
            ingest_name=to_copy_from.base_entity.name,
            ingest_desc=to_copy_from.base_entity.desc,
            ingest_mode=to_copy_from.ingest_mode,
            ingest_type=to_copy_from.ingest_type,
            ingest_config=to_copy_from.ingest_config,
            file_id=newfile
        )

    @staticmethod
    def exists(existing_ingest_id):
        tbl = kml_get_tbl_handle(tbl_name=Ingest.table_name,
                                 schema=SCHEMA_SYS)

        loadedObject = tbl.get_records(limit=-9999,
            options={'expression': f"entity_id={existing_ingest_id}"}
        )
        if len(loadedObject) > 0:
            return True
        return False

    @staticmethod
    def validate_config(ingest_config):
        # TODO: Create ingest_config validation steps
        # TODO: Minimally, determine if it is a valid JSON
        # for now, we assume all ingest_configs are valid
        return True

    def get_entity_id(self):
        return self.base_entity.get_entity_id()

    def get_file(self):
        if not self.file_id:
            return None
        return KMLFile.load(self.file_id).get_file()

    def set_file(self, infile):
        kf1 = KMLFile.create(
            the_file_path=infile,
            file_name=os.path.basename(infile),
            file_desc=f"JumboXar for ingest id {self.get_entity_id()}"
        )
        self.file_id=kf1.file_id
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name,
                                        schema=SCHEMA_SYS)
        sink_table.update_records(
            expressions=[f"entity_id={self.base_entity.get_entity_id()}"],
            new_values_maps={'file_id': str(kf1.file_id)}
        )
        return True

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['ingest_mode'] = self.ingest_mode
        serialization['ingest_type'] = self.ingest_type
        serialization['ingest_config'] = self.ingest_config
        serialization['file_id'] = self.file_id
        return serialization

    def __str__(self):
        return "Ingest %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True))


class ModelInstance:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_model_inst
    table_name='sys_kml_model_inst'
    table_schema=TBL_SCHEMA_sys_kml_model_inst

    MODEL_TYPES=['TENSORFLOW', 'BLACKBOX','RAPIDS', 'H2O', 'TURI']

    def __init__(self, base_entity, model_inst_config, resource_config, featureset_id, file_id_servable_mdl_binary=None, file_id_training_logs=None, file_id_xformer_ref=None, metrics={}):
        self.base_entity = base_entity
        self.model_inst_config = model_inst_config
        self.resource_config = resource_config
        self.featureset_id=featureset_id
        self.file_id_servable_mdl_binary=file_id_servable_mdl_binary
        self.file_id_training_logs=file_id_training_logs
        self.file_id_xformer_ref=file_id_xformer_ref
        self.metrics=metrics

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def create(model_inst_name, model_inst_desc, model_inst_config, resource_config=None, featureset_id=None, file_id_servable_mdl_binary=None, file_id_training_logs=None, file_id_xformer_ref=None, metrics={}):

        if not ModelInstance.validate_config(model_inst_config):
            raise ValueError('Model Instance Configuration given is invalid ' + str(model_inst_config))

        if file_id_servable_mdl_binary:
            if not KMLFile.exists(file_id_servable_mdl_binary):
                raise ValueError('Unknown KMLFile with file_id_servable_mdl_binary ' + str(file_id_servable_mdl_binary))

        if file_id_training_logs:
            if not KMLFile.exists(file_id_training_logs):
                raise ValueError('Unknown KMLFile with file_id_training_logs ' + str(file_id_training_logs))

        if file_id_xformer_ref:
            if not KMLFile.exists(file_id_xformer_ref):
                raise ValueError('Unknown KMLFile with file_id ' + str(file_id_xformer_ref))

        if featureset_id:
            if not FeatureSet.exists(featureset_id):
                raise ValueError('Unknown featureset_id %s' % str(featureset_id))

        base_entity = Entity.create(entity_name=model_inst_name,
            entity_desc=model_inst_desc,
            entity_type="MODEL_INSTANCE")
        supermodel_instance = ModelInstance(base_entity=base_entity,
            model_inst_config=model_inst_config,
            resource_config=resource_config,
            featureset_id=featureset_id,
            file_id_servable_mdl_binary = file_id_servable_mdl_binary,
            file_id_training_logs = file_id_training_logs,
            file_id_xformer_ref = file_id_xformer_ref,
            metrics=metrics)
        supermodel_instance.persist_insert()
        return supermodel_instance

    @staticmethod
    def load(existing_model_inst_id):
        tbl = kml_get_tbl_handle(tbl_name=ModelInstance.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"entity_id={existing_model_inst_id}" })
        if len(loadedObject)==0:
            raise LookupError('No Model Instance found with id ' + str(existing_model_inst_id))
        if len(loadedObject)>1:
            # this should NEVER happen, this would indicate very, very bad data corruption
            raise ValueError('Possible system corruption, more than one Model Instance found with id '
                + str(existing_model_inst_id))
        if len(loadedObject)==1:
            lo = loadedObject[0]
            mitemp=ModelInstance(
                base_entity=Entity.load(existing_model_inst_id),
                model_inst_config=json.loads(lo['model_inst_config']),
                resource_config=json.loads(lo['resource_config']),
                featureset_id=lo['featureset_id'],
                file_id_servable_mdl_binary = lo['file_id_servable_mdl_binary'],
                file_id_training_logs = lo['file_id_training_logs'],
                file_id_xformer_ref = lo['file_id_xformer_ref'],
                metrics=json.loads(lo['metrics']))
            return mitemp

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    def persist_insert(self):
        model_inst_datum = collections.OrderedDict()
        model_inst_datum["entity_id"] = self.base_entity.get_entity_id()
        model_inst_datum["model_inst_config"] = json.dumps(self.model_inst_config)
        model_inst_datum["resource_config"] = json.dumps(self.resource_config)
        model_inst_datum["featureset_id"] = self.featureset_id
        model_inst_datum["file_id_servable_mdl_binary"] = self.file_id_servable_mdl_binary
        model_inst_datum["file_id_xformer_ref"] = self.file_id_xformer_ref
        model_inst_datum["file_id_training_logs"] = self.file_id_training_logs
        model_inst_datum["metrics"] = self.metrics
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(model_inst_datum)

    @staticmethod
    def exists(existing_model_inst_id):
        tbl = kml_get_tbl_handle(tbl_name=ModelInstance.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999,
            options={"expression": f"entity_id={existing_model_inst_id}"}
        )

        if len(loadedObject)>0:
            return True
        return False

    @staticmethod
    def clone(existing_model_inst_id):
        # throws upwards all errors that the load function throws
        # we inherit checks and get them for free
        to_copy_from=ModelInstance.load(existing_model_inst_id)

        newxformerfile=None
        if to_copy_from.file_id_xformer_ref:
            newxformerfile=KMLFile.clone(to_copy_from.file_id_xformer_ref).file_id
        return ModelInstance.create(
            model_inst_name=to_copy_from.base_entity.name,
            model_inst_desc=to_copy_from.base_entity.desc,
            model_inst_config=to_copy_from.model_inst_config,
            resource_config=to_copy_from.resource_config,
            featureset_id=to_copy_from.featureset_id,
            file_id_xformer_ref=newxformerfile)
            # For the remainder of inputs, we default to empty by not specifying

    @staticmethod
    def validate_config(model_instance_config):
        #TODO: Complete this update method
        return True

    # TODO: Rename this to something more descriptive
    def get_file(self):
        if not self.file_id_servable_mdl_binary:
            return None
        return KMLFile.load(self.file_id_servable_mdl_binary).get_file()

    def persist_trained_servable_model_binary(self, infile):
        kf1 = KMLFile.create(
            the_file_path=infile,
            file_name=os.path.basename(infile),
            file_desc="model file for model instance id " + str (self.get_entity_id())
        )

        self.file_id_servable_mdl_binary=kf1.file_id
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)

        sink_table.update_records(
            expressions=["entity_id=" +str(self.base_entity.get_entity_id())],
            new_values_maps={"file_id_servable_mdl_binary": str(self.file_id_servable_mdl_binary)}
        )
        return True

    def persist_training_logs(self, infile):
        kf1 = KMLFile.create(
            the_file_path=infile,
            file_name=os.path.basename(infile),
            file_desc="training logs for model instance id " + str (self.get_entity_id()))

        self.file_id_training_logs=kf1.file_id
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)

        sink_table.update_records(
            expressions=["entity_id=" +str(self.base_entity.get_entity_id())],
            new_values_maps={"file_id_training_logs": str(self.file_id_training_logs)}
            )
        return True

    def set_metrics(self, inmetrics):
        self.metrics = inmetrics
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)

        update_time = datetime.datetime.utcnow().isoformat(' ')[:-3]
        sink_table.update_records(
            expressions=["entity_id=" + str(self.base_entity.get_entity_id())],
            new_values_maps={"metrics": str(json.dumps(self.metrics))}
        )
        self.base_entity.renew_last_updated_timestamp()
        return True

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['metrics'] = self.metrics
        serialization['file_id_servable_mdl_binary'] = self.file_id_servable_mdl_binary
        serialization['file_id_training_logs'] = self.file_id_training_logs
        serialization['file_id_xformer_ref'] = self.file_id_xformer_ref
        serialization['model_inst_config'] = self.model_inst_config
        serialization['resource_config'] = self.resource_config
        #serialization['featureset_id'] = self.featureset_id
        #if self.featureset_id:
        #    serialization['featureset'] = FeatureSet.load(self.featureset_id).serialize()
        return serialization

    def __str__(self):
        return "Model Instance %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True))


class ModelDeployment:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_model_depl
    table_name = 'sys_kml_model_depl'
    table_schema = TBL_SCHEMA_sys_kml_model_depl

    def __init__(self, base_entity, base_model_inst_id, featureset_id, deployment_type, model_dep_config, resource_config):
        self.base_entity = base_entity
        self.base_model_inst_id = base_model_inst_id
        self.featureset_id = featureset_id
        self.deployment_type = deployment_type
        self.model_dep_config = model_dep_config
        self.resource_config = resource_config
        self.base_model_inst=ModelInstance.load(base_model_inst_id)

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    def persist_insert(self):
        model_depl_datum = collections.OrderedDict()
        model_depl_datum["entity_id"] = self.base_entity.get_entity_id()
        model_depl_datum["base_model_inst_id"] = self.base_model_inst_id
        model_depl_datum["featureset_id"] = self.featureset_id
        model_depl_datum["deployment_type"] = self.deployment_type
        model_depl_datum["model_dep_config"] = json.dumps(self.model_dep_config)
        model_depl_datum["resource_config"] = json.dumps(self.resource_config)
        sink_table = kml_get_tbl_handle(self.table_name, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(model_depl_datum)

    @staticmethod
    def create(deployment_name, deployment_desc, base_model_inst_id, featureset_id, deployment_type, model_dep_config, resource_config):

        if not ModelInstance.exists(base_model_inst_id):
            raise ValueError('Unknown base model with id ' + str(base_model_inst_id))

        if not ModelDeployment.validate_config(model_dep_config):
            raise ValueError('Model Deployment Configuration given is invalid ' + str(model_dep_config))


        base_entity = Entity.create(entity_name=deployment_name,
            entity_desc=deployment_desc,
            entity_type="DEPLOYMENT")
        md = ModelDeployment(base_entity, base_model_inst_id, featureset_id, deployment_type, model_dep_config, resource_config)
        md.persist_insert()
        return md

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")

    @staticmethod
    def load(existing_deployment_id):
        tbl = kml_get_tbl_handle(ModelDeployment.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression" : f"entity_id={existing_deployment_id}" } )

        if len(loadedObject)>0:
            for lo in loadedObject:
                return ModelDeployment(base_entity=Entity.load(lo['entity_id']),
                    base_model_inst_id=lo['base_model_inst_id'],
                    featureset_id=lo['featureset_id'],
                    deployment_type=lo['deployment_type'],
                    model_dep_config=json.loads(lo['model_dep_config']),
                    resource_config=json.loads(lo['resource_config']))
        else:
            raise LookupError('No Model Deployment found with id ' + str(existing_deployment_id))

    @staticmethod
    def exists(existing_deployment_id):
        tbl = kml_get_tbl_handle(ModelDeployment.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(options = { "expression": f"entity_id='{existing_deployment_id}" })
        if len(loadedObject)>0:
            return True
        return False

    @staticmethod
    def validate_config(model_dep_config):
        #TODO: Complete this update method
        #tovalidate=json.loads(model_dep_config)
        return True

    def get_base_model_inst(self):
        return self.base_model_inst

    def set_config(self, in_config):
        sink_table = kml_get_tbl_handle(self.table_name, schema=SCHEMA_SYS)
        sink_table.update_records(
            expressions=["entity_id=" +str(self.base_entity.get_entity_id())],
            new_values_maps={"model_dep_config": json.dumps(in_config)}
            )

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization["deployment_type"] = self.deployment_type
        serialization["featureset_id"] = self.featureset_id
        serialization['featureset'] = FeatureSet.load(self.featureset_id).serialize()
        serialization["model_dep_config"] = self.model_dep_config
        serialization["base_model_inst_id"] = self.base_model_inst_id
        serialization["base_model_inst"] = self.base_model_inst.serialize()
        serialization["resource_config"] = self.resource_config
        return serialization

    def __str__(self):
        return "Model Deployment %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True))


class Catalog:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_catalog
    table_name='sys_kml_catalog'
    table_schema=TBL_SCHEMA_sys_kml_catalog

    def __init__(self, tag_text, user_id):

        if not tag_text:
            raise ValueError('Catalog must have non-blank text entry, received: ' + str(tag_text))

        #TODO: Implement user id check
        if False:
            raise ValueError('user_id not found: ' + str(user_id))

        self.tag_id=Catalog.get_next_tag_id()
        self.tag_text = tag_text
        self.user_id = user_id

        tag_datum = collections.OrderedDict()
        tag_datum["tag_id"] = self.tag_id
        tag_datum["tag_text"] = self.tag_text
        tag_datum["user_id"] = self.user_id
        sink_table = kml_get_tbl_handle(self.table_name, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(tag_datum)

    @staticmethod
    def get_next_tag_id():
        # TODO: Need to wrap this into a mutex and have a global-next-available-id
        minmax=db.aggregate_min_max(table_name=Catalog.table_name, column_name="tag_id")
        if minmax["status_info"]["status"]=='ERROR':
            return 1
        return minmax["max"]+1

    def attach_entity(self, some_entity_id):
        table_name_link='sys_kml_catalog_links'
        table_schema_link=TBL_SCHEMA_sys_kml_catalog_links
        link_datum = collections.OrderedDict()
        link_datum["tag_id"] = self.tag_id
        link_datum["entity_id"] = some_entity_id
        sink_table = kml_get_tbl_handle(table_name_link, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(link_datum)

    def detach_entity(self, some_entity_id):
        tbl = kml_get_tbl_handle(self.table_schema, schema = SCHEMA_SYS)
        #tbl.delete_records(expressions=["tag_id="+str(self.tag_id), "entity_id="+str(some_entity_id)])
        #TODO: Expressions seem to be doing an OR -- need to do an AND so we dont delete across tags!
        tbl.delete_records(expressions=["entity_id="+str(some_entity_id)])
        pass

    def get_attached(self):
        attached_entities=[]
        tbl = kml_get_tbl_handle(self.table_schema, schema=SCHEMA_SYS)
        topts = {"collection_name": "kml", "ttl": 1}
        attachments = tbl.filter(expression="tag_id="+str(self.tag_id), options=topts)
        for a in attachments:
            attached_entities.append(a[1])
        return sorted(attached_entities)

    def load(tag_id):
        tbl = kml_get_tbl_handle(self.table_schema, schema=SCHEMA_SYS)
        topts = {"collection_name": "kml", "ttl": 1}
        found = tbl.filter(expression="tag_id="+str(self.tag_id), options=topts)
        if found.size()==0:
            raise ValueError('Catalog entry not found with tag_id: ' + str(tag_id))

    def set_tag_text(self):
        #TODO: Complete this update method
        pass

    def get_tag_id(self):
        return self.tag_id

    def serialize(self):
        serialization = {}
        serialization['tag_id'] = self.tag_id
        serialization['tag_text'] = self.tag_text
        serialization['user_id'] = self.user_id
        serialization['attached_entities'] = self.get_attached()
        return serialization

    def __str__(self):
        return "Tag %s %s" % (str(self.tag_id), json.dumps(self.serialize(), indent=4, sort_keys=True))


class KMLFile:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_file
    table_name='sys_kml_file'
    table_schema=TBL_SCHEMA_sys_kml_file

    def __init__(self, file_id, the_file, md5, create_dt, file_name, file_desc):
        self.file_id=file_id
        self.the_file = the_file
        self.md5 = md5
        self.create_dt = create_dt
        self.file_name=file_name
        self.file_desc = file_desc

    def persist_insert(self):
        file_datum = collections.OrderedDict()
        file_datum["file_id"] = self.file_id
        file_datum["the_file"] = self.the_file
        file_datum["md5"] = self.md5
        file_datum["create_dt"] = self.create_dt
        file_datum["file_name"] = self.file_name
        file_datum["file_desc"] = self.file_desc
        sink_table = kml_get_tbl_handle(tbl_name = self.table_name, schema=SCHEMA_SYS)

        insert_status=sink_table.insert_records(file_datum)
        return insert_status

    @staticmethod
    def universe():
        toret = {}

        tbl = kml_get_tbl_handle(tbl_name = KMLFile.table_name, schema=SCHEMA_SYS)
        #TODO: How do we do an unbounded pull, rather than using an unnecessary filter?
        loadedObject = tbl.get_records(limit=-9999)
        for lo in loadedObject:
            #TODO: Will this load the *entire file* if so, DO IT DIFFERENTLY!!!
            anitem = {}
            anitem['file_id'] = lo['file_id']
            anitem['file_name'] = lo['file_name']
            anitem['file_desc'] = lo['file_desc']
            #toret[lo['file_size']] = lo['file_size']
            anitem['md5'] = lo['md5']
            anitem['create_dt'] = lo['create_dt']
            toret[str(lo['file_id'])] = anitem
        return toret

    @staticmethod
    def create(the_file_path, file_name, file_desc):

        if not the_file_path:
            raise ValueError('Must have path to file')

        try:
            st = os.stat(the_file_path)
        except os.error:
            raise ValueError('File received was not found')

        if os.path.isdir(the_file_path):
            raise ValueError('File received is a directory, not a file')

        file_in_bytes=None
        with open(the_file_path, 'rb') as f:
            file_in_bytes=f.read()

        import hashlib
        m = hashlib.md5()
        m.update(file_in_bytes)

        import datetime
        newfile = KMLFile(file_id=KMLFile.get_next_file_id(),
            the_file=file_in_bytes,
            md5=m.hexdigest(),
            create_dt=datetime.datetime.utcnow().isoformat(' ')[:-3],
            file_name=file_name,
            file_desc=file_desc
            )
        newfile.persist_insert()
        return newfile

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")

    @staticmethod
    def get_next_file_id():
        from kml.api.kmldbops import get_aggregate_min_max as kml_get_aggregate_min_max
        # TODO: Need to wrap this into a mutex and have a global-next-available-id
        minmax=kml_get_aggregate_min_max(tbl_name=KMLFile.table_name, col_name="file_id", schema=SCHEMA_SYS)
        if minmax["status_info"]["status"]=='ERROR':
            return 1
        return int(minmax["max"])+1

    @staticmethod
    def clone(existing_file_id):
        # throws upwards all errors that the load function throws
        # we inherit checks and get them for free
        to_copy_from=KMLFile.load(existing_file_id)

        import datetime
        newfile = KMLFile(
            file_id=KMLFile.get_next_file_id(),
            the_file=to_copy_from.get_file(),
            md5=to_copy_from.md5,
            create_dt=datetime.datetime.utcnow().isoformat(' ')[:-3],
            file_name=to_copy_from.file_name,
            file_desc=to_copy_from.file_desc
            )
        newfile.persist_insert()
        return newfile

    @staticmethod
    def load(existing_file_id):
        tbl = kml_get_tbl_handle(tbl_name = KMLFile.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"file_id={existing_file_id}" })
        if len(loadedObject)>0:
            for lo in loadedObject:
                return KMLFile(file_id=lo['file_id'],
                    the_file=lo['the_file'],
                    md5=lo['md5'],
                    create_dt=lo['create_dt'],
                    file_name=lo['file_name'],
                    file_desc=lo['file_desc'])
        else:
            raise ValueError('No KML File found with id ' + str(existing_file_id))

    @staticmethod
    def exists(existing_file_id):
        tbl = kml_get_tbl_handle(tbl_name = KMLFile.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"file_id={existing_file_id}" })
        if len(loadedObject)>0:
            return True
        return False

    def get_file_id(self):
        return self.file_id

    def get_file(self):
        return self.the_file

    def get_file_contents(self):
        import tempfile
        mytmpfile = tempfile.NamedTemporaryFile()
        self.save_file(mytmpfile.name)
        tempcontents = None
        with open(mytmpfile.name, "rt") as content_file:
            tempcontents = content_file.read()
        return tempcontents

    def save_file(self, as_name):
        f = open(as_name, 'wb')
        f.write(self.the_file)
        f.close()

    def serialize(self):
        serialization = {}
        serialization['file_id'] = self.file_id
        serialization['file_name'] = self.file_name
        serialization['file_desc'] = self.file_desc
        serialization['file_size'] = sys.getsizeof(self.the_file)
        serialization['md5'] = self.md5
        serialization['create_dt'] = self.create_dt
        #TODO: Figure out how to serialize datetime for JSON
        #serialization['create_dt'] = json.dumps(datetime.datetime(self.create_dt))
        return serialization

    def __str__(self):
        return "KML File %s %s" % (str(self.file_id), json.dumps(self.serialize(), indent=4, sort_keys=True))


class FeatureSet:
    """FeatureSet consists of both relational and inline transformations."""

    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_featureset

    ENTITY_TYPE_ID = 'FEATURESET'
    table_name = 'sys_kml_featureset'
    table_schema = TBL_SCHEMA_sys_kml_featureset

    base_entity = None
    config_inline = None
    config_relational = None

    def __init__(self, base_entity, config_inline=None, config_relational=None):

        self.base_entity = base_entity
        self.config_inline = config_inline
        self.config_relational = config_relational

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def create(featureset_name, featureset_desc,
               config_inline=None, config_relational=None):

        base_entity = Entity.create(
            entity_name=featureset_name,
            entity_desc=featureset_desc,
            entity_type=FeatureSet.ENTITY_TYPE_ID
        )

        featset_inst = FeatureSet(
            base_entity=base_entity,
            config_inline=config_inline,
            config_relational=config_relational
        )

        featset_inst.persist_insert()
        return featset_inst

    def persist_insert(self):
        featureset_datum = collections.OrderedDict()
        featureset_datum["entity_id"] = self.base_entity.get_entity_id()
        featureset_datum["config_inline"] = json.dumps(self.config_inline)
        featureset_datum["config_relational"] = json.dumps(self.config_relational)

        #sink_table = gpudb.GPUdbTable( name = self.table_name, db = db)
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(featureset_datum)
        return insert_status

    @staticmethod
    def load(existing_featureset_id):
        tbl = kml_get_tbl_handle(tbl_name = FeatureSet.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999,
            options={"expression": f"entity_id={existing_featureset_id}"}
        )
        if len(loadedObject) > 0:
            for lo in loadedObject:
                loadedfs = FeatureSet(
                    base_entity=Entity.load(lo['entity_id']),
                    config_inline=json.loads(lo['config_inline']),
                    config_relational=json.loads(lo['config_relational'])
                )
                return loadedfs
        else:
            raise LookupError('No FeatureSet found with id ' + str(existing_featureset_id))

    @staticmethod
    def clone(existing_featureset_id):
        to_copy_from=FeatureSet.load(existing_featureset_id)
        return FeatureSet.create(
            featureset_name=to_copy_from.base_entity.name,
            featureset_desc=to_copy_from.base_entity.desc,
            config_inline=to_copy_from.config_inline,
            config_relational=to_copy_from.config_relational
        )

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    @staticmethod
    def exists(existing_featureset_id):
        tbl = kml_get_tbl_handle(tbl_name=FeatureSet.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999,
            options={"expression": f"entity_id={existing_featureset_id}"}
        )
        if len(loadedObject) > 0:
            return True
        return False

    @staticmethod
    def validate_config(feature_cofig):
        #TODO: Add more details validations
        return (True,"Valid Config")

    def grab(self, infile_loc):
        config_inline = self.config_inline
        config_relational = self.config_relational # Not currently exported
        feat_mem = {}
        if config_inline and "feat_mem" in config_inline:
            feat_mem = config_inline["feat_mem"]

        feat_mem["metadata"] = {}
        feat_mem["metadata"]["gen_date"] = datetime.datetime.utcnow().isoformat(' ')[:-3]
        feat_mem["metadata"]["src_entity_id"] = self.base_entity.get_entity_id()
        feat_mem["metadata"]["src_entity_name"] = self.base_entity.name
        feat_mem["metadata"]["src_entity_desc"] = self.base_entity.desc
        with open(infile_loc, 'w') as jsonfile:
            jsonfile.write(json.dumps(feat_mem, indent=4))

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['config_inline'] = self.config_inline
        serialization['config_relational'] = self.config_relational
        if 'dataset_id_fitting' in self.config_inline:
            serialization['config_inline']['dataset_def'] = Dataset.load(
                self.config_inline['dataset_id_fitting']
            ).serialize()

        return serialization

    def __str__(self):
        return "FeatureSet %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True)
        )


class Dataset:
    ENTITY_TYPE_ID="DATASET"
    table_name='sys_kml_dataset'
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_dataset
    table_schema=TBL_SCHEMA_sys_kml_dataset
    base_entity = None
    source_config = None

    def __init__(self, base_entity, source_config):
        self.base_entity = base_entity
        self.source_config = source_config

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def create(dataset_name, dataset_desc, source_config):
        # source_config is a dictionary of
        #    'tbl' (the source table)
        #    'cols' (the columns of interest (for a get_records_by_column))
        #    'filterex' (filter expressions, if any)
        logger.debug("Dataset creation: Validating configuration")
        (validation_status, validation_message) = Dataset.validate_config(source_config)
        if not validation_status:
            raise ValueError(f'Dataset Configuration given is invalid {validation_message}')
        if not source_config:
            raise ValueError('Dataset source_config is empty')
        if 'tbl' not in source_config:
            raise ValueError('Dataset source_config must contain a source table')
        if 'cols' not in source_config:
            raise ValueError('Dataset source_config must contain columns of interest')

        # TODO: Get this working, not sure why the len doesnt work
        #if len(source_config['cols']):
        #    raise ValueError('Dataset source_config must contain 1+ columns of interest')

        # TODO: More meaningful validation of the SQL, perhaps a DB endpoint for validation of SQL?
        # TODO: More meaningful sanity check, e.g., row-count? (or something not too expensive)

        if "filterex" not in source_config:
            source_config["filterex"]=""
            logger.warning("Note filterex not found in source_config, using no filter")
            logger.warning(source_config)

        logger.debug("Dataset creation: Creating Entity")
        base_entity = Entity.create(entity_name=dataset_name,
            entity_desc=dataset_desc,
            entity_type=Dataset.ENTITY_TYPE_ID)
        logger.debug("Dataset creation: Creating Dataset Object")
        dset_inst = Dataset(base_entity=base_entity,
            source_config=json.dumps(source_config))

        logger.debug("Dataset creation: Transitioning to CREATED state")
        dset_inst.get_base_entity().transition_to_created()

        logger.debug("Dataset creation: Persisting")
        dset_inst.persist_insert()

        logger.debug(f"Dataset creation: Done -- Entity ID {base_entity.get_entity_id()}")
        return dset_inst

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    @staticmethod
    def validate_config(source_config):
        import kml.api.utils

        if 'tbl' not in source_config:
            return (False, 'Configuration missing source table (tbl)')
        if 'cols' not in source_config:
            return (False, 'Configuration missing source columns (cols)')

        if len(source_config['cols']) == 0:
            return (False, f"Configuration missing columns references")

        # Does the underlying table exist?
        #  if not kml.api.utils.table_exists(source_config['tbl']):
        if not kml_has_table(source_config['tbl'], schema=None):
            return (False, f"Configuration references non-existent table {source_config['tbl']}")

        tbl_h_superset = kml_get_tbl_handle(source_config['tbl'])
        tbl_cols_superset = [c.name.strip() for c in tbl_h_superset.get_table_type().columns]
        superset_contains_subset = \
            all(elem in tbl_cols_superset for elem in source_config['cols'])

        if not superset_contains_subset:
            return (superset_contains_subset, "Configuration references non-existent column")

        return (superset_contains_subset,"Valid Config")

    def persist_insert(self):
        dset_datum = collections.OrderedDict()
        dset_datum["entity_id"] = self.base_entity.get_entity_id()
        dset_datum["source_config"] = json.dumps(self.source_config)
        sink_table = kml_get_tbl_handle (tbl_name=self.table_name, schema=SCHEMA_SYS)
        insert_status=sink_table.insert_records(dset_datum)
        return insert_status

    @staticmethod
    def load(existing_dataset_id):
        tbl = kml_get_tbl_handle(tbl_name=Dataset.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"entity_id={existing_dataset_id}" })
        if len(loadedObject)>0:
            for lo in loadedObject:
                return Dataset(base_entity=Entity.load(lo['entity_id']),
                    source_config=json.loads(lo['source_config']))
        else:
            raise LookupError('No existing Dataset found with id ' + str(existing_dataset_id))

    @staticmethod
    def clone(existing_dataset_id):
        to_copy_from=Dataset.load(existing_dataset_id)
        return Dataset.create(
            dataset_name=to_copy_from.base_entity.name,
            dataset_desc=to_copy_from.base_entity.desc,
            source_config=json.loads(to_copy_from.source_config))

    @staticmethod
    def exists(existing_dataset_id):
        tbl = kml_get_tbl_handle(tbl_name=Dataset.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(options = { "expression": f"entity_id={existing_dataset_id}" })
        if len(loadedObject)>0:
            return True
        return False

    def preview(self, preview_size=7):
        toret = []

        dsconfig = json.loads(self.source_config)
        tbl = kml_get_tbl_handle(tbl_name=dsconfig["tbl"], schema=None)
        loadedObject = tbl.get_records(limit=-9999, options = {"expression":dsconfig["filterex"]})

        for lo in loadedObject:
            mynext = {}
            for c in dsconfig["cols"]:
                mynext[c]=lo[c]
            toret.append(mynext)
            if len(toret)>=preview_size:
                return toret
        return toret

    @staticmethod
    def get_records_filtered_generator(db, tbl, cols, fltr, batch_size):

        current_offset=0
        total_records_retrieved=0
        tbl = kml_get_tbl_handle(tbl_name=tbl, schema=None)
        while True:
            tbl_filtered_recordbatch = tbl.get_records(offset=current_offset, limit=batch_size, options = { "expression":fltr })
            for t in tbl_filtered_recordbatch:
                arec = {}
                for c in cols:
                    arec[c]=t[c]
                yield arec
            total_records_retrieved += len(tbl_filtered_recordbatch)
            if len(tbl_filtered_recordbatch) < batch_size:
                break
            current_offset = current_offset + batch_size
        return

    # TODO: This will be replaced with an initialized page-able generator function
    #   obviously we dont want to pass the entire dataset in memory
    def grab(self, infile_loc):
        dsconfig = json.loads(self.source_config)
        rec_genner = Dataset.get_records_filtered_generator(db=db,
                                             tbl=dsconfig["tbl"],
                                             cols= dsconfig["cols"],
                                             fltr=dsconfig["filterex"],
                                             batch_size=500)
        HEADER_ROW=[c for c in dsconfig["cols"]]

        import csv
        with open(infile_loc, 'w') as csvfile:
            spamwriter = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            spamwriter.writerow(HEADER_ROW)
            for lo in rec_genner:
                NEXT_ROW = [lo[c] for c in dsconfig["cols"] ]
                spamwriter.writerow(NEXT_ROW)

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['source_config'] = json.loads(self.source_config)
        return serialization

    def __str__(self):
        return "Dataset %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True))


class Audit:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_audit
    MODEL_TYPES=['NEURAL_NET','REGRESSION','BAYESIAN','BLACKBOX']

    table_name = 'sys_kml_audit'
    table_schema = TBL_SCHEMA_sys_kml_audit

    def __init__(self, base_entity, audit_config):
        self.base_entity = base_entity
        self.audit_config = audit_config

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity

    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def create(audit_name, audit_desc, audit_config):
        """Create audit instance."""

        if not audit_config:
            raise ValueError('Parameter audit_config is empty')

        (validation_status, validation_message) = Audit.validate_config(audit_config)

        if not validation_status:
            raise ValueError(
                f"Audit configuration given is invalid: {validation_message}"
            )

        base_entity = Entity.create(entity_name=audit_name,
                                    entity_desc=audit_desc,
                                    entity_type='AUDIT')

        audit_inst = Audit(base_entity=base_entity,
                           audit_config=audit_config)

        audit_inst.get_base_entity().transition_to_created()
        audit_inst.persist_insert()

        return audit_inst

    @staticmethod
    def validate_config(audit_config):
        """Check creation parameters."""

        if audit_config.get('dates'):
            if type(audit_config['dates']) != dict:
                return (False, 'audit_config: dates is not a dictionary')

        if audit_config.get('fields'):
            if type(audit_config['fields']) != list:
                return (False, 'audit_config: fields is not a list')

        return (True, 'Valid Config')

    def set_config(self, updated_config):
        """Update search parameters based on filter search terms."""

        sink_table = kml_get_tbl_handle(self.table_schema, schema=SCHEMA_SYS)
        sink_table.update_records(
            expressions=[f"entity_id={self.base_entity.get_entity_id()}"],
            new_values_maps={'audit_config': json.dumps(updated_config)}
        )

    def persist_insert(self):
        audit_datum = collections.OrderedDict()
        audit_datum['entity_id'] = self.base_entity.get_entity_id()
        audit_datum['audit_config'] = json.dumps(self.audit_config)
        sink_table = kml_get_tbl_handle(self.table_name, schema=SCHEMA_SYS)
        insert_status = sink_table.insert_records(audit_datum)

        return insert_status

    @staticmethod
    def load(existing_audit_id):
        """Load an existing audit entry."""

        tbl = kml_get_tbl_handle(Audit.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999,
            options={'expression': f"entity_id={existing_audit_id}"}
        )
        if len(loadedObject) > 0:
            for lo in loadedObject:
                return Audit(base_entity=Entity.load(lo['entity_id']),
                             audit_config=json.loads(lo['audit_config']))
        else:
            raise LookupError(f"No Audit found with id {existing_audit_id}")

    @staticmethod
    def clone(existing_audit_id):
        """Create a new entity from an existing."""

        to_copy_from=Audit.load(existing_audit_id)

        return Audit.create(audit_name=to_copy_from.base_entity.name,
                            audit_desc=to_copy_from.base_entity.desc,
                            audit_config=json.loads(to_copy_from.audit_config))

    @staticmethod
    def exists(existing_audit_id):
        tbl = kml_get_tbl_handle(Audit.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999,
            options={"expression": f"entity_id={existing_audit_id}"}
        )
        if len(loadedObject) > 0:
            return True

        return False

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['audit_config'] = self.audit_config

        return serialization

    def __str__(self):
        return 'Ingest %s %s' % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True)
        )


class KCredential:
    from kml.api.systable_spec import TBL_SCHEMA_sys_kml_credential

    ENTITY_TYPE_ID = 'CREDENTIAL'
    table_name='sys_kml_credential'
    table_schema=TBL_SCHEMA_sys_kml_credential

    CREDENTIAL_TYPES=['KAFKA','AWS_S3','POSTGRES','KINETICA_DB','CUSTOM','SPARK', 'DOCKER']

    ###########################################################
    # TODO: This should be moved into a base class
    def publish(self):
        self.base_entity.publish()
    def tombstone(self):
        self.base_entity.tombstone()
    def get_base_entity(self):
        return self.base_entity
    def change_state(self, to_state):
        if self.is_state_change_acceptable(to_state):
            self.base_entity.change_state(to_state)
        else:
            logger.warning(f"State change to {to_state} failed")
    def get_entity_id(self):
        return self.base_entity.get_entity_id()
    ###########################################################

    @staticmethod
    def encrypt_secrets(cred_type, in_dict):
        import copy
        from kml.api.utils import kencrypt, kdecrypt
        config_dict = copy.deepcopy(in_dict)
        if cred_type == "KAFKA":
            config_dict["key_password"] = kencrypt(config_dict["key_password"])
        elif cred_type in ["POSTGRES", "KINETICA_DB", "DOCKER"]:
            config_dict["password"] = kencrypt(config_dict["password"])
        elif cred_type == "AWS_S3":
            config_dict["aws_secret_access_key"] = kencrypt(config_dict["aws_secret_access_key"])
        elif cred_type == "CUSTOM":
            for k in config_dict:
                config_dict[k] = kencrypt(config_dict[k])
        return config_dict

    @staticmethod
    def decrypt_secrets(cred_type, in_dict):
        from kml.api.utils import kencrypt, kdecrypt
        import copy
        config_dict = copy.deepcopy(in_dict)
        if cred_type == "KAFKA":
            config_dict["key_password"] = kdecrypt(config_dict["key_password"])
        elif cred_type in ["POSTGRES", "KINETICA_DB", "DOCKER"]:
            config_dict["password"] = kdecrypt(config_dict["password"])
        elif cred_type == "AWS_S3":
            config_dict["aws_secret_access_key"] = kdecrypt(config_dict["aws_secret_access_key"])
        elif cred_type == "CUSTOM":
            for k in config_dict:
                config_dict[k] = kdecrypt(config_dict[k])
        return config_dict

    def __init__(self, base_entity, credential_type, credential_config):
        self.base_entity = base_entity
        self.credential_type = credential_type
        self.credential_config = credential_config

    def get_decrypted_credential_config(self):
        return KCredential.decrypt_secrets(self.credential_type, self.credential_config)

    @staticmethod
    def create(credential_name, credential_desc, credential_type, credential_config):

        isvalid, validate_msg = KCredential.validate_config(credential_type, credential_config)
        if not isvalid:
            raise ValueError(f'Credential Configuration given is invalid {validate_msg}')

        if credential_type not in KCredential.CREDENTIAL_TYPES:
            raise ValueError(f"Unrecognized credential type {credential_type}")

        base_entity = Entity.create(entity_name=credential_name,
            entity_desc=credential_desc,
            entity_type="CREDENTIAL")

        base_entity.change_state(to_state="CREATED")
        base_entity.publish()

        cred_inst = KCredential(base_entity=base_entity,
            credential_type=credential_type,
            credential_config=KCredential.encrypt_secrets(credential_type, credential_config))
        cred_inst.persist_insert()

        return cred_inst

    def is_state_change_acceptable(self, to_state):
        #TODO: Validate if this change is acceptable
        return True

    def persist_insert(self):
        insertable = collections.OrderedDict()
        insertable["entity_id"] = self.base_entity.get_entity_id()
        insertable["credential_type"] = json.dumps(self.credential_type)
        insertable["credential_config"] = json.dumps(self.credential_config)
        sink_table = kml_get_tbl_handle(tbl_name=self.table_name, schema=SCHEMA_SYS)
        _=sink_table.insert_records(insertable)
        # TODO: Check, then provide feedback on whether insert was successful

    @staticmethod
    def load(existing_credential_id):
        tbl = kml_get_tbl_handle(tbl_name=KCredential.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"entity_id={existing_credential_id}" })
        if len(loadedObject)==0:
            raise LookupError('No Credential found with id ' + str(existing_credential_id))
        if len(loadedObject)>1:
            raise LookupError(f"Multiple credentials found with id {existing_credential_id} -- system error")
        return KCredential(base_entity=Entity.load(loadedObject[0]['entity_id']),
            credential_type=json.loads(loadedObject[0]['credential_type']),
            credential_config=json.loads(loadedObject[0]['credential_config']))

    @staticmethod
    def clone(existing_credential_id):
        to_copy_from=KCredential.load(existing_credential_id)
        return KCredential.create(
            credential_name=to_copy_from.base_entity.name,
            credential_desc=to_copy_from.base_entity.desc,
            credential_type=to_copy_from.credential_type,
            credential_config=to_copy_from.credential_config)

    @staticmethod
    def exists(existing_credential_id):
        tbl = kml_get_tbl_handle(tbl_name=KCredential.table_name, schema=SCHEMA_SYS)
        loadedObject = tbl.get_records(limit=-9999, options = { "expression": f"entity_id={existing_credential_id}" })
        if len(loadedObject)>0:
            return True
        return False

    @staticmethod
    def validate_config(credential_type, credential_config):
        if credential_type not in KCredential.CREDENTIAL_TYPES:
            raise ValueError(f"Unrecognized credential type {credential_type}")
        if credential_type == "KAFKA":
            if "connection_str" not in credential_config: return False, "KAFKA credential missing connection_str"
            if "topics" not in credential_config: return False, "KAFKA credential missing topics"
            if "auth_method" in credential_config:
                if credential_config["auth_method"]=="KEYSTORE":
                    if "keystore" not in credential_config: return False, "KAFKA Keystore credential missing keystore"
                    if "key_password" not in credential_config: return False, "KAFKA Keystore credential missing key_password"
                if credential_config["auth_method"]=="SASL_SSL__PLAIN":
                    if "username" not in credential_config: return False, "KAFKA SASL credential missing username"
                    if "key_password" not in credential_config: return False, "KAFKA SASL credential missing key_password"
            return True, "No validation issues"
        elif credential_type == "AWS_S3":
            if "s3_bucket" not in credential_config: return False, "AWS S3 credential missing s3_bucket"
            if "s3_path" not in credential_config: return False, "AWS S3 credential missing s3_path"
            if "aws_access_key_id" not in credential_config: return False, "AWS S3 credential missing aws_access_key_id"
            if "aws_secret_access_key" not in credential_config: return False, "AWS S3 credential missing aws_secret_access_key"
            return True, "No validation issues"
        elif credential_type == "POSTGRES":
            if "connection_str" not in credential_config: return False, "Postgres credential missing connection_str"
            if "user" not in credential_config: return False, "Postgres credential missing user"
            if "password" not in credential_config: return False, "Postgres credential missing password"
            if "pg_database" not in credential_config: return False, "Postgres credential missing pg_database"
            return True, "No validation issues"
        elif credential_type == "KINETICA_DB":
            if "connection_str" not in credential_config: return False, "KineticaDB credential missing connection_str"
            if "user" not in credential_config: return False, "KineticaDB credential missing user"
            if "password" not in credential_config: return False, "KineticaDB credential missing password"
            return True, "No validation issues"
        elif credential_type == "DOCKER":
            if "registry_uri" not in credential_config: return False, "Docker credential missing registry_uri"
            if "user" not in credential_config: return False, "Docker credential missing user"
            if "password" not in credential_config: return False, "Docker credential missing password"
            if "email" not in credential_config: return False, "Docker credential missing email"
            return True, "No validation issues"
        elif credential_type == "CUSTOM":
            if len(credential_config) == 0: return False, "Empty custom credential - no KV pairs found"
            return True, "No validation issues"

    def get_handle(self):
        from kml.api.naming_svc import get_k8s_secret_handle
        return get_k8s_secret_handle(in_credential_id=self.get_entity_id())

    def serialize(self):
        serialization = {}
        serialization['entity'] = self.base_entity.serialize()
        serialization['credential_type'] = self.credential_type
        serialization['credential_config'] = self.credential_config
        if self.credential_type == "DOCKER":
            serialization['k8s_secrets_handle'] = self.get_handle()
        return serialization

    def __str__(self):
        return "Credential %s %s" % (
            self.base_entity.get_entity_id(),
            json.dumps(self.serialize(), indent=4, sort_keys=True))
