"""Create or upgrade KML system tables at startup. Also run system checks."""

import sys
import traceback
import datetime
import logging

from kubernetes.client import V1Namespace
import gpudb

from kml.api.systable_spec import *
from kml.api.controller_utils import ensure_no_entity_id_conflict
from kml.api.kml_config import (
    get_db_conn,
    get_k8_conn,
    get_namespace,
    get_systables_ver,
    get_required_db_ver,
    get_public_connections,
    get_sys_schemas,
    running_in_k8s
)
from kml.api.kmldbops import (
    create_table as kml_create_table,
    has_table as kml_has_table,
    get_tbl_handle as kml_get_tbl_handle,
    create_schema_or_collection as kml_create_schema
)

module = 'initialize'
logger = logging.getLogger(module)
logger.setLevel(logging.INFO)
console_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter(
    '%(asctime)s %(levelname)s  (KML) %(name)s - %(message)s'
)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


SCHEMA_SYS, SCHEMA_AUDIT, SCHEMA_SYS_ARCH, SCHEMA_AUDIT_ARCH = get_sys_schemas()

# Config
db = get_db_conn()
k8_apps_v1_api, k8_core_v1_api, k8_custom_objects_api, k8_batch_v1_api, _ = get_k8_conn()
namespace = get_namespace()

# Version table details
required_ver = get_systables_ver()
ver_tbl_name = 'sys_kml_systables_version'
ver_tbl_scm = [
    ['version', 'string', 'char16'],
    ['installed', 'string', 'datetime'],
    ['updated', 'string', 'datetime']
]

# System table details (name, schema) tuple
generation_queue = [
    ('sys_kml_entity', TBL_SCHEMA_sys_kml_entity),
    ('sys_kml_file', TBL_SCHEMA_sys_kml_file),
    ('sys_kml_project', TBL_SCHEMA_sys_kml_project),
    ('sys_kml_model_inst', TBL_SCHEMA_sys_kml_model_inst),
    ('sys_kml_model_depl', TBL_SCHEMA_sys_kml_model_depl),
    ('sys_kml_dataset', TBL_SCHEMA_sys_kml_dataset),
    ('sys_kml_featureset', TBL_SCHEMA_sys_kml_featureset),
    ('sys_kml_ingest', TBL_SCHEMA_sys_kml_ingest),
    ('sys_kml_catalog', TBL_SCHEMA_sys_kml_catalog),
    ('sys_kml_catalog_links', TBL_SCHEMA_sys_kml_catalog_links),
    ('sys_kml_audit', TBL_SCHEMA_sys_kml_audit),
    ('sys_kml_credential', TBL_SCHEMA_sys_kml_credential),
    ('sys_kml_user_data', TBL_SCHEMA_user_data),
    ('sys_kml_events', TBL_SCHEMA_events),
    ('sys_kml_events_lifecycle', TBL_SCHEMA_events_lifecycle),
    ('sys_kml_events_metrics', TBL_SCHEMA_events_metrics),
    ('sys_kml_slipway', TBL_SCHEMA_slipway)
]

quick_views = {
    "sys_kml_inventory_quick_ingest": {
        "table_names": ['sys_kml_entity as ke', 'sys_kml_ingest as ki'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'ki.ingest_mode',
            'ki.ingest_config'
        ],
        "expressions": [
            'ke.entity_id = ki.entity_id',
            'entity_type = "INGEST"'
        ]
    },
    "sys_kml_inventory_quick_dataset": {
        "table_names": ['sys_kml_entity as ke', 'sys_kml_dataset as kd'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'kd.source_config'
        ],
        "expressions": [
            'ke.entity_id = kd.entity_id',
            'entity_type = "DATASET"'
        ]
    },
    "sys_kml_inventory_quick_featureset": {
        "table_names": ['sys_kml_entity as ke', 'sys_kml_featureset as kf'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'kf.config_relational',
            'kf.config_inline'
        ],
        "expressions": [
            'ke.entity_id = kf.entity_id',
            'ke.entity_type = "FEATURESET"'
        ]
    },
    "sys_kml_inventory_quick_model_instance": {
        "table_names": ['sys_kml_entity as ke', 'sys_kml_model_inst as kmi'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'kmi.model_inst_config'
        ],
        "expressions": [
            'ke.entity_id = kmi.entity_id',
            'ke.entity_type = "MODEL_INSTANCE"'
        ]
    },
    "sys_kml_inventory_quick_model_deployment": {
        "table_names": ['sys_kml_entity as ke', 'sys_kml_model_depl as kmd'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'kmd.deployment_type',
            'kmd.model_dep_config',
            'kmd.base_model_inst_id'
        ],
        "expressions": [
            'ke.entity_id = kmd.entity_id',
            'ke.entity_type = "DEPLOYMENT"'
        ]
    },
    "sys_kml_inventory_quick_credential": {
        "table_names": ['sys_kml_entity as ke', 'sys_kml_credential as kc'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'kc.credential_type',
            'kc.credential_config'
        ],
        "expressions": [
            'ke.entity_id = kc.entity_id',
            'entity_type = "CREDENTIAL"'
        ]
    },
    "sys_kml_inventory_inflight": {
        "table_names": ['sys_kml_entity as ke'],
        "column_names": [
            'ke.entity_id',
            'ke.entity_type',
            'ke.entity_name',
            'ke.entity_desc',
            'ke.entity_state',
            'ke.doc_state',
            'ke.visibility',
            'ke.create_dt',
            'ke.update_dt'
        ],
        "expressions": [
            "entity_state in ('PENDING','CREATED','STARTING','RUNNING','COMPLETED','FAILED')",
            "entity_type in ('MODEL_INSTANCE', 'DEPLOYMENT', 'INGEST')"
        ]
    }
}




def check_system_tables_exist():
    all_tables_exist = True
    for tbl_to_check, _ in generation_queue:
        try:
            table_exists_response = kml_has_table(tbl_to_check,
                                                  schema=SCHEMA_SYS)
            if 'table_exists' not in table_exists_response:
                all_tables_exist = False
                logger.info(f"Missing system table {tbl_to_check}")
            elif not table_exists_response['table_exists']:
                all_tables_exist = False
                logger.info(f"Missing system table {tbl_to_check}")

        except gpudb.gpudb.GPUdbException as ge:
            all_tables_exist = False
            logger.error(f"Error finding table {ge}")

    return all_tables_exist


def upgrade_systables_add_tables():
    """Add any missing tables."""

    missing = False

    for (tbl_name, tbl_scm) in generation_queue:
        if not kml_has_table(tbl_name, schema=SCHEMA_SYS):
            missing = True
            logger.info(f"Adding missing table {tbl_name}")
            _ = kml_create_table(tbl_type_list=tbl_scm,
                                 tbl_name=tbl_name,
                                 schema=SCHEMA_SYS)
    if not missing:
        logger.info('No missing system tables detected')


def upgrade_systables_add_columns():
    """Add new columns to an existing table."""

    missing = False

    for (tbl_name, tbl_scm) in generation_queue:

        tbl_hdl = kml_get_tbl_handle(tbl_name, schema=SCHEMA_SYS)
        existing_cols = [x.name for x in tbl_hdl.get_table_type().columns]
        mods = []

        # Build a list of alteration mappings
        for col in tbl_scm:
            if col[0] not in existing_cols:
                logger.info(f"Missing column {col[0]} detected in table {tbl_name}")
                mod = {}
                name = str(col[0])
                base_type = str(col[1])
                if 'nullable' in col:
                    dtype = ','.join([base_type, 'null'])
                else:
                    dtype = base_type
                non_prop = [name, base_type, 'nullable']
                props = [x for x in col if x not in non_prop]
                mod.update(column_name=name,
                           action='add_column',
                           column_type=dtype,
                           column_properties=','.join(props))
                mods.append(mod)

        # Add any missing table columns
        if mods:
            missing = True
            tbl_hdl.alter_table_columns(column_alterations=mods, options={})

            # Check that all columns have been added
            required_cols = [x[0] for x in tbl_scm]
            existing_cols = [x.name for x in tbl_hdl.get_table_type().columns]
            if required_cols.sort() == existing_cols.sort():
                logger.info(f"Columns successfully added to {tbl_name}")
            else:
                logger.info(f"Column addition on table {tbl_name} failed")
                msg = f"Column addition on table {tbl_name} failed, unrecoverable"
                raise Exception(msg)

    if not missing:
        logger.info('No missing columns in existing system tables detected')


def upgrade_systables_delete_columns():
    """Delete old columns from an existing table."""

    present = False

    for (tbl_name, tbl_scm) in generation_queue:

        tbl_hdl = kml_get_tbl_handle(tbl_name, schema=SCHEMA_SYS)
        required_cols = [x[0] for x in tbl_scm]
        existing_cols = [x.name for x in tbl_hdl.get_table_type().columns]

        mods = []

        # Build a list of alteration mappings
        for col in existing_cols:
            if col not in required_cols:
                logger.info(f"Extra column {col} detected in table {tbl_name}")
                mod = {}
                name = str(col)
                mod.update(column_name=name,
                           action='delete_column')
                mods.append(mod)

        # Add any missing table columns
        if mods:
            present = True
            tbl_hdl.alter_table_columns(column_alterations=mods, options={})

            # Check that all columns have been deleted
            required_cols = [x[0] for x in tbl_scm]
            existing_cols = [x.name for x in tbl_hdl.get_table_type().columns]
            if required_cols.sort() == existing_cols.sort():
                logger.info(f"Columns successfully deleted from {tbl_name}")
            else:
                logger.error(f"Column deletion on table {tbl_name} failed")
                msg = f"Column deletion on table {tbl_name} failed, unrecoverable"
                raise Exception(msg)

    if not present:
        logger.info('No extra columns in existing system tables detected')


def upgrade_systables_mod_schema():
    """Refresh or update table schema for every column."""

    logger.info('Refreshing type schema and column properties for system tables')

    for (tbl_name, tbl_scm) in generation_queue:

        tbl_hdl = kml_get_tbl_handle(tbl_name, schema=SCHEMA_SYS)
        mods = []

        # Build a list of alteration mappings
        for col in tbl_scm:
            mod = {}
            name = str(col[0])
            base_type = str(col[1])
            if 'nullable' in col:
                dtype = ','.join([base_type, 'null'])
            else:
                dtype = base_type
            non_prop = [name, base_type, 'nullable']
            props = [x for x in col if x not in non_prop]
            mod.update(column_name=name,
                       action='change_column',
                       column_type=dtype,
                       column_properties=','.join(props))
            mods.append(mod)

        # Modify all columns for existing schema
        tbl_hdl.alter_table_columns(column_alterations=mods, options={})


def build_system_tables():
    """On clean install, create all tables and collections."""

    logger.info('Creating Workbench system schemas/collections')
    kml_create_schema(scname=SCHEMA_SYS)
    #  kml_create_schema(scname=SCHEMA_AUDIT)
    #  kml_create_schema(scname=SCHEMA_SYS_ARCH)
    #  kml_create_schema(scname=SCHEMA_AUDIT_ARCH)

    logger.info('Creating Workbench system tables')
    for (tbl_name, tbl_scm) in generation_queue:
        kml_create_table(tbl_type_list=tbl_scm,
                         tbl_name=tbl_name,
                         schema=SCHEMA_SYS)


def update_version_table():
    """Create or update system tables version string."""

    dt = datetime.datetime.utcnow().isoformat(' ', 'seconds')

    if kml_has_table(ver_tbl_name, schema=SCHEMA_SYS):
        ver_tbl = kml_get_tbl_handle(ver_tbl_name, schema=SCHEMA_SYS)
        resp = ver_tbl.get_records_by_column(column_names='version')

        if resp:
            existing_ver = resp['version'][0]
            ver_tbl.update_records(
                expressions=[f'version="{existing_ver}"'],
                new_values_maps={'version': required_ver,
                                 'updated': dt}
            )
        else:
            ver_tbl.insert_records({'version': required_ver,
                                    'installed': dt,
                                    'updated': dt})

        logger.info(f"System tables versioning updated to {required_ver}")

    else:
        ver_tbl = kml_create_table(tbl_type_list=ver_tbl_scm,
                                   tbl_name=ver_tbl_name,
                                   schema=SCHEMA_SYS)

        ver_tbl.insert_records([{'version': required_ver,
                                 'installed': dt,
                                 'updated': dt}])

        logger.info(f"System tables versioning {required_ver} added")


def detect_systables_ver():
    """Get version of the sys tables and determine if upgrade required."""

    if kml_has_table(ver_tbl_name, schema=SCHEMA_SYS):
        # Version table exists, determine if upgrade required

        ver_tbl = kml_get_tbl_handle(ver_tbl_name, schema=SCHEMA_SYS)
        resp = ver_tbl.get_records_by_column(column_names='version')
        existing_ver = resp['version'][0] if resp else None

        upgrade = False if existing_ver == required_ver else True
        logger.info(f"System tables version detected: {existing_ver}")

        if not upgrade:
            for (tbl_name, tbl_schm) in generation_queue:
                if not kml_has_table(tbl_name, schema=SCHEMA_SYS):
                    upgrade = True

        logger.info(f"System tables upgrade required: {upgrade}")

    else:
        upgrade = True
        logger.info(f"System tables version not found")
        logger.info(f"System tables upgrade required: {upgrade}")

    clean = True
    for (tbl_name, tbl_schm) in generation_queue:
        if kml_has_table(tbl_name, schema=SCHEMA_SYS):
            clean = False

    return clean, upgrade


def check_db_version():
    """Confirm that gpudb meets minimum requirements."""

    db_sys_props = db.show_system_properties()
    gpudb_ver = (
        db_sys_props['property_map']['version.gpudb_core_version'].split('.')[:4]
    )

    req_ver = get_required_db_ver()
    if len(req_ver) != 4:
        msg = f"""
            Required DB version string "{'.'.join(req_ver)}" incorrectly
            formatted. Required version in system.ini must be a four
            digit version string. eg. 7.1.1.1
        """
        logger.info(msg)
        raise Exception(f"Required database version must be 4 digits")

    compatible = False
    for i in range(4):
        if int(gpudb_ver[i]) > int(req_ver[i]):
            compatible = True
            break

    if compatible:
        logger.info(
            f"Database version {'.'.join(gpudb_ver)} >= "
            f"{'.'.join(req_ver)}")
    else:
        msg = f"""
            Detected database version {'.'.join(gpudb_ver)}, Active Analytics
            Workbench requires at least version {'.'.join(req_ver)}
        """
        logger.info(msg)
        raise Exception(f"Database version >= {'.'.join(req_ver)} required")


def check_db_public_address():
    """Confirm that gpudb head node public IP address is set correctly."""

    error_msg = """
    Active Analytics Workbench requires the host0_public_address parameter
    to be set correctly. Please set the parameter and restart the DB.
    Then start Active Analytics Workbench again.
    """

    db_sys_props = db.show_system_properties()
    gpudb_pub_addr = (
        db_sys_props['property_map'].get('conf.host0_public_address').strip()
    )

    _, pub_url, _ = get_public_connections()
    aaw_conf_pub_addr = pub_url.split('//')[1].split(':')[0]

    if gpudb_pub_addr == aaw_conf_pub_addr:
        logger.info('Required GPUdb public host is correctly set')
    else:
        logger.error(error_msg)
        msg = (
            f"Required GPUdb parameter host0_public_address is incrorrectly "
            f"configured, setup halted.\n"
            f"GPUdb conf = {gpudb_pub_addr}, AAW conf = {aaw_conf_pub_addr}"
        )
        raise Exception(msg)


def check_text_search_enabled():
    """Confirm that gpudb text search is enabled."""

    error_msg = """
    Active Analytics Workbench requires DB feature [enable_text_search] enabled.
    Please set to DB config [enable_text_search] to TRUE and restart DB.
    Then start Active Analytics Workbench again.
    """

    db_sys_props = db.show_system_properties()
    search_enabled = db_sys_props['property_map'].get('conf.enable_text_search')

    if search_enabled.strip().upper() == 'TRUE':
        logger.info('Required GPUdb text search is enabled')
    else:
        logger.error(error_msg)
        msg = 'Required GPUdb text search must be enabled, setup halted'
        raise Exception(msg)


def create_inventory_view(view_name, table_names, column_names, expressions):
    """Create or refresh materialized views and joins."""
    from kmldbops import create_mv_jointbl as kml_create_mv_jointbl
    from kmldbops import clear_table as kml_clear_table

    kml_clear_table(tbl_name=view_name, schema=SCHEMA_SYS)
    kml_create_mv_jointbl(
        view_name=view_name,
        table_names=table_names,
        column_names=column_names,
        expressions=expressions,
        schema=SCHEMA_SYS
    )


def add_system_views(db=db):
    """Build materialized views from system base tables."""

    logger.info('Initializing system table materialized views')

    n_views = len(quick_views)
    for view_count, view in enumerate(quick_views):
        logger.info(f"Creating view ({view_count + 1}/{n_views}): {view} ")
        create_inventory_view(
            view_name=view,
            table_names=quick_views[view]['table_names'],
            column_names=quick_views[view]['column_names'],
            expressions=quick_views[view]['expressions']
        )


def add_namespaces():
    """Add required namespaces to the K8 cluster if they are not there already."""

    logger.info(f"Creating Kubernetes namespace {namespace}")

    error_msg = """
    Kubernetes is required infrastructure for Active Analytics Workbench. The cluster
    appears to be unreachable. Check the operation of K8s on this machine.
    """

    ns_list = []

    try:
        ns_resp = k8_core_v1_api.list_namespace()
        for rec in ns_resp.items:
            ns_list.append(rec.metadata.name)
    except Exception as e:
        logger.error(error_msg)
        msg = 'Kubernetes connection exception (list_ns), setup halted'
        logger.error(msg)
        raise Exception(msg)

    if namespace not in ns_list:
        try:
            ns_options = V1Namespace(
                api_version='v1',
                kind='Namespace',
                metadata={'name': namespace,
                          'labels': {'name': 'kml'}}
            )

            _ = k8_core_v1_api.create_namespace(body=ns_options)

        except Exception as e:
            logger.error(error_msg)
            msg = 'Kubernetes connection exception (create_ns), setup halted'
            logger.error(msg)
            raise Exception(msg)


def check_system_ids():
    """Ensure there is no conflict between system and audit table IDs."""

    if not ensure_no_entity_id_conflict():
        logger.error("ERROR: ID conflict between system record and audit tables.")
        logger.error("  Use the following script to archive the audit tables then restart:")
        logger.error("    python -m kml.api.manage_systables --audit-archive")
        logger.error("  Help for this script is available with:")
        logger.error("    python -m kml.api.manage_systables --help")
        sys.exit(1)

    logger.info('No internal entity ID conflicts detected')


def check_collections_exist():
    from kml.api.kmldbops import check_if_exists_schema_or_collection as kml_check_if_exists_schema_or_collection
    """Ensure the required audit and archive collections/schemas exist."""

    req_schemas_updated = False
    req_schemas = [SCHEMA_AUDIT, SCHEMA_SYS_ARCH, SCHEMA_AUDIT_ARCH]

    logger.info('Checking collection schemas')
    for schema in req_schemas:
        if not kml_check_if_exists_schema_or_collection(schema):
            logger.info(f"Creating collection schema {schema}")
            req_schemas_updated = True
            kml_create_schema(scname=schema)

    if not req_schemas_updated:
        logger.info('No missing schema collections detected')


def check_env_status():
    """Log environment status of KML."""

    if running_in_k8s():
        logger.info('Running in cloud native (K8s) environment')
    else:
        logger.info('Running in on-prem or cloud VM environment')

    logger.info(f"Using gpudb client version {db.get_version_info()}")


if __name__ == '__main__':
    """Control block."""

    try:
        check_env_status()
        check_text_search_enabled()
        #  check_db_public_address()
        check_db_version()
        add_namespaces()
        check_collections_exist()

        clean, upgrade = detect_systables_ver()

        if clean:
            logger.info('Performing clean install of system tables')
            build_system_tables()
            update_version_table()

        elif upgrade:
            logger.info('Updating system tables')
            upgrade_systables_add_tables()
            upgrade_systables_add_columns()
            upgrade_systables_delete_columns()
            upgrade_systables_mod_schema()
            update_version_table()

        else:
            logger.info('No system table updates required')

        add_system_views()
        check_system_ids()

    except Exception as e:
        logger.error('Fatal setup and configuration error')
        logger.error(e)
        logger.error(traceback.format_exc())
        sys.exit(1)
