import os
import random
import json

import requests
import gpudb

from kml.api.kml_config import (
    get_api_base,
    get_ingress_controller_details
)
from kml.api.aux import attach_log

KML_API_BASE = get_api_base()

logger = attach_log(module=__name__, log_name='kml', debug=True)

def string_is_not_blank(s):
    return bool(s and s.strip())

def is_json(input_str):
  try:
    json_object = json.loads(input_str)
  except ValueError as e:
    return False
  return True

def compute_resources_available(replicas, allocation_per_replica):
    req_res = replicas * allocation_per_replica

    avail_res = get_k8_resources()['total_available']['GPU']

    logger.info(f'Requested {req_res} GPUs ({replicas} replica * {allocation_per_replica} allocations) with {avail_res} available')
    if avail_res >= req_res:
        return True
    else:
        return False

def get_k8_resources():
    """Determine available Kubernetes resources."""

    from kml.api.kml_config import get_k8_conn, get_namespace
    namespace = get_namespace()
    _, k8_core_v1_api, _, _, _ = get_k8_conn()

    cluster = {}
    cluster["nodes"]=[]
    cluster["gpu_subscribers"]=[]
    cluster["total_capacity"]={}
    cluster["total_allocatable"]={}
    cluster["total_available"]={}
    cluster["total_capacity"]["CPU"]=0
    cluster["total_capacity"]["GPU"]=0
    cluster["total_capacity"]["TPU"]=0
    cluster["total_allocatable"]["CPU"]=0
    cluster["total_allocatable"]["GPU"]=0
    cluster["total_allocatable"]["TPU"]=0
    cluster["total_available"]["GPU"]=-1
    cluster["total_available"]["TPU"]=0

    try:

        for anode in k8_core_v1_api.list_node().items:

            thisnode={}
            thisnode["allocatable"]=anode.status.allocatable
            thisnode["capacity"]=anode.status.capacity
            cluster["nodes"].append(thisnode)

            if "cpu" in anode.status.capacity:
                cluster["total_capacity"]["CPU"] = cluster["total_capacity"]["CPU"] + int(anode.status.capacity["cpu"])
            if "nvidia.com/gpu" in anode.status.capacity:
                cluster["total_capacity"]["GPU"] = cluster["total_capacity"]["GPU"] + int(anode.status.capacity["nvidia.com/gpu"])

            # Convert available CPU resouces to fractional K8s milliCPU standard if not already
            if "cpu" in anode.status.allocatable:
                if anode.status.allocatable["cpu"].isdigit():
                    alloc_cpu = int(anode.status.allocatable["cpu"]) * 1000
                else:
                    alloc_cpu = int(anode.status.allocatable["cpu"].strip('m'))
                cluster["total_allocatable"]["CPU"] = cluster["total_allocatable"]["CPU"] + alloc_cpu

            if "nvidia.com/gpu" in anode.status.allocatable:
                cluster["total_allocatable"]["GPU"] = cluster["total_allocatable"]["GPU"] + int(anode.status.allocatable["nvidia.com/gpu"])

            # Run thru all pods and add up usage
            gpu_usage = 0
            pod_list = k8_core_v1_api.list_pod_for_all_namespaces()
            for pod in pod_list.items:
                for c in pod.spec.containers:
                    if c.resources and c.resources.limits:

                        if "nvidia.com/gpu" in c.resources.limits:

                            gpu_usage = gpu_usage + int(c.resources.limits["nvidia.com/gpu"])

                            asubscriber={}

                            asubscriber["pod_name"] = pod.metadata.name
                            asubscriber["namespace"] = pod.metadata.namespace
                            asubscriber["gpu_usage"] = int(c.resources.limits["nvidia.com/gpu"])
                            cluster["gpu_subscribers"].append(asubscriber)

            cluster["total_available"]["GPU"] = cluster["total_allocatable"]["GPU"] - gpu_usage

            return cluster

    except Exception as e:
        raise ValueError(e)


def kencrypt(text_raw):
    import subprocess
    from subprocess import PIPE
    cmd = ["/opt/gpudb/core/bin/gpudb_encrypt.sh", f"{text_raw}"]
    syscall = subprocess.run(cmd, stdout=PIPE, stderr=PIPE)
    return syscall.stdout.decode().strip()

def kdecrypt(text_encrypted):
    import subprocess
    from subprocess import PIPE
    cmd = ["/opt/gpudb/core/bin/gpudb_decrypt.sh", f"{text_encrypted}"]
    syscall = subprocess.run(cmd, stdout=PIPE, stderr=PIPE)
    return syscall.stdout.decode().strip()

def is_integer(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def helper_get_secrets_name_list():
    from kml.api.kml_config import get_k8_conn, get_namespace
    namespace = get_namespace()
    _, k8_core_v1_api, _, _, _ = get_k8_conn()
    api_response = k8_core_v1_api.list_namespaced_secret(namespace) #, field_selector="metadata.name")
    secrets_list = [i.metadata.name for i in api_response.items]
    return secrets_list


def get_public_k8s_ip():
    """Return the routable IP or DNS address of K8s."""

    # Assumption: If the API is running in a Docker container then we are
    # using the Developer Edition and can further assume that the attached
    # Kubernetes infrastructure is a single node KinD cluster (also a Docker
    # container). Therefore we don't require an external IP bind because
    # the standard http ports are open on this K8s container.

    #  if os.path.isfile('/tmp/cgroup'):
    #      with open('/tmp/cgroup') as f:
    #          if '/docker/' in f.read():
    #              logger.info('Detected Kinetica Dev-Edition')
    #              return 'localhost'
    if os.path.isfile('/proc/self/cgroup'):
        with open('/proc/self/cgroup') as f:
            if '/docker/' in f.read():
                logger.info('Detected Kinetica Dev-Edition')
                return 'localhost'

    # Otherwise, proceed with the expectation that ingress-nginx is running
    # with MetalLB on-prem or a cloud provided load balancer.
    from kml.api.kml_config import get_k8_conn
    _, k8_core_v1_api, _, _, _ = get_k8_conn()

    ing_svc_name, ing_svc_ns, _, _ = get_ingress_controller_details()

    service_desc = k8_core_v1_api.read_namespaced_service(
        name=ing_svc_name,
        namespace=ing_svc_ns
    )
    if service_desc:
        if service_desc.status.load_balancer.ingress[0].ip:
            ingress_ip = service_desc.status.load_balancer.ingress[0].ip
        else:
            ingress_ip = service_desc.status.load_balancer.ingress[0].hostname
        logger.info(f"Using ingress controller IP {ingress_ip}")
    else:
        raise ValueError(f"Ingress controller service {ing_svc_name} in "
                         f"namespace {ing_svc_ns} is not available")

    return ingress_ip


#  def get_events_single_pod(deploy_name):
#      """Track container launch and set terminal state."""

#      logger.debug(f"** Pod watcher (deploy_name)")

#      k8_apps_v1_api, k8_core_v1_api, _, _, _ = get_k8_conn()

#      rollout_tm_out= 120
#      init_tm_out= 120
#      init_c = 0
#      rollout_c = 0
#      po_found = False

#      while not po_found and init_c < init_tm_out:
#          time.sleep(1)
#          pods = k8_core_v1_api.list_namespaced_pod(namespace)
#          for pod in pods.items:
#              if pod.metadata.name.startswith(deploy_name):
#                  po_name = pod.metadata.name
#                  po_found = True
#                  logger.debug(f"Pod {po_name} initializing")
#          init_c += 1

#      if not po_found:
#          return False, 'timeout initializing'

#      while rollout_c < rollout_tm_out:
#          time.sleep(1)
#          po_desc = k8_core_v1_api.read_namespaced_pod(
#              name=po_name,
#              namespace=namespace
#          )

#          try:
#              if po_desc.status.container_statuses[0].state.running:
#                  logger.debug(f"Pod {po_name} running")
#                  return True, 'running'
#              elif po_desc.status.container_statuses[0].state.waiting.reason == 'ErrImagePull':
#                  logger.debug(f"Pod {po_name} rollout failed: ErrImagePull")
#                  msg = str(po_desc.status.container_statuses[0].state.waiting.message)
#                  return False, msg
#              elif po_desc.status.container_statuses[0].state.waiting.reason == 'ImagePullBackOff':
#                  logger.debug(f"Pod {po_name} rollout failed: ImagePullBackOff")
#                  msg = str(po_desc.status.container_statuses[0].state.waiting.message)
#                  return False, msg
#              elif po_desc.status.container_statuses[0].state.waiting.reason == 'CrashLoopBackOff':
#                  logger.debug(f"Pod {po_name} rollout failed: CrashLoopBackOff")
#                  msg = str(po_desc.status.container_statuses[0].state.waiting.message)
#                  return False, msg

#          except Exception as _:
#              logger.debug(f"Pod {po_name} state undetectable")
#              pass

#          rollout_c += 1

