import os
import sys
import time
import json
import requests
import random
import logging
from logging.handlers import RotatingFileHandler
from pprint import pprint as pp

import gpudb

from kml.api.aux import attach_log
from kml.api.kml_config import get_public_connections
from kml.api.kml_config import get_api_base, get_db_credentials
from kml.api.validators import (validate_db_isaccessible
    ,validate_api_base
    ,validate_rest_api_isaccessible
    ,validate_k8s_isaccessible
    ,validate_docker_daemon
    ,validate_image_prepull
    ,validate_image_prepull_nonexistent
    ,validate_sys_containers_available
    ,validate_test_containers_available
    ,validate_db_latency
    ,validate_zmq_protocol
    ,validate_db_isaccessible_public_conxn
    ,validate_zmq_isaccessible_private_conxn
    ,validate_zmq_isaccessible_public_conxn)


logger = attach_log(module='kmlvalidate', log_name='kml', debug=True)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)

def validation_wrapper(desc, meth, params, suggestions=None):

    import importlib
    print("...")
    logger.info(f"Validating: {desc}")
    mod_name, func_name = meth.rsplit('.',1)
    mod = importlib.import_module(mod_name)
    func = getattr(mod, func_name)
    try:
        result = func(*params)
        if result:
            logger.info(f"Validating: {desc} - Success")
        else:
            logger.error(f"Validating: {desc} - Failure: Unknown reason")
            if suggestions:
                print(suggestions)
    except Exception as e:
        result = False
        reasons = str(e)
        logger.error(f"Validating: {desc} - Failure: {reasons}")
        if suggestions:
            print(suggestions)

def validation_suite():

    try:
        # --------------------------------------------------------------------
        # Step 1: Is the kml.ini file even filled out correctly?

        from kml.api.kml_config import get_public_connections
        public_api_base, _, _ = get_public_connections()
        validation_wrapper(desc="Public API Base URI",
                           meth="kml.api.validators.validate_api_base",
                           params=[public_api_base])

        _, _, public_tm_conn = get_public_connections()
        validation_wrapper(desc="ZMQ connection protocol",
                           meth="kml.api.validators.validate_zmq_protocol",
                           params=[public_tm_conn])

        from kml.api.kml_config import get_tm_conn_str
        tm_conn_str = get_tm_conn_str()
        validation_wrapper(desc="ZMQ connection protocol",
                           meth="kml.api.validators.validate_zmq_protocol",
                           params=[tm_conn_str])

        # --------------------------------------------------------------------
        # Step 2: Are the configured settings on the kml.ini reachable?
        validation_wrapper(desc="Accessible k8s",
                           meth="kml.api.validators.validate_k8s_isaccessible",
                           params=[])

        private_api_base = get_api_base()
        validation_wrapper(desc="Accessible REST API",
                           meth="kml.api.validators.validate_rest_api_isaccessible",
                           params=[private_api_base])

        from kml.api.kml_config import get_public_connections
        public_api_base, _, _ = get_public_connections()
        validation_wrapper(desc="Accessible DB Connection public",
                           meth="kml.api.validators.validate_rest_api_isaccessible",
                           params=[public_api_base])

        validation_wrapper(desc="Accessible DB Connection private",
                           meth="kml.api.validators.validate_db_isaccessible_public_conxn",
                           params=[])


    except Exception as e:
        logger.error(e)

if __name__ == '__main__':
    logger.info("KML System Validator...starting")
    validation_suite()
    logger.info("KML System Validator...completed")
