import os
import time
import json
import requests
import random
from pprint import pprint as pp

import gpudb

from kml.api.kml_config import get_api_base, get_db_credentials

KML_API_BASE = get_api_base()
db_user, db_pass = get_db_credentials()

def get_request_cred():
    db_user, db_pass = get_db_credentials()
    if db_user and db_pass:
        return (db_user, db_pass)
    return None

requests_cred = get_request_cred()

#RTO = helper_get_test_configs()["timeout_requests"]
#KTO = helper_get_test_configs()["timeout_k8s"]
#CTO = helper_get_test_configs()["timeout_container_pulls"]

RTO, KTO, CTO = 30,30,30

def validate_api_base(base_uri):
    from urllib.parse import urlparse
    assert base_uri, "Empty/blank API BASE configuration"

    try:
        result = urlparse(base_uri)
        assert result.scheme, f"Invalid API BASE scheme {result.scheme}"
        assert result.netloc, f"Invalid API BASE netloc {result.netloc}"
    except:
        return (False, f"Invalid API BASE {base_uri}")

    assert base_uri.startswith("http"), "Expecting http/https API BASE"
    return True

def validate_db_isaccessible():
    from kml.api.kml_config import get_db_conn
    db = get_db_conn()
    db_info = db.show_system_properties()
    assert "property_map" in db_info
    assert "version.gpudb_core_version" in db_info["property_map"]
    assert "status_info" in db_info
    assert "status" in db_info["status_info"]
    assert db_info["status_info"]["status"] == "OK"
    return True

def validate_rest_api_isaccessible(rest_api_uri):
    r = requests.get(f'{rest_api_uri}/ping', timeout=RTO, auth=requests_cred)
    assert r.status_code == 200, r.text
    retz = r.json()
    assert "success" in retz
    assert retz["success"]
    return True

def validate_k8s_isaccessible():
    import kubernetes as k8
    from kml.api.kml_config import get_k8_conn

    try:
        k8.config.load_kube_config()
        k8_core_v1_api = k8.client.CoreV1Api()
        for anode in k8_core_v1_api.list_node().items:
            assert "cpu" in anode.status.capacity, "No CPUs avaialble on k8s cluster"
            assert int(anode.status.capacity["cpu"]) > 0, "Insufficient k8s compute capacity"
    except:
        assert False, f"Problems with k8s connection"

    return True

def validate_docker_daemon():
    import docker
    from docker import APIClient
    from kml.api.kml_config import get_docker_daemon_uri
    DOCKER_DAEMON_URI = get_docker_daemon_uri()
    cli = APIClient(base_url=DOCKER_DAEMON_URI)
    assert cli.ping()
    return True

def helper_prepull_container(container_ref):
    from kml.api.kml_config import get_k8_conn, get_namespace
    k8_apps_v1_api, _, _, _, _ = get_k8_conn()
    namespace = get_namespace()

    # positive case
    r = requests.post(f'{KML_API_BASE}/admin/cluster/prepull',
                    json={"container_ref": container_ref},
                    timeout=CTO,
                    auth=requests_cred)
    assert r.status_code == 200, r.content
    retz = r.json()
    assert "success" in retz, retz
    assert retz["success"], retz
    assert "name" in retz["response"], retz

    _ = k8_apps_v1_api.delete_namespaced_daemon_set(name=retz["response"]["name"], namespace=namespace)

    return retz["success"]

def validate_image_prepull():

    from kml.api.kml_config import get_k8_conn, get_namespace
    k8_apps_v1_api, _, _, _, _ = get_k8_conn()
    namespace = get_namespace()

    # positive case
    r = requests.post(f'{KML_API_BASE}/admin/cluster/prepull',
                    json={"container_ref": "busybox"},
                    timeout=CTO,
                    auth=requests_cred)
    assert r.status_code == 200, r.text
    retz = r.json()
    assert "success" in retz
    assert retz["success"]
    assert "name" in retz["response"]

    _ = k8_apps_v1_api.delete_namespaced_daemon_set(name=retz["response"]["name"], namespace=namespace)
    return True

def validate_image_prepull_nonexistent():
    from kml.api.kml_config import get_k8_conn, get_namespace
    k8_apps_v1_api, _, _, _, _ = get_k8_conn()
    namespace = get_namespace()

    # negative case
    r = requests.post(f'{KML_API_BASE}/admin/cluster/prepull',
                    json={"container_ref": "busybox-187"},
                    timeout=RTO,
                    auth=requests_cred)
    assert r.status_code == 400, r.text
    retz = r.json()
    assert "success" in retz
    assert not retz["success"]
    assert "name" in retz["response"]

    _ = k8_apps_v1_api.delete_namespaced_daemon_set(name=retz["response"]["name"], namespace=namespace)
    return True

def validate_sys_containers_available():
    from kml.api.kml_config import get_sys_container

    print("\nPulling containers, this may take a while...")
    for sc in ['ctnr_kio',
                'ctnr_kafkaconnector',
                'ctnr_bb_basesdk',
                'ctnr_kml_depmgr',
                'ctnr_kml_rapids_jupyter',
                'ctnr_kml_input_mgr']:
            ctnr_uri = get_sys_container(sc)
            print(f"\tPulling system container {ctnr_uri}")
            assert helper_prepull_container(container_ref=ctnr_uri), f"Error pre-pulling container {ctnr_uri}"
    return True

def validate_test_containers_available():

    container_q = []
    container_q_skipped = []
    with open('container_refs.json') as cref:
        cref_j = json.load(cref)
        for testkey in cref_j:
            test_obj = cref_j[testkey]["test_objective"]
            ctnr_uri = cref_j[testkey]["container"]
            if "private" not in test_obj.lower():
                if ctnr_uri not in container_q:
                    print(f"\tQueueing image for pull {ctnr_uri}")
                    container_q.append(ctnr_uri)
            else:
                if ctnr_uri not in container_q_skipped:
                    print(f"\tSkipping potentially private image {ctnr_uri}")
                    container_q_skipped.append(ctnr_uri)

    print("\nPulling containers, this may take a while...")
    for ctnr_uri in set(container_q):
        print(f"\tPulling image {ctnr_uri}")
        assert helper_prepull_container(container_ref=ctnr_uri)
    return True

def validate_db_latency():
    NUM_RECORDS = 1000

    # how slow is *too* slow?
    # anything over 20sec for the full battery of latency quick tests is unacceptable
    TOO_MUCH_LATENCY_SHORT_CIRCUIT = 30

    r = requests.get(f'{KML_API_BASE}/latency/{NUM_RECORDS}', timeout=RTO, auth=requests_cred)
    assert r.status_code == 200, r.text
    retz = r.json()
    assert "success" in retz
    assert retz["success"]

    assert "response" in retz
    assert "total_db_ops_seconds" in retz["response"]
    assert "processing_time" in retz

    # ABSOLUTE LEVEL OF NON-FUNCTIONALITY
    assert retz["response"]["total_db_ops_seconds"] < TOO_MUCH_LATENCY_SHORT_CIRCUIT
    assert retz["processing_time"] < TOO_MUCH_LATENCY_SHORT_CIRCUIT

    print("Latency Metrics")
    pp(r.json())
    return True

def validate_zmq_protocol(zmq_cnxn_uri):

    from urllib.parse import urlparse
    assert zmq_cnxn_uri, "Empty/blank API BASE configuration"

    try:
        result = urlparse(base_uri)
        assert result.scheme, f"Invalid ZMQ connection string scheme {result.scheme}"
        assert result.netloc, f"Invalid ZMQ connection string netloc {result.netloc}"
    except:
        return (False, f"Invalid ZMQ connection string: {zmq_cnxn_uri}")

    assert zmq_cnxn_uri.startswith("tcp"), "Expecting tcp connection"
    return True



def validate_db_isaccessible_public_conxn():
    from kml.api.kml_config import get_public_connections
    from kml.api.utils import kdecrypt

    _, public_db_conn, _ = get_public_connections()

    db_user, db_pass = get_db_credentials()

    #  print('==== validators.py ====')
    try:
        db = gpudb.GPUdb(
            encoding='BINARY',
            host=public_db_conn,
            primary_host=public_db_conn,
            username=db_user,
            password=db_pass
        )
    except:
        #  print('==== Catch in validators.py ====')
        raise Exception

    db_info = db.show_system_properties()
    assert "property_map" in db_info, "DB missing property map"
    assert "version.gpudb_core_version" in db_info["property_map"], "DB missing version"
    assert "status_info" in db_info, "DB missing status_info"
    assert "status" in db_info["status_info"], f"DB missing status {db_info['status_info']}"
    assert db_info["status_info"]["status"] == "OK", f"DB status not OK {db_info['status_info']}"
    return True

def validate_zmq_isaccessible_private_conxn():
    from kml.api.kml_config import get_tm_conn_str
    import zmq
    import threading
    import requests
    from zmq.eventloop.future import Context

    tm_conn_str = get_tm_conn_str()

    try:
        zmq_context = zmq.eventloop.future.Context()
        mysocket = zmq_context.socket(zmq.SUB)
        mysocket.connect(tm_conn_str)
        topicstring = "ABC"
        topicfilter = bytes(topicstring, 'utf-8').decode()
        mysocket.setsockopt_string(zmq.SUBSCRIBE, topicfilter)
        mysocket.close()
        zmq_context.term()
    except:
        assert False, f"Problems with ZMQ Communication using {tm_conn_str}"
    return True

def validate_zmq_isaccessible_public_conxn():
    from kml.api.kml_config import get_public_connections
    import zmq
    import threading
    import requests
    from zmq.eventloop.future import Context

    _, _, public_tm_conn = get_public_connections()

    try:
        zmq_context = zmq.eventloop.future.Context()
        mysocket = zmq_context.socket(zmq.SUB)
        mysocket.connect(public_tm_conn)
        topicstring = "ABC"
        topicfilter = bytes(topicstring, 'utf-8').decode()
        mysocket.setsockopt_string(zmq.SUBSCRIBE, topicfilter)
        mysocket.close()
        zmq_context.term()
    except:
        assert False, f"Problems with ZMQ Communication using {public_tm_conn}"
    return True
