import os
import yaml

from kml.api.aux import get_k8_secrets_hash
from kml.api.kml_config import (
    get_namespace,
    get_sys_container,
    get_image_pull_policy,
    running_in_k8s
)
from kml.api.aux import attach_log

namespace = get_namespace()
logger = attach_log(module=__name__, log_name='kml', debug=True)
global_pull_policy, user_pull_policy = get_image_pull_policy()


def customize_yaml_byoc_ing(ing_id, byoc_container, dest_tabl, envvars={}):

    from kml.api.naming_svc import get_cluster_name_ingest
    from kml.api.kml_config import get_public_connections
    from kml.api.kml_config import get_db_credentials

    public_api_base, public_db_conn, public_tm_conn = get_public_connections()

    path = os.path.dirname(os.path.realpath(__file__))
    bbyaml = yaml.load(
        open('/'.join([path, 'yaml/tmplt-kml-ing-byoc-cnts.yaml'])),
        Loader=yaml.FullLoader
    )

    depname = get_cluster_name_ingest(in_ingest_id=ing_id)

    bbyaml['spec']['template']['spec']['containers'][0]['name'] = depname
    bbyaml['spec']['template']['spec']['containers'][0]['image'] = byoc_container
    bbyaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = user_pull_policy
    bbyaml['metadata']['name'] = depname
    bbyaml['metadata']['namespace'] = namespace

    db_user, db_pass = get_db_credentials()

    # Standard BYOC env vars
    bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
        {"name": "KML_CONN_STR", "value": public_api_base}
    )
    bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
        {"name": "DB_CONN_STR", "value": public_db_conn}
    )
    bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
        {"name": "DB_USER", "value": "no_cred"}
    )
    bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
        {"name": "DB_PASS", "value": "no_cred"}
    )
    bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
        {"name": "DB_TABL", "value": dest_tabl}
    )

    # Custom BYOC env vars
    for k, v in envvars.items():
        bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': str(v)}
        )

    # Remove default CPU node pool affiity if not on cloud
    if not running_in_k8s():
        bbyaml['spec']['template']['spec'].pop('nodeSelector')

    return bbyaml


def customize_yaml_input_mgr(depl_id, envvars={}, pod_resources=None,
                             use_ingress=False, slug=None):

    from kml.api.naming_svc import (
            get_cluster_name_input_mgr,
            get_cluster_name_service,
            get_cluster_name_ingress,
    )

    ref_container = get_sys_container("ctnr_kml_input_mgr")
    path = os.path.dirname(os.path.realpath(__file__))
    input_mgr_name = get_cluster_name_input_mgr(in_deployment_id=depl_id)
    svc_yaml = {}

    inp_yaml = yaml.load(
        open('/'.join([path, 'yaml/tmplt-kml-input-mgr.yaml'])),
        Loader=yaml.FullLoader
    )
    inp_yaml['spec']['template']['spec']['containers'][0]['name'] = input_mgr_name
    inp_yaml['spec']['template']['spec']['containers'][0]['image'] = ref_container
    inp_yaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = global_pull_policy
    inp_yaml['metadata']['name'] = input_mgr_name
    inp_yaml['metadata']['namespace'] = namespace
    for k, v in envvars.items():
        inp_yaml['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': str(v)}
        )

    # Assign pod affinity on cloud (default to CPU pool)
    if not running_in_k8s():
        inp_yaml['spec']['template']['spec'].pop('nodeSelector')
    else:
        if pod_resources['compute_target'] == 'GPU':
            inp_yaml['spec']['template']['spec']\
                ['nodeSelector']['app.kinetica.com/pool'] = 'aaw-gpu'

    # Default is no ingress definition
    igr_yaml = None

    if use_ingress:
        svc_name = get_cluster_name_service(depl_id)
        igr_name = get_cluster_name_ingress(depl_id)

        # Add ports and labels to deployment
        inp_yaml['spec']['selector']['matchLabels']['run'] = input_mgr_name
        inp_yaml['spec']['template']['metadata']['labels']['run'] = input_mgr_name
        inp_yaml['spec']['template']['spec']['containers'][0]['ports'][0]['containerPort'] = 8080
        inp_yaml['spec']['template']['spec']['containers'][0]['ports'][0]['protocol'] = 'TCP'

        # Define service spec
        svc_yaml = yaml.load(
            open('/'.join([path, 'yaml/tmplt-kml-on-demand-svc.yaml'])),
            Loader=yaml.FullLoader
        )
        svc_yaml['metadata']['name'] = svc_name
        svc_yaml['spec']['selector']['run'] = input_mgr_name
        #  svc_yaml['metadata']['labels']['run'] = input_mgr_name

        # Create ingress resource definition
        if not slug:
            raise ValueError('Ingress controller requires a named path (slug) but it was '
                             'not included in the manifest create payload')
        igr_yaml = yaml.load(
            open('/'.join([path, 'yaml/tmplt-kml-on-demand-ingress.yaml'])),
            Loader=yaml.FullLoader
        )
        igr_yaml['metadata']['name'] = igr_name
        igr_yaml['spec']['rules'][0]['http']['paths'][0]['path'] = f"/{slug}(/|$)(.*)"
        igr_yaml['spec']['rules'][0]['http']['paths'][0]['backend']['serviceName'] = svc_name

    #  logger.debug('==== Input Mgr ====')
    #  logger.debug(yaml.dump(inp_yaml))
    return inp_yaml, svc_yaml, igr_yaml


def customize_yaml_depmgr(depl_id, envvars={}, pod_resources=None):

    from kml.api.naming_svc import get_cluster_name_dep_mgr

    ref_container = get_sys_container("ctnr_kml_depmgr")
    path = os.path.dirname(os.path.realpath(__file__))
    bbyaml = yaml.load(
        open('/'.join([path, 'yaml/tmplt-kml-dep-mgr.yaml'])),
        Loader=yaml.FullLoader
    )

    depname = get_cluster_name_dep_mgr(in_deployment_id=depl_id)

    bbyaml['spec']['template']['spec']\
        ['containers'][0]['name'] = depname
    bbyaml['spec']['template']['spec']\
        ['containers'][0]['image'] = ref_container
    bbyaml['spec']['template']['spec']\
        ['containers'][0]['imagePullPolicy'] = global_pull_policy
    bbyaml['metadata']['name'] = depname
    bbyaml['metadata']['namespace'] = namespace
    for k, v in envvars.items():
        bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': str(v)}
        )

    # Assign pod affinity on cloud (default to CPU pool)
    if not running_in_k8s():
        bbyaml['spec']['template']['spec'].pop('nodeSelector')
    else:
        if pod_resources['compute_target'] == 'GPU':
            bbyaml['spec']['template']['spec']\
                ['nodeSelector']['app.kinetica.com/pool'] = 'aaw-gpu'

    #  logger.debug('==== Deployment Mgr ====')
    #  logger.debug(yaml.dump(bbyaml))
    return bbyaml


def customize_yaml_depl_bbx_worker(bbx_container, depl_id, envvars={},
                                   pod_resources=None, credential_id=None,
                                   secrets_handle=None):

    from kml.api.naming_svc import get_cluster_name_deployment_blackbox

    ref_container = get_sys_container("ctnr_bb_basesdk")

    path = os.path.dirname(os.path.realpath(__file__))
    bbyaml = yaml.load(
        open('/'.join([path, 'yaml/template-bb-worker.yaml'])),
        Loader=yaml.FullLoader
    )

    bbyaml['spec']['template']['spec']['containers'][0]['image'] = ref_container
    bbyaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = user_pull_policy

    depname_bbx = get_cluster_name_deployment_blackbox(in_deployment_id=depl_id)

    repo = bbx_container[:bbx_container.rfind('/')]
    secrets_hash = get_k8_secrets_hash(repo)

    bbyaml['metadata']['name'] = depname_bbx
    bbyaml['metadata']['namespace'] = namespace

    bbyaml['spec']['replicas'] = pod_resources['replicas']
    bbyaml['spec']['template']['spec']['containers'][0]['image'] = bbx_container

    # Inject memory resource requirements
    bbyaml['spec']['template']['spec']['containers'][0]\
        ['resources']['requests']['memory'] = f"{pod_resources['mem_request']}Mi"
    bbyaml['spec']['template']['spec']['containers'][0]\
        ['resources']['limits']['memory'] = f"{pod_resources['mem_limit']}Mi"

    # Inject compute target resource requirements (CPU)
    if pod_resources['compute_target'] == 'CPU':
        bbyaml['spec']['template']['spec']['containers'][0]\
            ['resources']['requests']['cpu'] = f"{pod_resources['cpu_request']}m"
        bbyaml['spec']['template']['spec']['containers'][0]\
            ['resources']['limits']['cpu'] = f"{pod_resources['cpu_limit']}m"
        bbyaml['spec']['template']['spec']['containers'][0]\
            ['resources']['limits'].pop('nvidia.com/gpu')

    # Alternatively inject compute target resource requirements (NVIDIA GPU)
    elif pod_resources['compute_target'] == 'GPU':
        bbyaml['spec']['template']['spec']['containers'][0]\
            ['resources']['limits']['nvidia.com/gpu'] = pod_resources['gpu_limit']
        bbyaml['spec']['template']['spec']['containers'][0]\
            ['resources']['requests'].pop('cpu')
        bbyaml['spec']['template']['spec']['containers'][0]\
            ['resources']['limits'].pop('cpu')

    else:
        raise ValueError(f"Compute target error. {pod_resources['compute_target']} "
                         f"unknown or undefined")

    # Inject credential as K8s secret
    if credential_id:
        # TODO: validate this credential ID exists
        # TODO: validate this credential is a private docker registry credential
        # TODO: pull secrets_handle from within this credential ID
        secrets_handle = ""
        bbyaml['spec']['template']['spec']['imagePullSecrets'][0]['name'] = secrets_handle
    elif secrets_handle:
        bbyaml['spec']['template']['spec']['imagePullSecrets'][0]['name'] = secrets_handle

    # TODO: not sure if having the templated value here is harmful
    # TODO: probably best to snip out this branch if credential_id is None
    if not credential_id and not secrets_handle:
        _ = bbyaml['spec']['template']['spec'].pop('imagePullSecrets')

    # Add all environment variables to the manifest
    bbyaml['spec']['template']['spec']['containers'][0]['env'] = []
    for k, v in envvars.items():
        bbyaml['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': str(v)}
        )

    # Assign pod affinity on cloud (default to CPU pool)
    if not running_in_k8s():
        bbyaml['spec']['template']['spec'].pop('nodeSelector')
    else:
        if pod_resources['compute_target'] == 'GPU':
            bbyaml['spec']['template']['spec']\
                ['nodeSelector']['app.kinetica.com/pool'] = 'aaw-gpu'

    #  logger.debug('==== Worker ====')
    #  logger.debug(yaml.dump(bbyaml))
    return bbyaml


def customize_yaml_ingest_batch(job_id, envvars={}):
    # TODO: Currently this code is in controller_ingest
    pass


def customize_yaml_ingest_continuous(dep_id, envvars={}):
    # TODO: Currently this code is in controller_ingest
    pass


def customize_yaml_rapids_jupyter(profile, envvars, slug):
    """Construct the deployment specs for Jupyter Lab and the service pod."""

    from kml.api.naming_svc import (
            get_cluster_name_jupyter,
            get_cluster_name_service,
            get_cluster_name_ingress,
    )

    path = os.path.dirname(os.path.realpath(__file__))
    jupyter_yaml = yaml.load(
        open('/'.join([path, 'yaml/tmplt-lab.yaml'])),
        Loader=yaml.FullLoader
    )
    svc_yaml = yaml.load(
        open('/'.join([path, 'yaml/tmplt-lab-svc.yaml'])),
        Loader=yaml.FullLoader
    )
    ingress_yaml = yaml.load(
        open('/'.join([path, 'yaml/tmplt-lab-ingress.yaml'])),
        Loader=yaml.FullLoader
    )

    jupyter_ctnr = get_sys_container("ctnr_kml_rapids_jupyter")
    jupyter_yaml['spec']['template']['spec']['containers'][0]['image'] = jupyter_ctnr
    jupyter_yaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = global_pull_policy

    unique_dep_name = f'kml-jpt-{profile}'
    jpt_dep_name = get_cluster_name_jupyter(profile)
    jpt_svc_name = get_cluster_name_service(profile)
    jpt_igr_name = get_cluster_name_ingress(profile)

    # RapidsLab deployment config
    jupyter_yaml['metadata']['name'] = jpt_dep_name
    jupyter_yaml['spec']['selector']['matchLabels']['run'] = jpt_dep_name
    jupyter_yaml['spec']['template']['metadata']['labels']['run'] = jpt_dep_name

    jupyter_yaml['metadata']['namespace'] = namespace

    jupyter_yaml['spec']['template']['spec']['containers'][0]['env'] = []

    for k, v in envvars.items():
        jupyter_yaml['spec']['template']['spec']['containers'][0]['env'].append(
            {'name': k, 'value': v}
        )

    # RapidsLab service config
    svc_yaml['metadata']['name'] = jpt_svc_name
    svc_yaml['spec']['selector']['run'] = jpt_dep_name
    svc_yaml['metadata']['labels']['run'] = jpt_dep_name

    # RapidsLab ingress config
    ingress_yaml['metadata']['name'] = jpt_igr_name
    ingress_yaml['metadata']['annotations']['nginx.ingress.kubernetes.io/app-root'] = f"/{slug}"
    ingress_yaml['spec']['rules'][0]['http']['paths'][0]['path'] = f"/{slug}"
    ingress_yaml['spec']['rules'][0]['http']['paths'][0]['backend']['serviceName'] = jpt_svc_name
    #  ingress_yaml['spec']['rules'][0]['http']['paths'][1]['backend']['serviceName'] = jpt_svc_name

    return jupyter_yaml, svc_yaml, ingress_yaml


def customize_yaml_ctnr_prepull(container_ref):
    """Construct the deployment specs for pre-pull daemon-set."""
    import random

    SRC_YAML = 'yaml/tmplt-kml-prepull.yaml'

    path = os.path.dirname(os.path.realpath(__file__))
    myyaml = yaml.load(open('/'.join([path, SRC_YAML])), Loader=yaml.FullLoader)

    unique_dep_name = f'kml-prepull-{random.randint(1000000, 9999999)}'

    myyaml['metadata']['name'] = unique_dep_name
    myyaml['metadata']['namespace'] = namespace
    myyaml['spec']['template']['spec']['containers'][0]['name'] = unique_dep_name
    myyaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = global_pull_policy
    myyaml['spec']['template']['spec']['containers'][0]['image'] = container_ref
    myyaml['spec']['template']['metadata']['labels']['name'] = "kml-prepull"
    myyaml['spec']['selector']['matchLabels']['name'] = "kml-prepull"

    return myyaml


def customize_yaml_introspect_bbx_worker(bbx_container, credential_id=None, secrets_handle=None):

    import random
    from kml.api.naming_svc import get_cluster_name_deployment_blackbox

    ref_container = get_sys_container("ctnr_bb_basesdk")

    path = os.path.dirname(os.path.realpath(__file__))
    bbyaml = yaml.load(
        open('/'.join([path, 'yaml/template-bb-introspectable.yaml'])),
        Loader=yaml.FullLoader
    )

    bbyaml['spec']['template']['spec']['containers'][0]['image'] = ref_container
    bbyaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = user_pull_policy

    unique_dep_name = f'kml-inspect-{random.randint(1000000, 9999999)}'

    repo = bbx_container[:bbx_container.rfind('/')]
    secrets_hash = get_k8_secrets_hash(repo)

    bbyaml['metadata']['name'] = unique_dep_name
    bbyaml['metadata']['namespace'] = namespace
    bbyaml['spec']['template']['spec']['containers'][0]['image'] = bbx_container

    if credential_id:
        # TODO: validate this credential ID exists
        # TODO: validate this credential is a private docker registry credential
        # TODO: pull secrets_handle from within this credential ID
        secrets_handle = ""
        bbyaml['spec']['template']['spec']['imagePullSecrets'][0]['name'] = secrets_handle
    elif secrets_handle:
        bbyaml['spec']['template']['spec']['imagePullSecrets'][0]['name'] = secrets_handle

    # TODO: not sure if having the templated value here is harmful
    # TODO: probably best to snip out this branch if credential_id is None
    if not credential_id and not secrets_handle:
        _ = bbyaml['spec']['template']['spec'].pop('imagePullSecrets')

    return bbyaml


def customize_yaml_ctnr_introspect(container_ref, secrets_handle=None):
    """Construct the deployment specs for pre-pull daemon-set."""
    import random

    SRC_YAML = 'yaml/template-bb-introspectable.yaml'

    path = os.path.dirname(os.path.realpath(__file__))
    myyaml = yaml.load(open('/'.join([path, SRC_YAML])), Loader=yaml.FullLoader)

    unique_dep_name = f'kml-its-{random.randint(1000000, 9999999)}'

    myyaml['metadata']['name'] = unique_dep_name
    myyaml['metadata']['namespace'] = namespace
    myyaml['spec']['template']['spec']['containers'][0]['name'] = unique_dep_name
    myyaml['spec']['template']['spec']['containers'][0]['image'] = container_ref

    myyaml['spec']['template']['spec']['containers'][0]['imagePullPolicy'] = user_pull_policy
    myyaml['spec']['template']['metadata']['labels']['name'] = "kml-introspect"
    myyaml['spec']['selector']['matchLabels']['name'] = "kml-introspect"

    if secrets_handle:
        logger.debug(f"Using secrets handle {secrets_handle}")
        myyaml['spec']['template']['spec']['imagePullSecrets'][0]['name'] = secrets_handle
    else:
        logger.debug(f"Introspecting w/o credential secrets")
        _ = myyaml['spec']['template']['spec'].pop('imagePullSecrets')

    # print(myyaml)
    return myyaml
